/*
 * ksi.c
 * main () for stand alone ksi interpreter
 *
 * Copyright (C) 1997-2000,2001,2002, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Sun Aug 17 20:55:20 1997
 * Last Update:   Fri Jun  7 10:11:51 2002
 *
 */

/* Comments:
 *
 */

static char rcsid[] = "$Id: ksi.c,v 1.10 2002/06/07 03:35:58 ksion Exp $";

#define KSI_IMPL_DEFS 1
#include <ksi.h>
#include <ksi_type.h>
#include <ksi_prim.h>

#include <signal.h>

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#  include <fcntl.h>
#endif

#ifdef _MSC_VER
#  include <io.h>
#endif


#if defined(MSWIN32) || defined(OS2) || defined(MSDOS) \
    || defined(__CYGWIN32__) || defined(__CYGWIN__)

#  define EXE_SUFFIX ".exe"
#  define IS_DIR(x)  ((x) == '/' || (x) == '\\' || (x) == ':')

#else

#  define EXE_SUFFIX ".static"
#  define IS_DIR(x)  ((x) == '/')

#endif


struct Ksi_Data
{
  int		interactive;
  ksi_obj	input, output, error;
  char		*app_name, *prompt;

  ksi_obj	intr_prim, intr_evt;
  int		do_intr;

#if HAVE_LIBREADLINE
  ksi_obj	complit_list;
  char		*history_file;
  int		history_size;
  char		*readline;
#endif
};


static struct Ksi_Data *ksi_data;


#if HAVE_LIBREADLINE

typedef int Function ();

extern char *readline (char *prompt);
extern int  rl_parse_and_bind (char *line);
extern int  rl_variable_bind (char *var, char *val);

extern void add_history (char *line);
extern int  read_history (char *filename);
extern int  write_history (char *filename);
extern int  append_history (int nelements, char *filename);
extern void *remove_history (int which);
extern int  where_history ();
extern int  history_search (char *string, int direction);
extern int  history_truncate_file (char *filename, int nlines);
extern void stifle_history (int max);

extern char *rl_readline_name;
extern char *rl_basic_word_break_characters;
extern char *rl_basic_quote_characters;
extern char *rl_completer_word_break_characters;
extern char *rl_completer_quote_characters;
extern Function *rl_completion_entry_function;
extern Function *rl_event_hook;


static const char*
rl_name (ksi_port x)
{
  return "readline";
}

static int
rl_close (ksi_port x)
{
  return 0;
}

static int
rl_read (ksi_port x, char *buf, int num)
{
  int i;

  while (ksi_data->readline == 0)
    {
      ksi_data->readline = readline (ksi_data->prompt);
      if (ksi_data->readline == 0)
	return 0;

      if (ksi_data->readline[0] == 0)
	{
	  free (ksi_data->readline);
	  ksi_data->readline = 0;
	}
      else
	{
	  add_history (ksi_data->readline);
	}
    }

  for (i = 0; i < num; i++)
    {
      if (ksi_data->readline[i] != 0)
	buf[i] = ksi_data->readline[i];
      else
	{
	  free (ksi_data->readline);
	  ksi_data->readline = 0;
	  buf[i++] = '\n';
	  break;
	}
    }

  if (ksi_data->readline)
    {
      num = strlen (ksi_data->readline) - i;
      memmove (ksi_data->readline, ksi_data->readline + i, num + 1);
    }

  return i;
}

static int
rl_write (ksi_port x, const char* buf, int num)
{
  return num;
}

static int
rl_ready (ksi_port x)
{
  return 1;
}

static int
rl_flush (ksi_port x)
{
  return 0;
}

static int
rl_fd (ksi_port x)
{
  return -1;
}

static struct Ksi_Port_Tag rl_port_tag =
{
  rl_name,
  rl_close,
  rl_read,
  rl_write,
  rl_ready,
  rl_ready,
  rl_flush,
  rl_fd,
  rl_fd
};

static ksi_obj
make_rl_port (void)
{
  struct Ksi_Port *port;
  port = (struct Ksi_Port*) ksi_malloc (sizeof *port);
  port->itag            = KSI_TAG_PORT;
  port->ops             = &rl_port_tag;
  port->input           = 1;
  return (ksi_obj) port;
}


static char*
ksi_compliter (char *text, int num)
{
  if (!text)
    text = "";

  if (num == 0)
    ksi_data->complit_list = ksi_abbrev (text, strlen (text));

  if (ksi_data->complit_list != ksi_nil)
    {
      char *ptr = ksi_aprintf ("%w", KSI_CAR (ksi_data->complit_list));
      int len = strlen (ptr);
      ksi_data->complit_list = KSI_CDR (ksi_data->complit_list);

      if (ptr[0] == '#' && ptr[1] == '|')
	{
	  ptr[len - 1] = '\0';
	  ptr += 2;
	}
      else if (ptr[0] == '#' && ptr[1] == ':' && ptr[2] == '|')
	{
	  ptr[len - 1] = '\0';
	  ptr += 3;
	}

      text = (char*) malloc (len + 1);
      memcpy (text, ptr, len + 1);
      return text;
    }

  return 0;
}

static int
idle_work ()
{
  if (ksi_wait_event (ksi_zero) == ksi_false)
    return ksi_gcollect (0);
  return 0;
}

static void
init_readline (void)
{
  /* Allow conditional parsing of the ~/.inputrc file. */
  rl_readline_name = ksi_data->app_name;

  rl_basic_word_break_characters = " \n\t()[]{}'`,;\"";
  rl_basic_quote_characters = "|";
  rl_completer_quote_characters = "|";
  rl_completion_entry_function = (Function*) ksi_compliter;
  rl_event_hook = idle_work;

  stifle_history (ksi_data->history_size);
  read_history (ksi_data->history_file);
}

#endif


static ksi_obj
intr_proc (ksi_obj evt)
{
  if (ksi_data->do_intr)
    ksi_quit ();
  return ksi_void;
}

static int
repl_proc (void *_data)
{
  struct Ksi_Data *data = (struct Ksi_Data*) _data;
  double t1, t2, t3, t4, t5, t6;
  ksi_obj val;

  for (;;)
    {
#if !HAVE_LIBREADLINE
      if (data->interactive)
	{
	  ksi_printf (data->output, "%s", data->prompt);
	  ksi_flush_port (data->output);
	}
#endif

      data->do_intr = 0;
      val = ksi_read_const (data->input);
      data->do_intr = 1;

      if (val == ksi_eof)
	ksi_exit (ksi_true);

      if (data->interactive)
	{
	  t5 = ksi_real_time ();
	  t3 = ksi_eval_time ();
	  t1 = ksi_cpu_time ();
	}

      val = ksi_eval (val, 0);

      if (data->interactive)
	{
	  t2 = ksi_cpu_time ();
	  t4 = ksi_eval_time ();
	  t6 = ksi_real_time ();

	  if (val != ksi_void)
	    ksi_printf (data->output, "%w\n", val);
	  ksi_flush_port (data->output);

	  ksi_printf (data->error,
		      "\n; cpu/eval/real time: %.4f/%.4f/%.4f sec",
		      t2-t1, t4-t3, t6-t5);
	  ksi_printf (data->error,
		      "\n; heap/free size: %d/%d bytes\n",
		      ksi_heap_size (), ksi_heap_free ());
	  ksi_flush_port (data->error);
	}
    }

  return 0;
}

static int
error_proc (void *data)
{
  return KSI_NORMAL_EXIT_CODE;
}


main (int argc, char* argv[])
{
  ksi_obj val;
  char *ptr, *app;
  int i;

  ((void) rcsid);
  ksi_init_std_ports (0, 1, 2);

  ptr = argv [0];
  for (i = strlen (ptr); i > 0; --i)
    if (IS_DIR (ptr[i-1]))
      {
	ptr += i;
	break;
      }

  app = (char*) ksi_malloc_data (strlen (ptr) + 1);
  strcpy (app, ptr);
#ifdef EXE_SUFFIX
  if (ksi_has_suffix (app, EXE_SUFFIX))
    app [strlen (app) - strlen (EXE_SUFFIX)] = '\0';
#endif

  ksi_data = (struct Ksi_Data*) ksi_malloc_eternal (sizeof *ksi_data);
  ksi_data->interactive   = isatty (0);
  ksi_data->input         = ksi_current_input_port ();
  ksi_data->output        = ksi_current_output_port ();
  ksi_data->error         = ksi_current_error_port ();
  ksi_data->app_name      = app;
  ksi_data->prompt        = ksi_aprintf ("%s> ", app);

  if (ksi_data->interactive)
    {
      ksi_data->intr_prim = ksi_defun (0, intr_proc, KSI_CALL_ARG1, 1);
      ksi_data->intr_evt  = ksi_signal_event (KSI_MK_SINT (SIGINT), ksi_data->intr_prim);

#if HAVE_LIBREADLINE
      ptr = getenv ("HISTSIZE");
      ksi_data->history_file = ksi_expand_file_name ("~/.ksi_history");
      ksi_data->history_size = (ptr ? atoi (ptr) : 200);
      ksi_data->complit_list = ksi_nil;
      ksi_data->input        = make_rl_port ();
      init_readline ();
#endif
      ksi_interactive (ksi_true);
    }

  ksi_init (app, argc, argv, &argc);
  if (ksi_data->intr_evt)
    ksi_start_event (ksi_data->intr_evt);
  i = ksi_call_with_dynamic_context (repl_proc, error_proc, ksi_data);

  if (ksi_data->interactive)
    {
#if HAVE_LIBREADLINE
      if ((ptr = getenv ("HISTFILESIZE")) != 0)
	ksi_data->history_size = atoi (ptr);

      write_history (ksi_data->history_file);
      history_truncate_file (ksi_data->history_file, ksi_data->history_size);
#endif

      ksi_printf (ksi_data->error, ";EXIT %d\n", i);
    }

  ksi_term ();
  exit (i);
}


 /* End of code */
