/*
 * ksi_base.c
 * basic functions
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Mon Dec  8 09:18:17 1997
 * Last Update:   Fri Apr 13 23:01:18 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_comp.h"

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#  include <fcntl.h>
#endif

#ifdef _MSC_VER
#  include <io.h>
#endif

static char rcsid[] = "$Id: ksi_base.c,v 1.10 2001/04/13 16:56:21 demakov Exp $";


struct Ksi_Interp*	ksi_interp;
const char*		ksi_wna_s = "%s: missing or extra args";


int
ksi_default_tag_equal (ksi_tag tag, ksi_obj x1, ksi_obj x2, int deep)
{
  return 0;
}

const char*
ksi_default_tag_print (ksi_tag tag, ksi_obj x, int slashify)
{
  if (x -> itag == KSI_TAG_BROKEN)
    return ksi_aprintf ("#<broken %s %p>", tag->type_name, x);
  else
    return ksi_aprintf ("#<%s %p>", tag->type_name, x);
}


/*
 * Function ksi_version
 * Function ksi_major_version
 * Function ksi_minor_version
 * Function ksi_patch_level
 *	 KSI_VERSION, KSI_MAJOR_VERSION, KSI_MINOR_VERSION 
 *	KSI_PATCH_LEVEL	   ksi_ver.h,
 *	   Makefile'
 *	    VERSION.
 *
 * Return
 *	 .
 */

const char*
ksi_version ()
{
  return KSI_VERSION;
}

int
ksi_major_version ()
{
  return KSI_MAJOR_VERSION;
}

int
ksi_minor_version ()
{
  return KSI_MINOR_VERSION;
}

int
ksi_patch_level ()
{
  return KSI_PATCH_LEVEL;
}

const char*
ksi_init_path ()
{
  return KSI->ksi_init_path;
}

const char*
ksi_site_path ()
{
  return KSI->ksi_site_path;
}

const char*
ksi_libs_path ()
{
  return KSI->ksi_libs_path;
}

const char*
ksi_app_path ()
{
  return KSI->ksi_app_path;
}

const char*
ksi_app ()
{
  return KSI->ksi_app;
}


int
ksi_std_ports_inited ()
{
  if (!ksi_interp)
    return 0;

  return ksi_interp -> ports_inited;
}

void
ksi_init_std_ports (int in, int out, int err)
{
  ksi_port port;

  BASE_CHECK_INIT;

  if (in >= 0)
    {
      port = (ksi_port) ksi_make_fd_port (in, 0, "r");

      ksi_interp -> input_port = port;
      ksi_interp -> std_input_port = port;
      if (KSI_INITED)
	ksi_defsym ("*stdin*", (ksi_obj) port, 0);
    }

  if (out >= 0)
    {
      port = (ksi_port) ksi_make_fd_port (out, 0, "w");

      ksi_interp -> output_port = port;
      ksi_interp -> std_output_port = port;
      if (KSI_INITED)
	ksi_defsym ("*stdout*", (ksi_obj) port, 0);
    }

  if (err >= 0)
    {
      port = (ksi_port) ksi_make_fd_port (err, 0, "w");

      ksi_interp -> error_port = port;
      ksi_interp -> std_error_port = port;
      if (KSI_INITED)
	ksi_defsym ("*stderr*", (ksi_obj) port, 0);
    }

  ksi_interp -> ports_inited = 1;
}


/* ksi_add_boot_initializer (ksi_init_function proc, void* data)
 * ksi_add_initializer (ksi_init_function proc, void* data)
 *
 *   .
 *  --  ,    
 * ,     `Init.scm'.
 *      .
 *
 *  ,   (  )
 *       .
 *   ,     ,
 *   ,   
 *  --  ,     ..
 *  ,    ,  
 *   `Init.scm' (,   , 
 *    ),  
 *  -.
 *     `Init.scm'.   ,  
 *     ,    ,  
 *    , 
 * .
 */

void
ksi_add_boot_initializer (ksi_init_function proc, void* data)
{
  struct Ksi_Init_Proc *curr, **last;

  BASE_CHECK_INIT;

  last = & KSI->boot_procs;
  for (curr = KSI->boot_procs; curr; curr = curr->next)
    {
      /* Return if this initializer is registered allready. */
      if (curr->proc == proc && curr->data == data)
	return;
      last = & curr->next;
    }

  curr = NEW_TYPED (struct Ksi_Init_Proc, next, data);
  curr->proc = proc;
  curr->data = data;

  *last = curr;
}

void
ksi_add_initializer (ksi_init_function proc, void* data)
{
  struct Ksi_Init_Proc *curr, **last;

  BASE_CHECK_INIT;

  last = & KSI->init_procs;
  for (curr = KSI->init_procs; curr; curr = curr->next)
    {
      /* Return if this initializer is registered allready. */
      if (curr->proc == proc && curr->data == data)
	return;
      last = & curr->next;
    }

  curr = NEW_TYPED (struct Ksi_Init_Proc, next, data);
  curr->proc = proc;
  curr->data = data;

  *last = curr;
}

void
ksi_add_library (const char *path, const char *file,
		 ksi_init_function proc, void* data)
{
  struct Ksi_Stat_Lib *curr, **last;
  char *name;
  int  len;

  BASE_CHECK_INIT;

  if (path)
    {
      len  = strlen (path) + strlen (DIR_SEP) + strlen (file);
      name = (char*) alloca (len + 1);
      strcpy (name, path);
      strcat (name, DIR_SEP);
      strcat (name, file);
    }
  else
    {
      len  = strlen (file);
      name = (char*) file;
    }

  name = ksi_expand_file_name (name);
  len  = strlen (name);

  last = & KSI->stat_libs;
  for (curr = KSI->stat_libs; curr; curr = curr->next)
    {
      if (strpathcmp (name, curr->name) == 0)
	{
	  curr->proc = proc;
	  curr->data = data;
	  return;
	}

      last = & curr->next;
    }

  curr = NEW_TYPED_EXT (struct Ksi_Stat_Lib, len, next, data);
  curr->data = data;
  curr->proc = proc;
  strcpy (curr->name, name);

  *last = curr;
}


#if defined(MSWIN32)

int
ksi_win32s ()
{
  DWORD win_ver = GetVersion();

  /* Check that this is not NT, and Windows major version <= 3	*/
  return ((win_ver & 0x80000000) && (win_ver & 0xff) <= 3);
}

int
ksi_winnt ()
{
  DWORD win_ver = GetVersion();

  return !(win_ver & 0x80000000);
}

int
ksi_win_console ()
{
  return KSI->win_console;
}

const char*
ksi_win_name ()
{
  DWORD win_ver = GetVersion();

  if (((win_ver & 0x80000000) && (win_ver & 0xff) <= 3))
    return "win32s";

  if ((win_ver & 0x80000000))
    return "windows-95";

  return "nt";
}

void
ksi_init_console (char* console_title)
{
  HANDLE Fin, Fout, Ferr;
  int in, out, err;

  if (!ksi_win32s ())
    {
      AllocConsole ();
      SetConsoleTitle (console_title);

      Fin  = GetStdHandle (STD_INPUT_HANDLE);
      Fout = GetStdHandle (STD_OUTPUT_HANDLE);
      Ferr = GetStdHandle (STD_ERROR_HANDLE);

      in  = _open_osfhandle ((int) Fin,  O_RDONLY);
      out = _open_osfhandle ((int) Fout, O_WRONLY);
      err = _open_osfhandle ((int) Ferr, O_WRONLY);

      ksi_init_std_ports (in, out, err);
    }
}

const char*
ksi_win_error_text (char *msg, unsigned long err)
{
  char buf[256];

  buf[0] = '\0';
  buf[sizeof buf - 1] = '\0';
  FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM
		 | FORMAT_MESSAGE_IGNORE_INSERTS
		 | FORMAT_MESSAGE_MAX_WIDTH_MASK,
		 NULL,
		 err,
		 0,
		 buf,
		 sizeof buf - 1,
		 NULL);

  if (buf[0])
    {
      char *ptr = (char*) ksi_malloc_data (strlen (buf));
      strcpy (ptr, buf);
      if (KSI->win_console)
	CharToOem (ptr, ptr);
      return ptr;
    }

  return ksi_aprintf ("%s: error code 0x%x", msg, err);
}

const char*
ksi_get_last_error (char *msg)
{
  return ksi_win_error_text (msg, GetLastError ());
}


int
ksi_tolower (int x)
{
  unsigned char buf [2];
  buf [0] = (unsigned char) x;
  buf [1] = '\0';

  if (!KSI->win_console)
    CharLower (buf);
  else
    {
      OemToCharBuff (buf, buf, 1);
      CharLower (buf);
      CharToOemBuff (buf, buf, 1);
    }

  return buf [0];
}

int
ksi_toupper (int x)
{
  unsigned char buf [2];
  buf [0] = (unsigned char) x;
  buf [1] = '\0';

  if (!KSI->win_console)
    CharUpper (buf);
  else
    {
      OemToCharBuff (buf, buf, 1);
      CharUpper (buf);
      CharToOemBuff (buf, buf, 1);
    }

  return buf [0];
}

int
ksi_isalpha (int x)
{
  if (KSI->win_console)
    {
      unsigned char buf [2];
      buf [0] = (unsigned char) x;

      OemToCharBuff (buf, buf, 1);
      x = buf [0];
    }

  return IsCharAlpha ((TCHAR) x);
}

int
ksi_isalnum (int x)
{
  if (KSI->win_console)
    {
      unsigned char buf [2];
      buf [0] = (unsigned char) x;

      OemToCharBuff (buf, buf, 1);
      x = buf [0];
    }

  return IsCharAlphaNumeric ((TCHAR) x);
}

int
ksi_isupper (int x)
{
  if (KSI->win_console)
    {
      unsigned char buf [2];
      buf [0] = (unsigned char) x;

      OemToCharBuff (buf, buf, 1);
      x = buf [0];
    }

  return IsCharUpper ((TCHAR) x);
}

int
ksi_islower (int x)
{
  if (KSI->win_console)
    {
      unsigned char buf [2];
      buf [0] = (unsigned char) x;

      OemToCharBuff (buf, buf, 1);
      x = buf [0];
    }

  return IsCharLower ((TCHAR) x);
}

#endif

/*
 * ksi_base_init
 *	  .
 *	    
 *	ksi_init, , ,     ,
 *	   Ksi_Interp.
 *
 * Side effect
 *	C  Ksi_Interp   
 *	.
 *	  ksi_interp    
 *	.
 */

struct Ksi_Interp*
ksi_base_init (void)
{
  KSI_STACK_ELEM top_stack;

  ksi_ignore (rcsid);

  if (!ksi_interp)
    {
      ksi_init_gc ();

      KSI = (struct Ksi_Interp*) ksi_malloc_eternal (sizeof *KSI);
      bzero (KSI, sizeof *KSI);

      KSI -> ksi_libs_path = KSI_LIB_DIR;
      KSI -> ksi_init_path = KSI_LIB_DIR;
      KSI -> ksi_site_path = KSI_LIB_DIR;
      KSI -> ksi_app_path  = KSI_LIB_DIR;
      KSI -> ksi_app = "ksi";
      KSI -> ksi_app_sym = ksi_false;
      KSI -> stack = &top_stack;
      KSI -> exit_handlers = ksi_nil;
      KSI -> error_handlers = ksi_nil;
      KSI -> curr_module = ksi_false;
      KSI -> modules = ksi_nil;
      KSI -> null_port = ksi_new_nul_port ();
      KSI -> protected_objs = ksi_nil;
      KSI -> glob_env = (ksi_obj) ksi_new_env (1000);
      KSI -> errlog_priority = ERRLOG_ERROR;
      KSI -> module_priority = ksi_nil;

#ifdef MSWIN32
      if (GetStdHandle (STD_INPUT_HANDLE) == INVALID_HANDLE_VALUE)
	KSI -> win_console = 0;
      else
	KSI -> win_console = 1;
#endif

#if !defined(MSWIN32) || !defined(KSIDLL)
      atexit (ksi_term);
#endif
    }

  return ksi_interp;
}


#if defined(MSWIN32) && defined(KSIDLL)

int APIENTRY
DllMain (HINSTANCE hInstance, DWORD fdwReason, PVOID pvReserved)
{
  switch (fdwReason)
    {
    case DLL_PROCESS_ATTACH:
      ksi_interp = ksi_base_init ();
      break;

    case DLL_PROCESS_DETACH:
      ksi_term ();
      break;

    case DLL_THREAD_ATTACH:
    case DLL_THREAD_DETACH:
      break;
    }

  return TRUE;
}

#endif

 /* End of code */
