/*
 * ksi_char.c
 * chars
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Fri Jan 24 23:11:51 1997
 * Last Update:   Sat Nov 24 20:13:57 2001
 *
 * Comments:
 *
 */

#include "ksi_int.h"
#include <locale.h>


static char rcsid[] = "$Id: ksi_char.c,v 1.4 2001/11/25 14:30:01 demakov Exp $";


const char *ksi_char_names[] =
{
  "#\\nul","#\\soh","#\\stx","#\\etx","#\\eot","#\\enq","#\\ack","#\\bel",
  "#\\bs", "#\\ht", "#\\nl", "#\\vt", "#\\np", "#\\cr", "#\\so", "#\\si",
  "#\\dle","#\\dc1","#\\dc2","#\\dc3","#\\dc4","#\\nak","#\\syn","#\\etb",
  "#\\can", "#\\em","#\\sub","#\\esc", "#\\fs", "#\\gs", "#\\rs", "#\\us",
  "#\\space", "#\\newline", "#\\tab", "#\\backspace",
  "#\\return", "#\\page", "#\\escape", "#\\del",
  "#\\null", // for end null char in char_nums
  NULL
};

static char ksi_char_code[] =
  "\000\001\002\003\004\005\006\007"
  "\010\011\012\013\014\015\016\017"
  "\020\021\022\023\024\025\026\027"
  "\030\031\032\033\034\035\036\037"
  " \n\t\b\r\f\033\177";

#define KSI_CHAR_NUMS (sizeof (ksi_char_code))


ksi_obj
ksi_char_p (ksi_obj x)
{
  return KSI_CHAR_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_str2char (const char* name, int len)
{
  int i, j, n;

  if (len > 2 && name[0] == '#' && name[1] == '\\')
    {
      name += 2;
      len -= 2;
    }

  for (i = 0; i < KSI_CHAR_NUMS; ++i)
    {
      if (strlen (ksi_char_names[i]) == len + 2)
	{
	  for (j = 0; j < len; j++)
	    if (tolower (ksi_char_names[i][j+2]) != tolower (name[j]))
	      goto next;
	  return KSI_MK_CHAR (ksi_char_code[i]);
	}
    next:
      ;
    }

  if (name[0] == 'o' || name[0] == '0')
    {
      for (n = i = 0; ++i < len; )
        if ('0' <= name[i] && name[i] <= '7')
          n = n * 8 + name[i] - '0';
        else
          return ksi_false;

      return KSI_MK_CHAR (n);
    }

  if (name[0] == 'x' || name[0] == 'h')
    {
      for (n = i = 0; ++i < len; )
        if ('0' <= name[i] && name[i] <= '9')
          n = n * 16 + name[i] - '0';
        else if ('a' <= name[i] && name[i] <= 'f')
          n = n * 16 + (name[i] - 'a' + 10);
        else if ('A' <= name[i] && name[i] <= 'F')
          n = n * 16 + (name[i] - 'A' + 10);
        else
          return ksi_false;

      return KSI_MK_CHAR (n);
    }

  return ksi_false;
}

const char*
ksi_char2str (ksi_obj x)
{
  int i, c;
  char* ptr;

  c = KSI_CHAR_COD (x);
  for (i = 0; i < KSI_CHAR_NUMS; ++i)
    if (ksi_char_code[i] == c)
      return ksi_char_names[i];

  if (c < 128
      || (c < 255 && isalpha (c))
      || KSI_TRUE_P (ksi_spec_val ("*print-8bit-chars*", ksi_false, 0)))
    {
      ptr = (char*) ksi_malloc_data (4);
      ptr[0] = '#'; ptr[1] = '\\'; ptr[2] = (char) c; ptr[3] = '\0';
    }
  else
    {
      ptr = ksi_aprintf ("#\\o%03o", c);
    }

  return ptr;
}


#define CONCAT(a,b) a##b

#define DEF_COMP(op, test, name)					\
ksi_obj									\
CONCAT (ksi_char_, op) (int ac, ksi_obj* av)				\
{									\
  int c1, c2, i;							\
  if (ac < 1)								\
    return ksi_true;							\
									\
  TYPE_CHECK (av[0], KSI_CHAR_P (av[0]), name ": invalid char");	\
  c1 = KSI_CHAR_COD (av[0]);						\
  for (i = 1; i < ac; i++)						\
    {									\
      TYPE_CHECK (av[i], KSI_CHAR_P (av[i]), name ": invalid char");	\
      c2 = KSI_CHAR_COD (av[i]);					\
      if (!(test))							\
	return ksi_false;						\
      c1 = c2;								\
    }									\
  return ksi_true;							\
}

#ifdef HAVE_STRCOLL
static inline int chrcmp (char c1, char c2)
{
  char s1[2], s2[2];
  s1[0] = c1; s1[1] = '\0';
  s2[0] = c2; s2[1] = '\0';
  return strcoll (s1, s2);
}

#  define CHRCMP(c1, op, c2) (chrcmp ((c1), (c2)) op 0)
#else
#  define CHRCMP(c1, op, c2) ((c1) op (c2))
#endif

DEF_COMP (eq_p, c1 == c2, "char=?")
DEF_COMP (lt_p, CHRCMP (c1, <, c2), "char<?")
DEF_COMP (gt_p, CHRCMP (c1, >, c2), "char>?")
DEF_COMP (le_p, CHRCMP (c1, <=, c2), "char<=?")
DEF_COMP (ge_p, CHRCMP (c1, >=, c2), "char>=?")
DEF_COMP (ci_eq_p, tolower (c1) == tolower (c2), "char=?")
DEF_COMP (ci_lt_p, CHRCMP (tolower (c1), <, tolower (c2)), "char-ci<?")
DEF_COMP (ci_gt_p, CHRCMP (tolower (c1), >, tolower (c2)), "char-ci>?")
DEF_COMP (ci_le_p, CHRCMP (tolower (c1), <=, tolower (c2)), "char-ci<=?")
DEF_COMP (ci_ge_p, CHRCMP (tolower (c1), >=, tolower (c2)), "char-ci>=?")


#define DEF_TEST(op, test, name)				\
ksi_obj								\
CONCAT (ksi_char_, op) (ksi_obj o)				\
{								\
  int c;							\
  TYPE_CHECK (o, KSI_CHAR_P (o), name ": invalid char");	\
								\
  c = KSI_CHAR_COD (o);						\
  if (test)							\
    return ksi_true;						\
  return ksi_false;						\
}

DEF_TEST (digit_p, isdigit (c), "char-numeric?")
DEF_TEST (space_p, isspace (c), "char-whitespace?")
DEF_TEST (alpha_p, isalpha (c), "char-alphabetic?")
DEF_TEST (upper_p, isupper (c), "char-upper-case?")
DEF_TEST (lower_p, islower (c), "char-lower-case?")


ksi_obj
ksi_char2integer (ksi_obj x)
{
  TYPE_CHECK (x, KSI_CHAR_P (x), "char->integer: invalid char");

  return KSI_MK_SINT (KSI_CHAR_COD (x));
}

ksi_obj
ksi_integer2char (ksi_obj x)
{
  int i = 0;

  TYPE_CHECK (x, ksi_integer_p (x) == ksi_true,
	      "integer->char: invalid integer");
  if (KSI_SINT_P (x))
    {
      i = KSI_SINT_COD (x);
      if (! (0 <= i && i <= 255))
	goto bad;
    }
  else if (KSI_FLONUM_P (x))
    {
      double d = ((ksi_flonum) x) -> real;
      if (! (0.0 <= d && d <= 255.0))
	goto bad;
      i = (int) d;
    }
  else
    {
    bad:
      ksi_exn_error ("range", x, "integer->char: integer out of range");
    }

  return KSI_MK_CHAR (i);
}

ksi_obj
ksi_char_upcase (ksi_obj x)
{
  int c;
  TYPE_CHECK (x, KSI_CHAR_P (x), "char-upcase: invalid char");
  c = KSI_CHAR_COD (x);
  return KSI_MK_CHAR (toupper (c));
}

ksi_obj
ksi_char_downcase (ksi_obj x)
{
  int c;
  TYPE_CHECK (x, KSI_CHAR_P (x), "char-downcase: invalid char");
  c = KSI_CHAR_COD (x);
  return KSI_MK_CHAR (tolower (c));
}

static ksi_obj
ksi_setlocale (ksi_obj category, ksi_obj locale)
{
  const char *p = 0;
  int c = 0;

  if (KSI_SYM_P (category))
    p = KSI_SYM_PTR (category);
  else if (KSI_KEY_P (category))
    p = KSI_KEY_PTR (category);
  else
    KSI_EXN_ERR ("type", category, "setlocale: invalid category in arg1");

  if (strcasecmp (p, "LC_ALL") == 0)
    c = LC_ALL;
  else if (strcasecmp (p, "LC_COLLATE") == 0)
    c = LC_COLLATE;
  else if (strcasecmp (p, "LC_CTYPE") == 0)
    c = LC_CTYPE;
  else if (strcasecmp (p, "LC_MONETARY") == 0)
    c = LC_MONETARY;
  else if (strcasecmp (p, "LC_NUMERIC") == 0)
    c = LC_NUMERIC;
  else if (strcasecmp (p, "LC_TIME") == 0)
    c = LC_TIME;
  else
    KSI_EXN_ERR ("type", category, "setlocale: invalid category in arg1");

  if (!locale)
    p = 0;
  else if (KSI_STR_P (locale))
    p = KSI_STR_PTR (locale);
  else
    KSI_EXN_ERR ("type", locale, "setlocale: invalid locale in arg2");

  p = setlocale (c, p);

  return (p ? ksi_str02string (p) : ksi_false);
}


static struct Ksi_Prim_Def defs[] =
{
  { "char?",		ksi_char_p,		KSI_CALL_ARG1, 1 },

  { "char=?",		ksi_char_eq_p,		KSI_CALL_REST0, 0 },
  { "char<?",		ksi_char_lt_p,		KSI_CALL_REST0, 0 },
  { "char>?",		ksi_char_gt_p,		KSI_CALL_REST0, 0 },
  { "char<=?",		ksi_char_le_p,		KSI_CALL_REST0, 0 },
  { "char>=?",		ksi_char_ge_p,		KSI_CALL_REST0, 0 },
  { "char-ci=?",	ksi_char_ci_eq_p,	KSI_CALL_REST0, 0 },
  { "char-ci<?",	ksi_char_ci_lt_p,	KSI_CALL_REST0, 0 },
  { "char-ci>?",	ksi_char_ci_gt_p,	KSI_CALL_REST0, 0 },
  { "char-ci<=?",	ksi_char_ci_le_p,	KSI_CALL_REST0, 0 },
  { "char-ci>=?",	ksi_char_ci_ge_p,	KSI_CALL_REST0, 0 },

  { "char-alphabetic?",	ksi_char_alpha_p,	KSI_CALL_ARG1, 1 },
  { "char-numeric?",	ksi_char_digit_p,	KSI_CALL_ARG1, 1 },
  { "char-whitespace?",	ksi_char_space_p,	KSI_CALL_ARG1, 1 },
  { "char-upper-case?",	ksi_char_upper_p,	KSI_CALL_ARG1, 1 },
  { "char-lower-case?",	ksi_char_lower_p,	KSI_CALL_ARG1, 1 },

  { "integer->char",	ksi_integer2char,	KSI_CALL_ARG1, 1 },
  { "char->integer",	ksi_char2integer,	KSI_CALL_ARG1, 1 },
  { "char-upcase",	ksi_char_upcase,	KSI_CALL_ARG1, 1 },
  { "char-downcase",	ksi_char_downcase,	KSI_CALL_ARG1, 1 },

  /**/
  { "setlocale",	ksi_setlocale,		KSI_CALL_ARG2, 1 },

  { 0 }
};

void
ksi_init_chars ()
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}


 /* End of code */
