/* -*-mode:C++-*- */
/*
 * ksi_comp.h
 * low level control structures
 *
 * Copyright (C) 1997-2000,2001, Ivan Demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Wed Oct 22 20:44:58 1997
 * Last Update:   Wed Mar 28 04:24:05 2001
 *
 * 
 * $Id: ksi_comp.h,v 1.3 2001/04/02 19:49:05 demakov Exp $
 *
 */

#ifndef KSI_COMP_H
#define KSI_COMP_H

union Ksi_Var;
struct Ksi_EnvRec;
struct Ksi_Varbox;
struct Ksi_Freevar;

typedef union Ksi_Var*		ksi_var;
typedef struct Ksi_EnvRec*	ksi_envrec;
typedef struct Ksi_Varbox*	ksi_varbox;
typedef struct Ksi_Freevar*	ksi_freevar;


struct Ksi_EnvRec
{
  ksi_obj	sym;
  ksi_obj	val;
};

struct Ksi_Environ
{
  unsigned	itag;
  ksi_obj	module;
  ksi_valtab	valtab;
};

struct Ksi_Frame
{
  int		num;
  ksi_frame	next;
  ksi_obj	env;
  ksi_obj	vals[1];
};

struct Ksi_Func
{
  unsigned	itag;
  int		reqv, opts, rest;
  ksi_obj	env;
  ksi_obj	code;
};

struct Ksi_Closure
{
  unsigned	itag;
  int		nums, nary, opts;
  ksi_frame	frm;
  ksi_obj	doc;
  ksi_obj	body;
};

struct Ksi_Promise
{
  unsigned	itag;
  ksi_obj	val;
  ksi_frame	frm;
  ksi_obj	body;
};

struct Ksi_Values
{
  unsigned	itag;
  ksi_obj	vals;
};

struct Ksi_Code
{
  unsigned	itag;
  int		num;
  ksi_obj	val [1];
};

struct Ksi_Varbox
{
  unsigned	itag;
  int		lev;
  int		num;
};

struct Ksi_Freevar
{
  unsigned	itag;
  ksi_obj	sym;
  ksi_obj*	val;
};

union Ksi_Var
{
  unsigned		itag;
  struct Ksi_Code	code;
  struct Ksi_Freevar	var;
  struct Ksi_Varbox	box;
};


#define CODE_P(x)	(KSI_OBJ_P (x)					\
			 && KSI_TAG_FIRST_CODE <= KSI_OBJ_TAG (x)	\
			 && KSI_OBJ_TAG (x) <= KSI_TAG_LAST_CODE)

#define FREEVAR_P(x)	(KSI_OBJ_P (x)					\
			 && KSI_TAG_FREEVAR <= KSI_OBJ_TAG (x)		\
			 && KSI_OBJ_TAG (x) <= KSI_TAG_GLOBAL)

#define LOCVAR_P(x)	(KSI_OBJ_P (x)					\
			 && KSI_TAG_VAR0 <= KSI_OBJ_TAG (x)		\
			 && KSI_OBJ_TAG (x) <= KSI_TAG_VARN)

#define IMMOP_P(x)	(KSI_OBJ_P (x)					\
			 && KSI_TAG_FIRST_OP <= KSI_OBJ_TAG (x)		\
			 && KSI_OBJ_TAG (x) <= KSI_TAG_LAST_OP)


#define ENV_P(x)	(KSI_OBJ_IS ((x), KSI_TAG_ENVIRON))
#define ENV_MOD(x)	(((ksi_env) (x)) -> module)

#define VAR_NUM(x)	(((ksi_varbox) (x)) -> num)
#define VAR_LEV(x)	(((ksi_varbox) (x)) -> lev)
#define VAR_SYM(x)	(((ksi_freevar) (x)) -> sym)
#define VAR_VAL(x)	(*(((ksi_freevar) (x)) -> val))

#define CODE_NUM(x)	(((ksi_code) (x)) -> num)
#define CODE_VAL(x,n)	(((ksi_code) (x)) -> val [n])

#define QUOTE_VAL(x)	CODE_VAL ((x), 0)

#define IF_TEST(x)	CODE_VAL ((x), 0)
#define IF_THEN(x)	CODE_VAL ((x), 1)
#define IF_ELSE(x)	CODE_VAL ((x), 2)

#define CLOS_NUMS(x)	(((ksi_closure) (x)) -> nums)
#define CLOS_NARY(x)	(((ksi_closure) (x)) -> nary)
#define CLOS_OPTS(x)	(((ksi_closure) (x)) -> opts)
#define CLOS_FRM(x)	(((ksi_closure) (x)) -> frm)
#define CLOS_DOC(x)	(((ksi_closure) (x)) -> doc)
#define CLOS_BODY(x)	(((ksi_closure) (x)) -> body)

#define PROMISE_P(x)	(KSI_OBJ_IS ((x), KSI_TAG_PROMISE))
#define PROMISE_VAL(x)	(((ksi_promise) (x)) -> val)
#define PROMISE_FRM(x)	(((ksi_promise) (x)) -> frm)
#define PROMISE_BODY(x)	(((ksi_promise) (x)) -> body)

#define VALUES_P(x)	(KSI_OBJ_IS ((x), KSI_TAG_VALUES))
#define VALUES_VALS(x)	(((ksi_values) (x)) -> vals)


ksi_frame
ksi_new_frame (int size, ksi_frame frm);

ksi_env
ksi_new_env (int size);

ksi_obj
ksi_new_code (int n, int tag);

ksi_obj
ksi_new_freevar (ksi_obj sym);

ksi_obj
ksi_new_varbox (int lev, int num);

ksi_obj
ksi_new_closure (int nums, int nary, int opts, ksi_frame frm,
		 ksi_obj body, ksi_obj doc);

ksi_obj
ksi_comp (ksi_obj form, ksi_obj env);


#endif

 /* End of file */
