/*
 * ksi_ctrl.c
 * Control utils
 *
 * Copyright (C) 1997-2000,2001, Ivan Demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Wed Aug  6 22:27:10 1997
 * Last Update:   Sat Mar 24 09:37:52 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_comp.h"

static char rcsid[] = "$Id: ksi_ctrl.c,v 1.9 2001/04/04 16:59:07 demakov Exp $";


ksi_obj
ksi_dynamic_wind (ksi_obj pre, ksi_obj act, ksi_obj post)
{
  ksi_obj result;
  ksi_wind wind;

  ksi_apply_0 (pre);

  wind = (ksi_wind) ksi_malloc (sizeof *wind);
  wind->cont = KSI->wind;
  wind->the_catch = 0;
  wind->pre = pre;
  wind->post = post;
  KSI->wind = wind;

  result = ksi_apply_0 (act);

  KSI->wind = KSI->wind->cont;
  ksi_apply_0 (post);
  return result;
}


ksi_obj
ksi_call_cc (ksi_obj proc)
{
  ksi_obj val, cont;

  TYPE_CHECK (proc, PROC_P (proc), "call/cc: invalid procedure");
  val = ksi_continuation (&cont);
  return val ? val : ksi_apply_1 (proc, cont);
}


ksi_obj
ksi_catch (ksi_obj tag, ksi_obj body, ksi_obj handler)
{
  volatile ksi_wind wind;
  TYPE_CHECK (body, PROC_P (body), "catch: invalid procedure in arg2");
  TYPE_CHECK (handler, !handler || PROC_P (handler),
	      "catch: invalid procedure in arg3");

  wind = ksi_add_catch (tag, handler, 0);

  FLUSH_REGISTER_WINDOWS;
  if (setjmp (wind -> the_catch -> jmp . j_buf) == 0)
    {
      ksi_obj result = ksi_apply_0 (body);
      ksi_del_catch (wind);
      return result;
    }

  return wind -> the_catch -> value;
}

ksi_obj
ksi_catch_with_retry (ksi_obj tag, ksi_obj body, ksi_obj handler)
{
  volatile ksi_wind wind;
  TYPE_CHECK (body, PROC_P (body),
	      "catch-with-retry: invalid procedure in arg2");
  TYPE_CHECK (handler, !handler || PROC_P (handler),
	      "catch-with-retry: invalid procedure in arg3");

  wind = ksi_add_catch (tag, handler, 1);

  FLUSH_REGISTER_WINDOWS;
  if (setjmp (wind -> the_catch -> jmp . j_buf) == 0)
    {
      ksi_obj result = ksi_apply_0 (body);
      ksi_del_catch (wind);
      return result;
    }

  return wind -> the_catch -> value;
}

ksi_obj
ksi_throw (ksi_obj tag, ksi_obj val)
{
  ksi_wind wind;

  if (tag == ksi_err)
    ksi_throw_error (val);

  wind = ksi_find_catch (tag);
  if (wind == 0)
    ksi_exn_error ("misc", tag, "throw: uncatched thrown object");

  return ksi_throw_to_catch (wind, 1, tag, val, KSI_RETHROW_NORMAL);
}


ksi_obj
ksi_add_exit_handler (ksi_obj proc)
{
  BASE_CHECK_INIT;
  TYPE_CHECK (proc, PROC_P (proc), "add-exit-handler: invalid procedure");

  KSI->exit_handlers = CONS (proc, KSI->exit_handlers);
  return ksi_void;
}

ksi_obj
ksi_add_error_handler (ksi_obj proc)
{
  BASE_CHECK_INIT;
  TYPE_CHECK (proc, PROC_P (proc), "add-error-handler: invalid procedure");

  KSI->error_handlers = CONS (proc, KSI->error_handlers);
  return ksi_void;
}

ksi_obj
ksi_remove_error_handler (ksi_obj proc)
{
  ksi_obj *loc;

  BASE_CHECK_INIT;
  loc = &KSI->error_handlers;
  while (*loc != ksi_nil)
    {
      if (CAR (*loc) == proc)
	{
	  *loc = CDR (*loc);
	  break;
	}
      loc = & CDR (*loc);
    }

  return ksi_void;
}

ksi_obj
ksi_new_values (int num, ksi_obj* val)
{
  ksi_values x;

  if (num == 1)
    return val [0];

  x = NEW_TYPED (struct Ksi_Values, vals, vals);
  x->itag = KSI_TAG_VALUES;
  x->vals = ksi_new_list (num, val);

  return (ksi_obj) x;
}

ksi_obj
ksi_call_with_values (ksi_obj thunk, ksi_obj proc)
{
  ksi_obj x;
  TYPE_CHECK (thunk, PROC_P (thunk),
	      "call-with-values: invalid procedure in arg1");
  TYPE_CHECK (proc, PROC_P (proc),
	      "call-with-values: invalid procedure in arg2");

  x = ksi_apply_0 (thunk);
  if (VALUES_P (x))
    return ksi_apply (proc, VALUES_VALS (x));

  return ksi_apply_1 (proc, x);
}

ksi_obj
ksi_eval (ksi_obj form, ksi_obj env)
{
  if (PAIR_P (form))
    {
      ksi_obj expand = ksi_spec_val ("*system-expander*", 0, env);
      if (PROC_P (expand))
	form = ksi_apply_1 (expand, form);
    }

  return ksi_eval_sexp (form, env);
}

ksi_obj
ksi_eval_with_catch (ksi_obj form, ksi_obj env)
{
  volatile ksi_wind wind = ksi_add_catch (ksi_true, 0, 0);

  FLUSH_REGISTER_WINDOWS;
  if (setjmp (wind -> the_catch -> jmp . j_buf) == 0)
    {
      ksi_obj result = ksi_eval (form, env);
      ksi_del_catch (wind);
      return result;
    }

  return wind -> the_catch -> value;
}

ksi_obj
ksi_eval_str (const char* scheme_code)
{
  ksi_obj x, port;

  port = (ksi_obj) ksi_new_str_port (ksi_str02string (scheme_code));
  ((ksi_port) port) -> input = 1;
  x = ksi_read (port);
  ksi_close_port (port);
  return ksi_eval (x, 0);
}

ksi_obj
ksi_eval_str_with_catch (const char* scheme_code)
{
  volatile ksi_wind wind = ksi_add_catch (ksi_true, 0, 0);

  FLUSH_REGISTER_WINDOWS;
  if (setjmp (wind -> the_catch -> jmp . j_buf) == 0)
    {
      ksi_obj result = ksi_eval_str (scheme_code);
      ksi_del_catch (wind);
      return result;
    }

  return wind -> the_catch -> value;
}

ksi_obj
ksi_apply_proc_with_catch (ksi_obj proc, int argc, ksi_obj* argv)
{
  volatile ksi_wind wind = ksi_add_catch (ksi_true, 0, 0);

  FLUSH_REGISTER_WINDOWS;
  if (setjmp (wind -> the_catch -> jmp . j_buf) == 0)
    {
      ksi_obj result = ksi_apply_proc (proc, argc, argv);
      ksi_del_catch (wind);
      return result;
    }

  return wind -> the_catch -> value;
}

ksi_obj
ksi_apply_0 (ksi_obj proc)
{
  return ksi_apply_proc (proc, 0, 0);
}

ksi_obj
ksi_apply_1 (ksi_obj proc, ksi_obj arg1)
{
  return ksi_apply_proc (proc, 1, &arg1);
}

ksi_obj
ksi_apply_2 (ksi_obj proc, ksi_obj arg1, ksi_obj arg2)
{
  ksi_obj args[2];
  args[0] = arg1;
  args[1] = arg2;
  return ksi_apply_proc (proc, 2, args);
}

ksi_obj
ksi_apply_3 (ksi_obj proc, ksi_obj arg1, ksi_obj arg2, ksi_obj arg3)
{
  ksi_obj args[3];
  args[0] = arg1;
  args[1] = arg2;
  args[2] = arg3;
  return ksi_apply_proc (proc, 3, args);
}

ksi_obj
ksi_apply (ksi_obj proc, ksi_obj alist)
{
  ksi_obj *av;
  int i, len = LIST_LEN (alist);
  TYPE_CHECK (alist, len >= 0, "ksi_apply: invalid argument list");

  av = (ksi_obj*) alloca (len * sizeof *av);
  for (i = 0; i < len; i++)
    {
      av [i] = CAR (alist);
      alist = CDR (alist);
    }

  return ksi_apply_proc (proc, len, av);
}

ksi_obj
ksi_apply_0_with_catch (ksi_obj proc)
{
  return ksi_apply_proc_with_catch (proc, 0, 0);
}

ksi_obj
ksi_apply_1_with_catch (ksi_obj proc, ksi_obj arg1)
{
  return ksi_apply_proc_with_catch (proc, 1, &arg1);
}

ksi_obj
ksi_apply_2_with_catch (ksi_obj proc, ksi_obj arg1, ksi_obj arg2)
{
  ksi_obj args[2];
  args[0] = arg1;
  args[1] = arg2;
  return ksi_apply_proc_with_catch (proc, 2, args);
}

ksi_obj
ksi_apply_3_with_catch (ksi_obj proc, ksi_obj arg1, ksi_obj arg2, ksi_obj arg3)
{
  ksi_obj args[3];
  args[0] = arg1;
  args[1] = arg2;
  args[2] = arg3;
  return ksi_apply_proc_with_catch (proc, 3, args);
}

ksi_obj
ksi_apply_with_catch (ksi_obj proc, ksi_obj alist)
{
  ksi_obj *av;
  int i, len = LIST_LEN (alist);
  TYPE_CHECK (alist, len >= 0, "ksi_apply_with_catch: invalid argument list");

  av = (ksi_obj*) alloca (len * sizeof *av);
  for (i = 0; i < len; i++)
    {
      av [i] = CAR (alist);
      alist = CDR (alist);
    }

  return ksi_apply_proc_with_catch (proc, len, av);
}


static struct Ksi_Prim_Def defs [] =
{
  { "values",			ksi_new_values,		KSI_CALL_REST0, 0 },

  { "add-exit-handler",		ksi_add_exit_handler,	KSI_CALL_ARG1, 1 },
  { "add-error-handler",	ksi_add_error_handler,	KSI_CALL_ARG1, 1 },
  { "remove-error-handler",	ksi_remove_error_handler, KSI_CALL_ARG1, 1 },
  { "exit",			ksi_exit,		KSI_CALL_ARG1, 0 },

  { "dynamic-wind",		ksi_dynamic_wind,	KSI_CALL_ARG3, 3 },
  { "catch",			ksi_catch,		KSI_CALL_ARG3, 2 },
  { "catch-with-retry",		ksi_catch_with_retry,	KSI_CALL_ARG3, 2 },
  { "throw",			ksi_throw,		KSI_CALL_ARG2, 1 },
  { "eval",			ksi_eval,		KSI_CALL_ARG2, 1 },

  { 0 }
};

void
ksi_init_ctrl ()
{
  ksi_obj x;

  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
  ksi_defsym ("apply", ksi_new_imm (KSI_TAG_APPLY), 0);
  ksi_defsym ("call-with-values", ksi_new_imm (KSI_TAG_CALL_WITH_VALUES), 0);

  x = ksi_new_imm (KSI_TAG_CALL_CC);
  ksi_defsym ("call/cc", x, 0);
  ksi_defsym ("call-with-current-continuation", x, 0);
}

 /* End of code */
