/*
 * ksi_dump.c
 * 
 *
 * Copyright (C) 1998-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written
 * agreement is hereby granted, provided that the above copyright notice
 * and this paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Wed Jul  1 20:59:28 1998
 * Last Update:   Sat Nov 10 00:36:00 2001
 *
 */

/* Comments:
 *
 * The code is not very portable.
 * Only 32-bit Intel processor is fully supported.
 * It might not be too hard to support processors with:
 *   1) number of bits in char == 8
 *   2) sizeof(char) < sizeof(int) <= sizeof(ksi_obj).
 * 
 */

#include "ksi_int.h"
#include "ksi_comp.h"
#include "ksi_numb.h"
#include <math.h>

static char rcsid[] = "$Id: ksi_dump.c,v 1.8 2001/11/25 14:30:01 demakov Exp $";

#if !(KSI_INT_SIZE == 32 || KSI_INT_SIZE == 64)
#  error unsupported int size
#endif

#if KSI_INT_SIZE == 64
#  define LONG int
#elif SIZEOF_LONG >= SIZEOF_INT * 2
#  define LONG long
#elif defined(LONG_LONG) && (SIZEOF_LONG_LONG >= SIZEOF_INT * 2)
#  define LONG LONG_LONG
#endif


#define D_END		0x00
#define D_FALSE		0x01
#define D_TRUE		0x02
#define D_NIL		0x03
#define D_EOF		0x04 /* ^D */
#define D_VOID		0x05
#define D_ERR		0x06

#define D_ZERO		0x10
#define D_ONE		0x11
#define D_TWO		0x12
#define D_INT1		0x13
#define D_INT2		0x14
#define D_INT4		0x15
#define D_INT8		0x16
#define D_BIG1		0x17
#define D_BIG2		0x18
#define D_BIG4		0x19
#define D_BIG8		0x1a
#define D_FLOREAL	0x1b
#define D_FLOCOMP	0x1c
#define D_RATIONAL	0x1d
#define D_COMPLEX	0x1e
#define D_NUMBER	0x1f

#define D_CHAR		0x20
#define D_SYM1		0x21
#define D_SYM2		0x22
#define D_SYM4		0x23
#define D_SYM8		0x24
#define D_KEY1		0x25
#define D_KEY2		0x26
#define D_KEY4		0x27
#define D_KEY8		0x28
#define D_STR1		0x29
#define D_STR2		0x2a
#define D_STR4		0x2b
#define D_STR8		0x2c
#define D_CSTR1		0x2d
#define D_CSTR2		0x2e
#define D_CSTR4		0x2f
#define D_CSTR8		0x30
#define D_VEC1		0x31
#define D_VEC2		0x32
#define D_VEC4		0x33
#define D_VEC8		0x34
#define D_CVEC1		0x35
#define D_CVEC2		0x36
#define D_CVEC4		0x37
#define D_CVEC8		0x38
#define D_PAIR		0x39
#define D_CPAIR		0x3a
#define D_DATA1		0x3b
#define D_DATA2		0x3c
#define D_DATA4		0x3d
#define D_DATA8		0x3e

#define D_CODE1		0x60
#define D_CODE2		0x61
#define D_CODE4		0x62
#define D_CODE8		0x63
#define D_FREEVAR	0x64
#define D_VAR1		0x65
#define D_VAR2		0x66
#define D_VAR4		0x67
#define D_VAR8		0x68
#define D_IMMOP		0x69


static int
dump_int4 (int x, ksi_port port)
{
  int n = 0;
  n += ksi_port_putc (port, (int) (x & 0xff));
  n += ksi_port_putc (port, (int) ((x >> 8)  & 0xff));
  n += ksi_port_putc (port, (int) ((x >> 16) & 0xff));
  n += ksi_port_putc (port, (int) ((x >> 24) & 0xff));
  return n;
}

#if defined(LONG)
static int
dump_long (LONG x, ksi_port port)
{
  int n = 0;
  n += ksi_port_putc (port, (int) (x & 0xff));
  n += ksi_port_putc (port, (int) ((x >> 8) & 0xff));
  n += ksi_port_putc (port, (int) ((x >> 16) & 0xff));
  n += ksi_port_putc (port, (int) ((x >> 24) & 0xff));
  n += ksi_port_putc (port, (int) ((x >> 32) & 0xff));
  n += ksi_port_putc (port, (int) ((x >> 40) & 0xff));
  n += ksi_port_putc (port, (int) ((x >> 48) & 0xff));
  n += ksi_port_putc (port, (int) ((x >> 56) & 0xff));
  return n;
}
#endif

static int
dump_tag (int tag, ksi_port port)
{
#if KSI_TAG_LAST_OP <= 0x7f
  return ksi_port_putc (port, tag);
#elif KSI_TAG_LAST_OP <= 0x7fff
  int n = 0;
  n += ksi_port_putc (port, tag & 0xff);
  n += ksi_port_putc (port, (tag >> 8) & 0xff);
  return n;
#else
  int n = 0;
  n += ksi_port_putc (port, tag & 0xff);
  n += ksi_port_putc (port, (tag >> 8) & 0xff);
  n += ksi_port_putc (port, (tag >> 16) & 0xff);
  n += ksi_port_putc (port, (tag >> 24) & 0xff);
  return n;
#endif
}

static int
dump_len (int tag, int len, ksi_port port)
{
  int n = 0;

  if (ABS (len) <= 0x7f)
    {
      n += ksi_port_putc (port, tag);
      n += ksi_port_putc (port, len);
    }
  else if (ABS (len) <= 0x7fff)
    {
      n += ksi_port_putc (port, tag+1);
      n += ksi_port_putc (port, len & 0xff);
      n += ksi_port_putc (port, (len >> 8) & 0xff);
    }
  else
#if KSI_INT_SIZE == 64
    if (ABS (len) <= 0x7fffffff)
#endif
      {
	n += ksi_port_putc (port, tag+2);
	n += ksi_port_putc (port, len & 0xff);
	n += ksi_port_putc (port, (len >> 8) & 0xff);
	n += ksi_port_putc (port, (len >> 16) & 0xff);
	n += ksi_port_putc (port, (len >> 24) & 0xff);
      }
#if KSI_INT_SIZE == 64
    else
      {
	n += ksi_port_putc (port, tag+3);
	n += ksi_port_putc (port, len & 0xff);
	n += ksi_port_putc (port, (len >> 8) & 0xff);
	n += ksi_port_putc (port, (len >> 16) & 0xff);
	n += ksi_port_putc (port, (len >> 24) & 0xff);
	n += ksi_port_putc (port, (len >> 32) & 0xff);
	n += ksi_port_putc (port, (len >> 40) & 0xff);
	n += ksi_port_putc (port, (len >> 48) & 0xff);
	n += ksi_port_putc (port, (len >> 56) & 0xff);
      }
#endif

  return n;
}

static int
dump_double (double x, ksi_port port)
{
  int exp, val1, val2, n;

  x = frexp (x, &exp);
  val1 = (int) (x * 0x40000000);
  val2 = (int) ((x - val1) * 0x40000000);

  n  = dump_int4 (val1, port);
  n += dump_int4 (val2, port);
  n += dump_int4 (exp, port);

  return n;
}

static int
dump (ksi_obj x, ksi_port port)
{
  int num = 0;

again:

  if (x == ksi_void)  return num + ksi_port_putc (port, D_VOID);
  if (x == ksi_false) return num + ksi_port_putc (port, D_FALSE);
  if (x == ksi_true)  return num + ksi_port_putc (port, D_TRUE);
  if (x == ksi_eof)   return num + ksi_port_putc (port, D_EOF);
  if (x == ksi_nil)   return num + ksi_port_putc (port, D_NIL);
  if (x == ksi_err)   return num + ksi_port_putc (port, D_ERR);

  if (x == ksi_zero)  return ksi_port_putc (port, D_ZERO);
  if (x == ksi_one)   return ksi_port_putc (port, D_ONE);
  if (x == ksi_two)   return ksi_port_putc (port, D_TWO);

  if (KSI_SINT_P (x))
    {
#if (KSI_SIZEOF_INT_TYPE * CHAR_BIT == 32)
      return num + dump_len (D_INT1, KSI_SINT_COD (x), port);
#elif (KSI_SIZEOF_INT_TYPE * CHAR_BIT == 64)
      KSI_INT_TYPE val = KSI_SINT_COD (x);
      if (-0x7fffffff <= val && val <= 0x7fffffff)
	return num + dump_len (D_INT1, val, port);
      num += ksi_port_putc (port, D_INT8);
      return num + dump_long (val, port);
#endif
    }

  if (KSI_BIGNUM_P (x))
    {
      int sign = ((ksi_bignum) x) -> sign;
      int len  = ABS (sign);
      int i;

#if KSI_INT_SIZE == 32
      if (len == 1)
	{
	  unsigned val = ((ksi_bignum) x) -> digs [0];
	  if (val <= 0x7fffffff)
	    {
	      if (sign < 0) val = (unsigned) (-(int) val);
	      return num + dump_len (D_INT1, val, port);
	    }
	}
#ifdef LONG
      if (len <= 2)
	{
	  unsigned LONG val = (len == 1 ? 0 : ((ksi_bignum) x) -> digs [1]);
	  val = (val << 32) | ((ksi_bignum) x) -> digs [0];
	  if (val <= 0x7fffffffffffffff)
	    {
	      if (sign < 0) val = (unsigned LONG) (-(LONG) val);
	      num += ksi_port_putc (port, D_INT8);
	      return num + dump_long (val, port);
	    }
	}
#endif /* LONG */

      num += dump_len (D_BIG1, sign, port);
      for (i = 0; i < len; i++)
	num += dump_int4 (((ksi_bignum) x) -> digs [i], port);

#elif KSI_INT_SIZE == 64
      if (len == 1)
	{
	  unsigned val = ((ksi_bignum) x) -> digs [0];
	  if (val <= 0x7fffffffffffffff)
	    {
	      if (sign < 0) val = (unsigned) (-(int) val);
	      return num + dump_len (D_INT1, val, port);
	    }
	}

      num += dump_len (D_BIG1, 2*sign, port);
      for (i = 0; i < len; i++)
	{
	  unsigned val = ((ksi_bignum) x) -> digs [i];
	  num += dump_int4 (val, port);
	  num += dump_int4 (val >> 32, port);
	}
#endif /* KSI_INT_SIZE == 64 */

      return num;
    }

  if (KSI_FLONUM_P (x))
    {
      num += ksi_port_putc (port, D_FLOREAL);
      num += dump_double (KSI_REPART (x), port);
      return num;
    }

  if (KSI_COMPLEX_P (x))
    {
      num += ksi_port_putc (port, D_FLOCOMP);
      num += dump_double (KSI_REPART (x), port);
      num += dump_double (KSI_IMPART (x), port);
      return num;
    }

  if (KSI_CHAR_P (x))
    {
      num += ksi_port_putc (port, D_CHAR);
      return num + ksi_port_putc (port, KSI_CHAR_COD (x));
    }

  if (KSI_SYM_P (x))
    {
      num += dump_len (D_SYM1, KSI_SYM_LEN (x), port);
      return num + ksi_port_write (port, KSI_SYM_PTR (x), KSI_SYM_LEN (x));
    }

  if (KSI_KEY_P (x))
    {
      num += dump_len (D_KEY1, KSI_KEY_LEN (x), port);
      return num + ksi_port_write (port, KSI_KEY_PTR (x), KSI_KEY_LEN (x));
    }

  if (KSI_C_STR_P (x))
    {
      num += dump_len (D_CSTR1, KSI_STR_LEN (x), port);
      return num + ksi_port_write (port, KSI_STR_PTR (x), KSI_STR_LEN (x));
    }

  if (KSI_STR_P (x))
    {
      num += dump_len (D_STR1, KSI_STR_LEN (x), port);
      return num + ksi_port_write (port, KSI_STR_PTR (x), KSI_STR_LEN (x));
    }

  if (KSI_C_VEC_P (x))
    {
      int i;
      num += dump_len (D_CVEC1, KSI_VEC_LEN (x), port);
      for (i = 0; i < KSI_VEC_LEN (x); i++)
	num += dump (KSI_VEC_REF (x, i), port);
      return num;
    }

  if (KSI_VEC_P (x))
    {
      int i;
      num += dump_len (D_VEC1, KSI_VEC_LEN (x), port);
      for (i = 0; i < KSI_VEC_LEN (x); i++)
	num += dump (KSI_VEC_REF (x, i), port);
      return num;
    }

  if (KSI_C_PAIR_P (x))
    {
      num += ksi_port_putc (port, D_CPAIR);
      num += dump (KSI_CAR (x), port);
      x = KSI_CDR (x);
      goto again;
    }

  if (KSI_PAIR_P (x))
    {
      num += ksi_port_putc (port, D_PAIR);
      num += dump (KSI_CAR (x), port);
      x = KSI_CDR (x);
      goto again;
    }

  if (KSI_CPTR_P (x))
    {
      num += dump_len (D_DATA1, KSI_CPTR_SZ (x), port);
      return num + ksi_port_write (port, KSI_CPTR_VAL (x), KSI_CPTR_SZ (x));
    }

  if (CODE_P (x))
    {
      int i, len = CODE_NUM (x) + 1;
      num += dump_len (D_CODE1, len, port);
      num += dump_tag (x->itag, port);
      for (i = 0; i < len; i++)
	num += dump (CODE_VAL (x, i), port);
      return num;
    }

  if (FREEVAR_P (x))
    {
      num += ksi_port_putc (port, D_FREEVAR);
      return num + dump (VAR_SYM (x), port);
    }

  if (LOCVAR_P (x))
    {
      num += dump_len (D_VAR1, VAR_LEV (x), port);
      num += dump_int4 (VAR_NUM (x), port);
      return num;
    }

  if (IMMOP_P (x))
    {
      num += ksi_port_putc (port, D_IMMOP);
      num += dump_tag (x->itag, port);
      return num;
    }

  ksi_exn_error ("misc", x, "object->dump: cannot dump object");
  return num;
}


static int
undump_int1 (ksi_port port, const char *proc)
{
  int c = ksi_port_getc (port);
  if (c < 0)
    ksi_error ("%s: unexpected end of data", proc);
  return ((c & 0x80) ? (int) ((signed char) c) : c);
}

static int
undump_int2 (ksi_port port, const char *proc)
{
  int c1 = ksi_port_getc (port);
  int c2 = ksi_port_getc (port);
  if (c1 < 0 || c2 < 0)
    ksi_error ("%s: unexpected end of data", proc);
  c1 = (c2 << 8) | c1;
  return ((c1 & 0x8000) ? (int) ((signed short) c1) : c1);
}

static int
undump_int4 (ksi_port port, const char *proc)
{
  int c1 = ksi_port_getc (port);
  int c2 = ksi_port_getc (port);
  int c3 = ksi_port_getc (port);
  int c4 = ksi_port_getc (port);
  if (c1 < 0 || c2 < 0 || c3 < 0 || c4 < 0)
    ksi_error ("%s: unexpected end of data", proc);
  c1 = (c4 << 24) | (c3 << 16) | (c2 << 8) | c1;
#if KSI_INT_SIZE == 32
  return c1;
#elif KSI_INT_SIZE == 64
  return ((c1 & 0x80000000) ? -((c1 ^ 0xffffffff) + 1) : c1);
#endif
}

#if defined(LONG)
static LONG
undump_int8 (ksi_port port, const char *proc)
{
  int c1 = ksi_port_getc (port);
  int c2 = ksi_port_getc (port);
  int c3 = ksi_port_getc (port);
  int c4 = ksi_port_getc (port);
  int c5 = ksi_port_getc (port);
  int c6 = ksi_port_getc (port);
  int c7 = ksi_port_getc (port);
  int c8 = ksi_port_getc (port);
  if (c1 < 0 || c2 < 0 || c3 < 0 || c4 < 0 || c5 < 0 || c6 < 0 || c7 < 0 || c8 < 0)
    ksi_error ("%s: unexpected end of data", proc);

  c1 = (c4 << 24) | (c3 << 16) | (c2 << 8) | c1;
  c2 = (c8 << 24) | (c7 << 16) | (c6 << 8) | c5;
  return (((LONG) c2) << 32) | c1;
}
#endif

#if KSI_TAG_LAST_OP <= 0x7f
#  define undump_tag undump_int1
#elif KSI_TAG_LAST_OP <= 0x7fff
#  define undump_tag undump_int2
#else
#  define undump_tag undump_int4
#endif

static double
undump_double (ksi_port port, const char *proc)
{
  int val1 = undump_int4 (port, proc);
  int val2 = undump_int4 (port, proc);
  int exp  = undump_int4 (port, proc);
  return ldexp ((((double) val2 / 0x40000000) + val1) / 0x40000000, exp);
}

static void
undump_str (ksi_port port, void *ptr, int len, const char *proc)
{
  while (len > 0)
    {
      int n = ksi_port_read (port, ptr, len);
      if (n <= 0)
	ksi_error ("%s: unexpected end of data", proc);
      ptr += n;
      len -= n;
    }
}

static ksi_obj
undump (ksi_port port, const char *proc)
{
  int len;
  int cod = ksi_port_getc (port);
  switch (cod)
    {
    case -1:		return ksi_eof;
    case D_FALSE:	return ksi_false;
    case D_TRUE:	return ksi_true;
    case D_NIL:		return ksi_nil;
    case D_EOF:		return ksi_eof;
    case D_VOID:	return ksi_void;
    case D_ERR:		return ksi_err;
    case D_ZERO:	return ksi_zero;
    case D_ONE:		return ksi_one;
    case D_TWO:		return ksi_two;

    case D_INT1:	return KSI_MK_SINT (undump_int1 (port, proc));
    case D_INT2:	return KSI_MK_SINT (undump_int2 (port, proc));
    case D_INT4:	return ksi_int2num (undump_int4 (port, proc));
    case D_INT8:
#ifdef LONG
      return ksi_long2num (undump_int8 (port, proc));
#else
      {
	ksi_obj v1 = ksi_int2num (undump_int4 (port, proc));
	ksi_obj v2 = ksi_int2num (undump_int4 (port, proc));
	return ksi_logior (ksi_ash (v1, KSI_MK_SINT (32)), v2);
      }
#endif

    case D_BIG1:
      len = undump_int1 (port, proc);
    d_big:
      {
	int i, sign;
	unsigned *digs;

	sign = len;
	len  = ABS (len);
#if KSI_INT_SIZE == 32
	digs = (unsigned*) ksi_malloc_data (len * sizeof *digs);
	for (i = 0; i < len; i++)
	  digs[i] = (unsigned) undump_int4 (port, proc);
	return ksi_make_bignum (sign, digs);
#elif KSI_INT_SIZE == 64
	digs = (unsigned*) ksi_malloc_data ((len+1)/2 * sizeof *digs);
	for (i = 0; i < len; i++)
	  {
	    unsigned tmp = (unsigned) undump_int4 (port, proc);
	    if (i & 1)
	      digs[i/2] |= tmp << 32;
	    else
	      digs[i/2] = tmp;
	  }
	return ksi_make_bignum (sign, digs);
#endif
      }
    case D_BIG2:
      len = undump_int2 (port, proc);
      goto d_big;
    case D_BIG4:
      len = undump_int4 (port, proc);
      goto d_big;
#if KSI_INT_SIZE == 64
    case D_BIG8:
      len = undump_int8 (port, proc);
      goto d_big;
#endif

    case D_FLOREAL:
      return ksi_double2num (undump_double (port, proc));

    case D_FLOCOMP:
      {
	double re = undump_double (port, proc);
	double im = undump_double (port, proc);
	return ksi_rectangular (re, im);
      }

    case D_CHAR:
      return KSI_MK_CHAR (undump_int1 (port, proc));

    case D_SYM1:
      len = undump_int1 (port, proc);
    d_sym:
      {
	char *ptr = (char*) alloca (len);
	undump_str (port, ptr, len, proc);
	return ksi_str2sym (ptr, len);
      }
    case D_SYM2:
      len = undump_int2 (port, proc);
      goto d_sym;
    case D_SYM4:
      len = undump_int4 (port, proc);
      goto d_sym;
#if KSI_INT_SIZE == 64
    case D_SYM8:
      len = undump_int8 (port, proc);
      goto d_sym;
#endif

    case D_KEY1:
      len = undump_int1 (port, proc);
    d_key:
      {
	char *ptr = (char*) alloca (len);
	undump_str (port, ptr, len, proc);
	return ksi_str2key (ptr, len);
      }
    case D_KEY2:
      len = undump_int2 (port, proc);
      goto d_key;
    case D_KEY4:
      len = undump_int4 (port, proc);
      goto d_key;
#if KSI_INT_SIZE == 64
    case D_KEY8:
      len = undump_int8 (port, proc);
      goto d_key;
#endif

    case D_STR1:
      len = undump_int1 (port, proc);
    d_str:
      {
	char *ptr = (char*) alloca (len);
	undump_str (port, ptr, len, proc);
	return ksi_str2string (ptr, len);
      }
    case D_STR2:
      len = undump_int2 (port, proc);
      goto d_str;
    case D_STR4:
      len = undump_int4 (port, proc);
      goto d_str;
#if KSI_INT_SIZE == 64
    case D_STR8:
      len = undump_int8 (port, proc);
      goto d_str;
#endif

    case D_CSTR1:
      len = undump_int1 (port, proc);
    d_cstr:
      {
	ksi_obj x;
	char *ptr = (char*) alloca (len);
	undump_str (port, ptr, len, proc);
	x = ksi_str2string (ptr, len);
	x->itag = KSI_TAG_CONST_STRING;
	return x;
      }
    case D_CSTR2:
      len = undump_int2 (port, proc);
      goto d_cstr;
    case D_CSTR4:
      len = undump_int4 (port, proc);
      goto d_cstr;
#if KSI_INT_SIZE == 64
    case D_CSTR8:
      len = undump_int8 (port, proc);
      goto d_cstr;
#endif

    case D_VEC1:
      len = undump_int1 (port, proc);
    d_vec:
      {
	int i;
	ksi_obj x = ksi_alloc_vector (len, KSI_TAG_VECTOR);
	for (i = 0; i < len; i++)
	  KSI_VEC_REF (x, i) = undump (port, proc);
	return x;
      }
    case D_VEC2:
      len = undump_int2 (port, proc);
      goto d_vec;
    case D_VEC4:
      len = undump_int4 (port, proc);
      goto d_vec;
#if KSI_INT_SIZE == 64
    case D_VEC8:
      len = undump_int8 (port, proc);
      goto d_vec;
#endif

    case D_CVEC1:
      len = undump_int1 (port, proc);
    d_cvec:
      {
	int i;
	ksi_obj x = ksi_alloc_vector (len, KSI_TAG_CONST_VECTOR);
	for (i = 0; i < len; i++)
	  KSI_VEC_REF (x, i) = undump (port, proc);
	return x;
      }
    case D_CVEC2:
      len = undump_int2 (port, proc);
      goto d_cvec;
    case D_CVEC4:
      len = undump_int4 (port, proc);
      goto d_cvec;
#if KSI_INT_SIZE == 64
    case D_CVEC8:
      len = undump_int8 (port, proc);
      goto d_cvec;
#endif

    case D_PAIR:
      {
	ksi_obj x1 = undump (port, proc);
	ksi_obj x2 = undump (port, proc);
	return ksi_cons (x1, x2);
      }
    case D_CPAIR:
      {
	ksi_obj x1 = undump (port, proc);
	ksi_obj x2 = undump (port, proc);
	x1 = ksi_cons (x1, x2);
	x1->itag = KSI_TAG_CONST_PAIR;
	return x1;
      }

    case D_DATA1:
      len = undump_int1 (port, proc);
    d_data:
      {
	char *ptr = (char*) ksi_malloc_data (len);
	undump_str (port, ptr, len, proc);
	return ksi_make_pointer (ptr, len);
      }
    case D_DATA2:
      len = undump_int2 (port, proc);
      goto d_data;
    case D_DATA4:
      len = undump_int4 (port, proc);
      goto d_data;
#if KSI_INT_SIZE == 64
    case D_DATA8:
      len = undump_int8 (port, proc);
      goto d_data;
#endif

    case D_CODE1:
      len = undump_int1 (port, proc);
    d_code:
      {
	int i, tag = undump_tag (port, proc);
	ksi_obj x = ksi_new_code (len, tag);
	for (i = 0; i < len; i++)
	  CODE_VAL (x, i) = undump (port, proc);
	return x;
      }
    case D_CODE2:
      len = undump_int2 (port, proc);
      goto d_code;
    case D_CODE4:
      len = undump_int4 (port, proc);
      goto d_code;
#if KSI_INT_SIZE == 64
    case D_CODE8:
      len = undump_int8 (port, proc);
      goto d_code;
#endif

    case D_FREEVAR:
      {
	ksi_obj sym = undump (port, proc);
	return ksi_new_freevar (sym);
      }

    case D_VAR1:
      len = undump_int1 (port, proc);
    d_var:
      return ksi_new_varbox (len, undump_int4 (port, proc));
    case D_VAR2:
      len = undump_int2 (port, proc);
      goto d_var;
    case D_VAR4:
      len = undump_int4 (port, proc);
      goto d_var;
#if KSI_INT_SIZE == 64
    case D_VAR8:
      len = undump_int8 (port, proc);
      goto d_var;
#endif

    case D_IMMOP:
      return ksi_new_imm (undump_tag (port, proc));

    default:
      ksi_error ("%s: unknown data type %x", proc, cod);
    }

  return 0;
}


void*
ksi_dump (ksi_obj x, int* size)
{
  ksi_port p = ksi_new_str_port (0);
  p->output = 1;
  dump (x, p);

  x = ksi_port_string ((ksi_obj) p);
  if (size)
    *size = KSI_STR_LEN (x);
  return KSI_STR_PTR (x);
}

ksi_obj
ksi_undump (void* data, int size)
{
  ksi_obj  x = ksi_ptr2string (data, size);
  ksi_port p = ksi_new_str_port (x);
  p->input = 1;

  return undump (p, "ksi_undump");
}

ksi_obj
ksi_write_dump (ksi_obj x, ksi_obj p)
{
  TYPE_CHECK (p, KSI_OUTPUT_PORT_P (p), "write-dump: invalid output port in arg2");

  dump (x, (ksi_port) p);
  return ksi_void;
}

ksi_obj
ksi_read_dump (ksi_obj p)
{
  TYPE_CHECK (p, KSI_INPUT_PORT_P (p), "read-dump: invalid input port in arg1");
  return undump ((ksi_port) p, "read-dump");
}

ksi_obj
ksi_obj2dump (ksi_obj x)
{
  ksi_port p = ksi_new_str_port (0);
  p->output = 1;
  dump (x, p);
  return ksi_port_string ((ksi_obj) p);
}

ksi_obj
ksi_dump2obj (ksi_obj x)
{
  ksi_port p;

  if (KSI_CPTR_P (x))
    x = ksi_ptr2string (KSI_CPTR_VAL (x), KSI_CPTR_SZ (x));

  TYPE_CHECK (x, KSI_STR_P (x), "dump->object: invalid string in arg1");
  p = ksi_new_str_port (x);
  p->input = 1;
  return undump (p, "dump->object");
}


static struct Ksi_Prim_Def defs [] =
{
  { "object->dump",	ksi_obj2dump,		KSI_CALL_ARG1, 1 },
  { "dump->object",	ksi_dump2obj,		KSI_CALL_ARG1, 1 },
  { "read-dump",	ksi_read_dump,		KSI_CALL_ARG1, 1 },
  { "write-dump",	ksi_write_dump,		KSI_CALL_ARG2, 2 },

  { 0 }
};

void
ksi_init_dump (void)
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}

 /* End of code */
