/*
 * ksi_env.c
 * environment
 *
 * Copyright (C) 1998-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Sun Jan  4 20:31:17 1998
 * Last Update:   Mon Jul  2 18:54:13 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_vtab.h"
#include "ksi_comp.h"
#include "ksi_mod.h"


static char rcsid[] = "$Id: ksi_env.c,v 1.9 2001/07/21 12:57:32 demakov Exp $";


static unsigned
hash_env (void* r, unsigned n, void* data)
{
  ksi_obj sym = ((ksi_envrec) r) -> sym;
#if 1
  return ksi_hash_str (KSI_SYM_PTR (sym), KSI_SYM_LEN (sym), n);
#else
  return ((((unsigned) sym) * ((unsigned) data)) % n);
#endif
}

static int
cmp_env (void *r1, void *r2, void *data)
{
  return (((ksi_envrec) r1) -> sym) - (((ksi_envrec) r2) -> sym);
}

ksi_env
ksi_new_env (int size)
{
  ksi_env env;
  int r;

  for (r = 0; r == 0; r = rand ());

  env = NEW_TYPED (struct Ksi_Environ, module, valtab);
  env -> itag = KSI_TAG_ENVIRON;
  env -> module = 0;
  env -> valtab = ksi_new_valtab (size, hash_env, cmp_env, (void*) r);

  return env;
}

ksi_obj*
ksi_lookup_env (ksi_env env, ksi_obj sym)
{
  struct Ksi_EnvRec r, *p;
  KSI_CHECK2 (sym, SYM_P (sym), "ksi_lookup_env: invalid symbol");

  r.sym = sym;
  p = (ksi_envrec) ksi_lookup_vtab (env->valtab, &r);
  return (p ? &p->val : 0);
}

ksi_obj*
ksi_append_env (ksi_env env, ksi_obj sym, ksi_obj val)
{
  ksi_envrec p;

  KSI_CHECK2 (sym, SYM_P (sym), "ksi_append_env: invalid symbol");

  p = (ksi_envrec) ksi_malloc (sizeof *p);
  p->sym = sym;
  p->val = val;
  ksi_append_vtab (env->valtab, p);

  return &p->val;
}


ksi_obj
ksi_env_p (ksi_obj x)
{
  return (KSI_OBJ_IS (x, KSI_TAG_ENVIRON) ? ksi_true : ksi_false);
}

ksi_obj
ksi_global_env ()
{
  return KSI->glob_env;
}

ksi_obj
ksi_current_env ()
{
  if (MOD_P (KSI->curr_module))
    return MOD_ENV (KSI->curr_module);
  return KSI->glob_env;
}


static int
copy_env (void *val, void *env)
{
  ksi_append_vtab (((ksi_env) env) -> valtab, val);
  return 0;
}

ksi_obj
ksi_make_env (ksi_obj src_env)
{
  ksi_env env;

  if (!src_env)
    return (ksi_obj) ksi_new_env (0);

  TYPE_CHECK (src_env, ENV_P (src_env), "@make-environment: invalid environ in arg1");

  env = ksi_new_env (((ksi_env) src_env) -> valtab -> size);
  ksi_iterate_vtab (((ksi_env) src_env) -> valtab, copy_env, env);
  return (ksi_obj) env;
}


ksi_obj
ksi_define (ksi_obj sym, ksi_obj val, ksi_obj env)
{
  ksi_obj *ptr;

  TYPE_CHECK (sym, SYM_P (sym), "@define: invalid symbol in arg1");
  if (!env)
    env = ksi_current_env ();
  else
    {
      if (MOD_P (env))
	env = MOD_ENV (env);

      TYPE_CHECK (env, ENV_P (env), "@define: invalid environ in arg3");
    }

  ptr = ksi_lookup_env ((ksi_env) env, sym);
  if (ptr)
    *ptr = val;
  else
    ksi_append_env ((ksi_env) env, sym, val);

  return ksi_void;
}

ksi_obj
ksi_defined_p (ksi_obj sym, ksi_obj env)
{
  TYPE_CHECK (sym, SYM_P (sym), "@defined?: invalid symbol in arg1");
  if (!env)
    env = ksi_current_env ();
  else
    {
      if (MOD_P (env))
	env = MOD_ENV (env);
      else
	TYPE_CHECK (env, ENV_P (env), "@defined?: invalid environ in arg2");
    }

  if (ksi_lookup_env ((ksi_env) env, sym))
    return ksi_true;

  if (ENV_MOD (env))
    {
      ksi_obj use;
      for (use = MOD_USES (ENV_MOD (env)); use != ksi_nil; use = CDR (use))
	{
	  ksi_obj exp = MOD_EXPORT (CAR (use));
	  if (exp == ksi_true || KSI_TRUE_P (ksi_memq (sym, exp)))
	    {
	      if (ksi_lookup_env ((ksi_env) MOD_ENV (CAR (use)), sym))
		return ksi_true;
	    }
	}
    }

  if (env != KSI->glob_env && ksi_lookup_env ((ksi_env) KSI->glob_env, sym))
    return ksi_true;

  return ksi_false;
}


ksi_obj
ksi_var_p (ksi_obj env, ksi_obj sym)
{
  if (MOD_P (env))
    env = MOD_ENV (env);
  else
    TYPE_CHECK (env, ENV_P (env), "@variable?: invalid environ in arg1");
  TYPE_CHECK (sym, SYM_P (sym), "@variable?: invalid symbol in arg2");

  if (ksi_lookup_env ((ksi_env) env, sym))
    return ksi_true;

  return ksi_false;
}

ksi_obj
ksi_var_ref (ksi_obj env, ksi_obj sym)
{
  ksi_obj *rec;

  if (MOD_P (env))
    env = MOD_ENV (env);
  else
    TYPE_CHECK (env, ENV_P (env), "@variable-ref: invalid environ in arg1");
  TYPE_CHECK (sym, SYM_P (sym), "@variable-ref: invalid symbol in arg2");

  if ((rec = ksi_lookup_env ((ksi_env) env, sym)) != 0)
    return *rec;

  KSI_EXN_ERR ("variable", sym, "@variable-ref: no variable in current module");
  return ksi_void;
}

ksi_obj
ksi_var_set (ksi_obj env, ksi_obj sym, ksi_obj val)
{
  ksi_obj *rec;

  if (MOD_P (env))
    env = MOD_ENV (env);
  else
    TYPE_CHECK (env, ENV_P (env), "@variable-set!: invalid environ in arg1");
  TYPE_CHECK (sym, SYM_P (sym), "@variable-set!: invalid symbol in arg2");

  if ((rec = ksi_lookup_env ((ksi_env) env, sym)) != 0)
    {
      *rec = val;
      return ksi_void;
    }

  KSI_EXN_ERR ("variable", sym, "@variable-set!: no variable in current module");
  return ksi_void;
}


ksi_obj
ksi_defsym (const char* name, ksi_obj val, ksi_obj env)
{
  ksi_obj sym = ksi_str02sym (name);
  ksi_define (sym, val, env);
  return sym;
}

ksi_obj
ksi_glob_val (ksi_obj sym, ksi_obj def)
{
  ksi_obj *rec;

  KSI_CHECK2 (sym, SYM_P (sym), "@global-value: invalid symbol in arg1");
  rec = ksi_lookup_env ((ksi_env) KSI->glob_env, sym);
  return rec ? *rec : def;
}

ksi_obj
ksi_spec_val (const char* name, ksi_obj def, ksi_obj env)
{
  ksi_obj sym, *rec;

  if (!env)
    env = ksi_current_env ();
  else
    TYPE_CHECK (env, ENV_P (env), "@special-value: invalid environ");

  sym = ksi_find_symbol (name, strlen (name));
  if (sym)
    {
      rec = ksi_lookup_env ((ksi_env) env, sym);
      if (rec)
	return *rec;

      rec = ksi_lookup_env ((ksi_env) KSI->glob_env, sym);
      if (rec)
	return *rec;
    }

  return def ? def : ksi_false;
}

static ksi_obj
ksi_env_mod (ksi_obj env)
{
  TYPE_CHECK (env, ENV_P (env), "@environment-module: invalid environ");
  return (ENV_MOD (env) ? ENV_MOD (env) : ksi_false);
}


static int
iter_func (void *r, void *p)
{
  ksi_apply_2 ((ksi_obj) p, ((ksi_envrec) r) -> sym, ((ksi_envrec) r) -> val);
  return 0;
}

static ksi_obj
ksi_env_for_each (ksi_obj proc, ksi_obj env)
{
  TYPE_CHECK (env, ENV_P (env), "@environment-for-each: invalid environ");
  ksi_iterate_vtab (((ksi_env) env) -> valtab, iter_func, proc);
  return ksi_void;
}

static struct Ksi_Prim_Def defs [] = 
{
  { "@environment?",		ksi_env_p,		KSI_CALL_ARG1, 1 },
  { "@global-environment",	ksi_global_env,		KSI_CALL_ARG0, 0 },
  { "@current-environment",	ksi_current_env,	KSI_CALL_ARG0, 0 },
  { "@make-environment",	ksi_make_env,		KSI_CALL_ARG1, 0 },
  { "@define",			ksi_define,		KSI_CALL_ARG3, 2 },
  { "@defined?",		ksi_defined_p,		KSI_CALL_ARG2, 1 },
  { "@variable?",		ksi_var_p,		KSI_CALL_ARG2, 2 },
  { "@variable-ref",		ksi_var_ref,		KSI_CALL_ARG2, 2 },
  { "@variable-set!",		ksi_var_set,		KSI_CALL_ARG3, 3 },
  { "@environment-module",	ksi_env_mod,		KSI_CALL_ARG1, 1 },
  { "@environment-for-each",	ksi_env_for_each,	KSI_CALL_ARG2, 2 },

  { 0 }
};

void
ksi_init_env ()
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}


 /* End of code */
