/*
 * ksi_eval.c
 * eval
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Fri Jan 24 20:49:47 1997
 * Last Update:   Mon Apr 16 02:11:42 2001
 */

/* Comments:
 *
 */


#include "ksi_int.h"
#include "ksi_vtab.h"
#include "ksi_comp.h"
#include "ksi_mod.h"
#include "ksi_klos.h"

static char rcsid[] = "$Id: ksi_eval.c,v 1.11 2001/04/29 11:22:06 demakov Exp $";

static ksi_obj ksi_eval_int (ksi_obj x, ksi_frame env);

#define EVAL(form,frm)	(ksi_eval_int (form, frm))



static void
ksi_recomp_var (ksi_obj x, ksi_frame frm, char *name)
{
  ksi_obj *val, sym, use, env;

  sym = VAR_SYM (x);
  env = (frm ? frm->env : ksi_current_env ());

  val = ksi_lookup_env ((ksi_env) env, sym);
  if (val)
    {
      x->itag = KSI_TAG_LOCAL;
      ((ksi_var) x) -> var.val = val;
      return;
    }

  if (ENV_MOD (env))
    {
      for (use = MOD_USES (ENV_MOD (env)); use != ksi_nil; use = CDR (use))
	{
	  ksi_obj exp = MOD_EXPORT (CAR (use));
	  if (exp == ksi_true || KSI_TRUE_P (ksi_memq (sym, exp)))
	    {
	      val = ksi_lookup_env ((ksi_env) MOD_ENV (CAR (use)), sym);
	      if (val)
		{
		  if (exp == ksi_true)
		    x->itag = KSI_TAG_LOCAL;
		  else
		    x->itag = KSI_TAG_IMPORTED;
		  ((ksi_var) x) -> var.val = val;
		  return;
		}
	    }
	}
    }

  val = ksi_lookup_env ((ksi_env) KSI->glob_env, sym);
  if (val)
    {
      x->itag = KSI_TAG_GLOBAL;
      ((ksi_var) x) -> var.val = val;
      return;
    }

  ksi_exn_error ("variable", sym, "%s: unbound variable", name);
}

static void
ksi_eval_set (ksi_obj form, ksi_frame frm)
{
  ksi_obj	res, var, env;
  int		i;

  var = CODE_VAL (form, 0);
  res = EVAL (CODE_VAL (form, 1), frm);

again:

  switch (var->itag)
    {
    case KSI_TAG_VAR0:
      frm->vals [VAR_NUM (var)] = res;
      break;

    case KSI_TAG_VAR1:
      frm->next->vals [VAR_NUM (var)] = res;
      break;

    case KSI_TAG_VAR2:
      frm->next->next->vals [VAR_NUM (var)] = res;
      break;

    case KSI_TAG_VARN:
      i = VAR_LEV (var);
      while (--i >= 0)
	frm = frm->next;
      frm->vals [VAR_NUM (var)] = res;
      break;

    case KSI_TAG_LOCAL:
      VAR_VAL (var) = res;
      break;

    case KSI_TAG_IMPORTED:
      KSI_EXN_ERR ("variable", VAR_SYM (var), "set!: cannot assign imported variable");
      break;

    case KSI_TAG_GLOBAL:
      env = (frm ? frm->env : ksi_current_env ());
      KSI_EXN_CHECK ("variable", VAR_SYM (var), ENV_MOD (env) == 0,
		     "set!: cannot assign global variable");
      VAR_VAL (var) = res;
      break;

    case KSI_TAG_FREEVAR:
      ksi_recomp_var (var, frm, "set!");
      if (var->itag == KSI_TAG_FREEVAR)
	KSI_EXN_ERR ("variable", VAR_SYM (var), "set!: cannot assign pseudo variable");
      goto again;
    }
}

static void
ksi_eval_def (ksi_obj form, ksi_frame frm)
{
  ksi_obj sym, val, *rec, env;

  sym = CODE_VAL (form, 0);
  val = EVAL (CODE_VAL (form, 1), frm);
  env = (frm ? frm->env : ksi_current_env ());
  rec = ksi_lookup_env ((ksi_env) env, sym);
  KSI_WARN (sym, rec == 0, "define: variable redefined");

  if (KSI_OBJ_IS (val, KSI_TAG_CLOSURE))
    {
      if (CLOS_DOC (val) == ksi_false)
	CLOS_DOC (val) = sym;
      else if (KSI_STR_P (CLOS_DOC (val)))
	CLOS_DOC (val) = CONS (sym, CLOS_DOC (val));
    }

  if (rec)
    *rec = val;
  else
    ksi_append_env ((ksi_env) env, sym, val);
}


static ksi_frame
clos_frm (ksi_obj proc, int argc, ksi_obj* argv)
{
  ksi_frame	frm;
  ksi_obj	args;
  int		nary;

  frm  = ksi_new_frame (CLOS_NUMS (proc), CLOS_FRM (proc));
  nary = CLOS_NARY (proc);

  if (CLOS_OPTS (proc))
    {
      nary -= 1;
      WNA_CHECK (argc >= nary, proc, ksi_obj2str (proc));

      args = ksi_nil;
      while (--argc >= nary)
	args = CONS (argv [argc], args);
      frm->vals [nary] = args;
    }
  else
    {
      WNA_CHECK (argc == nary, proc, ksi_obj2str (proc));
    }

  while (--nary >= 0)
    frm->vals [nary] = argv [nary];

  return frm;
}

#define ALLOCARR(vec,n) do {						\
      if ((n) > arr_sz)							\
	  vec = (ksi_obj*) alloca ((arr_sz = (n)) * sizeof *vec);	\
    } while (0)

#ifdef __GNUC__
#  if defined(__i386__) || defined(__i486__) || defined(__i586__)
#    define REG asm("%edi")
#  else
#    define REG
#  endif
#else
#  define REG
#endif

static ksi_obj
ksi_eval_int (ksi_obj form, ksi_frame frm)
{
  register ksi_obj	*arr REG;
  int			i, arr_sz;
  ksi_obj		val;

  arr    = 0;
  arr_sz = 0;

again:

  KSI_CHECK2 (form, form && KSI_OBJ_P (form), "eval: invalid object");

  CHECK_EVENTS;
  switch (form->itag)
    {
    case KSI_TAG_SYMBOL:
    case KSI_TAG_PAIR:
    case KSI_TAG_CONST_PAIR:
    case KSI_TAG_VECTOR:
    case KSI_TAG_CONST_VECTOR:
    case KSI_TAG_STRING:
    case KSI_TAG_CONST_STRING:
    case KSI_TAG_KEYWORD:
    case KSI_TAG_BIGNUM:
    case KSI_TAG_FLONUM:
    case KSI_TAG_RATIONAL:
    case KSI_TAG_COMPLEX:
      return form;

    case KSI_TAG_FREEVAR:
      ksi_recomp_var (form, frm, "eval");
    case KSI_TAG_GLOBAL:
    case KSI_TAG_LOCAL:
    case KSI_TAG_IMPORTED:
      return (VAR_VAL (form));

    case KSI_TAG_VAR0:
      return (frm->vals [VAR_NUM (form)]);

    case KSI_TAG_VAR1:
      return (frm->next->vals [VAR_NUM (form)]);

    case KSI_TAG_VAR2:
      return (frm->next->next->vals [VAR_NUM (form)]);

    case KSI_TAG_VARN:
      for (i = 0; i < VAR_LEV (form); i++)
	frm = frm->next;
      return (frm->vals [VAR_NUM (form)]);

    case KSI_TAG_QUOTE:
      return (QUOTE_VAL (form));

    case KSI_TAG_SET:
      ksi_eval_set (form, frm);
      return (ksi_void);

    case KSI_TAG_DEFINE:
      ksi_eval_def (form, frm);
      return (ksi_void);

    case KSI_TAG_AND:
      for (i = 0; i < CODE_NUM (form); i++)
	{
	  if (EVAL (CODE_VAL (form, i), frm) == ksi_false)
	    {
	      return (ksi_false);
	    }
	}
      form = CODE_VAL (form, i);
      goto again;

    case KSI_TAG_OR:
      for (i = 0; i < CODE_NUM (form); i++)
	if ((val = EVAL (CODE_VAL (form, i), frm)) != ksi_false)
	  return val;
      form = CODE_VAL (form, i);
      goto again;

    case KSI_TAG_BEGIN:
      for (i = 0; i < CODE_NUM (form); i++)
	EVAL (CODE_VAL (form, i), frm);
      form = CODE_VAL (form, i);
      goto again;

    case KSI_TAG_IF2:
      if (EVAL (IF_TEST (form), frm) == ksi_false)
	return ksi_void;
      form = IF_THEN (form);
      goto again;

    case KSI_TAG_IF3:
      if (EVAL (IF_TEST (form), frm) == ksi_false)
	form = IF_ELSE (form);
      else
	form = IF_THEN (form);
      goto again;

    case KSI_TAG_LAMBDA:
      return ksi_new_closure (KSI_SINT_COD (CODE_VAL (form, 0)),
			      KSI_SINT_COD (CODE_VAL (form, 1)),
			      KSI_SINT_COD (CODE_VAL (form, 2)),
			      frm,
			      CODE_VAL (form, 4),
			      CODE_VAL (form, 3));

    case KSI_TAG_DELAY:
      val = (ksi_obj) NEW_TYPED (struct Ksi_Promise, val, body);
      val->itag = KSI_TAG_PROMISE;
      PROMISE_VAL (val) = 0;
      PROMISE_FRM (val) = ksi_new_frame (KSI_SINT_COD (CODE_VAL (form, 0)),
					 frm);
      PROMISE_BODY (val) = CODE_VAL (form, 1);
      return val;

    case KSI_TAG_FRAME:
      frm = ksi_new_frame (KSI_SINT_COD (CODE_VAL (form, 0)), frm);
      form = CODE_VAL (form, 1);
      goto again;

    case KSI_TAG_CALL:
      i = CODE_NUM (form) + 1;
      ALLOCARR (arr, i);
      {
	ksi_obj *tmp = & CODE_VAL (form, 0);
	while (--i >= 0)
	  {
#if 1
	    switch ((*tmp)->itag)
	      {
	      case KSI_TAG_SYMBOL:
	      case KSI_TAG_PAIR:
	      case KSI_TAG_CONST_PAIR:
	      case KSI_TAG_VECTOR:
	      case KSI_TAG_CONST_VECTOR:
	      case KSI_TAG_STRING:
	      case KSI_TAG_CONST_STRING:
	      case KSI_TAG_KEYWORD:
	      case KSI_TAG_BIGNUM:
	      case KSI_TAG_FLONUM:
	      case KSI_TAG_RATIONAL:
	      case KSI_TAG_COMPLEX:	
		*arr++ = *tmp;
		break;
	      case KSI_TAG_QUOTE:
		*arr++ = QUOTE_VAL (*tmp);
		break;
	      case KSI_TAG_GLOBAL:
	      case KSI_TAG_LOCAL:
	      case KSI_TAG_IMPORTED:
		*arr++ = VAR_VAL (*tmp);
		break;
	      case KSI_TAG_VAR0:
		*arr++ = frm->vals [VAR_NUM (*tmp)];
		break;
	      case KSI_TAG_VAR1:
		*arr++ = frm->next->vals [VAR_NUM (*tmp)];
		break;
	      case KSI_TAG_VAR2:
		*arr++ = frm->next->next->vals [VAR_NUM (*tmp)];
		break;
	      default:
		*arr++ = ksi_eval_int (*tmp, frm);
	      }
	    tmp++;
#else
	    *arr++ = ksi_eval_int (*tmp++, frm);
#endif
	  }
	i = CODE_NUM (form) + 1;
	arr -= i;
      }

    apply_call:
      /* arr [0]		::= function
       * arr [1 .. (i-1)]	::= arguments
       * i			::= number of arguments (include function)
       */

      if (!KSI_OBJ_P (arr [0]))
	{
	bad_apply:
	  KSI_EXN_ERR ("type", arr[0], "apply: invalid object to apply");
	}

      switch (arr[0]->itag)
	{
	case KSI_TAG_PRIM:
	  return ksi_apply_prim (arr[0], i-1, arr+1);

	case KSI_TAG_PRIM_0:
	  WNA_CHECK (i == 1, arr[0], KSI_PRIM_NAME (arr[0]));
	  return ((ksi_call_arg0) KSI_PRIM_PROC (arr[0])) ();

	case KSI_TAG_PRIM_1:
	  WNA_CHECK (i == 2, arr[0], KSI_PRIM_NAME (arr[0]));
	  return ((ksi_call_arg1) KSI_PRIM_PROC (arr[0])) (arr[1]);

	case KSI_TAG_PRIM_2:
	  WNA_CHECK (i == 3, arr[0], KSI_PRIM_NAME (arr[0]));
	  return ((ksi_call_arg2) KSI_PRIM_PROC (arr[0])) (arr[1], arr[2]);

	case KSI_TAG_PRIM_r:
	  WNA_CHECK (i-1 >= KSI_PRIM_REQV (arr[0]), arr[0], KSI_PRIM_NAME (arr[0]));
	  return ((ksi_call_rest0) KSI_PRIM_PROC (arr[0])) (i-1, arr+1);

	case KSI_TAG_PRIM_CLOSURE:
	  return ksi_apply_prim_closure (arr[0], i-1, arr+1);

	case KSI_TAG_CLOSURE:
	  frm  = clos_frm (arr[0], i-1, arr+1);
	  form = CLOS_BODY (arr[0]);
	  goto again;

	case KSI_TAG_APPLY:
	  WNA_CHECK (i >= 3, arr[0], "apply");
	apply:
	  {
	    int n = i - 2;
	    ksi_obj *tmp = arr + 1;
	    val = tmp[n];
	    i = LIST_LEN (val);
	    TYPE_CHECK (val, i >= 0, "apply: not a list in last arg");

	    ALLOCARR (arr, i+n);
	    for (i = 0; --n >= 0; /**/)
	      arr[i++] = *tmp++;
	    for (/**/; val != ksi_nil; val = CDR (val))
	      arr[i++] = CAR (val);

	    goto apply_call;

	  } /* case KSI_TAG_APPLY */

	case KSI_TAG_CALL_CC:
	  WNA_CHECK (i == 2, arr[0], "call/cc");
#if 0
	  arr[0] = arr[1];
	  val = ksi_continuation (&arr[1]);
	  if (val)
	    return val;
#else
	  arr[0] = arr[1];
#ifdef KSI_STACK_GROWS_UP
	  arr[1] = ksi_new_continuation ((ksi_stack) &arr[-1]);
#else
	  arr[1] = ksi_new_continuation ((ksi_stack) &arr[arr_sz]);
#endif
	  FLUSH_REGISTER_WINDOWS;
	  if (setjmp (((ksi_jump) (((ksi_prim_closure) arr[1]) -> argv[0])) -> j_buf))
	    {
	      val = KSI->jump_val;
	      KSI->jump_val = 0;
	      return val;
	    }
#endif
	  goto apply_call;

	case KSI_TAG_CALL_WITH_VALUES:
	  WNA_CHECK (i == 3, arr[0], "call-with-values");
	  val = ksi_apply_0 (arr[1]);
	  if (!VALUES_P (val))
	    {
	      arr[0] = arr[2];
	      arr[1] = val;
	      i = 2;
	      goto apply_call;
	    }
	  arr[1] = arr[2];
	  arr[2] = VALUES_VALS (val);
	  /* i = 3; */
	  goto apply;

	case KSI_TAG_NOT:
	  WNA_CHECK (i == 2, arr[0], "not");
	  return (arr[1] == ksi_false ? ksi_true : ksi_false);

	case KSI_TAG_EQP:
	  WNA_CHECK (i == 3, arr[0], "eq?");
	  return (KSI_EQ_P (arr[1], arr[2]) ? ksi_true : ksi_false);

	case KSI_TAG_EQVP:
	  WNA_CHECK (i == 3, arr[0], "eqv?");
	  return (KSI_EQV_P (arr[1], arr[2]) ? ksi_true : ksi_false);

	case KSI_TAG_EQUALP:
	  WNA_CHECK (i == 3, arr[0], "equal?");
	  return (KSI_EQUAL_P (arr[1], arr[2]) ? ksi_true : ksi_false);

	case KSI_TAG_MEMQ:
	  WNA_CHECK (i == 3, arr[0], "memq");
	  return ksi_memq (arr[1], arr[2]);

	case KSI_TAG_MEMV:
	  WNA_CHECK (i == 3, arr[0], "memv");
	  return ksi_memv (arr[1], arr[2]);

	case KSI_TAG_MEMBER:
	  WNA_CHECK (i == 3, arr[0], "member");
	  return ksi_member (arr[1], arr[2], 0);

	case KSI_TAG_CONS:
	  WNA_CHECK (i == 3, arr[0], "cons");
	  return (CONS (arr[1], arr[2]));

	case KSI_TAG_CAR:
	  WNA_CHECK (i == 2, arr[0], "car");
	  TYPE_CHECK (arr[1], PAIR_P (arr[1]), "car: not pair in arg1");
	  return (CAR (arr[1]));

	case KSI_TAG_CDR:
	  WNA_CHECK (i == 2, arr[0], "cdr");
	  TYPE_CHECK (arr[1], PAIR_P (arr[1]), "cdr: not pair in arg1");
	  return (CDR (arr[1]));

	case KSI_TAG_LIST:
	  return (ksi_new_list (i-1, arr+1));

	case KSI_TAG_NULLP:
	  WNA_CHECK (i == 2, arr[0], "null?");
	  return (arr[1] == ksi_nil ? ksi_true : ksi_false);

	case KSI_TAG_PAIRP:
	  WNA_CHECK (i == 2, arr[0], "pair?");
	  return (PAIR_P (arr[1]) ? ksi_true : ksi_false);

	case KSI_TAG_LISTP:
	  WNA_CHECK (i == 2, arr[0], "list?");
	  return (KSI_LIST_P (arr[1]) ? ksi_true : ksi_false);

	case KSI_TAG_VECTORP:
	  WNA_CHECK (i == 2, arr[0], "vector?");
	  return (KSI_VEC_P (arr[1]) ? ksi_true : ksi_false);

	case KSI_TAG_INSTANCE:
	  if (PURE_GENERIC_P (arr[0]))
	    {
	      form = ksi_new_list (i-1, arr+1);
	      val  = ksi_compute_effective_method (arr[0], form);

	      if (val == ksi_nil)
		{
		  val = arr[0];
		  ALLOCARR (arr, 3); i = 3;
		  arr[0] = ksi_glob_val (KSI->sym_no_applicable_method, 0);
		  arr[1] = val;
		  arr[2] = form;
		  if (!arr[0])
		    ksi_exn_error ("misc", arr[1], "no applicable methods");
		  goto apply_call;
		}

	      goto call_next;
	    }

	  if (GENERIC_P (arr[0]))
	    {
	      form = ksi_new_list (i-1, arr+1);
	      val = arr[0];
	      ALLOCARR (arr, 3); i = 3;
	      arr[0] = ksi_glob_val (KSI->sym_apply_generic, 0);
	      arr[1] = val;
	      arr[2] = form;
	      if (!arr[0])
		ksi_exn_error ("misc", arr[1], "apply-generic not defined");
	      goto apply_call;
	    }
	  goto bad_apply;

	case KSI_TAG_NEXT_METHOD:
	  val  = arr[0];
	  form = (i > 1 ? ksi_new_list (i-1, arr+1) : NEXT_ARGS (val));

	  if (NEXT_PROCS (val) == ksi_nil)
	    {
	      ALLOCARR (arr, 3); i = 3;
	      arr[0] = ksi_glob_val (KSI->sym_no_next_method, 0);
	      arr[1] = NEXT_GF (val);
	      arr[2] = form;
	      if (!arr[0])
		ksi_exn_error ("misc", arr[1], "no next method");
	      goto apply_call;
	    }

	call_next:
	  ALLOCARR (arr, 5); i = 5;
	  arr[1] = CAR (NEXT_PROCS (val));
	  arr[2] = ksi_new_next (NEXT_GF (val),
				 form,
				 CDR (NEXT_PROCS (val)));
	  arr[3] = (PAIR_P (CDR (NEXT_PROCS (val)))
		    ? KSI->true_proc : KSI->false_proc);
	  arr[4] = form;
	  goto apply;

	} /* switch (arr[0]->itag) */

      goto bad_apply;

    } /*  switch (form->itag) */

  ksi_exn_error ("type", form, "eval: form not eval'able");
  return 0;

}

#undef ALLOCARR
#undef LABEL
#undef SWITCH


ksi_obj
ksi_eval_sexp (ksi_obj form, ksi_obj new_env)
{
  if (!new_env)
    new_env = ksi_current_env ();
  else
    TYPE_CHECK (new_env, ENV_P (new_env), "eval: invalid environ in arg2");

#if STRICT_R5RS
  if (form == ksi_nil)
    ksi_exn_error ("misc", form, "eval: no subexpresion in procedure call");
#endif

  if (!KSI_OBJ_P (form))
    return form;

#if 0
  if (!CODE_P (form) && !FREEVAR_P (form) && !LOCVAR_P (form))
    form = ksi_comp (form, ksi_false);
#else
  if (!CODE_P (form) && !FREEVAR_P (form))
    form = ksi_comp (form, ksi_false);
#endif

  return ksi_eval_int (form, 0);
}


ksi_obj
ksi_apply_proc (ksi_obj proc, int argc, ksi_obj* argv)
{
  ksi_obj	val;
  ksi_frame	frm;
  int		i;

  if (KSI_OBJ_P (proc))
    {
      switch (proc->itag)
	{
	case KSI_TAG_LIST:
	  return ksi_new_list (argc, argv);

	case KSI_TAG_NOT:
	  WNA_CHECK (argc == 1, proc, "not");
	  return (argv[0] == ksi_false ? ksi_true : ksi_false);

	case KSI_TAG_NULLP:
	  WNA_CHECK (argc == 1, proc, "null?");
	  return (argv[0] == ksi_nil ? ksi_true : ksi_false);

	case KSI_TAG_PAIRP:
	  WNA_CHECK (argc == 1, proc, "pair?");
	  return (PAIR_P (argv[0]) ? ksi_true : ksi_false);

	case KSI_TAG_LISTP:
	  WNA_CHECK (argc == 1, proc, "list?");
	  return (KSI_LIST_P (argv[0]) ? ksi_true : ksi_false);

	case KSI_TAG_VECTORP:
	  WNA_CHECK (argc == 1, proc, "vector?");
	  return (KSI_VEC_P (argv[0]) ? ksi_true : ksi_false);

	case KSI_TAG_EQP:
	  WNA_CHECK (argc == 2, proc, "eq?");
	  return KSI_EQ_P (argv[0], argv[1]) ? ksi_true : ksi_false;

	case KSI_TAG_EQVP:
	  WNA_CHECK (argc == 2, proc, "eqv?");
	  return KSI_EQV_P (argv[0], argv[1]) ? ksi_true : ksi_false;

	case KSI_TAG_EQUALP:
	  WNA_CHECK (argc == 2, proc, "equal?");
	  return KSI_EQUAL_P (argv[0], argv[1]) ? ksi_true : ksi_false;

	case KSI_TAG_MEMQ:
	  WNA_CHECK (argc == 2, proc, "memq");
	  return ksi_memq (argv[0], argv[1]);

	case KSI_TAG_MEMV:
	  WNA_CHECK (argc == 2, proc, "memv");
	  return ksi_memv (argv[0], argv[1]);

	case KSI_TAG_MEMBER:
	  WNA_CHECK (argc == 2, proc, "member");
	  return ksi_member (argv[0], argv[1], 0);

	case KSI_TAG_CONS:
	  WNA_CHECK (argc == 2, proc, "cons");
	  return CONS (argv[0], argv[1]);

	case KSI_TAG_CAR:
	  WNA_CHECK (argc == 1, proc, "car");
	  TYPE_CHECK (argv[0], PAIR_P (argv[0]), "car: invalid pair");
	  return CAR (argv[0]);

	case KSI_TAG_CDR:
	  WNA_CHECK (argc == 1, proc, "cdr");
	  TYPE_CHECK (argv[0], PAIR_P (argv[0]), "cdr: invalid pair");
	  return CDR (argv[0]);

	case KSI_TAG_PRIM_0:
	  WNA_CHECK (argc == 0, proc, KSI_PRIM_NAME (proc));
	  val = ((ksi_call_arg0) KSI_PRIM_PROC (proc)) ();
	  return val;

	case KSI_TAG_PRIM_1:
	  WNA_CHECK (argc == 1, proc, KSI_PRIM_NAME (proc));
	  val = ((ksi_call_arg1) KSI_PRIM_PROC (proc)) (argv[0]);
	  return val;

	case KSI_TAG_PRIM_2:
	  WNA_CHECK (argc == 2, proc, KSI_PRIM_NAME (proc));
	  val = ((ksi_call_arg2) KSI_PRIM_PROC (proc)) (argv[0], argv[1]);
	  return val;

	case KSI_TAG_PRIM_r:
	  WNA_CHECK (argc >= KSI_PRIM_REQV (proc), proc, KSI_PRIM_NAME (proc));
	  val = ((ksi_call_rest0) KSI_PRIM_PROC (proc)) (argc, argv);
	  return val;

	case KSI_TAG_PRIM:
	  return ksi_apply_prim (proc, argc, argv);

	case KSI_TAG_CLOSURE:
	  frm = clos_frm (proc, argc, argv);
	  val = ksi_eval_int (CLOS_BODY (proc), frm);
	  return val;
	} /* switch */
    }

  val = ksi_new_code (argc + 1, KSI_TAG_CALL);
  CODE_VAL (val, 0) = ksi_new_quote (proc);
  for (i = 1; i <= argc; i++)
    CODE_VAL (val, i) = ksi_new_quote (*argv++);

  val = ksi_eval_int (val, 0);
  return val;
}


ksi_obj
ksi_force (ksi_obj x)
{
  ksi_obj val;

  if (!PROMISE_P (x))
    return x;

  if (PROMISE_VAL (x) == 0)
    {
      val = ksi_eval_int (PROMISE_BODY (x), PROMISE_FRM (x));
      if (PROMISE_VAL (x) == 0)
	{
	  PROMISE_VAL  (x) = val;
	  PROMISE_FRM  (x) = 0;
	  PROMISE_BODY (x) = 0;
	}
    }

  return PROMISE_VAL (x);
}


static struct Ksi_Prim_Def defs [] =
{
  { "force",			ksi_force,		KSI_CALL_ARG1, 1 },
  { "@eval-sexp",		ksi_eval_sexp,		KSI_CALL_ARG2, 1 },

  { 0 }
};

void
ksi_init_eval (void)
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}

/* End of code */
