/*
 * ksi_exep.c
 * exceptions
 *
 * Copyright (C) 1998-2000, Ivan Demakov.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written
 * agreement is hereby granted, provided that the above copyright notice
 * and this paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Wed Jun 17 00:20:22 1998
 * Last Update:   Wed Nov  8 23:39:15 2000
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"

static char rcsid[] = "$Id: ksi_exep.c,v 1.5 2001/04/04 16:59:07 demakov Exp $";


static ksi_obj
cont_prim (struct Ksi_Jump* jmp, int argc, ksi_obj* argv)
{
  ksi_long_jump (jmp, ksi_new_values (argc, argv));
  return ksi_void;
}

ksi_obj
ksi_new_continuation (ksi_stack here)
{
  struct Ksi_Jump *jmp = (struct Ksi_Jump*) ksi_malloc (sizeof *jmp);
  ksi_init_jump (jmp, KSI->stack, here);
  return ksi_close_prim (ksi_defun (0, cont_prim, KSI_CALL_REST1, 1),
			 1, (ksi_obj*) &jmp);
}

ksi_obj
ksi_continuation (ksi_obj *esc)
{
  struct Ksi_Jump *jmp = (struct Ksi_Jump*) ksi_malloc (sizeof *jmp);
  ksi_init_jump (jmp, KSI->stack, (ksi_stack) &jmp);

  *esc = ksi_close_prim (ksi_defun (0, cont_prim, KSI_CALL_REST1, 1),
			 1, (ksi_obj*) &jmp);

  FLUSH_REGISTER_WINDOWS;
  if (setjmp (jmp->j_buf))
    return KSI->jump_val;
  return 0;
}

ksi_wind
ksi_add_catch (ksi_obj tag, ksi_obj handler, int may_retry)
{
  ksi_wind wind;
  struct Ksi_Catch* the_catch;

  the_catch = (struct Ksi_Catch*) ksi_malloc (sizeof *the_catch);
  the_catch->tag = tag;
  the_catch->handler = handler;
  the_catch->retry = (may_retry ? ksi_false : 0);

  ksi_init_jump (&the_catch->jmp, 0, 0);

  wind = (ksi_wind) ksi_malloc (sizeof *wind);
  wind->cont = KSI -> wind;
  wind->the_catch = the_catch;
  KSI->wind = wind;

  return wind;
}


void
ksi_del_catch (ksi_wind wind)
{
  KSI_ASSERT (wind && wind == KSI->wind);
  KSI->wind = KSI->wind->cont;
}


ksi_wind
ksi_find_catch (ksi_obj tag)
{
  struct Ksi_Catch* the_catch;
  ksi_wind wind;
  ksi_obj x;

  for (wind = KSI->wind; wind; wind = wind->cont)
    {
      if (wind == KSI->exit_catch)
	return wind;

      the_catch = wind->the_catch;
      if (the_catch)
	{
	  if (the_catch->tag == ksi_true || the_catch->tag == tag)
	    return wind;

	  for (x = the_catch->tag; PAIR_P (x); x = CDR (x))
	    if (CAR (x) == tag)
	      return wind;
	}
    }

  return 0;
}


static ksi_obj
throw_prim (ksi_wind wind)
{
  ksi_obj tag = wind->the_catch->thrown;
  ksi_obj val = wind->the_catch->value;
  ksi_obj ret = wind->the_catch->retry;

  /* we are in post already */
  wind->post = 0;
  wind->the_catch->value = ksi_err;

  if (ret)
    {
      wind->the_catch->retry = ksi_false;
      val = ksi_apply_3 (wind->the_catch->handler, ret, tag, val);
    }
  else
    {
      val = ksi_apply_2 (wind->the_catch->handler, tag, val);
    }

  wind->the_catch->value = val;
  return ksi_void;
}

ksi_obj
ksi_throw_to_catch (ksi_wind wind, int retry, ksi_obj tag, ksi_obj val, ksi_rethrow_t rethrow)
{
  ksi_obj result;
  struct Ksi_Catch *the_catch;

  the_catch = wind->the_catch;
  KSI_ASSERT (the_catch != 0);

  the_catch->thrown  = tag;
  the_catch->value   = (val ? val : ksi_void);
  the_catch->rethrow = rethrow;

  if (the_catch->handler)
    {
      if (retry && the_catch->retry)
	{
	  result = ksi_continuation (&the_catch->retry);
	  if (result)
	    {
	      the_catch->retry = ksi_false;
	      return result;
	    }
	}

      wind->post = ksi_close_prim (ksi_defun (0, throw_prim, KSI_CALL_ARG1, 1),
				   1, (ksi_obj*) &wind);
    }

  ksi_long_jump (&the_catch->jmp, ksi_true);

  /* never returns */
  return 0;
}

ksi_obj
ksi_rethrow (ksi_wind wind)
{
  ksi_obj tag = wind->the_catch->thrown;
  ksi_obj val = wind->the_catch->value;

  switch (wind->the_catch->rethrow)
    {
    case KSI_RETHROW_NORMAL: return ksi_throw (tag, val);
    case KSI_RETHROW_ERROR:  return ksi_throw_error (val);
    case KSI_RETHROW_FATAL:  return ksi_throw_fatal (val);
    case KSI_RETHROW_EXIT:   return ksi_exit (val);
    }

  ksi_error ("ksi_rethrow: internal error");
  return 0;
}

int
ksi_handle_error (ksi_obj tag, ksi_obj exn)
{
  ksi_obj x;

  if (tag != ksi_err)
    exn = ksi_make_exn ("misc", tag, "uncatched thrown object", errlog_error);
  else if (!KSI_EXN_P (exn))
    exn = ksi_make_exn ("misc", exn, "unspecified error", errlog_error);

  if (KSI && KSI->error_handlers && KSI->error_handlers != ksi_nil)
    {
      for (x = KSI->error_handlers; x != ksi_nil; x = CDR (x))
	ksi_apply_1_with_catch (CAR (x), exn);
    }
  else
    {
      ksi_errlog_msg (KSI_SINT_COD (KSI_EXN_PRI (exn)),
		      KSI_STR_PTR (KSI_EXN_MSG (exn)));
    }

  return 0;
}

ksi_obj
ksi_throw_error (ksi_obj exn)
{
  ksi_wind wind;

  if (!exn || !KSI_EXN_P (exn))
    exn = ksi_make_exn ("misc", exn, "unspecified error", errlog_error);

  wind = ksi_find_catch (ksi_err);
  if (wind)
    ksi_throw_to_catch (wind, 0, ksi_err, exn, KSI_RETHROW_ERROR);

  ksi_handle_error (ksi_err, exn);
  ksi_term ();
  exit (KSI_ERROR_EXIT_CODE);

#ifndef __GNUC__
  return 0;
#endif
}

ksi_obj
ksi_throw_fatal (ksi_obj exn)
{
  if (!exn || !KSI_EXN_P (exn))
    exn = ksi_make_exn ("fatal", exn, "fatal error", errlog_alert);

  if (KSI->exit_catch)
    ksi_throw_to_catch (KSI->exit_catch, 0, ksi_err, exn, KSI_RETHROW_FATAL);

  ksi_handle_error (ksi_err, exn);
  ksi_term ();
  exit (KSI_ERROR_EXIT_CODE);

#ifndef __GNUC__
  return 0;
#endif
}

ksi_obj
ksi_exit (ksi_obj x)
{
  int code;

  if (x == ksi_null)
    x = ksi_true;

  if (x == ksi_false)
    code = KSI_ERROR_EXIT_CODE;
  else if (x == ksi_true)
    code = KSI_NORMAL_EXIT_CODE;
  else
    {
      TYPE_CHECK (x, KSI_SINT_P (x), "exit: invalid exit code");
      code = KSI_SINT_COD (x);
    }

  if (KSI && KSI->exit_catch)
    ksi_throw_to_catch (KSI->exit_catch, 0, KSI_MK_EXIT (code), x, KSI_RETHROW_EXIT);

  ksi_term ();
  exit (code);

#ifndef __GNUC__
  return ksi_void;
#endif
}

ksi_obj
ksi_quit (void)
{
  ksi_ignore (rcsid);

  if (KSI && KSI->exit_catch)
    ksi_throw_to_catch (KSI->exit_catch, 0, ksi_err, KSI->user_quit_exn, KSI_RETHROW_FATAL);

  ksi_handle_error (ksi_err, KSI->user_quit_exn);
  ksi_term ();
  exit (KSI_ERROR_EXIT_CODE);

#ifndef __GNUC__
  return ksi_void;
#endif
}


 /* End of code */
