/*
 * ksi_init.c
 * ksi initialization
 *
 * Copyright (C) 1997-2001,2002, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Thu Jul 31 23:14:21 1997
 * Last Update:   Sat Jan 19 20:08:12 2002
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_prim.h"
#include "ksi_sign.h"
#include "ksi_mod.h"
#include "ksi_mod.h"
#include "ksi_klos.h"

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#  include <fcntl.h>
#endif

#ifdef _MSC_VER
#  include <io.h>
#  define F_OK 0
#endif


static char rcsid[] = "$Id: ksi_init.c,v 1.17 2002/04/25 15:21:10 ksion Exp $";


static ksi_obj
ksi_copyright ()
{
  static char str [] = "Copyright (C) 1997-2002, Ivan Demakov <demakov@users.sourceforge.net>";
  return ksi_ptr02string (str);
}

static ksi_obj
ksi_banner ()
{
  static char str [] = "Welcome to the Ksi Scheme interpreter.";
  return ksi_ptr02string (str);
}

static ksi_obj
ksi_obj_version ()
{
  return ksi_ptr02string (KSI_VERSION);
}

static ksi_obj
ksi_obj_major_version ()
{
  return ksi_int2num (KSI_MAJOR_VERSION);
}

static ksi_obj
ksi_obj_minor_version ()
{
  return ksi_int2num (KSI_MINOR_VERSION);
}

static ksi_obj
ksi_obj_patch_level ()
{
  return ksi_int2num (KSI_PATCH_LEVEL);
}

static ksi_obj
ksi_obj_cpu ()
{
  return ksi_ptr02sym (KSI_CPU);
}

static ksi_obj
ksi_obj_os ()
{
#if defined(MSWIN32)
  return ksi_ptr02sym (ksi_win_name ());
#else
  return ksi_ptr02sym (KSI_OS);
#endif
}

static ksi_obj
ksi_obj_host ()
{
  return ksi_ptr02sym (KSI_HOST);
}

static ksi_obj
ksi_console_p ()
{
#if defined(MSWIN32)
  return KSI -> win_console ? ksi_true : ksi_false;
#else
  return ksi_true;
#endif
}

ksi_obj
ksi_interactive (ksi_obj x)
{
  int i = KSI->interactive;
  if (x != ksi_null)
    KSI->interactive = (x == ksi_false ? 0 : 1);

  return i ? ksi_true : ksi_false;
}

static ksi_obj
init_path ()
{
  return (KSI->ksi_init_path
	  ? ksi_ptr02string (KSI->ksi_init_path)
	  : ksi_false);
}

static ksi_obj
site_path ()
{
  return (KSI->ksi_site_path
	  ? ksi_ptr02string (KSI->ksi_site_path)
	  : ksi_false);
}

static ksi_obj
libs_path ()
{
  return (KSI->ksi_libs_path
	  ? ksi_ptr02string (KSI->ksi_libs_path)
	  : ksi_false);
}

static ksi_obj
app_path ()
{
  return (KSI->ksi_app_path
	  ? ksi_ptr02string (KSI->ksi_app_path)
	  : ksi_false);
}

static ksi_obj
app ()
{
  return KSI->ksi_app_sym;
}

static ksi_obj
inst_include_dir ()
{
#ifdef KSI_INSTAL_INCLUDE_DIR
  return ksi_ptr02string (KSI_INSTAL_INCLUDE_DIR);
#else
  return ksi_ptr02string ("/usr/local/include");
#endif
}

static ksi_obj
inst_lib_dir ()
{
#ifdef KSI_INSTAL_LIB_DIR
  return ksi_ptr02string (KSI_INSTAL_LIB_DIR);
#else
  return ksi_ptr02string ("/usr/local/lib");
#endif
}

static ksi_obj
inst_bin_dir ()
{
#ifdef KSI_INSTAL_BIN_DIR
  return ksi_ptr02string (KSI_INSTAL_BIN_DIR);
#else
  return ksi_ptr02string ("/usr/local/bin");
#endif
}

static ksi_obj
inst_info_dir ()
{
#ifdef KSI_INSTAL_INFO_DIR
  return ksi_ptr02string (KSI_INSTAL_INFO_DIR);
#else
  return ksi_ptr02string ("/usr/local/info");
#endif
}

static ksi_obj
inst_man_dir ()
{
#ifdef KSI_INSTAL_MAN_DIR
  return ksi_ptr02string (KSI_INSTAL_MAN_DIR);
#else
  return ksi_ptr02string ("/usr/local/man");
#endif
}

static ksi_obj
build_cflags ()
{
#ifdef KSI_BUILD_CFLAGS
  return ksi_ptr02string (KSI_BUILD_CFLAGS);
#else
  return ksi_ptr02string ("");
#endif
}

static ksi_obj
build_libs ()
{
#ifdef KSI_BUILD_LIBS
  return ksi_ptr02string (KSI_BUILD_LIBS);
#else
  return ksi_ptr02string ("");
#endif
}

static ksi_obj
loading_file ()
{
  return (KSI->loading_file
	  ? ksi_str02string (KSI->loading_file)
	  : ksi_false);
}

static struct Ksi_Prim_Def defs [] =
{
  { "ksi:version",		ksi_obj_version,	KSI_CALL_ARG0, 0 },
  { "ksi:major-version",	ksi_obj_major_version,	KSI_CALL_ARG0, 0 },
  { "ksi:minor-version",	ksi_obj_minor_version,	KSI_CALL_ARG0, 0 },
  { "ksi:patch-level",		ksi_obj_patch_level,	KSI_CALL_ARG0, 0 },
  { "ksi:cpu",			ksi_obj_cpu,		KSI_CALL_ARG0, 0 },
  { "ksi:os",			ksi_obj_os,		KSI_CALL_ARG0, 0 },
  { "ksi:host",			ksi_obj_host,		KSI_CALL_ARG0, 0 },
  { "ksi:copyright",		ksi_copyright,		KSI_CALL_ARG0, 0 },
  { "ksi:banner",		ksi_banner,		KSI_CALL_ARG0, 0 },
  { "ksi:interactive",		ksi_interactive,	KSI_CALL_ARG1, 0 },
  { "ksi:libs-path",		libs_path,		KSI_CALL_ARG0, 0 },
  { "ksi:init-path",		init_path,		KSI_CALL_ARG0, 0 },
  { "ksi:site-path",		site_path,		KSI_CALL_ARG0, 0 },
  { "ksi:app-path",		app_path,		KSI_CALL_ARG0, 0 },
  { "ksi:application",		app,			KSI_CALL_ARG0, 0 },
  { "ksi:console?",		ksi_console_p,		KSI_CALL_ARG0, 0 },
  { "ksi:loading-file",		loading_file,		KSI_CALL_ARG0, 0 },

  { "ksi:install-include-dir",	inst_include_dir,	KSI_CALL_ARG0, 0 },
  { "ksi:install-lib-dir",	inst_lib_dir,		KSI_CALL_ARG0, 0 },
  { "ksi:install-bin-dir",	inst_bin_dir,		KSI_CALL_ARG0, 0 },
  { "ksi:install-info-dir",	inst_info_dir,		KSI_CALL_ARG0, 0 },
  { "ksi:install-man-dir",	inst_man_dir,		KSI_CALL_ARG0, 0 },
  { "ksi:build-cflags",		build_cflags,		KSI_CALL_ARG0, 0 },
  { "ksi:build-libs",		build_libs,		KSI_CALL_ARG0, 0 },

  { 0 }
};

void
ksi_init_interp ()
{
  BASE_CHECK_INIT;

  ksi_init_symtab ();
  ksi_init_signals ();
  ksi_init_events ();
  ksi_init_symbols ();
  ksi_init_bools ();
  ksi_init_chars ();
  ksi_init_strings ();
  ksi_init_keywords ();
  ksi_init_numbers ();
  ksi_init_pairs ();
  ksi_init_vectors ();
  ksi_init_ports ();
  ksi_init_misc ();
  ksi_init_procs ();
  ksi_init_ctrl ();
  ksi_init_env ();
  ksi_init_load ();
  ksi_init_hash ();
  ksi_init_time ();
  ksi_init_error ();
  ksi_init_shel ();
  ksi_init_read ();
  ksi_init_write ();
  ksi_init_dump ();
  ksi_init_comp ();
  ksi_init_eval ();
  ksi_init_socket ();
  ksi_init_klos ();

#ifdef DYNAMIC_LINKING
  ksi_init_dynl ();
#endif

  ksi_init_modules ();

  KSI->sym_and			= ksi_ptr02sym ("and");
  KSI->sym_begin		= ksi_ptr02sym ("begin");
  KSI->sym_case			= ksi_ptr02sym ("case");
  KSI->sym_cond			= ksi_ptr02sym ("cond");
  KSI->sym_define		= ksi_ptr02sym ("define");
  KSI->sym_delay		= ksi_ptr02sym ("delay");
  KSI->sym_do			= ksi_ptr02sym ("do");
  KSI->sym_else			= ksi_ptr02sym ("else");
  KSI->sym_if			= ksi_ptr02sym ("if");
  KSI->sym_lambda		= ksi_ptr02sym ("lambda");
  KSI->sym_frame		= ksi_ptr02sym ("frame");
  KSI->sym_let			= ksi_ptr02sym ("let");
  KSI->sym_letrec		= ksi_ptr02sym ("letrec");
  KSI->sym_letstar		= ksi_ptr02sym ("let*");
  KSI->sym_or			= ksi_ptr02sym ("or");
  KSI->sym_quasiquote		= ksi_ptr02sym ("quasiquote");
  KSI->sym_quote		= ksi_ptr02sym ("quote");
  KSI->sym_set			= ksi_ptr02sym ("set!");
  KSI->sym_syntax		= ksi_ptr02sym ("syntax");
  KSI->sym_unquote		= ksi_ptr02sym ("unquote");
  KSI->sym_unquote_splicing	= ksi_ptr02sym ("unquote-splicing");
  KSI->sym_arrow		= ksi_ptr02sym ("=>");
  KSI->sym_dots			= ksi_ptr02sym ("...");
  KSI->sym_call			= ksi_ptr02sym ("call");
  KSI->sym_locvar		= ksi_ptr02sym ("locvar");
  KSI->sym_local		= ksi_ptr02sym ("local");
  KSI->sym_global		= ksi_ptr02sym ("global");
  KSI->sym_freevar		= ksi_ptr02sym ("freevar");
  KSI->sym_function		= ksi_ptr02sym ("function");
  KSI->sym_prefix		= ksi_ptr02sym ("prefix");
  KSI->sym_postfix		= ksi_ptr02sym ("postfix");
  KSI->sym_cons			= ksi_ptr02sym ("cons");
  KSI->sym_car			= ksi_ptr02sym ("car");
  KSI->sym_cdr			= ksi_ptr02sym ("cdr");
  KSI->sym_not			= ksi_ptr02sym ("not");
  KSI->sym_eq_p			= ksi_ptr02sym ("eq?");
  KSI->sym_eqv_p		= ksi_ptr02sym ("eqv?");
  KSI->sym_equal_p		= ksi_ptr02sym ("equal?");
  KSI->sym_memq			= ksi_ptr02sym ("memq");
  KSI->sym_memv			= ksi_ptr02sym ("memv");
  KSI->sym_member		= ksi_ptr02sym ("member");
  KSI->sym_null_p		= ksi_ptr02sym ("null?");
  KSI->sym_pair_p		= ksi_ptr02sym ("pair?");
  KSI->sym_list_p		= ksi_ptr02sym ("list?");
  KSI->sym_list			= ksi_ptr02sym ("list");
  KSI->sym_apply		= ksi_ptr02sym ("apply");
  KSI->sym_call_cc		= ksi_ptr02sym ("call/cc");
  KSI->sym_call_vs		= ksi_ptr02sym ("call/vs");
  KSI->sym_vector_p		= ksi_ptr02sym ("vector?");
  KSI->sym_inactive		= ksi_ptr02sym ("inactive");
  KSI->sym_wait			= ksi_ptr02sym ("wait");
  KSI->sym_ready		= ksi_ptr02sym ("ready");
  KSI->sym_timeout		= ksi_ptr02sym ("timeout");

  ksi_reg_unit (defs);

  ksi_defsym ("*load-path*",		ksi_nil, 0);
  ksi_defsym ("*system-expander*",	ksi_false, 0);
  ksi_defsym ("*load-verbose*",		ksi_false, 0);

  ksi_defsym ("*read-c-number*",	ksi_false, 0);
  ksi_defsym ("*read-sharp-hook*",	ksi_false, 0);
  ksi_defsym ("*read-case-sensitive*",	ksi_false, 0);
  ksi_defsym ("*read-keywords*",	ksi_false, 0);

  ksi_defsym ("*double-precision*",	KSI_MK_SINT (14), 0);
  ksi_defsym ("*print-closure-body*",	ksi_false, 0);
  ksi_defsym ("*print-pointer-body*",	ksi_false, 0);
  ksi_defsym ("*print-8bit-chars*",	ksi_false, 0);

  ksi_defsym ("*gensym-with-prefix*",	ksi_true, 0);
}

static void
ksi_init_lib_path (char* arg0)
{
  ksi_obj init_str, site_str, libs_str;
  char *lib, *app_lib = 0;

  lib = getenv (KSI_ENV_LIB);

#if defined(MSWIN32) && defined(KSIDLL)
  if (lib == 0)
    {
      int i;
      char *dir, *p;
      HMODULE hm = GetModuleHandle ("ksi.dll");
      if (hm)
	{
	  dir = (char*) ksi_malloc_data (MAX_PATH_LENGTH);
	  if (GetModuleFileName (hm, dir, MAX_PATH_LENGTH) > 0)
	    {
	      /* find the dll directory */ 
	      for (i = strlen (dir) - 1; i > 0; --i)
		if (IS_DIR_SUFFIX (dir[i]))
		  break;
	      dir[i] = '\0';

	      /* init-subdir exists in the dll directory? */
	      p = ksi_aprintf ("%s\\%d.%d",
			       dir,
			       KSI_MAJOR_VERSION, KSI_MINOR_VERSION);

	      if (access (p, F_OK) == 0)
		lib = ksi_expand_file_name (dir);
	      else
		{
		  /* now try parent directory */
		  dir = ksi_aprintf ("%s\\..", dir);
		  p   = ksi_aprintf ("%s\\%d.%d",
				     dir,
				     KSI_MAJOR_VERSION, KSI_MINOR_VERSION);

		  if (access (p, F_OK) == 0)
		    lib = ksi_expand_file_name (dir);
		}
	    }
	}
    }
#endif

#if defined(MSDOS) || defined(MSWIN32) || defined(OS2)
  if (lib)
    {
      char *p = ksi_aprintf ("%s\\%s\\%s", lib, APP_DIR, KSI->ksi_app);
      if (access (p, F_OK) == 0)
	app_lib = lib;
    }

  if (!app_lib && ABSOLUTE_PATH (arg0))
    {
      int i;
      char *p, *dir;

      /* find the executable directory */ 
      dir = (char*) ksi_malloc_data (strlen (arg0) + 1);
      strcpy (dir, arg0);
      for (i = strlen (dir) - 1; i > 0; --i)
	if (IS_DIR_SUFFIX (dir[i]))
	  break;
      dir[i] = '\0';

      /* app-dir exists in the executable directory? */
      p = ksi_aprintf ("%s\\%s\\%s", dir, APP_DIR, KSI->ksi_app);
      if (access (p, F_OK) == 0)
	app_lib = ksi_expand_file_name (app_lib);
      else
	{
	  /* now try parent directory */
	  dir = ksi_aprintf ("%s\\..", dir);
	  p   = ksi_aprintf ("%s\\%s\\%s", dir, APP_DIR, KSI->ksi_app);
	  if (access (p, F_OK) == 0)
	    app_lib = ksi_expand_file_name (app_lib);
	}

      if (lib == 0)
	lib = app_lib;
    }
#endif

  if (lib == 0)
    lib = KSI_LIB_DIR;
  if (app_lib == 0)
    app_lib = lib;

  KSI->ksi_libs_path = ksi_aprintf ("%s", lib);
  KSI->ksi_site_path = ksi_aprintf ("%s%s%s", lib, DIR_SEP, SITE_DIR);
  KSI->ksi_app_path  = ksi_aprintf ("%s%s%s", app_lib, DIR_SEP, APP_DIR);
  KSI->ksi_init_path = ksi_aprintf ("%s%s%d.%d", lib, DIR_SEP,
				    KSI_MAJOR_VERSION, KSI_MINOR_VERSION);

  init_str = ksi_ptr02string (KSI->ksi_init_path);
  site_str = ksi_ptr02string (KSI->ksi_site_path);
  libs_str = ksi_ptr02string (KSI->ksi_libs_path);

  ksi_defsym ("*load-path*", LIST1 (init_str), 0);
}

void
ksi_init (const char* app, int argc, char** argv, ksi_stack top_ptr)
{
  struct Ksi_Init_Proc* init;
  ksi_obj tmp;
  int i;

  ksi_ignore (rcsid);
  BASE_CHECK_INIT;

  if (app)
    {
      KSI->ksi_app = (char*) ksi_malloc_data (strlen (app) + 1);
      strcpy ((char*) KSI->ksi_app, app);
      KSI->ksi_app_sym = ksi_ptr02sym (KSI->ksi_app);
    }

  KSI->stack = (top_ptr ? top_ptr : &argc);

  if (KSI_INITED)
    return;

  ksi_init_interp ();
  KSI->inited = 1;

  if (argc >= 2 && strcmp (argv [1], "-b") == 0)
    {
      ksi_defsym ("*load-path*", LIST1 (ksi_ptr02string (".")), 0);

      /* build arg-list in reverse order */
      tmp = ksi_nil;
      for (i = argc; i-- > 0; )
	tmp = ksi_cons (ksi_str02string (argv[i]), tmp);
      ksi_defsym ("*argv*", tmp, 0);

      for (init = KSI->boot_procs; init; init = init->next)
	init->proc (init->data);
      for (init = KSI->init_procs; init; init = init->next)
	init->proc (init->data);

      if (argc >= 3)
	ksi_load (ksi_str02string (argv [2]));
    }
  else
    {
      ksi_init_lib_path (argc <= 0 ? 0 : argv[0]);

      /* build arg-list */
      tmp = ksi_nil;
      while (argc-- > 0)
	tmp = ksi_cons (ksi_str02string (argv[argc]), tmp);
      ksi_defsym ("*argv*", tmp, 0);

      for (init = KSI->boot_procs; init; init = init->next)
	init->proc (init->data);

      ksi_load_init_file (KSI_BOOT_FILE);
      ksi_load_init_file (KSI_KLOS_FILE);

      /* Modules should be initialized last */
      ksi_load_init_file (KSI_MODULES_FILE);
      if (!MOD_P (KSI->curr_module))
	{
	  KSI->curr_module = ksi_make_module (ksi_ptr02sym ("top"), ksi_nil, ksi_nil, 0);
	  ksi_set_module_opt (KSI->curr_module, ksi_ptr02sym ("loaded"), ksi_true);
	}

      for (init = KSI->init_procs; init; init = init -> next)
	init->proc (init->data);

      /* last stage of initializing */
      KSI->inited = 2;
      ksi_load_init_file (KSI_INIT_FILE);
    }

  KSI->inited = 3;
}

void
ksi_term (void)
{
  if (!KSI || !KSI->inited)
    return;

  KSI->inited = 0;
  while (KSI->exit_handlers != ksi_nil)
    {
      ksi_obj proc = KSI_CAR (KSI->exit_handlers);
      KSI->exit_handlers = KSI_CDR (KSI->exit_handlers);
      ksi_apply_0 (proc);
    }

#ifdef DYNAMIC_LINKING
  ksi_term_dynl ();
#endif
  ksi_term_ports ();
  ksi_term_events ();
  ksi_term_signals ();

  ksi_free (KSI);
  KSI = 0;
  ksi_gcollect (1);
}


 /* End of code */
