/* -*-mode:C++-*- */
/*
 * ksi_int.h
 * Internal structs
 *
 * Copyright (C) 1997-2000,2001,2002, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Thu Jul 31 19:19:17 1997
 * Last Update:   Sat Jan 19 20:07:38 2002
 *
 *
 * $Id: ksi_int.h,v 1.21 2002/01/19 20:27:15 ksion Exp $
 *
 */

#ifndef KSI_INT_H
#define KSI_INT_H

#define KSI_IMPL_DEFS 1

#include "ksi.h"
#include "ksi_type.h"
#include "ksi_jump.h"
#include "ksi_prim.h"

#ifdef MSWIN32
#  include <windows.h>
#endif


struct Ksi_FdPort
{
  struct Ksi_Port	kp;
  char			*name;		/* port name */
  char			*r_buf;		/* read ahead buffer */
  char			*w_buf;		/* pending write buffer */
  ksi_obj		evt;            /* async event */
  int			fd;		/* file desc */
  int			pg_size;	/* min buffer size */
  int			r_pos;		/* current read position */
  int			r_len;		/* number of bytes in r_buf */
  int			w_num;		/* number of pending bytes */
  int			w_size;		/* size of w_buf */
  struct Ksi_FdPort	*next;		/* list of all fd ports */

#if defined(MSWIN32)
  const char*	(*error) (struct Ksi_FdPort*);
  int		(*close) (struct Ksi_FdPort*);
  int		(*r_ready) (struct Ksi_FdPort*, int wait);
  int		(*w_ready) (struct Ksi_FdPort*, int wait);
  int		(*read) (struct Ksi_FdPort*, char *buf, int len);
  int		(*write) (struct Ksi_FdPort*, const char *buf, int len);
  int		(*set_async) (struct Ksi_FdPort*, int async);
#endif
};

struct Ksi_StringPort
{
  struct Ksi_Port	kp;
  ksi_string		str;
  int			size, pos;
};

struct Ksi_Dyn_Lib
{
  struct Ksi_Dyn_Lib	*next;
  char			*name;
  void			*handle;
  unsigned		count;
};

struct Ksi_Stat_Lib
{
  struct Ksi_Stat_Lib	*next;
  void			*data;
  ksi_init_function	proc;
  char			name[1];
};

struct Ksi_Init_Proc
{
  struct Ksi_Init_Proc	*next;
  void			*data;
  ksi_init_function	proc;
};


struct Ksi_Interp
{
  volatile int		have_event;
  ksi_event_mgr		event_mgr;
  ksi_event		waiting_events, pending_events, active_events;

  int			inited;
  int			interactive;

  ksi_valtab		obarray;
  ksi_valtab		keyarray;

  char*			ksi_libs_path;
  char*			ksi_init_path;
  char*			ksi_site_path;
  char*			ksi_app_path;
  const char*		ksi_app;
  ksi_obj		ksi_app_sym;

  ksi_obj		sym_and;
  ksi_obj		sym_begin;
  ksi_obj		sym_case;
  ksi_obj		sym_cond;
  ksi_obj		sym_define;
  ksi_obj		sym_delay;
  ksi_obj		sym_do;
  ksi_obj		sym_else;
  ksi_obj		sym_if;
  ksi_obj		sym_lambda;
  ksi_obj		sym_frame;
  ksi_obj		sym_let;
  ksi_obj		sym_letstar;
  ksi_obj		sym_letrec;
  ksi_obj		sym_or;
  ksi_obj		sym_set;
  ksi_obj		sym_quote;
  ksi_obj		sym_quasiquote;
  ksi_obj		sym_unquote;
  ksi_obj		sym_unquote_splicing;
  ksi_obj		sym_arrow;
  ksi_obj		sym_dots;
  ksi_obj		sym_syntax;
  ksi_obj		sym_call;
  ksi_obj		sym_locvar;
  ksi_obj		sym_local;
  ksi_obj		sym_global;
  ksi_obj		sym_freevar;
  ksi_obj		sym_function;
  ksi_obj		sym_prefix;
  ksi_obj		sym_postfix;
  ksi_obj		sym_cons;
  ksi_obj		sym_car;
  ksi_obj		sym_cdr;
  ksi_obj		sym_not;
  ksi_obj		sym_eq_p;
  ksi_obj		sym_eqv_p;
  ksi_obj		sym_equal_p;
  ksi_obj		sym_memq;
  ksi_obj		sym_memv;
  ksi_obj		sym_member;
  ksi_obj		sym_null_p;
  ksi_obj		sym_pair_p;
  ksi_obj		sym_list_p;
  ksi_obj		sym_list;
  ksi_obj		sym_apply;
  ksi_obj		sym_call_cc;
  ksi_obj		sym_call_vs;
  ksi_obj		sym_vector_p;
  ksi_obj		sym_inactive;
  ksi_obj		sym_wait;
  ksi_obj		sym_ready;
  ksi_obj		sym_timeout;

  ksi_obj		sym_apply_generic;
  ksi_obj		sym_no_next_method;
  ksi_obj		sym_no_applicable_method;
  ksi_obj		sym_proc_slot;
  ksi_obj		Top, Object, Class;
  ksi_obj		Generic, Method, Proc, Entity;
  ksi_obj		Boolean, Char, String, Symbol;
  ksi_obj		Number, Complex, Real, Integer;
  ksi_obj		Vector, Pair, List, Null;
  ksi_obj		Procedure, Keyword, Unknown;
  ksi_obj		Record, Record_rtd;

  struct Ksi_FdPort	*fd_ports;
  ksi_port		input_port, std_input_port;
  ksi_port		output_port, std_output_port;
  ksi_port		error_port, std_error_port;
  ksi_port		null_port;
  int			ports_inited;

  ksi_obj		void_proc;
  ksi_obj		true_proc;
  ksi_obj		false_proc;
  ksi_obj		list_proc;

  ksi_stack		stack;
  ksi_wind		wind;
  ksi_wind		exit_catch;
  ksi_obj		no_mem_exn;
  ksi_obj		user_quit_exn;
  ksi_obj		jump_val;

  ksi_obj		glob_env;
  ksi_obj		gensym_num;
  ksi_obj		protected_objs;
  ksi_obj		curr_module;
  ksi_obj		modules;

  struct Ksi_Stat_Lib	*stat_libs;
  struct Ksi_Dyn_Lib	*dyn_libs;
  struct Ksi_Init_Proc	*boot_procs;
  struct Ksi_Init_Proc	*init_procs;
  ksi_obj		exit_handlers;
  ksi_obj		error_handlers;

  ksi_obj		errlog_port;
  ksi_obj		module_priority;
  int			close_errlog_port;
  int			errlog_priority;

  int			load_indent;
  const char*		load_err_msg;
  ksi_port		curr_port;
  char*			loading_file;

#ifdef MSWIN32
  int			win_console;
#endif
};


/* initialization utils */

struct Ksi_Interp*
ksi_base_init (void);

void
ksi_init_interp (void);

void
ksi_init_gc (void);

void
ksi_init_time (void);

void
ksi_init_error (void);

void
ksi_init_shel (void);

void
ksi_init_read (void);

void
ksi_init_write (void);

void
ksi_init_symtab (void);

void
ksi_init_symbols (void);

void
ksi_init_bools (void);

void
ksi_init_chars (void);

void
ksi_init_strings (void);

void
ksi_init_numbers (void);

void
ksi_init_keywords (void);

void
ksi_init_pairs (void);

void
ksi_init_vectors (void);

void
ksi_init_ports (void);

void
ksi_term_ports (void);

void
ksi_init_procs (void);

void
ksi_init_misc (void);

void
ksi_init_syntax (void);

void
ksi_init_env (void);

void
ksi_init_ctrl (void);

void
ksi_init_load (void);

void
ksi_init_dynl (void);

void
ksi_term_dynl (void);

void
ksi_init_hash (void);

void
ksi_init_dump (void);

void
ksi_init_comp (void);

void
ksi_init_core (void);

void
ksi_init_simp (void);

void
ksi_init_eval (void);

void
ksi_init_socket (void);

void
ksi_term_socket (void);

void
ksi_init_events (void);

void
ksi_term_events (void);


/* Internal object creation:
 *	- `alloc' functions allocate memory for object;
 *	- `new' functions allocate and initialize object;
 */

ksi_obj
ksi_alloc_vector (int dim, int tag);

ksi_obj
ksi_new_vector (int argc, ksi_obj* argv);

ksi_obj
ksi_new_list (int argc, ksi_obj* argv);

ksi_obj
ksi_new_string (int argc, ksi_obj* argv);

ksi_obj
ksi_new_imm (int tag);

ksi_port
ksi_new_nul_port (void);

ksi_port
ksi_new_str_port (ksi_obj str);

ksi_port
ksi_new_fd_port (int fd, const char* name, int no_init);

ksi_obj
ksi_open_fd_port_int (const char* fname, const char* mode, const char* proc);

ksi_obj
ksi_new_values (int num, ksi_obj* val);

ksi_obj
ksi_new_continuation (ksi_stack here);

ksi_obj
ksi_continuation (ksi_obj *esc);

ksi_obj
ksi_new_quote (ksi_obj x);


/* misc utils */

unsigned
ksi_hash_obj (ksi_obj obj, unsigned n);

ksi_obj*
ksi_lookup_env (ksi_env env, ksi_obj sym);

ksi_obj*
ksi_append_env (ksi_env env, ksi_obj sym, ksi_obj val);

ksi_wind
ksi_add_catch (ksi_obj tag, ksi_obj handler, int may_retry);

void
ksi_del_catch (ksi_wind the_catch);

ksi_wind
ksi_find_catch (ksi_obj tag);

ksi_obj
ksi_throw_to_catch (ksi_wind wind, int retry, ksi_obj tag, ksi_obj val, ksi_rethrow_t rethrow);

ksi_obj
ksi_rethrow (ksi_wind wind);

int
ksi_handle_error (ksi_obj tag, ksi_obj exn);

void
ksi_errlog_msg (int pri, const char *msg);

const char*
ksi_dynload_file (char* fname);

const char*
ksi_mk_filename (ksi_obj x, char* name);

int
ksi_do_events (void);

void
ksi_cancel_port_events (ksi_port port);

ksi_obj
ksi_apply_prim (ksi_obj prim, int ac, ksi_obj *av);

ksi_obj
ksi_close_prim (ksi_obj prim, int argc, ksi_obj *argv);

ksi_obj
ksi_apply_prim_closure (ksi_obj clos, int argc, ksi_obj *argv);

ksi_obj
ksi_eval_sexp (ksi_obj sexp, ksi_obj env);


extern struct Ksi_Interp*	ksi_interp;
extern const char*		ksi_char_names[];
extern const char*		ksi_wna_s;


#define ksi_ignore(x) (void) (x)

#define KSI		ksi_interp

#define KSI_INITED	(ksi_interp && ksi_interp->inited)

#define BASE_CHECK_INIT				\
  do {						\
       if (!ksi_interp) ksi_base_init ();	\
     } while (0)

#define SYMTAB_CHECK_INIT						\
  do {									\
       if (!KSI->obarray) KSI->obarray = ksi_new_symtab (3000);	\
     } while (0)

#define KEYTAB_CHECK_INIT						\
  do {									\
       if (!KSI->keyarray) KSI->keyarray = ksi_new_keytab (300);	\
     } while (0)

#define CHECK_EVENTS do {if (KSI->have_event) ksi_do_events ();} while (0)


#ifdef KSI_STACK_GROWS_UP
#  define STK_CMP(a,b) ((ksi_stack) (a) > (ksi_stack) (b))
#else
#  define STK_CMP(a,b) ((ksi_stack) (a) < (ksi_stack) (b))
#endif


#define SYM_P		KSI_SYM_P
#define SYM_PTR		KSI_SYM_PTR
#define SYM_LEN		KSI_SYM_LEN

#define PAIR_P		KSI_PAIR_P
#define CAR		KSI_CAR
#define CDR		KSI_CDR
#define LIST_LEN	KSI_LIST_LEN

#define NEW_TYPED	KSI_NEW_TYPED
#define NEW_TYPED_EXT	KSI_NEW_TYPED_EXT
#define NEW_TYPED_ARR	KSI_NEW_TYPED_ARR

#define LIST1		KSI_LIST1
#define LIST2		KSI_LIST2
#define LIST3		KSI_LIST3
#define LIST4		KSI_LIST4

#define PROC_P(x)	(ksi_procedure_p ((x)) == ksi_true)

#ifdef __GNUC__
static ksi_obj inline
ksi_cons_inline (ksi_obj car, ksi_obj cdr)
{
  ksi_pair x = NEW_TYPED (struct Ksi_Pair, car, cdr);
  x -> itag = KSI_TAG_PAIR;
  x -> car = car;
  x -> cdr = cdr;
  return (ksi_obj) x;
}
#  define CONS(x,y)	ksi_cons_inline ((ksi_obj) (x), (ksi_obj) (y))
#else
#  define CONS(x,y)	ksi_cons ((ksi_obj) (x), (ksi_obj) (y))
#endif


#define TYPE_CHECK        KSI_TYPE_CHECK
#define RANGE_CHECK       KSI_RANGE_CHECK
#define WNA_CHECK(c,p,s)  ((void)((c)||ksi_exn_error("arity",p,ksi_wna_s,s)))


#endif

/* End of code */
