/*
 * ksi_key.c
 * keywords
 *
 * Copyright (C) 1997-2000, Ivan Demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Sat Jun 14 16:13:16 1997
 * Last Update:   Sat Jul 29 18:24:58 2000
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_vtab.h"

static char rcsid[] = "$Id: ksi_key.c,v 1.2 2000/08/14 18:07:58 demakov Exp $";


/*
 * ksi_str2key
 *	     NAME   LEN   
 *	 .     ,  
 *
 * Return
 *	    .
 *
 * Side effects
 *	      ,   ,
 *	       .
 */

ksi_obj
ksi_str2key (const char* name, int len)
{
  ksi_keyword key;

  BASE_CHECK_INIT;
  KEYTAB_CHECK_INIT;

  key = ksi_lookup_key (KSI->keyarray, name, len);
  if (key == 0)
    {
      key = (ksi_keyword) ksi_malloc_data (len + 1 + sizeof *key);
      key -> itag = KSI_TAG_KEYWORD;
      key -> len = len;
      key -> ptr = (char*) key + sizeof *key;
      if (len)
	memcpy ((char*) key -> ptr, name, len);
      ((char*) key -> ptr) [len] = 0;

      ksi_append_key (KSI->keyarray, key);
    }

  return (ksi_obj) key;
}


/*
 * ksi_ptr2key
 *	     NAME   LEN   
 *	 .     ,  
 *
 * Return
 *	    
 *
 * Side effects
 *	      ,   ,
 *	     .
 */

ksi_obj
ksi_ptr02key (const char* name)
{
  ksi_keyword key;
  int len = strlen (name);

  BASE_CHECK_INIT;
  KEYTAB_CHECK_INIT;

  key = ksi_lookup_key (KSI->keyarray, name, len);
  if (key == 0)
    {
      key = NEW_TYPED (struct Ksi_Keyword, ptr, ptr);
      key -> itag = KSI_TAG_KEYWORD;
      key -> len = len;
      key -> ptr = name;

      ksi_append_key (KSI->keyarray, key);
    }

  return (ksi_obj) key;
}

ksi_obj
ksi_key_p (ksi_obj x)
{
  return KSI_KEY_P (x) ? ksi_true : ksi_false;
}

const char*
ksi_key2str (ksi_obj key)
{
  int i, k, len;
  const char* name;
  char* ptr;

  TYPE_CHECK (key, KSI_KEY_P (key), "ksi_key2str: invalid keyword");

  len = KSI_KEY_LEN (key);
  name = KSI_KEY_PTR (key);
  ptr = (char*) ksi_malloc_data (len + 3);

  ptr [k=0] = '#';
  ptr [++k] = ':';
  for (i = 0; i < len; ++i)
    if (name[i])
      ptr [++k] = name [i];
  ptr [++k] = '\0';

  return ptr;
}

ksi_obj
ksi_string2keyword (ksi_obj str)
{
  TYPE_CHECK (str, KSI_STR_P (str), "string->keyword: invalid string");

  if (KSI_C_STR_P (str))
    return ksi_ptr02key (KSI_STR_PTR (str));
  return ksi_str2key (KSI_STR_PTR (str), KSI_STR_LEN (str));
}

ksi_obj
ksi_keyword2string (ksi_obj key)
{
  TYPE_CHECK (key, KSI_KEY_P (key), "keyword->string: invalid keyword");

  return ksi_ptr02string (KSI_KEY_PTR (key));
}

ksi_obj
ksi_symbol2keyword (ksi_obj sym)
{
  TYPE_CHECK (sym, KSI_SYM_P (sym), "symbol->keyword: invalid symbol");

  return ksi_ptr02key (KSI_SYM_PTR (sym));
}

ksi_obj
ksi_keyword2symbol (ksi_obj key)
{
  TYPE_CHECK (key, KSI_KEY_P (key), "keyword->symbol: invalid keyword");

  return ksi_ptr02sym (KSI_KEY_PTR (key));
}

ksi_obj
ksi_make_keyword (ksi_obj x)
{
  if (KSI_SYM_P (x))
    return ksi_symbol2keyword (x);
  if (KSI_STR_P (x))
    return ksi_string2keyword (x);
  if (KSI_KEY_P (x))
    return x;

  KSI_EXN_ERR ("type", x, "make-keyword: invalid object in arg1");
  return 0;
}

static struct Ksi_Prim_Def defs[] =
{
  { "keyword?",		ksi_key_p,		KSI_CALL_ARG1, 1 },
  { "string->keyword",	ksi_string2keyword,	KSI_CALL_ARG1, 1 },
  { "symbol->keyword",	ksi_symbol2keyword,	KSI_CALL_ARG1, 1 },
  { "keyword->string",	ksi_keyword2string,	KSI_CALL_ARG1, 1 },
  { "keyword->symbol",	ksi_keyword2symbol,	KSI_CALL_ARG1, 1 },
  { "make-keyword",	ksi_make_keyword,	KSI_CALL_ARG1, 1 },

  { 0 }
};

void
ksi_init_keywords ()
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}


 /* End of code */
