/*
 * ksi_klos.c
 * ksi object system
 *
 * Copyright (C) 1997-2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Thu Jun 12 19:50:51 1997
 * Last Update:   Sat Nov 10 00:48:47 2001
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_klos.h"

static char rcsid[] = "$Id: ksi_klos.c,v 1.9 2001/11/25 14:30:01 demakov Exp $";

/*
 * instance
 * --------
 *  (instance)  klos   Ksi_Instance.
 *     
 *  `klass' -    ;
 *  `flags' -  ;
 *  `slots' -  .
 *
 *   --      .
 *     (,    
 * ,   ,    --    
 * ).        `<class>'.
 *
 *  `flags'    --   
 * :
 *	I_CLASS		-   
 *	I_GENERIC	-   generic-
 *	I_METHOD	-   
 *	I_PURE_CLASS	-      
 *			       ,
 *			    
 *	I_PURE_GENERIC	-    --   <generic>.
 *			       
 *			  -     .
 *			  (   Stk).
 *
 * I_PURE_CLASS  I_PURE_GENERIC    , 
 *   I_CLASS  I_GENERIC 
 * (.  `ksi_alloc_instance').
 *
 *        (..
 * ,    ).
 *        (..   
 * ,     ,   ).
 *
 * 
 * -----
 *   klos      `<class>'.
 * (`<class>',   ,    --   
 *  ).
 *
 *   `<class>' (  )  klos   :
 *	name	-  ;
 *	dsupers	-   ;
 *	dslots	-    ( );
 *	defargs	-    ;
 *	cpl	-    (class precedence list);
 *	nfields	-      ;
 *	slots	-     ,   
 *		  ,     --  ,
 *		    --    (  );
 *	gns	- ,     .
 *		       ,  
 *		       5.
 *		  1   --  ,
 *                2                 -- allocation
 *		  2                 -- getter
 *		  3                 -- setter
 *		  4                 --  ,
 *		  5                 --  ,
 *
 *   --   (   ,   
 *    ,      ,
 *   , ,   "klos.scm",  
 *   (`define-class'  ..)   ).
 *
 * allocation --      
 *
 * Getter  setter --      (
 *     ),    ,
 * .
 *
 *  --  ,   , 
 *        
 * ,   ,      .
 *
 *       
 *  .      ,  
 *  ,   `type?'  
 *    . (.    `type?').
 *
 *    `name', `dsupers'  `dslots'  
 *   .     
 *   .     klos  
 * ,      
 * (   )  ,  
 *     ,  , 
 *   .
 *
 *
 *  .
 * --------------
 *  ``type? VAL TYPE''     VAL
 *  TYPE.
 *
 *        TYPE           (`type?' VAL TYPE)
 *    -----------       ----------------
 *	#t		#t
 *	#f		#f
 *	class		(subclass? (class-of VAL) TYPE)
 *	procedure	(TYPE VAL)
 *	list	        (memv VAL TYPE)
 *
 *  , 
 *
 *  -  TYPE  ,      
 *    ;
 *
 *  -  TYPE --  ,      , 
 *     VAL --       ;
 *
 *  -  TYPE --  ,    
 *          VAL;
 *
 *  -  TYPE --  ,       ,
 *     VAL  (  eqv?)    
 *    .
 */


#define CONCAT(a,b) a##b

#define CLASS_SLOT_REF(x,i,s) (INST_IS ((x), I_PURE_CLASS) ?	\
			       SLOT_REF ((x), i) :		\
			       ksi_slot_ref ((x), s))

#define GF_SLOT_REF(x,i,s) (INST_IS ((x), I_PURE_GENERIC) ?	\
			    SLOT_REF ((x), i) :			\
			    ksi_slot_ref ((x), s))

#define METHOD_SLOT_REF(x,i,s) (INST_IS ((x), I_PURE_METHOD) ?	\
			        SLOT_REF ((x), i) :		\
			        ksi_slot_ref ((x), s))

#define CNAME_OF(x)	CLASS_SLOT_REF ((x), S_CNAME, sym_cname)
#define DSUPERS_OF(x)	CLASS_SLOT_REF ((x), S_DSUPERS, sym_dsupers)
#define DSLOTS_OF(x)	CLASS_SLOT_REF ((x), S_DSLOTS, sym_dslots)
#define DEFARGS_OF(x)	CLASS_SLOT_REF ((x), S_DEFARGS, sym_defargs)
#define CPL_OF(x)	CLASS_SLOT_REF ((x), S_CPL, sym_cpl)
#define SLOTS_OF(x)	CLASS_SLOT_REF ((x), S_SLOTS, sym_slots)
#define NFIELDS_OF(x)	CLASS_SLOT_REF ((x), S_NFIELDS, sym_nfields)
#define GNS_OF(x)	CLASS_SLOT_REF ((x), S_GNS, sym_gns)

#define GNAME_OF(x)	GF_SLOT_REF ((x), S_GNAME, sym_gname)
#define METHODS_OF(x)	GF_SLOT_REF ((x), S_METHODS, sym_methods)
#define ARITY_OF(x)	GF_SLOT_REF ((x), S_ARITY, sym_arity)

#define GF_OF(x)	METHOD_SLOT_REF ((x), S_GF, sym_gf)
#define SPECS_OF(x)	METHOD_SLOT_REF ((x), S_SPECS, sym_specs)
#define PROC_OF(x)	METHOD_SLOT_REF ((x), S_PROC, sym_proc)
#define COMBINATION_OF(x) METHOD_SLOT_REF ((x), S_COMBINATION, sym_combination)
#define SUBCLASS_P(x,c)	(ksi_memq (c, CPL_OF (x)) != ksi_false)

#define S_cname		"name"
#define S_dsupers	"direct-supers"
#define S_dslots	"direct-slots"
#define S_defargs	"default-initargs"
#define S_cpl		"cpl"
#define S_slots		"slots"
#define S_nfields	"nfields"
#define S_gns		"get-n-set"

#define S_gname		"name"
#define S_methods	"methods"
#define S_arity		"arity"

#define S_gf		"generic-function"
#define S_specs		"specializers"
#define S_proc		"procedure"
#define S_combination	"combination"

#define K_initform      "initform"
#define K_initarg       "initarg"
#define K_defargs	"default-initargs"
#define K_type          "type"
#define K_name          "name"
#define K_dsupers       "supers"
#define K_dslots        "slots"
#define K_specs         "specializers"
#define K_proc          "procedure"
#define K_gf            "generic-function"
#define K_arity		"arity"
#define K_combination	"combination"
#define K_before	"before"
#define K_after		"after"
#define K_around	"around"
#define K_primary	"primary"

static ksi_obj sym_cname;
static ksi_obj sym_dsupers;
static ksi_obj sym_dslots;
static ksi_obj sym_defargs;
static ksi_obj sym_cpl;
static ksi_obj sym_slots;
static ksi_obj sym_nfields;
static ksi_obj sym_gns;

static ksi_obj sym_gname;
static ksi_obj sym_methods;
static ksi_obj sym_arity;

static ksi_obj sym_gf;
static ksi_obj sym_specs;
static ksi_obj sym_proc;
static ksi_obj sym_combination;

static ksi_obj key_initform;
static ksi_obj key_initarg;
static ksi_obj key_defargs;
static ksi_obj key_type;
static ksi_obj key_name;
static ksi_obj key_dsupers;
static ksi_obj key_dslots;
static ksi_obj key_specs;
static ksi_obj key_proc;
static ksi_obj key_gf;
static ksi_obj key_arity;
static ksi_obj key_combination;
static ksi_obj key_after;
static ksi_obj key_before;
static ksi_obj key_around;
static ksi_obj key_primary;


#define Top		KSI->Top
#define Object		KSI->Object
#define Class		KSI->Class
#define Generic		KSI->Generic
#define Method		KSI->Method
#define Proc		KSI->Proc
#define Entity		KSI->Entity
#define Boolean		KSI->Boolean
#define Char		KSI->Char
#define String		KSI->String
#define Symbol		KSI->Symbol
#define Number		KSI->Number
#define Complex		KSI->Complex
#define Real		KSI->Real
#define Integer		KSI->Integer
#define Vector		KSI->Vector
#define Pair		KSI->Pair
#define List		KSI->List
#define Null		KSI->Null
#define Procedure	KSI->Procedure
#define Keyword		KSI->Keyword
#define Unknown		KSI->Unknown
#define Record		KSI->Record
#define Record_rtd	KSI->Record_rtd


/* ksi_get_arg
 *
 * Search KEY in arg-lists ARGS.
 *
 * Returns value for key, or ksi_void if not found.
 */

ksi_obj
ksi_get_arg (ksi_obj key, ksi_obj args, ksi_obj def)
{
  if (key == ksi_void || args == ksi_nil || args == ksi_false)
    return def ? def : ksi_void;

  while (PAIR_P (key))
    {
      if (CDR (key) == ksi_nil)
	key = CAR (key);
      else
	{
	  ksi_obj x = ksi_get_arg (CAR (key), args, 0);
	  if (x != ksi_void)
	    return x;
	  key = CDR (key);
	}
    }

  while (args != ksi_nil)
    {
      TYPE_CHECK (args, PAIR_P (args), "@get-arg: improper list in arg2");
      TYPE_CHECK (args, PAIR_P (CDR (args)), "@get-arg: no value for key");

      if (key == CAR (args))
	return CAR (CDR (args));

      args = CDR (CDR (args));
    }
  
  return def ? def : ksi_void;
}


/*
 * some klos predicates
 */

ksi_obj
ksi_instance_p (ksi_obj x)
{
  return INST_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_class_p (ksi_obj x)
{
  return CLASS_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_generic_p (ksi_obj x)
{
  return GENERIC_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_method_p (ksi_obj x)
{
  return METHOD_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_new_next (ksi_obj gf, ksi_obj args, ksi_obj procs)
{
  ksi_next_method next;
  next = NEW_TYPED (struct Ksi_NextMethod, procs, gf);
  next -> itag = KSI_TAG_NEXT_METHOD;
  next -> gf = gf;
  next -> args = args;
  next -> procs = procs;
  return (ksi_obj) next;
}

/* ksi_new_instance
 *	allocate object ``instance''
 */

static ksi_instance
ksi_new_instance (ksi_instance cls, int n, int flags)
{
  ksi_instance inst;
  int i = sizeof *inst + (n * sizeof (ksi_obj));

  inst = (ksi_instance) ksi_malloc (i);
  inst -> itag = KSI_TAG_INSTANCE;
  inst -> flags = flags;
  inst -> klass = cls;
  inst -> slots = (ksi_obj*) (inst + 1);
  for (i = 0; i < n; i++)
    inst -> slots [i] = ksi_void;

  return inst;
}


/* ksi_obj build_gns (ksi_obj slot_list)
 * 	build `getters-n-seters' for given SLOT_LIST.
 *	This function doesn't handle slot options.
 *	It serves only for booting top classes.
 *	Real procedure will be defined in `klos.scm'.
 */

static ksi_obj
build_gns (ksi_obj slot_list)
{
  int i;
  ksi_obj res = ksi_nil;
  for (i = 0; slot_list != ksi_nil; i++)
    {
      ksi_obj slot = KSI_CAR (slot_list);
      ksi_obj name = slot;
      ksi_obj init = ksi_void;
      ksi_obj type = ksi_void;
      if (PAIR_P (slot))
	{
	  name = CAR (slot);
	  init = ksi_get_arg (key_initform, CDR (slot), 0);
	  type = ksi_get_arg (key_type, CDR (slot), 0);
	}
      if (type == ksi_void)
	type = ksi_true;

      slot = ksi_alloc_vector (GNS_SIZE, KSI_TAG_VECTOR);
      KSI_VEC_REF (slot, GNS_NAME)   = name;
      KSI_VEC_REF (slot, GNS_ALLOC)  = ksi_str02key ("instance");
      KSI_VEC_REF (slot, GNS_GETTER) = KSI_MK_SINT (i);
      KSI_VEC_REF (slot, GNS_SETTER) = KSI_MK_SINT (i);
      KSI_VEC_REF (slot, GNS_INIT)   = init;
      KSI_VEC_REF (slot, GNS_TYPE)   = type;

      res = ksi_cons (slot, res);
      slot_list = KSI_CDR (slot_list);
    }

  return res;
}


/* ksi_instance build_top (char* name, ksi_obj dsuper, ksi_instance klass)
 *	Allocate and initialize top class.
 *	Also variable NAME is bound to created class object.
 *	This function doesn't handle multiple inheritance,
 *	and classes that have slots.
 *	It serves only for booting top classes.
 *
 * Arguments:
 *	NAME	- name of class
 *	DSUPER	- superclass of created class
 *	klass	- class of created class (metaclass).
 */

static ksi_obj
build_top (char* name, ksi_obj dsuper, ksi_obj kls)
{
  ksi_instance res;
  ksi_obj tmp, cpl = ksi_nil, *loc = &cpl;

  for (tmp = dsuper; tmp != ksi_nil; /**/)
    {
      *loc = CONS (tmp, ksi_nil);
      loc = & (CDR (*loc));

      tmp = SLOT_REF (tmp, S_DSUPERS);
      if (tmp != ksi_nil)
	tmp = CAR (tmp);
    }

  res = ksi_new_instance ((ksi_instance) kls,
			  NUMBER_OF_CLASS_SLOTS,
			  I_CLASS | I_PURE_CLASS);

  SLOT_REF (res, S_CNAME)    = ksi_defsym (name, (ksi_obj) res, 0);
  SLOT_REF (res, S_DSUPERS)  = (dsuper == ksi_nil ? ksi_nil : LIST1 (dsuper));
  SLOT_REF (res, S_DSLOTS)   = ksi_nil;
  SLOT_REF (res, S_DEFARGS)  = ksi_nil;
  SLOT_REF (res, S_CPL)      = CONS (res, cpl);
  SLOT_REF (res, S_SLOTS)    = ksi_nil;
  SLOT_REF (res, S_NFIELDS)  = ksi_zero;
  SLOT_REF (res, S_GNS)      = ksi_nil;

  return (ksi_obj) res;
}


/* void init_top_classes
 *	Initialize top classes.
 */

static void
init_top_classes ()
{
  ksi_obj slots;

  /* alloc `<class>' class */
  Class = (ksi_obj) ksi_new_instance (0,
				      NUMBER_OF_CLASS_SLOTS,
				      I_CLASS | I_PURE_CLASS);
  CLASS_OF (Class) = (ksi_instance) Class;

  /* build top classes */
  Top	    = build_top ("<top>",		ksi_nil, Class);
  Object    = build_top ("<object>",		Top,	Class);

  Boolean   = build_top ("<boolean>",		Top,	Class);
  Char      = build_top ("<char>",		Top,	Class);
  String    = build_top ("<string>",		Top,	Class);
  Symbol    = build_top ("<symbol>",		Top,	Class);
  Keyword   = build_top ("<keyword>",		Top,	Class);

  List      = build_top ("<list>",		Top,	Class);
  Pair      = build_top ("<pair>",		List,	Class);
  Null      = build_top ("<null>",		List,	Class);

  Vector    = build_top ("<vector>",		Top,	Class);

  Number    = build_top ("<number>",		Top,	Class);
  Complex   = build_top ("<complex>",		Number,	Class);
  Real      = build_top ("<real>",		Complex, Class);
  Integer   = build_top ("<integer>",		Real,	Class);

  Unknown   = build_top ("<unknown>",		Top,	Class);


  /* we can initialize `<class>' now. */
  slots = KSI_LIST8 (KSI_LIST5 (sym_cname,
				key_initarg, key_name,
				key_type, Symbol),
		     KSI_LIST5 (sym_dsupers,
				key_initarg, key_dsupers,
				key_type, List),
		     KSI_LIST5 (sym_dslots,
				key_initarg, key_dslots,
				key_type, List),
		     KSI_LIST7 (sym_defargs,
				key_initarg, key_defargs,
				key_initform, ksi_nil,
				key_type, List),
		     KSI_LIST3 (sym_cpl,
				key_type, List),
		     KSI_LIST3 (sym_slots,
				key_type, List),
		     KSI_LIST5 (sym_nfields,
				key_initform, ksi_zero,
				key_type, Integer),
		     KSI_LIST3 (sym_gns,
				key_type, List));

  SLOT_REF (Class, S_CNAME)	= ksi_defsym ("<class>", Class, 0);
  SLOT_REF (Class, S_DSUPERS)	= KSI_LIST1 (Object);
  SLOT_REF (Class, S_DSLOTS)	= slots;
  SLOT_REF (Class, S_DEFARGS)	= ksi_nil;
  SLOT_REF (Class, S_CPL)	= KSI_LIST3 (Class, Object, Top);
  SLOT_REF (Class, S_SLOTS)	= slots;
  SLOT_REF (Class, S_NFIELDS)	= KSI_MK_SINT (NUMBER_OF_CLASS_SLOTS);
  SLOT_REF (Class, S_GNS)	= build_gns (slots);


  /* build record classes */
  Record_rtd = build_top ("<record-rtd>", Class, Class);
  SLOT_REF (Record_rtd, S_SLOTS)	= slots;
  SLOT_REF (Record_rtd, S_NFIELDS)	= KSI_MK_SINT (NUMBER_OF_CLASS_SLOTS);
  SLOT_REF (Record_rtd, S_GNS)		= build_gns (slots);

  Record    = build_top ("<record>", Top, Class);

  /* build procedure classes */

  Proc = build_top ("<procedure-class>", Class, Class);
  SLOT_REF (Proc, S_SLOTS)	= slots;
  SLOT_REF (Proc, S_NFIELDS)	= KSI_MK_SINT (NUMBER_OF_CLASS_SLOTS);
  SLOT_REF (Proc, S_GNS)	= build_gns (slots);

  Entity = build_top ("<entity-class>", Proc, Class);
  SLOT_REF (Entity, S_SLOTS)	= slots;
  SLOT_REF (Entity, S_NFIELDS)	= KSI_MK_SINT (NUMBER_OF_CLASS_SLOTS);
  SLOT_REF (Entity, S_GNS)	= build_gns (slots);

  Procedure = build_top ("<procedure>", Top, Proc);

  /* build <method> */
  slots = KSI_LIST4 (KSI_LIST3 (sym_gf,
				key_initarg, key_gf),
		     KSI_LIST3 (sym_specs,
				key_initarg, key_specs),
		     KSI_LIST5 (sym_combination,
				key_initarg, key_combination,
				key_initform, key_primary),
		     KSI_LIST3 (sym_proc,
				key_initarg, key_proc));

  Method = build_top ("<method>", Object, Class);
  SLOT_REF (Method, S_DSLOTS)	= slots;
  SLOT_REF (Method, S_SLOTS)	= slots;
  SLOT_REF (Method, S_NFIELDS)	= KSI_MK_SINT (NUMBER_OF_METHOD_SLOTS);
  SLOT_REF (Method, S_GNS)	= build_gns (slots);

  /* build <generic> */
  slots = KSI_LIST3 (KSI_LIST3 (sym_gname,
				key_initarg, key_name),
		     KSI_LIST3 (sym_methods,
				key_initform, ksi_nil),
		     KSI_LIST5 (sym_arity,
				key_initarg, key_arity,
				key_initform, ksi_true));

  Generic = build_top ("<generic>", Object, Entity);
  SLOT_REF (Generic, S_DSLOTS)	= slots;
  SLOT_REF (Generic, S_SLOTS)	= slots;
  SLOT_REF (Generic, S_NFIELDS)	= KSI_MK_SINT (NUMBER_OF_GENERIC_SLOTS);
  SLOT_REF (Generic, S_GNS)	= build_gns (slots);
}


static ksi_obj
find_slot (ksi_obj gns, ksi_obj name)
{
  ksi_obj slot;

  for (/**/; gns != ksi_nil; gns = CDR (gns))
    {
      KSI_CHECK2 (gns, PAIR_P (gns), "find_slot: invalid gns");

      slot = CAR (gns);
      KSI_CHECK2 (gns, KSI_VEC_P (slot) && KSI_VEC_LEN (slot) >= GNS_SIZE,
		  "find_slot: invalid gns");

      if (KSI_VEC_REF (slot, GNS_NAME) == name)
	return slot;
    }

  return 0;
}

static ksi_obj
get_slot_value (ksi_obj obj, ksi_obj name, ksi_obj (*apply) (ksi_obj, ksi_obj))
{
  ksi_obj cls, slot, val;

  cls = (ksi_obj) CLASS_OF (obj);
  if (KSI_SINT_P (name))
    {
      int i = KSI_SINT_COD (name);
      if (0 <= i && i < KSI_SINT_COD (NFIELDS_OF (cls)))
	val = SLOT_REF (obj, i);
      else
	val = ksi_inst_slot_missing (cls, obj, name, 0);
    }
  else
    {
      slot = find_slot (GNS_OF (cls), name);
      if (!slot)
	val = ksi_inst_slot_missing (cls, obj, name, 0);
      else
	{
	  slot = KSI_VEC_REF (slot, GNS_GETTER);
	  if (KSI_SINT_P (slot))
	    val = SLOT_REF (obj, KSI_SINT_COD (slot));
	  else
	    val = apply (slot, obj);
	}
    }

  return val;
}


/* ksi_slot_ref
 */

ksi_obj
ksi_slot_ref (ksi_obj obj, ksi_obj name)
{
  ksi_obj val;
  TYPE_CHECK (obj, INST_P (obj), "slot-ref: invalid instance in arg1");

  val = get_slot_value (obj, name, ksi_apply_1);
  if (val == ksi_void)
    return ksi_inst_slot_unbound ((ksi_obj) CLASS_OF (obj), obj, name);

  return val;
}


ksi_obj
ksi_slot_set (ksi_obj obj, ksi_obj name, ksi_obj val)
{
  ksi_obj cls, slot, type;
  TYPE_CHECK (obj, INST_P (obj), "slot-set!: invalid instance in arg1");

  cls = (ksi_obj) CLASS_OF (obj);
  if (KSI_SINT_P (name))
    {
      int i = KSI_SINT_COD (name);
      if (0 <= i && i < KSI_SINT_COD (NFIELDS_OF (cls)))
	SLOT_REF (obj, KSI_SINT_COD (name)) = val;
      else
	ksi_inst_slot_missing (cls, obj, name, val);
    }
  else
    {
      slot = find_slot (GNS_OF (cls), name);
      if (!slot)
	ksi_inst_slot_missing (cls, obj, name, val);
      else
	{
	  type = KSI_VEC_REF (slot, GNS_TYPE);
	  slot = KSI_VEC_REF (slot, GNS_SETTER);

	  if (type != ksi_void && ksi_type_p (val, type) == ksi_false)
	    {
	      ksi_exn_error ("type", val,
			     "slot-set!: invalid value for slot `%s' in %s",
			     ksi_obj2str (name), ksi_obj2str (cls));
	    }

	  if (KSI_SINT_P (slot))
	    SLOT_REF (obj, KSI_SINT_COD (slot)) = val;
	  else
	    ksi_apply_2 (slot, obj, val);
	}
    }

  return ksi_void;
}

ksi_obj
ksi_slot_bound_p (ksi_obj obj, ksi_obj name)
{
  ksi_obj val;
  TYPE_CHECK (obj, INST_P (obj), "slot-bound?: invalid instance in arg1");

  val = get_slot_value (obj, name, ksi_apply_1_with_catch);
  if (val == ksi_void || KSI_EXN_P (val))
    return ksi_false;

  return ksi_true;
}


ksi_obj
ksi_slot_exist_p (ksi_obj obj, ksi_obj name)
{
  TYPE_CHECK (obj, INST_P (obj), "slot-exist?: invalid instance in arg1");
  return ksi_slot_exist_in_class_p ((ksi_obj) CLASS_OF (obj), name);
}


ksi_obj
ksi_slot_exist_in_class_p (ksi_obj cls, ksi_obj name)
{
  TYPE_CHECK (cls, CLASS_P (cls), "slot-exist-in-class?: invalid class in arg1");

  if (KSI_SINT_P (name))
    {
      int i = KSI_SINT_COD (name);
      return (0 <= i && i < KSI_SINT_COD (NFIELDS_OF (cls)) ? ksi_true : ksi_false);
    }

  return (find_slot (GNS_OF (cls), name) ? ksi_true : ksi_false);
}


/* slot_num
 *	return offset of slot NAME in instance slot array, if slot is
 *	local slot, and `-1' otherwise.
 */

static int
slot_num (ksi_obj gns, ksi_obj name)
{
  ksi_obj slot = find_slot (gns, name);
  if (slot)
    {
      ksi_obj getter = KSI_VEC_REF (slot, GNS_GETTER);
      ksi_obj setter = KSI_VEC_REF (slot, GNS_SETTER);
      if (KSI_SINT_P (getter) && KSI_SINT_P (setter))
	return KSI_SINT_COD (slot);
    }

  return -1;
}


/*
 * ksi_alloc_instance (ksi_obj cls)
 *	allocate instance of class CLS.
 *
 * Scheme procedure
 *	@allocate-instance cls
 *
 * Note:
 *	this procedure is used as standard procedure for allocating 
 *	instance in klos.
 */

static ksi_obj
ksi_alloc_instance (ksi_obj cls)
{
  int type, n;
  ksi_obj cpl;
  TYPE_CHECK (cls, CLASS_P (cls), "@allocate-instance: invalid class in arg1");

  /* first, compute type of instance */

  if (cls == Generic)
    type = I_PURE_GENERIC | I_GENERIC | I_METHODS_SORTED;
  else
    {
      type = 0;

      /* look type in cpl of class */
      for (cpl = CPL_OF (cls); cpl != ksi_nil; cpl = KSI_CDR (cpl))
	{
	  ksi_obj x = KSI_CAR (cpl);

	  if (x == Generic)
	    type |= I_GENERIC;
	  else if (x == Method)
	    type |= I_METHOD;
	  else if (x == Class)
	    type |= I_CLASS;
	}

      if (type & I_CLASS)
	{
	  /* if we can access all class slots via constant offsets
	   * in the instance slot array, set I_PURE_CLASS flag.
	   */
	  ksi_obj gets = GNS_OF (cls);
	  if (slot_num (gets, sym_cname) == S_CNAME
	      && slot_num (gets, sym_dsupers) == S_DSUPERS
	      && slot_num (gets, sym_dslots) == S_DSLOTS
	      && slot_num (gets, sym_defargs) == S_DEFARGS
	      && slot_num (gets, sym_cpl) == S_CPL
	      && slot_num (gets, sym_slots) == S_SLOTS
	      && slot_num (gets, sym_nfields) == S_NFIELDS
	      && slot_num (gets, sym_gns) == S_GNS)
	    {
	      type |= I_PURE_CLASS;
	    }
	}

      if (type & I_METHOD)
	{
	  ksi_obj gets = GNS_OF (cls);
	  if (slot_num (gets, sym_gf) == S_GF
	      && slot_num (gets, sym_specs) == S_SPECS
	      && slot_num (gets, sym_combination) == S_COMBINATION
	      && slot_num (gets, sym_proc) == S_PROC)
	    {
	      type |= I_PURE_METHOD;
	    }
	}
    }

  /* for last, allocate memory for instance */
  n = KSI_SINT_COD (NFIELDS_OF (cls));
  return (ksi_obj) ksi_new_instance ((ksi_instance) cls, n, type);
}


/*
 * ksi_modify_instance (ksi_obj old, ksi_obj new)
 *	Set class and slots of OLD the same as in NEW
 *
 * Scheme procedure
 *	@modify-instance! old new
 *
 * Note:
 *	this procedure is used as helper of change-class
 */

static ksi_obj
ksi_modify_instance (ksi_obj o, ksi_obj n)
{
  TYPE_CHECK (o, INST_P (o), "@modify-instance: invalid instance in arg1");
  TYPE_CHECK (n, INST_P (n), "@modify-instance: invalid instance in arg2");

  CLASS_OF (o) = CLASS_OF (n);
  SLOTS_PTR (o) = SLOTS_PTR (n);
  return o;
}


/*
 * Implementation of Pure Generic Functions
 */

/* specs_conform ARITY SPECS
 *
 * test if method specializers are conforms the generic function arity.
 */

static ksi_obj
specs_conform (ksi_obj arity, ksi_obj specs)
{
  ksi_obj a, s;

  while (PAIR_P (arity))
    {
      if (!PAIR_P (specs))
	return ksi_false;

      a = CAR (arity);
      s = CAR (specs);

      TYPE_CHECK (s, (s == ksi_true || CLASS_P (s)
		      || PROC_P (s) || KSI_LIST_P (s)),
		  "@add-method: invalid specializer");

      if (a != ksi_true)
	{
	  if (CLASS_P (a))
	    {
	      if (!CLASS_P (s) || !SUBCLASS_P (s, a))
		return ksi_false;
	    }
	  else if (PAIR_P (a))
	    {
	      if (!PAIR_P (s))
		return ksi_false;

	      while (PAIR_P (s))
		{
		  if (ksi_memv (CAR (s), a) == ksi_false)
		    return ksi_false;
		  s = CDR (s);
		}
	    }
	  else if (a != s)
	    return ksi_false;
	}

      arity = CDR (arity);
      specs = CDR (specs);
    }

  if (arity == ksi_nil && specs != ksi_nil)
    return ksi_false;
  return ksi_true;
}

static int
lists_congr (ksi_obj l1, ksi_obj l2)
{
  ksi_obj common = ksi_nil;
  ksi_obj uniq = ksi_nil;

  while (PAIR_P (l1))
    {
      if (ksi_memv (CAR (l1), l2) == ksi_false)
	uniq = CONS (CAR (l1), uniq);
      else
	common = CONS (CAR (l1), common);
      l1 = CDR (l1);
    }

  if (common == ksi_nil)
    return 0;

  if (uniq == ksi_nil)
    {
      while (l2 != ksi_nil)
	{
	  if (ksi_memv (CAR (l2), common) == ksi_false)
	    return -1;
	  l2 = CDR (l2);
	}

      return 1;
    }

  return -1;
}


/* more_specific_p M1 M2 ARGS
 *
 * Test if method M1 is more specific than method M2.
 * Both M1 and M2 should be different and applicable to argmuments ARGS,
 * otherwise results would be unpredictable.
 *
 * If ARGS is NULL, the procedure tries to check that M1 is more specific
 * than M2 for any possible arguments.  This can be done if a type of
 * an argument is not taken into account (corresponding specializer is not
 * a class), or if the corresponding specializer of one method
 * is a superclass of the specializer of another method.
 *
 * Returns
 *	1  if M1 is more specific than M2;
 *	0  if not;
 *     -1  if can't determine (this possible only if ARGS == 0);
 */

static int
more_specific_p (ksi_obj m1, ksi_obj m2, ksi_obj args)
{
  /*
   * Note:
   * Specializer list of M1 and M2 can have different length
   * (i.e. one can be longer than the other when we have a dotted
   * parameter list).  For instance, with the call:
   *
   *   (M 1)
   *
   * with the methods:
   *
   *   1) (define-method M (a . l) ....)
   *   2) (define-method M (a b . l) ...)
   *   3) (define-method M (a b) ....)
   *
   * we consider that the 2'nd method is more specific than 1'st,
   * and 3'rd is more specific than second.
   *
   */

  ksi_obj s1 = SPECS_OF (m1), s2 = SPECS_OF (m2);
  for (/**/; /**/; s1 = CDR (s1), s2 = CDR (s2))
    {
      ksi_obj cs1, cs2, l;

      if (s1 == ksi_nil) return 1;
      if (s2 == ksi_nil) return 0;
      if (!PAIR_P (s1)) return 0;
      if (!PAIR_P (s2)) return 1;

      cs1 = CAR (s1), cs2 = CAR (s2);
      if (cs1 != cs2)
        {
	  /* Only one method should have specializer that is not class.
	   * It is error if both method have non-class specializers.
	   * Method with list specializer is more specific and
	   * method with procedure specializer is less specific.
	   */

	  if (PAIR_P (cs1)) return 1;
	  if (PAIR_P (cs2)) return 0;
	  if (!CLASS_P (cs1) && !CLASS_P (cs2)) goto next;
	  if (!CLASS_P (cs1)) return 0; 
	  if (!CLASS_P (cs2)) return 1;

	  /* Now both specializers are a classes.
	   * More specific method is a method with the specializer
	   * that occur first in the class precedence list of the argument.
	   */

	  if (args)
	    {
	      l = CPL_OF (ksi_class_of (CAR (args)));
	      while (l != ksi_nil)
		{
		  if (cs1 == CAR (l)) return 1;
		  if (cs2 == CAR (l)) return 0;
		  l = CDR (l);
		}
	    }
	  else
	    {
	      if (SUBCLASS_P (cs1, cs2)) return 1;
	      if (SUBCLASS_P (cs2, cs1)) return 0;
	      return -1;
	    }

	  /* for applicable methods next line will never execute,
	   * because generic function cannot have two methods that
	   * doesn't have any difference in specializers.
	   */
          break;
        }
    next:
      if (args)
	args = CDR (args);
    }

  ksi_error ("@method-more-specific?: not applicable methods");
  return 0; /* should not occur! */
}

static ksi_obj
method_more_specific_p (ksi_obj m1, ksi_obj m2, ksi_obj al)
{
  TYPE_CHECK (m1, METHOD_P (m1), "@method-more-specific?: invalid method");
  TYPE_CHECK (m2, METHOD_P (m2), "@method-more-specific?: invalid method");
  TYPE_CHECK (al, KSI_LIST_P (al), "@method-more-specific?: invalid list");

  return more_specific_p (m1, m2, al) ? ksi_true : ksi_false;
}


/* applicable_p METHOD ARGS
 *
 * Test if the METHOD is applicable to ARGS.
 */

static int
applicable_p (ksi_obj method, ksi_obj args)
{
  ksi_obj specs = SPECS_OF (method);
  for (/**/; args != ksi_nil; specs = CDR (specs), args = CDR (args))
    {
      TYPE_CHECK (args, PAIR_P (args),
		  "@method-applicable?: invalid list in arg2");

      /* spec list ends but args not, the method isn't applicable */
      if (specs == ksi_nil)
	return 0;

      /* in case of dotted pair in spec list, 
       * the method is applicable to any rest args.
       */
      if (!PAIR_P (specs))
	return 1;

      if (ksi_type_p (CAR (args), CAR (specs)) == ksi_false)
	return 0;
    }

  /* The args ends; and now if spec list ends too or has dotted pair at
   * this point, the method is applicable, otherwise not.
   */
  return (PAIR_P (specs) ? 0 : 1);
}


/* ksi_applicable_p METHOD ARGS
 *
 * Scheme primitive to call C-function `applicable_p'.
 */

static ksi_obj
ksi_applicable_p (ksi_obj method, ksi_obj args)
{
  TYPE_CHECK (method, METHOD_P (method),
	      "@method-applicable?: invalid method in arg1");

  return applicable_p (method, args) ? ksi_true : ksi_false;
}


/* add_method GF METHOD
 *
 *  Append the METHOD to list of methods of generic function GF,
 *  if the list has not method with the same specializers,
 *  and replace existance nethod otherwise.
 *
 * Return the METHOD.
 *
 * Used from `klos.scm'.
 */

static ksi_obj
add_method (ksi_obj gf, ksi_obj method)
{
  int tmp;
  ksi_obj arity, specs, comb, mthds, l, l1, l2, s1, s2, *loc;

  TYPE_CHECK (gf, GENERIC_P (gf),
	      "@add-method: invalid generic function in arg1");
  TYPE_CHECK (method, METHOD_P (method),
	      "@add-method: invalid method in arg2");

  arity = ARITY_OF (gf);
  specs = SPECS_OF (method);
  if (specs_conform (arity, specs) == ksi_false)
    ksi_exn_error ("type", method, "@add-method: method is not conform %s",
		   ksi_obj2str (gf));

  comb = COMBINATION_OF (method);
  mthds = METHODS_OF (gf);
  for (l = mthds; l != ksi_nil; l = CDR (l))
    {
      if (comb != COMBINATION_OF (CAR (l)))
	continue;

      l1 = specs;
      l2 = SPECS_OF (CAR (l));
      for (/**/; /**/; l1 = CDR (l1), l2 = CDR (l2))
        {
	  if (!PAIR_P (l1) && !PAIR_P (l2) && l1 == l2)
            {
	      /* If both specializers lists ends or has dotted cdr at end,
	       * lists are congruent (i.e methods can't be distinguished).
	       * Old method is replaced by new.
	       */
              CAR (l) = method;
              return method;
            }

	  /* if any of spec list ends, spec lists are noncongruent. */
          if (!PAIR_P (l1) || !PAIR_P (l2))
            break;

	  s1 = CAR (l1);
	  s2 = CAR (l2);
	  if ((s1 == Top || s1 == ksi_true) && (s2 == Top || s2 == ksi_true))
	    continue;

	  if (s1 != s2)
	    {
	      /* If any (or both) specializer is a class, spec lists are
	       * noncongruent (the methods can be distinguised simply).
	       * If one or both specializers are procedures, spec lists are
	       * considered noncongruent (it is error to provide such
	       * procedure specializers that introduce conflicts).
	       * If both spesializers are lists and if all elements
	       * of the lists are equal, the methods are congruent.
	       * If all elements are not equal, the methods
	       * are noncongruent. Otherwise (some elements equal and
	       * some not equal), the methods are conflicting,
	       * and error signaled.
	       */

	      if (!PAIR_P (s1) || !PAIR_P (s2))
		break;

	      tmp = lists_congr (s1, s2);
	      if (tmp < 0)
		ksi_exn_error ("type", method,
			       "@add-method: method conflicts with %s",
			       ksi_obj2str (CAR (l)));
	      if (tmp == 0)
		break;
	    }
        }
    }

  /* If we are here, we have not encountered a method
   * with congruent specializers. (i.e. new method can (or, at least, should)
   * be distinguished from any other method of the generic function).
   */

  loc = &mthds;
  if ((((ksi_instance) gf) -> flags & I_METHODS_SORTED) != 0)
    {
      for (l = mthds; l != ksi_nil; l = CDR (l))
	{
	  if (comb != COMBINATION_OF (CAR (l)))
	    continue;

	  tmp = more_specific_p (method, CAR (l), 0);
	  if (tmp > 0)
	    loc = & CDR (l);
	  else if (tmp < 0)
	    {
	      ((ksi_instance) gf) -> flags &= ~I_METHODS_SORTED;
	      break;
	    }
	}
    }

  *loc = CONS (method, *loc);
  ksi_slot_set (gf, sym_methods, mthds);
  return method;
}


/* ksi_compute_applicable_methods GF ARGS COMBINATION
 *
 * Builds a list of applicable methods of GF for the argument-list ARGS.
 * In the list more specific methods preceed less specific methods.
 *
 * Return builden list, or ksi_nil if there are no one applicable method in 
 * generic function.
 */

ksi_obj
ksi_compute_applicable_methods (ksi_obj gf, ksi_obj args, ksi_obj combination)
{
  ksi_obj l, m, x, *loc, res;

  l = METHODS_OF (gf);

  /* find any applicable method */
  for (res = 0; l != ksi_nil; l = CDR (l))
    {
      m = CAR (l);
      if (COMBINATION_OF (m) == combination && applicable_p (m, args))
	{
	  res = m;
	  break;
	}
    }

  if (!res)
    return ksi_nil;

  res = CONS (res, ksi_nil);
  while ((l = CDR (l)) != ksi_nil)
    {
      m = CAR (l);
      if (COMBINATION_OF (m) == combination && applicable_p (m, args))
	{
	  if ((((ksi_instance) gf) -> flags & I_METHODS_SORTED) != 0)
	    {
	      KSI_ASSERT (more_specific_p (m, CAR (res), args));
	      res = CONS (m, res);
	    }
	  else
	    {
	      loc = &res;
	      for (x = res; x != ksi_nil; x = CDR (x))
		{
		  if (more_specific_p (m, CAR (x), args))
		    break;
		  loc = & CDR (x);
		}
	      *loc = CONS (m, *loc);
	    }
	}
    }

  return res;
}


/* compute_methods
 *
 * Scheme primitive that call ksi_compute_applicable_methods
 */

static ksi_obj
compute_methods (ksi_obj gf, ksi_obj args, ksi_obj combination)
{
  TYPE_CHECK (gf, GENERIC_P (gf),
	      "@compute-applicable-methods: invalid generic in arg1");
  TYPE_CHECK (args, KSI_LIST_P (args),
	      "@compute-applicable-methods: invalid list in arg2");

  return ksi_compute_applicable_methods (gf, args, combination);
}

static ksi_obj
get_procs (ksi_obj methods, int reverse)
{
  ksi_obj res = ksi_nil, *loc = &res;
  while (methods != ksi_nil)
    {
      ksi_obj proc = PROC_OF (CAR (methods));
      if (reverse)
	res = CONS (proc, res);
      else
	{
	  *loc = CONS (proc, ksi_nil);
	  loc = & CDR (*loc);
	}
      methods = CDR (methods);
    }
  return res;
}

static ksi_obj
call_first (ksi_obj data, int argc, ksi_obj* argv)
{
  ksi_obj proc, res;

  proc = CAR (data);
  while (proc != ksi_nil)
    {
      ksi_apply_proc (CAR (proc), argc-2, argv+2);
      proc = CDR (proc);
    }

  proc = CAR (CDR (data));
  res  = ksi_apply_proc (proc, argc, argv);

  proc = CAR (CDR (CDR (data)));
  while (proc != ksi_nil)
    {
      ksi_apply_proc (CAR (proc), argc-2, argv+2);
      proc = CDR (proc);
    }

  return res;
}

ksi_obj
ksi_compute_effective_method (ksi_obj gf, ksi_obj args)
{
  ksi_obj vals [3];
  ksi_obj primary, around, before, after;

  primary = ksi_compute_applicable_methods (gf, args, key_primary);
  around  = ksi_compute_applicable_methods (gf, args, key_around);
  if (primary == ksi_nil)
    {
      if (around == ksi_nil)
	return ksi_nil;

      return ksi_new_next (gf, args, get_procs (around, 0));
    }

  before = ksi_compute_applicable_methods (gf, args, key_before);
  after  = ksi_compute_applicable_methods (gf, args, key_after);
  if (before == ksi_nil && after == ksi_nil)
    {
      if (around == ksi_nil)
	return ksi_new_next (gf, args, get_procs (primary, 0));

      vals [0] = get_procs (around, 0);
      vals [1] = get_procs (primary, 0);
      return ksi_new_next (gf, args, ksi_append_x (2, vals));
    }

  primary = get_procs (primary, 0);
  around  = get_procs (around, 0);
  before  = get_procs (before, 0);
  after   = get_procs (after, 1);

  vals[0] = around;
  vals[2] = CDR (primary);

  primary = LIST3 (before, CAR (primary), after);
  primary = ksi_close_prim (ksi_defun (0, call_first, KSI_CALL_REST1, 3),
			    1, &primary);
  vals[1] = LIST1 (primary);

  return ksi_new_next (gf, args, ksi_append_x (3, vals));
}

ksi_obj
ksi_inst_slot_unbound (ksi_obj cls, ksi_obj obj, ksi_obj slot)
{
  ksi_obj sym = ksi_ptr02sym ("slot-unbound");
  ksi_obj proc = ksi_glob_val (sym, ksi_false);

  if (PROC_P (proc))
    return ksi_apply_3 (proc, cls, obj, slot);

  ksi_error ("slot-ref: unbound slot: `%s\' in %s",
	     ksi_obj2str (slot), ksi_obj2str (obj));

  return ksi_false;
}

ksi_obj
ksi_inst_slot_missing (ksi_obj cls, ksi_obj obj, ksi_obj slot, ksi_obj val)
{
  ksi_obj vals[4];
  ksi_obj sym = ksi_ptr02sym ("slot-missing");
  ksi_obj proc = ksi_glob_val (sym, ksi_false);

  if (PROC_P (proc))
    {
      vals[0] = cls;
      vals[1] = obj;
      vals[2] = slot;
      vals[3] = val;
      return ksi_apply_proc (proc, val ? 4 : 3, vals);
    }

  ksi_error ("%s: no slot with name `%s' in %s",
	     (val ? "slot-set!" : "slot-ref"), 
	     ksi_obj2str (slot), ksi_obj2str (cls));

  return ksi_false;
}

ksi_obj
ksi_inst_eqv_p (ksi_obj o1, ksi_obj o2)
{
  ksi_obj sym = ksi_ptr02sym ("instance-eqv?");
  ksi_obj proc = ksi_glob_val (sym, ksi_false);

  if (PROC_P (proc))
    return ksi_apply_2 (proc, o1, o2);

  return ksi_false;
}

ksi_obj
ksi_inst_equal_p (ksi_obj o1, ksi_obj o2)
{
  ksi_obj sym = ksi_ptr02sym ("instance-equal?");
  ksi_obj proc = ksi_glob_val (sym, ksi_false);

  if (PROC_P (proc))
    return ksi_apply_2 (proc, o1, o2);

  return ksi_false;
}

unsigned int
ksi_hash_inst (ksi_obj x, unsigned n)
{
  ksi_obj sym = ksi_ptr02sym ("hash-instance");
  ksi_obj proc = ksi_glob_val (sym, ksi_false);

  if (PROC_P (proc))
    {
      x = ksi_apply_2 (proc, x, KSI_MK_SINT (n));
      return KSI_SINT_COD (x);
    }

  return ((unsigned int) x) % n;
}

void
ksi_write_inst (ksi_obj x, ksi_port p, int slashify)
{
  ksi_obj sym, proc;

  sym  = ksi_ptr02sym (slashify ? "write-instance" : "display-instance");
  proc = ksi_glob_val (sym, ksi_false);

  if (PROC_P (proc))
    ksi_apply_2 (proc, x, (ksi_obj) p);
  else
    ksi_printf ((ksi_obj) p, "#<instance %p>", x);
}


/*
 * ksi_obj ksi_class_of (ksi_obj x)
 *	Return class of X.
 *	If X isn't instance, return one of predefined top classes.
 */

ksi_obj
ksi_class_of (ksi_obj x)
{
  if (INST_P (x))
    return (ksi_obj) CLASS_OF (x);

  if (KSI_BOOL_P (x))
    return Boolean;
  if (x == ksi_nil)
    return Null;
  if (KSI_PAIR_P (x))
    return Pair;
  if (KSI_CHAR_P (x))
    return Char;
  if (KSI_SYM_P (x))
    return Symbol;
  if (KSI_KEY_P (x))
    return Keyword;
  if (KSI_STR_P (x))
    return String;
  if (KSI_VEC_P (x))
    return Vector;
  if (KSI_SINT_P (x) || KSI_BIGNUM_P (x))
    return Integer;
  if (KSI_FLONUM_P (x))
    return Real;
  if (KSI_COMPLEX_P (x))
    return Complex;
  if (PROC_P (x))
    return Procedure;
  return Unknown;
}


/* ksi_obj ksi_type_of (ksi_obj x)
 *
 * Returns a symbol that names the type of X.
 */

ksi_obj
ksi_type_of (ksi_obj x)
{
  if (x == ksi_nil)
    return ksi_ptr02sym ("null");
  if (x == ksi_eof)
    return ksi_ptr02sym ("eof-object");
  if (KSI_BOOL_P (x))
    return ksi_ptr02sym ("boolean");
  if (KSI_CHAR_P (x))
    return ksi_ptr02sym ("char");
  if (KSI_NUM_P (x))
    return ksi_ptr02sym ("number");
  if (KSI_STR_P (x))
    return ksi_ptr02sym ("string");
  if (KSI_SYM_P (x))
    return ksi_ptr02sym ("symbol");
  if (KSI_KEY_P (x))
    return ksi_ptr02sym ("keyword");
  if (KSI_PORT_P (x))
    return ksi_ptr02sym ("port");
  if (KSI_PAIR_P (x))
    {
      if (KSI_LIST_P (x))
	return ksi_ptr02sym ("list");
      return ksi_ptr02sym ("pair");
    }
  if (KSI_VEC_P (x))
    return ksi_ptr02sym ("vector");
  if (PROC_P (x))
    return ksi_ptr02sym ("procedure");
  if (INST_P (x))
    {
      if (CLASS_P (x))
	return ksi_ptr02sym ("class");
      if (SUBCLASS_P (ksi_class_of (x), Record))
	return ksi_ptr02sym ("record");
      return ksi_ptr02sym ("instance");
    }

  return ksi_ptr02sym ("unknown");
}


/* ksi_type_p X TYPE
 *
 * Test if X has the type TYPE.
 *
 * Used to test a value in ksi_slot_set,
 * and to determine applicabability of methods.
 */

ksi_obj
ksi_type_p (ksi_obj x, ksi_obj type)
{
  if (KSI_BOOL_P (type))
    return type;

  if (CLASS_P (type))
    {
      if (INST_P (x))
	return (SUBCLASS_P ((ksi_obj) CLASS_OF (x), type)
		? ksi_true : ksi_false);
      return SUBCLASS_P (ksi_class_of (x), type) ? ksi_true : ksi_false;
    }

  if (PROC_P (type))
    return ksi_apply_1 (type, x);

  while (PAIR_P (type))
    {
      if (KSI_EQV_P (x, CAR (type)))
	return ksi_true;
      type = CDR (type);
    }

  return ksi_false;
}


/*
 * Some useful getters for <class> slots.
 * 
 * Note, that this getters are primitives, not generic functions.
 * And so, they can be used while generic functions are not initialized yet.
 */

#define DEF_GETTER(c,n,s,i)					\
static ksi_obj							\
CONCAT(class_,c) (ksi_obj obj)					\
{								\
  if (!CLASS_P (obj))						\
    ksi_exn_error ("type", obj, "class-%s: invalid class:", n);	\
  if (INST_IS (obj, I_PURE_CLASS))				\
    return SLOT_REF (obj, i);					\
  return ksi_slot_ref (obj, s);					\
}

DEF_GETTER (name, S_cname, sym_cname, S_CNAME)
DEF_GETTER (dsupers, S_dsupers, sym_dsupers, S_DSUPERS)
DEF_GETTER (dslots, S_dslots, sym_dslots, S_DSLOTS)
DEF_GETTER (defargs, S_defargs, sym_defargs, S_DEFARGS)
DEF_GETTER (cpl, S_cpl, sym_cpl, S_CPL)
DEF_GETTER (slots, S_slots, sym_slots, S_SLOTS)

static ksi_obj
class_get_n_set (ksi_obj cls, ksi_obj slot)
{
  ksi_obj gns;

  TYPE_CHECK (cls, CLASS_P (cls), "class-get-n-set: invalid class in arg1");

  gns = GNS_OF (cls);
  if (slot)
    gns = find_slot (gns, slot);

  return gns ? gns : ksi_false;
}


/* ksi_merge_args
 *
 * Merge argument lists.
 * Argument list is devided in parts of two elements.
 * In each part the first element is a key and the second is a value.
 * Key ussually is a symbol or a keyword.  Value may be any type.
 */

static ksi_obj
ksi_merge_args (int argc, ksi_obj* argv)
{
  ksi_obj l, x, res = ksi_nil;
  while (--argc >= 0)
    {
      for (l = *argv++; l != ksi_nil; l = CDR (CDR (l)))
	{
	  TYPE_CHECK (l, PAIR_P (l), "@merge-args: improper list");
	  TYPE_CHECK (l, PAIR_P (CDR (l)), "@merge-args: no value for key");

	  for (x = res; x != ksi_nil; x = CDR (CDR (x)))
	    if (CAR (x) == CAR (l))
	      goto next;

	  res = CONS (CAR (l), CONS (CAR (CDR (l)), res));
	next:;
	}
    }

  return res;
}


static struct Ksi_Prim_Def defs [] =
{
  { "instance?",		ksi_instance_p,			KSI_CALL_ARG1, 1 },
  { "class?",			ksi_class_p,			KSI_CALL_ARG1, 1 },
  { "generic?",			ksi_generic_p,			KSI_CALL_ARG1, 1 },
  { "method?",			ksi_method_p,			KSI_CALL_ARG1, 1 },

  { "class-of",			ksi_class_of,			KSI_CALL_ARG1, 1 },
  { "type-of",			ksi_type_of,			KSI_CALL_ARG1, 1 },
  { "type?",			ksi_type_p,			KSI_CALL_ARG2, 2 },

  { "slot-ref",			ksi_slot_ref,			KSI_CALL_ARG2, 2 },
  { "slot-set!",		ksi_slot_set,			KSI_CALL_ARG3, 3 },
  { "slot-bound?",		ksi_slot_bound_p,		KSI_CALL_ARG2, 2 },
  { "slot-exist?",		ksi_slot_exist_p,		KSI_CALL_ARG2, 2 },
  { "slot-exist-in-class?",	ksi_slot_exist_in_class_p,	KSI_CALL_ARG2, 2 },

  { "class-name",		class_name,			KSI_CALL_ARG1, 1 },
  { "class-direct-supers", 	class_dsupers,			KSI_CALL_ARG1, 1 },
  { "class-direct-slots",	class_dslots,			KSI_CALL_ARG1, 1 },
  { "class-precedence-list",	class_cpl,			KSI_CALL_ARG1, 1 },
  { "class-slots",		class_slots,			KSI_CALL_ARG1, 1 },
  { "class-default-initargs",	class_defargs,			KSI_CALL_ARG1, 1 },
  { "class-get-n-set",		class_get_n_set,		KSI_CALL_ARG2, 1 },

  { "@allocate-instance",	ksi_alloc_instance,		KSI_CALL_ARG1, 1 },
  { "@modify-instance!",	ksi_modify_instance,		KSI_CALL_ARG2, 2 },
  { "@add-method",		add_method,			KSI_CALL_ARG2, 2 },
  { "@method-applicable?",	ksi_applicable_p,		KSI_CALL_ARG2, 2 },
  { "@method-more-specific?",	method_more_specific_p,		KSI_CALL_ARG3, 3 },
  { "@compute-applicable-methods", compute_methods,		KSI_CALL_ARG3, 3 },

  { "@merge-args",		ksi_merge_args,			KSI_CALL_REST0, 0 },
  { "@get-arg",			ksi_get_arg,			KSI_CALL_ARG3,  2 },

  { 0 }
};

void
ksi_init_klos ()
{
  ksi_ignore (rcsid);
  sym_cname = ksi_ptr02sym (S_cname);
  sym_gname = ksi_ptr02sym (S_gname);
  sym_dsupers = ksi_ptr02sym (S_dsupers);
  sym_dslots = ksi_ptr02sym (S_dslots);
  sym_defargs = ksi_ptr02sym (S_defargs);
  sym_cpl = ksi_ptr02sym (S_cpl);
  sym_slots = ksi_ptr02sym (S_slots);
  sym_nfields = ksi_ptr02sym (S_nfields);
  sym_gns = ksi_ptr02sym (S_gns);
  sym_gf = ksi_ptr02sym (S_gf);
  sym_specs = ksi_ptr02sym (S_specs);
  sym_proc = ksi_ptr02sym (S_proc);
  sym_methods = ksi_ptr02sym (S_methods);
  sym_arity = ksi_ptr02sym (S_arity);
  sym_combination = ksi_ptr02sym (S_combination);

  key_initform = ksi_ptr02key (K_initform);
  key_initarg = ksi_ptr02key (K_initarg);
  key_defargs = ksi_ptr02key (K_defargs);
  key_type = ksi_ptr02key (K_type);
  key_name = ksi_ptr02key (K_name);
  key_dsupers = ksi_ptr02key (K_dsupers);
  key_dslots = ksi_ptr02key (K_dslots);
  key_specs = ksi_ptr02key (K_specs);
  key_proc = ksi_ptr02key (K_proc);
  key_gf = ksi_ptr02key (K_gf);
  key_arity = ksi_ptr02key (K_arity);
  key_combination = ksi_ptr02key (K_combination);
  key_after = ksi_ptr02key (K_after);
  key_before = ksi_ptr02key (K_before);
  key_around = ksi_ptr02key (K_around);
  key_primary = ksi_ptr02key (K_primary);

  ksi_reg_unit (defs);
  init_top_classes ();

  KSI->sym_apply_generic = ksi_ptr02sym ("apply-generic");
  KSI->sym_no_applicable_method = ksi_ptr02sym ("no-applicable-method");
  KSI->sym_no_next_method = ksi_ptr02sym ("no-next-method");
  KSI->sym_proc_slot = sym_proc;

  ksi_defsym ("gns-name",       KSI_MK_SINT (GNS_NAME),   0);
  ksi_defsym ("gns-allocation", KSI_MK_SINT (GNS_ALLOC),  0);
  ksi_defsym ("gns-getter",     KSI_MK_SINT (GNS_GETTER), 0);
  ksi_defsym ("gns-setter",     KSI_MK_SINT (GNS_SETTER), 0);
  ksi_defsym ("gns-init",       KSI_MK_SINT (GNS_INIT),   0);
  ksi_defsym ("gns-type",       KSI_MK_SINT (GNS_TYPE),   0);
  ksi_defsym ("gns-size",       KSI_MK_SINT (GNS_SIZE),   0);
}


 /* End of code */
