/* -*-mode:C++-*- */
/*
 * ksi_klos.h
 * ksi object system -- internal defs
 *
 * Copyright (C) 1997-2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Thu Jun 12 19:51:55 1997
 * Last Update:   Fri Jul 27 21:36:54 2001
 *
 * $Id: ksi_klos.h,v 1.3 2001/07/27 18:48:05 demakov Exp $
 *
 */

#ifndef KSI_KLOS_H
#define KSI_KLOS_H


enum klos_instance_flags
{
  I_CLASS		= 1 << 0,
  I_GENERIC		= 1 << 1,
  I_METHOD		= 1 << 2,
  I_PURE_CLASS		= 1 << 3,
  I_PURE_GENERIC	= 1 << 4,
  I_PURE_METHOD		= 1 << 5,
  I_METHODS_SORTED	= 1 << 6
};

enum klos_class_slots
{
  S_CNAME,		/* class name */
  S_DSUPERS,		/* list of direct supers */
  S_DSLOTS,		/* list of direct slots */
  S_DEFARGS,		/* list of default initargs */
  S_CPL,		/* class precedence list */
  S_SLOTS,		/* list of all slots */
  S_NFIELDS,		/* number of fields */
  S_GNS,		/* get_n_sets */
  NUMBER_OF_CLASS_SLOTS
};

enum klos_generic_slots
{
  S_GNAME,		/* generic name */
  S_METHODS,		/* list of methods */
  S_ARITY,		/* list of formals */
  NUMBER_OF_GENERIC_SLOTS
};

enum klos_method_slots
{
  S_GF,			/* generic function */
  S_SPECS,		/* specializers */
  S_COMBINATION,	/* combination */
  S_PROC,		/* procedure */
  NUMBER_OF_METHOD_SLOTS
};

enum klos_gns_index
{
  GNS_NAME,
  GNS_ALLOC,
  GNS_GETTER,
  GNS_SETTER,
  GNS_INIT,
  GNS_TYPE,
  GNS_SIZE
};


struct Ksi_Instance
{
  unsigned	itag;
  int		flags;
  ksi_instance	klass;
  ksi_obj	*slots;
};

struct Ksi_NextMethod
{
  unsigned	itag;
  ksi_obj	procs;
  ksi_obj	args;
  ksi_obj	gf;
};

#define INST_P(x)		KSI_OBJ_IS ((x), KSI_TAG_INSTANCE)
#define NEXT_METHOD_P(x)	KSI_OBJ_IS ((x), KSI_TAG_NEXT_METHOD)

#define INST_IS(x,f)		(((ksi_instance) (x)) -> flags & (f))

#define METHOD_P(x)		(INST_P (x) && INST_IS ((x), I_METHOD))
#define GENERIC_P(x)		(INST_P (x) && INST_IS ((x), I_GENERIC))
#define CLASS_P(x)		(INST_P (x) && INST_IS ((x), I_CLASS))

#define PURE_GENERIC_P(x)	(INST_P (x) && INST_IS ((x), I_PURE_GENERIC))
#define PURE_CLASS_P(x)		(INST_P (x) && INST_IS ((x), I_PURE_CLASS))
#define PURE_METHOD_P(x)	(INST_P (x) && INST_IS ((x), I_PURE_METHOD))

#define CLASS_OF(x)		(((ksi_instance) (x)) -> klass)
#define SLOTS_PTR(x)		(((ksi_instance) (x)) -> slots)
#define SLOT_REF(x,r)		(SLOTS_PTR (x) [r])

#define NEXT_PROCS(x)		(((ksi_next_method) (x)) -> procs)
#define NEXT_ARGS(x)		(((ksi_next_method) (x)) -> args)
#define NEXT_GF(x)		(((ksi_next_method) (x)) -> gf)


void
ksi_init_klos (void);

ksi_obj
ksi_new_next (ksi_obj gf, ksi_obj args, ksi_obj procs);

void
ksi_write_inst (ksi_obj x, ksi_port p, int slashify);

ksi_obj
ksi_inst_eqv_p (ksi_obj o1, ksi_obj o2);

ksi_obj
ksi_inst_equal_p (ksi_obj o1, ksi_obj o2);

ksi_obj
ksi_inst_slot_unbound (ksi_obj cls, ksi_obj obj, ksi_obj slot);

ksi_obj
ksi_inst_slot_missing (ksi_obj cls, ksi_obj obj, ksi_obj slot, ksi_obj val);

unsigned int
ksi_hash_inst (ksi_obj x, unsigned n);

ksi_obj
ksi_compute_applicable_methods (ksi_obj gf, ksi_obj args, ksi_obj combination);

ksi_obj
ksi_compute_effective_method (ksi_obj gf, ksi_obj args);


#endif

/* End of code */
