/*
 * ksi_numb.c
 * numbers
 *
 * Copyright (C) 1997-1999,2000,2001,2004, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Sun Apr 27 14:17:11 1997
 * Last Update:   Thu Jul  8 19:51:25 2004
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_numb.h"


static char rcsid[] = "$Id: ksi_numb.c,v 1.11 2004/07/08 15:29:17 ksion Exp $";

static double ksi_fp_zero = 0.0;

#define BITSPERDIG (sizeof (unsigned) * CHAR_BIT)


#define TALLOC(x,t,s)					\
  do { x = (t*) alloca ((s) * sizeof (t));		\
     } while (0)


static inline ksi_bignum
alloc_big (int sz)
{
  ksi_bignum num;
  num = ksi_malloc_data ((sz - 1) * sizeof (unsigned) + sizeof *num);
  num -> itag = KSI_TAG_BIGNUM;
  return num;
}

static ksi_obj
new_big (int sign, unsigned* digs)
{
  ksi_bignum num;
  int sz = ABS (sign);

  while (sz > 0)
    {
      if (digs [--sz] != 0)
	{
	  sz++;
	  break;
	}
    }

  if (sz <= 0)
    return ksi_zero;

#if KSI_SIZEOF_INT_TYPE <= SIZEOF_INT
  if (sz == 1 && digs[0] < KSI_MAX_INT_TYPE)
    return KSI_MK_SINT (sign < 0 ? -digs[0] : digs[0]);
#elif KSI_SIZEOF_INT_TYPE <= 2 * SIZEOF_INT
  if (sz == 1)
    return KSI_MK_SINT (sign < 0 ? -digs[0] : digs[0]);
  if (sz == 2)
    {
      unsigned KSI_INT_TYPE v = digs[1];
      v = (v << KSI_INT_SIZE) | digs[0];
      if (v < KSI_MAX_INT_TYPE)
	return KSI_MK_SINT (sign < 0 ? -v : v);
    }
#endif

  num = alloc_big (sz);
  num->sign = (sign < 0 ? -sz : sz);
  while (--sz >= 0)
    num->digs[sz] = digs[sz];

  return (ksi_obj) num;
}

static ksi_obj
adjust_big (ksi_bignum num)
{
  int sz = ABS (num->sign);

  while (sz > 0)
    {
      if (num->digs[--sz] != 0)
	{
	  sz++;
	  break;
	}
    }

  if (sz <= 0)
    return ksi_zero;

#if KSI_SIZEOF_INT_TYPE <= SIZEOF_INT
  if (sz == 1 && num->digs[0] < KSI_MAX_INT_TYPE)
    return KSI_MK_SINT (num->sign < 0 ? -num->digs[0] : num->digs[0]);
#elif KSI_SIZEOF_INT_TYPE <= 2 * SIZEOF_INT
  if (sz == 1)
    return KSI_MK_SINT (num->sign < 0 ? -num->digs[0] : num->digs[0]);
  if (sz == 2)
    {
      unsigned KSI_INT_TYPE v = num->digs[1];
      v = (v << KSI_INT_SIZE) | num->digs[0];
      if (v < KSI_MAX_INT_TYPE)
	return KSI_MK_SINT (num->sign < 0 ? -v : v);
    }
#endif

  num->sign = (num->sign < 0 ? -sz : sz);
  return (ksi_obj) num;
}

static ksi_obj
add_big_ui (unsigned* dig, int sign, unsigned y)
{
  int absz = ABS (sign);
  ksi_bignum res = alloc_big (sign > 0 ? absz+1 : absz);
  unsigned *sum = res -> digs;

  if (sign > 0)
    {
      register unsigned a = *dig++ + y;
      *sum++ = a;
      --absz;
      if (a < y) /* carry */
        {
          while (--absz >= 0)
            {
              a = *dig++;
              *sum++ = ++a;
              if (a > 0)
		goto copy;
            }
          *sum = 1;
	  res -> sign = sign + 1;
          return adjust_big (res);
	}
    }
  else
    {
      register unsigned m = *dig++;
      y = m - y;
      *sum++ = y;
      --absz;
      if (y > m) /* carry */
        {
          while (1)
            {
              m = *dig++;
              *sum++ = m - 1;
              if (m > 0)
		{
		  if (absz == 0)
		    break;
		  goto copy;
		}
            }
	  res -> sign = sign + 1;
	  return adjust_big (res);
	}
    }

copy:
  while (--absz >= 0)
    *sum++ = *dig++;
  res -> sign = sign;
  return adjust_big (res);
}


static ksi_obj
sub_big_ui (unsigned* dig, int sign, unsigned y)
{
  int absz = ABS (sign);
  ksi_bignum res = alloc_big (sign < 0 ? absz+1 : absz);
  unsigned *sum = res -> digs;

  if (sign < 0)
    {
      register unsigned a = *dig++ + y;
      *sum++ = a;
      --absz;
      if (a < y) /* carry */
        {
          while (--absz >= 0)
            {
              a = *dig++;
              *sum++ = ++a;
              if (a > 0)
                goto copy;
            }
          *sum = 1;
	  res -> sign = sign - 1;
          return adjust_big (res);
        }
    }
  else
    {
      register unsigned m = *dig++;
      y = m - y;
      *sum++ = y;
      --absz;
      if (y > m) /* carry */
        {
          while (1)
            {
              m = *dig++;
              *sum++ = m - 1;
              if (m > 0)
		{
		  if (absz == 0)
		    break;
		  goto copy;
		}
            }
	  res -> sign = sign - 1;
	  return adjust_big (res);
        }
    }

copy:
  while (--absz >= 0)
    *sum++ = *dig++;
  res -> sign = sign;
  return adjust_big (res);
}

static ksi_obj
mul_big_i (int sign, unsigned* dig, int y)
{
  int mult_size = ABS (sign);
  ksi_bignum res;
  unsigned *prod, cr;

  if (y == 0)
    return ksi_zero;

  sign = PROD_SIGN (sign, y);
  y = ABS (y);

  res = alloc_big (mult_size + 1);
  res -> sign = (sign < 0 ? -mult_size : mult_size);
  prod = res -> digs;

  cr = 0;
  while (--mult_size >= 0)
    {
      unsigned p1, p0 = *dig++;
      MUL_LI (p1, p0, y, p0);
      p0 += cr;
      cr = p1 + (p0 < cr);
      *prod++ = p0;
    }
  if (cr != 0)
    {
      *prod = cr;
      res -> sign += (sign < 0 ? -1 : 1);
    }

  return adjust_big (res);
}

static ksi_obj
divmod_big_i (int sign, unsigned* div, int *pd)
{
  int y = *pd, size = ABS (sign), i;
  ksi_bignum res;
  unsigned *quot, r, n;

  sign = PROD_SIGN (sign, y);
  y = ABS (y);

  res = alloc_big (size);
  quot = res -> digs;

  i = size - 1;
  r = div [i];
  if (r >= y)
    r = 0;
  else
    --size, --i;

  for ( ; i >= 0; --i)
    {
      n = div [i];
      DIV_LI (quot [i], r, r, n, y);
    }
  *pd = r;

  res -> sign = (sign < 0 ? -size : size);
  return adjust_big (res);
}

static int
big_cmp (unsigned* p1, unsigned* p2, int sz)
{
  register unsigned cmp = 1;

  p1 += sz; p2 += sz;
  while (--sz >= 0)
    {
      if ((cmp = *(--p1) - *(--p2)) != 0)
	break;
    }

  return cmp;
}


/* add_big_u -- add two big numbers.
 *
 * Must be: (add1_size >= add2_size) && (add2_size >  0)
 */

static int
add_big_u (unsigned* sum_ptr,
           unsigned* add1_ptr, int add1_size,
           unsigned* add2_ptr, int add2_size)
{
  unsigned a1, a2, sum;

  /* No carry loop */
  do
    {
      a1 = *add1_ptr++;
      a2 = *add2_ptr++;
      sum = a1 + a2;
      *sum_ptr++ = sum;
      if (sum < a2)
        goto cr_loop;
    ncr_loop:
      --add1_size;
      --add2_size;
    }
  while (add2_size > 0);

copy_add1:
  while (--add1_size >= 0)
    *sum_ptr++ = *add1_ptr++;
  return 0;

  /* Carry loop */
  do
    {
      a1 = *add1_ptr++;
      a2 = *add2_ptr++;
      sum = a1 + a2 + 1;
      *sum_ptr++ = sum;
      if (sum > a2)
        goto ncr_loop;
    cr_loop:
      --add1_size;
      --add2_size;
    }
  while (add2_size > 0);

  while (--add1_size >= 0)
    {
      a1 = *add1_ptr++;
      sum = a1 + 1;
      *sum_ptr++ = sum;
      if (sum != 0)
        goto copy_add1;
    }

  return 1;
}


/* sub_big_u -- substuct two big numbers.
 *
 * Must be: ((min_size >= sub_size) && (sub_size > 0)
 */

static int
sub_big_u (unsigned* dif_ptr,
           unsigned* min_ptr, int min_size,
           unsigned* sub_ptr, int sub_size)
{
  unsigned m, s, dif;

  /* No carry loop */
  do
    {
      m = *min_ptr++;
      s = *sub_ptr++;
      dif = m - s;
      *dif_ptr++ = dif;
      if (dif > m)
        goto cr_loop;
    ncr_loop:
      --min_size;
      --sub_size;
    }
  while (sub_size > 0);

copy_min:

  if (min_size == 0)
    {
      while (*(--dif_ptr) == 0)
	--min_size;
      return min_size;
    }

  while (--min_size >= 0)
    *dif_ptr++ = *min_ptr++;
  return 0;

  /* Carry loop */
  do
    {
      m = *min_ptr++;
      s = *sub_ptr++;
      dif = m - s - 1;
      *dif_ptr++ = dif;
      if (dif < m)
        goto ncr_loop;
    cr_loop:
      --min_size;
      --sub_size;
    }
  while (sub_size > 0);

  if (min_size == 0)
    return 1;

  while (min_size > 0)
    {
      --min_size;
      m = *min_ptr++;
      *dif_ptr++ = m - 1;
      if (m != 0)
        goto copy_min;
    }

  ksi_error ("sub_big_u: internal error");
  return 0;
}


/* mul_big_u
 *
 * Must be: usize >= vsize && vsize > 0 wptr != uptr && wptr != vptr
 */

static int
mul_big_u (unsigned* wptr,
           unsigned* uptr, int usize,
           unsigned* vptr, int vsize)
{
  int j, i, k, cr = 0;
  for (k = 0; k < usize; k++)
    {
      unsigned p1, p0;
      MUL_LI (p1, p0, uptr[k], vptr[0]);
      wptr[k] = (p0 += cr);
      cr = p1 + (p0 < cr);
    }
  if ((wptr[k] = cr) != 0)
    k++;

  i = 0;
  while (++i < vsize)
    {
      cr = 0; k = i;
      for (j = 0; j < usize; k++, j++)
        {
          unsigned p1, p0, s;
          MUL_LI (p1, p0, uptr[j], vptr[i]);
          s = wptr[k] + p0;
          p1 += (s < p0);
          wptr[k] = (s += cr);
          cr = p1 + (s < cr);
        }
      if ((wptr[k] = cr) != 0)
        k++;
    }

  return k;
}

static int
rshift_big (unsigned* wp,
            unsigned* up, int usize,
            unsigned cnt)
{
  unsigned high_limb, low_limb;
  unsigned sh_1, sh_2;
  int i;

  if (usize == 0)
    return 0;

  sh_1 = cnt;
  if (sh_1 == 0)
    {
      if (wp != up)
        {
          /* Copy from low end to high end, to allow specified input/output
             overlapping.  */
          for (i = 0; i < usize; i++)
            wp[i] = up[i];
        }
      return usize;
    }

  wp -= 1;
  sh_2 = KSI_INT_SIZE - sh_1;
  high_limb = up[0];
  low_limb = high_limb;

  for (i = 1; i < usize; i++)
    {
      high_limb = up[i];
      wp[i] = (low_limb >> sh_1) | (high_limb << sh_2);
      low_limb = high_limb;
    }
  low_limb >>= sh_1;
  if (low_limb != 0)
    {
      wp[i] = low_limb;
      return usize;
    }

  return usize - 1;
}

static unsigned
lshift_big (unsigned* wp,
            unsigned* up, int usize,
            unsigned cnt)
{
  unsigned high_limb, low_limb;
  unsigned sh_1, sh_2;
  int i;
  unsigned retval;

  if (usize == 0)
    return 0;

  sh_1 = cnt;
  if (sh_1 == 0)
    {
      if (wp != up)
        {
          /* Copy from high end to low end, to allow specified input/output
             overlapping.  */
          for (i = usize - 1; i >= 0; i--)
            wp[i] = up[i];
        }
      return 0;
    }

  wp += 1;
  sh_2 = KSI_INT_SIZE - sh_1;
  i = usize - 1;
  low_limb = up[i];
  retval = low_limb >> sh_2;
  high_limb = low_limb;
  while (--i >= 0)
    {
      low_limb = up[i];
      wp[i] = (high_limb << sh_1) | (low_limb >> sh_2);
      high_limb = low_limb;
    }
  wp[i] = high_limb << sh_1;

  return retval;
}

static int
rshiftci_big (unsigned* wp,
              unsigned* up, int usize,
              unsigned cnt,
              unsigned carry_in)
{
  unsigned high_limb, low_limb;
  unsigned sh_1, sh_2;
  int i;

  if (usize <= 0)
    return 0;

  sh_1 = cnt;
  if (sh_1 == 0)
    {
      if (wp != up)
        {
          /* Copy from low end to high end, to allow specified input/output
             overlapping.  */
          for (i = 0; i < usize; i++)
            wp[i] = up[i];
        }
      return usize;
    }

  wp -= 1;
  sh_2 = KSI_INT_SIZE - sh_1;
  low_limb = up[0];
  for (i = 1; i < usize; i++)
    {
      high_limb = up[i];
      wp[i] = (low_limb >> sh_1) | (high_limb << sh_2);
      low_limb = high_limb;
    }
  low_limb = (low_limb >> sh_1) | (carry_in << sh_2);
  if (low_limb != 0)
    {
      wp[i] = low_limb;
      return usize;
    }

  return usize - 1;
}

static int
div_big_u (unsigned* quot_ptr,
           unsigned* num_ptr, int num_size,
           unsigned* den_ptr, int den_size)
{
  int q_is_long = 0;

  switch (den_size)
    {
    case 0:
      ksi_error ("div_big_u: internal error, devide by zero");

    case 1:
      {
        int i;
        unsigned n1, n0, d;

        d = den_ptr[0];
        i = num_size - 1;
        n1 = num_ptr[i];
        i--;

        if (n1 >= d)
          {
            q_is_long = 1;
            n1 = 0;
            i++;
          }

        for (; i >= 0; i--)
          {
            n0 = num_ptr[i];
            DIV_LI (quot_ptr[i], n1, n1, n0, d);
          }

        num_ptr[0] = n1;
      }
      break;

    case 2:
      {
        int i;
        unsigned n0, n1, n2, d0, d1;

        num_ptr += num_size - 2;
        d0 = den_ptr[1];
        d1 = den_ptr[0];
        n0 = num_ptr[1];
        n1 = num_ptr[0];

        if (n0 >= d0)
          {
            q_is_long = 1;
            n1 = n0;
            n0 = 0;
            num_ptr++;
            num_size++;
          }

        for (i = num_size - den_size - 1; i >= 0; i--)
          {
            unsigned q, r;

            num_ptr--;
            if (n0 == d0)
              {
                /* Q should be either 111..111 or 111..110.  Need special
                   treatment of this rare case as normal division would
                   give overflow.  */
                q = ~0;

                r = n1 + d0;
                if (r < d0)     /* Carry in the addition? */
                  {
                    n2 = num_ptr[0];

                    ADD_LI (n0, n1, r - d1, n2, 0, d1);
                    quot_ptr[i] = q;
                    continue;
                  }
                n0 = d1 - (d1 != 0);
                n1 = -d1;
              }
            else
              {
                DIV_LI (q, r, n0, n1, d0);
                MUL_LI (n0, n1, d1, q);
              }

            n2 = num_ptr[0];
          q_test:
            if (n0 > r || (n0 == r && n1 > n2))
              {
                /* The estimated Q was too large.  */
                q--;

                SUB_LI (n0, n1, n0, n1, 0, d1);
                r += d0;
                if (r >= d0)    /* If not carry, test q again.  */
                  goto q_test;
              }

            quot_ptr[i] = q;
            SUB_LI (n0, n1, r, n2, n0, n1);
          }
        num_ptr[1] = n0;
        num_ptr[0] = n1;
      }
      break;

    default:
      {
        int i;
        unsigned d0 = den_ptr[den_size - 1];
        unsigned d1 = den_ptr[den_size - 2];
        unsigned n0 = num_ptr[num_size - 1];
        int ugly_hack_flag = 0;

        if (n0 >= d0)
          {
            /* There's a problem with this case, which shows up later in the
               code.  q becomes 1 (and sometimes 0) the first time when
               we've been here, and w_cy == 0 after the main do-loops below.
               But c = num_ptr[j] reads rubbish outside the num_ptr vector!
               Maybe I can solve this cleanly when I fix the early-end
               optimization here in the default case.  For now, I change the
               add_back entering condition, to kludge.  Leaving the stray
               memref behind!

               HACK: Added ugly_hack_flag to make it work.  */

            q_is_long = 1;
            n0 = 0;
            num_size++;
            ugly_hack_flag = 1;
          }

        num_ptr += num_size;
        den_ptr += den_size;
        for (i = num_size - den_size - 1; i >= 0; i--)
          {
            unsigned q;
            unsigned n1;
            unsigned w_cy;
            unsigned d, c;
            int j;

            num_ptr--;
            if (n0 == d0)
              /* This might over-estimate q, but it's probably not worth
                 the extra code here to find out.  */
              q = ~0;
            else
              {
                unsigned r;

                DIV_LI (q, r, n0, num_ptr[-1], d0);
                MUL_LI (n1, n0, d1, q);

                while (n1 > r || (n1 == r && n0 > num_ptr[-2]))
                  {
                    q--;
                    r += d0;
                    if (r < d0) /* I.e. "carry in previous addition?"  */
                      break;
                    n1 -= n0 < d1;
                    n0 -= d1;
                  }
              }

            w_cy = 0;
            j = -den_size;
            do
              {
                d = den_ptr[j];
                c = num_ptr[j];
                MUL_LI (n1, n0, d, q);
                n0 += w_cy;
                w_cy = (n0 < w_cy) + n1;
                n0 = c - n0;
                num_ptr[j] = n0;
                if (n0 > c)
                  goto cy_loop;
              ncy_loop:
                j++;
              }
            while  (j < 0);

            if (ugly_hack_flag)
              {
                c = 0;
                ugly_hack_flag = 0;
              }
            else
              c = num_ptr[j];
            if (c >= w_cy)
              goto store_q;
            goto add_back;

            do
              {
                d = den_ptr[j];
                c = num_ptr[j];
                MUL_LI (n1, n0, d, q);
                n0 += w_cy;
                w_cy = (n0 < w_cy) + n1;
                n0 = c - n0 - 1;
                num_ptr[j] = n0;
                if (n0 < c)
                  goto ncy_loop;
              cy_loop:
                j++;
              }
            while  (j < 0);

            if (ugly_hack_flag)
              {
                c = 0;
                ugly_hack_flag = 0;
              }
            else
              c = num_ptr[j];
            w_cy++;
            if (c >= w_cy)
              goto store_q;

          add_back:
            j = -den_size;
            do
              {
                d = den_ptr[j];
                n0 = num_ptr[j] + d;
                num_ptr[j] = n0;
                if (n0 < d)
                  goto ab_cy_loop;
              ab_ncy_loop:
                j++;
              }
            while  (j < 0);
            ksi_error ("div_big_u: internal error: we should always have a carry out!");

            do
              {
                d = den_ptr[j];
                n0 = num_ptr[j] + d + 1;
                num_ptr[j] = n0;
                if (n0 > d)
                  goto ab_ncy_loop;
              ab_cy_loop:
                j++;
              }
            while  (j < 0);
            q--;

          store_q:
            quot_ptr[i] = q;
          }
      }
    }

  return q_is_long;
}

static int
sqrt_big_u (unsigned* root_ptr, unsigned* rem_ptr,
            unsigned* op_ptr, int op_size)
{
  /* Tables for initial approximation of the square root.  These are
     indexed with bits 1-8 of the operand for which the square root is
     calculated, where bit 0 is the most significant non-zero bit.  I.e.
     the most significant one-bit is not used, since that per definition
     is one.  Likewise, the tables don't return the highest bit of the
     result.  That bit must be inserted by or:ing the returned value with
     0x100.  This way, we get a 9-bit approximation from 8-bit tables!  */

  /* Table to be used for operands with an even total number of bits.
     (Exactly as in the decimal system there are similarities between the
     square root of numbers with the same initial digits and an even
     difference in the total number of digits.  Consider the square root
     of 1, 10, 100, 1000, ...)  */
  static unsigned char even_approx_tab[256] =
  {
    0x6a, 0x6a, 0x6b, 0x6c, 0x6c, 0x6d, 0x6e, 0x6e,
    0x6f, 0x70, 0x71, 0x71, 0x72, 0x73, 0x73, 0x74,
    0x75, 0x75, 0x76, 0x77, 0x77, 0x78, 0x79, 0x79,
    0x7a, 0x7b, 0x7b, 0x7c, 0x7d, 0x7d, 0x7e, 0x7f,
    0x80, 0x80, 0x81, 0x81, 0x82, 0x83, 0x83, 0x84,
    0x85, 0x85, 0x86, 0x87, 0x87, 0x88, 0x89, 0x89,
    0x8a, 0x8b, 0x8b, 0x8c, 0x8d, 0x8d, 0x8e, 0x8f,
    0x8f, 0x90, 0x90, 0x91, 0x92, 0x92, 0x93, 0x94,
    0x94, 0x95, 0x96, 0x96, 0x97, 0x97, 0x98, 0x99,
    0x99, 0x9a, 0x9b, 0x9b, 0x9c, 0x9c, 0x9d, 0x9e,
    0x9e, 0x9f, 0xa0, 0xa0, 0xa1, 0xa1, 0xa2, 0xa3,
    0xa3, 0xa4, 0xa4, 0xa5, 0xa6, 0xa6, 0xa7, 0xa7,
    0xa8, 0xa9, 0xa9, 0xaa, 0xaa, 0xab, 0xac, 0xac,
    0xad, 0xad, 0xae, 0xaf, 0xaf, 0xb0, 0xb0, 0xb1,
    0xb2, 0xb2, 0xb3, 0xb3, 0xb4, 0xb5, 0xb5, 0xb6,
    0xb6, 0xb7, 0xb7, 0xb8, 0xb9, 0xb9, 0xba, 0xba,
    0xbb, 0xbb, 0xbc, 0xbd, 0xbd, 0xbe, 0xbe, 0xbf,
    0xc0, 0xc0, 0xc1, 0xc1, 0xc2, 0xc2, 0xc3, 0xc3,
    0xc4, 0xc5, 0xc5, 0xc6, 0xc6, 0xc7, 0xc7, 0xc8,
    0xc9, 0xc9, 0xca, 0xca, 0xcb, 0xcb, 0xcc, 0xcc,
    0xcd, 0xce, 0xce, 0xcf, 0xcf, 0xd0, 0xd0, 0xd1,
    0xd1, 0xd2, 0xd3, 0xd3, 0xd4, 0xd4, 0xd5, 0xd5,
    0xd6, 0xd6, 0xd7, 0xd7, 0xd8, 0xd9, 0xd9, 0xda,
    0xda, 0xdb, 0xdb, 0xdc, 0xdc, 0xdd, 0xdd, 0xde,
    0xde, 0xdf, 0xe0, 0xe0, 0xe1, 0xe1, 0xe2, 0xe2,
    0xe3, 0xe3, 0xe4, 0xe4, 0xe5, 0xe5, 0xe6, 0xe6,
    0xe7, 0xe7, 0xe8, 0xe8, 0xe9, 0xea, 0xea, 0xeb,
    0xeb, 0xec, 0xec, 0xed, 0xed, 0xee, 0xee, 0xef,
    0xef, 0xf0, 0xf0, 0xf1, 0xf1, 0xf2, 0xf2, 0xf3,
    0xf3, 0xf4, 0xf4, 0xf5, 0xf5, 0xf6, 0xf6, 0xf7,
    0xf7, 0xf8, 0xf8, 0xf9, 0xf9, 0xfa, 0xfa, 0xfb,
    0xfb, 0xfc, 0xfc, 0xfd, 0xfd, 0xfe, 0xfe, 0xff,
  };

  /* Table to be used for operands with an odd total number of bits.
     (Further comments before previous table.)  */
  static unsigned char odd_approx_tab[256] =
  {
    0x00, 0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x03,
    0x03, 0x04, 0x04, 0x05, 0x05, 0x06, 0x06, 0x07,
    0x07, 0x08, 0x08, 0x09, 0x09, 0x0a, 0x0a, 0x0b,
    0x0b, 0x0c, 0x0c, 0x0d, 0x0d, 0x0e, 0x0e, 0x0f,
    0x0f, 0x10, 0x10, 0x10, 0x11, 0x11, 0x12, 0x12,
    0x13, 0x13, 0x14, 0x14, 0x15, 0x15, 0x16, 0x16,
    0x16, 0x17, 0x17, 0x18, 0x18, 0x19, 0x19, 0x1a,
    0x1a, 0x1b, 0x1b, 0x1b, 0x1c, 0x1c, 0x1d, 0x1d,
    0x1e, 0x1e, 0x1f, 0x1f, 0x20, 0x20, 0x20, 0x21,
    0x21, 0x22, 0x22, 0x23, 0x23, 0x23, 0x24, 0x24,
    0x25, 0x25, 0x26, 0x26, 0x27, 0x27, 0x27, 0x28,
    0x28, 0x29, 0x29, 0x2a, 0x2a, 0x2a, 0x2b, 0x2b,
    0x2c, 0x2c, 0x2d, 0x2d, 0x2d, 0x2e, 0x2e, 0x2f,
    0x2f, 0x30, 0x30, 0x30, 0x31, 0x31, 0x32, 0x32,
    0x32, 0x33, 0x33, 0x34, 0x34, 0x35, 0x35, 0x35,
    0x36, 0x36, 0x37, 0x37, 0x37, 0x38, 0x38, 0x39,
    0x39, 0x39, 0x3a, 0x3a, 0x3b, 0x3b, 0x3b, 0x3c,
    0x3c, 0x3d, 0x3d, 0x3d, 0x3e, 0x3e, 0x3f, 0x3f,
    0x40, 0x40, 0x40, 0x41, 0x41, 0x41, 0x42, 0x42,
    0x43, 0x43, 0x43, 0x44, 0x44, 0x45, 0x45, 0x45,
    0x46, 0x46, 0x47, 0x47, 0x47, 0x48, 0x48, 0x49,
    0x49, 0x49, 0x4a, 0x4a, 0x4b, 0x4b, 0x4b, 0x4c,
    0x4c, 0x4c, 0x4d, 0x4d, 0x4e, 0x4e, 0x4e, 0x4f,
    0x4f, 0x50, 0x50, 0x50, 0x51, 0x51, 0x51, 0x52,
    0x52, 0x53, 0x53, 0x53, 0x54, 0x54, 0x54, 0x55,
    0x55, 0x56, 0x56, 0x56, 0x57, 0x57, 0x57, 0x58,
    0x58, 0x59, 0x59, 0x59, 0x5a, 0x5a, 0x5a, 0x5b,
    0x5b, 0x5b, 0x5c, 0x5c, 0x5d, 0x5d, 0x5d, 0x5e,
    0x5e, 0x5e, 0x5f, 0x5f, 0x60, 0x60, 0x60, 0x61,
    0x61, 0x61, 0x62, 0x62, 0x62, 0x63, 0x63, 0x63,
    0x64, 0x64, 0x65, 0x65, 0x65, 0x66, 0x66, 0x66,
    0x67, 0x67, 0x67, 0x68, 0x68, 0x68, 0x69, 0x69,
  };

  /* R (root result) */
  unsigned* rp;		     /* Pointer to least significant word */
  int rsize;                 /* The size in words */

  /* T (OP shifted to the left a.k.a. normalized) */
  unsigned* tp;              /* Pointer to least significant word */
  int tsize;                 /* The size in words */
  unsigned* t_end_ptr;       /* Pointer right beyond most sign. word */
  unsigned t_high0, t_high1; /* The two most significant words */

  /* TT (temporary for numerator/remainder) */
  unsigned* ttp;             /* Pointer to least significant word */

  /* X (temporary for quotient in main loop) */
  unsigned* xp;              /* Pointer to least significant word */
  int xsize;                 /* The size in words */

  unsigned cnt;
  unsigned initial_approx;   /* Initially made approximation */
  int tsizes[KSI_INT_SIZE];  /* Successive calculation precisions */
  int tmp;
  int i;

  /* If OP is zero, both results are zero.  */
  if (op_size == 0)
    return 0;

  CNT_ZERO (cnt, op_ptr[op_size - 1]);
  tsize = op_size;
  if ((tsize & 1) != 0)
    {
      cnt += KSI_INT_SIZE;
      tsize++;
    }

  rsize = tsize / 2;
  rp = root_ptr;

  /* Shift OP an even number of bits into T, such that either the most or
     the second most significant bit is set, and such that the number of
     words in T becomes even.  This way, the number of words in R=sqrt(OP)
     is exactly half as many as in OP, and the most significant bit of R
     is set.

     Also, the initial approximation is simplified by this up-shifted OP.

     Finally, the Newtonian iteration which is the main part of this
     program performs division by R.  The fast division routine expects
     the divisor to be "normalized" in exactly the sense of having the
     most significant bit set.  */

  TALLOC (tp, unsigned, tsize);

  t_high0 = lshift_big (tp + cnt / KSI_INT_SIZE, op_ptr, op_size,
                        (cnt & ~1) % KSI_INT_SIZE);
  if (cnt >= KSI_INT_SIZE)
    tp[0] = 0;

  t_high0 = tp[tsize - 1];
  t_high1 = tp[tsize - 2];      /* Never stray.  TSIZE is >= 2.  */

  /* Get a 9 bit approximation from the tables.  The tables expect to
     be indexed with the 8 high bits right below the highest bit.
     Also, the highest result bit is not returned by the tables, and
     must be or:ed into the result.  The scheme gives 9 bits of start
     approximation with just 256-entry 8 bit tables.  */

  if ((cnt & 1) == 0)
    {
      /* The most sign bit of t_high0 is set.  */
      initial_approx = t_high0 >> (KSI_INT_SIZE - 8 - 1);
      initial_approx &= 0xff;
      initial_approx = even_approx_tab[initial_approx];
    }
  else
    {
      /* The most significant bit of T_HIGH0 is unset,
         the second most significant is set.  */
      initial_approx = t_high0 >> (KSI_INT_SIZE - 8 - 2);
      initial_approx &= 0xff;
      initial_approx = odd_approx_tab[initial_approx];
    }
  initial_approx |= 0x100;
  initial_approx <<= KSI_INT_SIZE - 8 - 1;

  /* Perform small precision Newtonian iterations to get a full word
     approximation.  For small operands, these iteration will make the
     entire job.  */
  if (t_high0 == ~0)
    initial_approx = t_high0;
  else
    {
      unsigned quot;

      if (t_high0 >= initial_approx)
        initial_approx = t_high0 + 1;

      /* First get about 18 bits with pure C arithmetics.  */
      quot = (t_high0 / (initial_approx >> KSI_INT_SIZE/2) <<
              KSI_INT_SIZE/2);
      initial_approx = (initial_approx + quot) / 2;
      initial_approx |= 1 << (KSI_INT_SIZE - 1);

      /* Now get a full word by one (or for > 36 bit machines) several
         iterations.  */
      for (i = 16; i < KSI_INT_SIZE; i <<= 1)
        {
          unsigned ignored_remainder;

          DIV_LI (quot, ignored_remainder,
                  t_high0, t_high1, initial_approx);
          initial_approx = (initial_approx + quot) / 2;
          initial_approx |= 1 << (KSI_INT_SIZE - 1);
        }
    }

  rp[0] = initial_approx;
  rsize = 1;

  TALLOC (xp, unsigned, tsize);
  TALLOC (ttp, unsigned, tsize);

  t_end_ptr = tp + tsize;

  if (tsize > 2)
    {
      /* Determine the successive precisions to use in the iteration.  We
         minimize the precisions, beginning with the highest (i.e. last
         iteration) to the lowest (i.e. first iteration).  */

      tmp = tsize / 2;
      for (i = 0;;i++)
        {
          tsize = (tmp + 1) / 2;
          if (tmp == tsize)
            break;
          tsizes[i] = tsize + tmp;
          tmp = tsize;
        }

      /* Main Newton iteration loop.  For big arguments, most of the
         time is spent here.  */

      /* Loop invariants:
       * R <= shiftdown_to_same_size(floor(sqrt(OP))) < R + 1.
       * X - 1 < shiftdown_to_same_size(floor(sqrt(OP))) <= X.
       * R <= shiftdown_to_same_size(X).
       */

      while (--i >= 0)
        {
          unsigned cy;
          tsize = tsizes[i];

          /* Need to copy the numerator into temporary space, as
             mpn_div overwrites its numerator argument with the
             remainder (which we currently ignore).  */
          {
	    int k;
            for (k = 0; k < tsize; ++k)
              ttp[k] = (t_end_ptr - tsize) [k];
          }
          cy = div_big_u (xp, ttp, tsize, rp, rsize);
          xsize = tsize - rsize;
          cy = cy ? xp[xsize] : 0;

          /* Add X and R with the most significant limbs aligned,
             temporarily ignoring at least one limb at the low end of X.  */
          tmp = xsize - rsize;
          cy += add_big_u (xp + tmp, rp, rsize, xp + tmp, rsize);

          /* If T begins with more than 2 x BITS_PER_MP_LIMB of ones, we get
             intermediate roots that'd need an extra bit.  We don't want to
             handle that since it would make the subsequent divisor
             non-normalized, so round such roots down to be only ones in the
             current precision.  */
          if (cy == 2)
            {
              int j;
              for (j = xsize; j >= 0; j--)
                xp[j] = ~0;
            }

          /* Divide X by 2 and put the result in R.  This is the new
             approximation.  Shift in the carry from the addition.  */
          rsize = rshiftci_big (rp, xp, xsize, 1, 1);
        }
    }

  /* We computed the square root of OP * 2**(2*floor(cnt/2)).
     This has resulted in R being 2**floor(cnt/2) to large.
     Shift it down here to fix that.  */
  rsize = rshift_big (rp, rp, rsize, cnt/2);

  /* Calculate the remainder.  */
  tsize = mul_big_u (tp, rp, rsize, rp, rsize);
  if (op_size == tsize)
    {
      int cmp;
      if ((cmp = big_cmp (op_ptr, tp, op_size)) == 0)
        return 0;
      if (cmp < 0)
        goto less;
    }
  else if (op_size < tsize)
    {
      unsigned one;
    less:
      /* R is too large.  Decrement it.  */
      one = 1;

      tsize = tsize + sub_big_u (tp, tp, tsize, rp, rsize);
      tsize = tsize + sub_big_u (tp, tp, tsize, rp, rsize);
      tsize = tsize + add_big_u (tp, tp, tsize, &one, 1);

      sub_big_u (rp, rp, rsize, &one, 1);
    }
  return op_size + sub_big_u (rem_ptr, op_ptr, op_size, tp, tsize);
}

static ksi_obj
add_big (ksi_bignum u, ksi_bignum v)
{
  int us, vs, ss = 0;
  unsigned *up, *vp;
  ksi_bignum res;

  vs = ABS (v -> sign);
  us = u -> sign;
  if (us >= 0)
    {
      if (us >= vs)
        {
          up = u -> digs;
          vp = v -> digs;
        }
      else
        {
          up = v -> digs; us = vs;
          vp = u -> digs; vs = u -> sign;
        }

      res = alloc_big (us+1);
      ss = us;

      if (add_big_u (res->digs, up, us, vp, vs))
        res -> digs [ss++] = 1;
      res -> sign = ss;
    }
  else
    {
      us = - us;
      up = u -> digs;
      vp = v -> digs;

      res = alloc_big (MAX (us, vs));
      if (us < vs || (us == vs && (ss = big_cmp (up, vp, us)) < 0))
	ss = vs + sub_big_u (res->digs, vp, vs, up, us);
      else if (us > vs || ss > 0)
	ss = - (us + sub_big_u (res->digs, up, us, vp, vs));
      else /* (us == vs && ss == 0) <=:=> (abs u) is equal (abs v) */
        return ksi_zero;

      res -> sign = ss;
    }

  return adjust_big (res);
}

static ksi_obj
sub_big (ksi_bignum u, ksi_bignum v)
{
  int us, vs, ss = 0;
  unsigned *up, *vp;
  ksi_bignum res;

  vs = ABS (v -> sign);
  us = u -> sign;
  if (us < 0)
    {
      us = - us;
      if (us >= vs)
        {
          up = u -> digs;
          vp = v -> digs;
        }
      else
        {
          up = v -> digs; us = vs;
          vp = u -> digs; vs = - (u -> sign);
        }

      res = alloc_big (us+1);
      ss = us;

      if (add_big_u (res->digs, up, us, vp, vs))
        res -> digs [ss++] = 1;
      res -> sign = -ss;
    }
  else
    {
      up = u -> digs;
      vp = v -> digs;

      res = alloc_big (MAX (us, vs));

      if (us < vs || (us == vs && (ss = big_cmp (up, vp, us)) < 0))
	ss = - (vs + sub_big_u (res->digs, vp, vs, up, us));
      else if (us > vs || ss > 0)
	ss = us + sub_big_u (res->digs, up, us, vp, vs);
      else /* (us == vs && ss == 0) <=:=> (abs u) is equal (abs v) */
        return ksi_zero;

      res -> sign = ss;
    }

  return adjust_big (res);
}

static ksi_obj
mul_big (ksi_bignum u, ksi_bignum v)
{
  ksi_bignum res;
  unsigned *vp, *up;
  int usize = u -> sign;
  int vsize = v -> sign;
  int sign = PROD_SIGN (usize, vsize);

  usize = ABS (usize);
  vsize = ABS (vsize);
  if (usize >= vsize)
    {
      vp = v -> digs;
      up = u -> digs;
    }
  else
    {
      vp = u -> digs; vsize = usize;
      up = v -> digs; usize = ABS (v -> sign);
    }

  res = alloc_big (usize+vsize);

  usize = mul_big_u (res -> digs, up, usize, vp, vsize);
  res->sign = (sign < 0 ? -usize : usize);

  return adjust_big (res);
}

static void
div_big (ksi_bignum num, ksi_bignum den, ksi_obj* quot, ksi_obj* rem)
{
  /* True : abs (num->sign) >= abs (den->size) */

  int nsize = num -> sign, dsize = den -> sign;
  int qsize, rsize, i;
  int sign = PROD_SIGN (nsize, dsize);
  unsigned *qp, *rp, *np, *dp, normalization_steps;
  ksi_bignum tmp;

  nsize = ABS (nsize);
  dsize = ABS (dsize);
  rsize = nsize + 1;
  qsize = rsize - dsize + 1;

  TALLOC (qp, unsigned, qsize);
  TALLOC (rp, unsigned, rsize);
  np = num -> digs;
  dp = den -> digs;

  CNT_ZERO (normalization_steps, dp[dsize - 1]);

  /* Normalize the denominator, i.e. make its most significant bit set by
     shifting it NORMALIZATION_STEPS bits to the left.  Also shift the
     numerator the same number of steps (to keep the quotient the same!).  */
  if (normalization_steps != 0)
    {
      unsigned *tp, ndigit;

      TALLOC (tp, unsigned, (dsize + 1));
      lshift_big (tp, dp, dsize, normalization_steps);
      dp = tp;

      ndigit = lshift_big (rp, np, nsize, normalization_steps);
      if (ndigit != 0)
        {
          rp[nsize] = ndigit;
          rsize = nsize + 1;
        }
      else
        rsize = nsize;
    }
  else
    {
      for (rsize = 0; rsize < nsize; rsize++)
        rp[rsize] = np[rsize];
      rsize = nsize;
    }

  qsize = rsize - dsize + div_big_u (qp, rp, rsize, dp, dsize);

  if (quot)
    {
      if (qsize > 0 && qp [qsize - 1] == 0)
        --qsize;

      tmp = alloc_big (qsize);
      tmp -> sign = sign < 0 ? -qsize : qsize;
      for (i = 0; i < qsize; i++)
	tmp -> digs [i] = qp [i];
      *quot = adjust_big (tmp);
    }

  if (rem)
    {
      rsize = dsize;
      while (rsize > 0)
        {
          if (rp [rsize - 1] != 0)
            break;
          --rsize;
        }

      tmp = alloc_big (rsize);
      if (normalization_steps != 0)
        rsize = rshift_big (tmp -> digs, rp, rsize, normalization_steps);
      else
	{
	  for (i = 0; i < rsize; i++)
	    tmp -> digs[i] = rp[i];
	}

      tmp -> sign = rsize;
      *rem = adjust_big (tmp);
    }
}

static ksi_obj
sqrt_big (ksi_bignum op, int* has_rem)
{
  int op_size = ABS (op -> sign);
  int root_size = (op_size + 1) / 2;
  int rem_size;
  unsigned *op_ptr = op -> digs;
  unsigned *root_ptr, *rem_ptr;

  TALLOC (root_ptr, unsigned, root_size);
  TALLOC (rem_ptr, unsigned, op_size);

  rem_size = sqrt_big_u (root_ptr, rem_ptr, op_ptr, op_size);

  if (has_rem)
    *has_rem = rem_size;

  return new_big (root_size, root_ptr);
}

/* --------------------------------------------------------------- */

/*
 * ksi_str2flo
 *	  ksi_str2num    
 *	   .
 *	RADIX    2, 8, 10, 16.
 *
 * Return
 *	     ,
 *	 ksi_false     .
 *	        RADIX.
 *	       `#' (
 *	  0),   `_' (),    
 *	     `+'  `-'.
 *	 `_'    .
 *	
 *	     .
 *	     ,
 *	  -  
 *
 *	  * [+-]{N}/{D} 
 *		 `/'    
 *		N  D      
 *		  N  D.
 *
 *	  * [+-][{N}][.{D}][e[+-]{E}]
 *			. -  
 *			e -  :   `E F L D S e f l d s'
 *	
 *
 * Bugs
 *	 RADIX  16,  `D E F d e f'  
 *	 ,     .
 *
 */

static ksi_obj
ksi_str2flo (const char* str, int radix)
{
  register int c, i = 0;
  double lead_sgn;
  double res = 0.0, tmp = 0.0;
  int flg = 0;
  int point = 0;
  ksi_obj second;

  if (*str == '-') lead_sgn = -1.0, ++str;
  else if (*str == '+') lead_sgn = 1.0, ++str;
  else lead_sgn = 0.0;

  if ((str[0] == 'n' || str[0] == 'N')
      && (str[1] == 'a' || str[1] == 'A')
      && (str[2] == 'n' || str[2] == 'N')
      && (str[3] == '.'))
    { /* handle [+-]nan. */
      res = ksi_fp_zero / ksi_fp_zero;
      str += 4;
      if (str[0] == '0') str++;
      goto done;
    }

  if ((str[0] == 'i' || str[0] == 'I')
      && (str[1] == 'n' || str[1] == 'N')
      && (str[2] == 'f' || str[2] == 'F')
      && (str[3] == '.'))
    { /* handle [+-]inf. */
      res = 1.0 / ksi_fp_zero;
      str += 4;
      if (str[0] == '0') str++;
      goto done;
    }

  if (*str == 'i' || *str == 'I')
    { /* handle `+i' and `-i'   */
      if (lead_sgn==0.0) return ksi_false;  /* must have leading sign */
      if (str[1] != '\0') return ksi_false; /* `i' not last character */
      return ksi_rectangular (0.0, lead_sgn);
    }

  for (/**/; str[i]; ++i) /* check initial digits */
    {
      switch (c = str[i])
        {
        case '#':
          c = 0;
          break;
        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9':
          c = c - '0';
          break;
        case 'D': case 'E': case 'F':
          if (radix <= 10) goto out1; /* must be exponent */
        case 'A': case 'B': case 'C':
          c = c - 'A' + 10;
          break;
        case 'd': case 'e': case 'f':
          if (radix <= 10) goto out1;
        case 'a': case 'b': case 'c':
          c = c - 'a' + 10;
          break;
        case '_':
          if (flg)
            continue;
        default:
          goto out1;
        }
      if (c >= radix) return ksi_false; /* bad digit for radix */
      res = res * radix + c;
      flg = 1;                  /* res is valid */
    }

out1:

  if (str[i] == '\0') goto done;

  /* By here, must have seen a digit,
   * or must have next char be a `.'
   */
  if (!flg && str[i] != '.')
    return ksi_false;

  if (str[i] == '/')
    {
      flg = 0;
      for (++i; str[i]; ++i)
        {
          switch (c = str[i])
            {
            case '#':
              c = 0;
              break;
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
              c = c - '0';
              break;
            case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
              c = c - 'A' + 10;
              break;
            case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
              c = c - 'a' + 10;
              break;
            case '_':
              if (flg)
                continue;
            default:
              goto out2;
            }
          if (c >= radix) return ksi_false;
          tmp = tmp * radix + c;
	  flg = 1; /* tmp is valid */
        }
    out2:
      if (!flg) return ksi_false;
      res /= tmp;
      goto done;
    }

  if (str[i] == '.')
    { /* decimal point notation */
      for (++i; str[i]; ++i)
        {
          switch (c = str[i])
            {
            case '#':
              c = 0;
              break;
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
              c = c - '0';
              break;
            case 'D': case 'E': case 'F':
              if (radix <= 10) goto out3; /* must be exponent */
            case 'A': case 'B': case 'C':
              c = c - 'A' + 10;
              break;
            case 'd': case 'e': case 'f':
              if (radix <= 10) goto out3;
            case 'a': case 'b': case 'c':
              c = c - 'a' + 10;
              break;
            case '_':
              if (flg)
                continue;
            default:
              goto out3;
            }
          if (c >= radix) return ksi_false; /* bad digit for radix */
          res = res * radix + c;
          point--;
          flg = 1;
        }
    out3:
      if (!flg) return ksi_false;/* no digits before or after decimal point */
      if (str[i] == '\0') goto adjust;
    }

  switch (str[i])
    { /* exponent */
    case 'd': case 'D':
    case 'e': case 'E':
    case 'f': case 'F':
    case 'l': case 'L':
    case 's': case 'S':
      {
        int expsgn = 1, expon = 0;
        if (!str[++i]) return ksi_false; /* bad exponent */
        switch (str[i])
          {
          case '-':  expsgn=(-1);
          case '+':  if (!str[++i]) return ksi_false; /* bad exponent */
          }
        for (; str[i]; ++i)
          {
            switch (c = str[i])
              {
              case '0': case '1': case '2': case '3': case '4':
              case '5': case '6': case '7': case '8': case '9':
                c = c - '0';
                break;
              case 'D': case 'E': case 'F': case 'A': case 'B': case 'C':
                c = c - 'A' + 10;
                break;
              case 'd': case 'e': case 'f': case 'a': case 'b': case 'c':
                c = c - 'a' + 10;
                break;
              case '_':
                if (flg)
                  continue;
              default:
                goto out4;
              }
            if (c >= radix) return ksi_false; /* bad digit for radix */
            expon = expon * radix + c;
          }
      out4:
        point += expsgn * expon;
      }
    }

adjust:
  if (point >= 0)
    while (point--) res *= radix;
  else
    while (point++) res /= radix;

done:
  /* at this point, we have a legitimate floating point result */
  if (lead_sgn==-1.0)  res = -res;
  if (str[i] == '\0')
    return ksi_double2num (res);

  if (str[i] == 'i' || str[i] == 'I')
    { /* pure imaginary number  */
      if (lead_sgn==0.0) return ksi_false;    /* must have leading sign */
      if (str[i+1] != '\0') return ksi_false; /* `i' not last character */
      return ksi_rectangular (0.0, res);
    }

  switch (str[i])
    {
    case '-': case '+':
      break;
    case '@':
      { /* polar input for complex number */
        /* get a `real' for angle */
        second = ksi_str2flo (&str[i+1], radix);
        if (!(KSI_FLONUM_P(second))) return ksi_false; /* not `real' */
	tmp = KSI_REPART (second);
        return ksi_polar (res, tmp);
      }
    default: return ksi_false;
    }

  second = ksi_str2flo (&str[i], radix);
  if (KSI_FLONUM_P (second))
    {
      /* handles x+0i and x-0i */
      tmp = KSI_REPART (second);
      if (tmp != 0.0) return ksi_false;
    }
  else
    {
      if (!KSI_COMPLEX_P (second)) return ksi_false;
      if (KSI_REPART (second) != 0.0) return ksi_false;
      tmp = KSI_IMPART (second);
    }

  return ksi_rectangular (res, tmp);
}


/*
 * ksi_str2int
 *	  ksi_str2num      .
 *	RADIX    2, 8, 10, 16.
 *
 * Return
 *	    ,    
 *	   .     ksi_false.
 */

static ksi_obj
ksi_str2int (const char* str, int radix)
{
  int neg = 0, c;
  unsigned long len, flg = 0;
  ksi_obj val = ksi_zero;

  if (*str == '-') ++neg, ++str;
  else if (*str == '+') ++str;

  /* estimate size of the resulting number */
  len = strlen (str);
  if (radix == 10)      len = 1 + (84 * len) / (KSI_INT_SIZE * 25);
  else if (radix == 8)  len = 1 + (3 * len) / KSI_INT_SIZE;
  else if (radix == 16) len = 1 + (4 * len) / KSI_INT_SIZE;

  while (*str)
    {
      switch (c = *str++)
        {
        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9':
          c = c - '0';
          break;
        case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
          c = c - 'a' + 10;
          break;
        case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
          c = c - 'A' + 10;
          break;
        case '_':
          if (flg)
            continue;
        default:
          return ksi_false;
        }
      if (c >= radix)
        return ksi_false;

      val = ksi_mul (val, KSI_MK_SINT (radix));
      if (neg)
        val = ksi_sub (val, KSI_MK_SINT (c));
      else
        val = ksi_add (val, KSI_MK_SINT (c));
      flg = 1;
    }

  return val;
}


/*
 * ksi_str02num
 *	   .
 *	RADIX    2, 8, 10, 16.
 *
 * Return
 *	       ,  ksi_false.
 *
 */

ksi_obj
ksi_str02num (const char* str, int radix)
{
  ksi_obj val;
  int rd = radix, ex = 0, cc = (radix == 0 ? 1 : 0);

  if (rd && rd != 2 && rd != 8 && rd != 10 && rd != 16)
    ksi_exn_error ("range", ksi_int2num (rd), "string->number: invalid radix");

  if (str == 0 || str[0] == '\0')
    return ksi_false;

  while (str[0] == '#')
    {
      cc = 0;
      switch (str[1])
        {
        case 'i': case 'I': if (ex) return ksi_false; ex = 2;  break;
        case 'e': case 'E': if (ex) return ksi_false; ex = 1;  break;
        case 'b': case 'B': if (rd) return ksi_false; rd = 2;  break;
        case 'o': case 'O': if (rd) return ksi_false; rd = 8;  break;
        case 'd': case 'D': if (rd) return ksi_false; rd = 10; break;
        case 'x': case 'X': if (rd) return ksi_false; rd = 16; break;
        default:            return ksi_false;
        }
      str += 2;
    }

  if ((str[0] == '-' || str[0] == '+') && str[1] == '\0')
    return ksi_false;

  if (rd == 0) rd = 10;
  if (ex == 0)
    {
      const char* p;
      for (p = str; *p; ++p)
        {
          if (strchr (rd <= 10 ? ".#@/isfdleISFDLE" : ".#@/islISL", *p))
            return ksi_str2flo (str, rd);
        }
    }
  if (ex == 2)
    return ksi_str2flo (str, rd);

  if (cc && ksi_spec_val ("*read-c-number*", ksi_false, 0) != ksi_false)
    {
      if (str [0] == '0' && str [1] != '\0')
	{
	  if (str [1] == 'x' || str [1] == 'X')
	    val = ksi_str2int (str+2, 16);
	  else
	    val = ksi_str2int (str+1, 8);
	  if (val != ksi_false)
	    return val;
	}
    }

  val = ksi_str2int (str, rd);
  if (ex == 0 && rd > 10 && val == ksi_false)
    val = ksi_str2flo (str, rd);

  return val;
}


static char*
double2str (double f)
{
  static double fx[] =
  {
    0.0,
    5e-1,  5e-2,  5e-3,  5e-4,  5e-5,
    5e-6,  5e-7,  5e-8,  5e-9,  5e-10,
    5e-11, 5e-12, 5e-13, 5e-14, 5e-15,
    5e-16, 5e-17, 5e-18, 5e-19, 5e-20
  };

  ksi_obj dbl_prec;
  int exp, wp, ch, efmt, dpt, d, i;
  char* a;

  if (f != f) return "nan.";
  if (f == f/2)
    {
      if (copysign (1.0, f) < 0.0)
	return (f == 0.0 ? "-0.0" : "-inf.");
      return (f == 0.0 ? "0.0" : "+inf.");
    }

  dbl_prec = ksi_spec_val ("*double-precision*", 0, 0);
  wp = (KSI_SINT_P (dbl_prec) ? KSI_SINT_COD (dbl_prec) : 14);
  a = (char*) ksi_malloc_data (wp * 2);
  exp = 0; ch = 0;

  if (f < 0.0) { a[ch++] = '-'; f = -f; }

  if (wp >= sizeof (fx) / sizeof (double))
    wp = sizeof (fx) / sizeof (double) - 1;

  while (f < 1.0) { f *= 10.0; exp--; }
  while (f > 10.0) { f *= 0.1; exp++; }
  if (f + fx[wp] >= 10.0) { f = 1.0; exp++; }

  efmt = (exp < -3) || (exp > wp+2);
  if (!efmt)
    {
      if (exp < 0)
	{
	  a[ch++] = '0';
	  a[ch++] = '.';
	  dpt = exp;
	  while (++dpt)  a[ch++] = '0';
	}
      else
	dpt = exp+1;
    }
  else
    dpt = 1;

  do
    {
      d = (int) f;
      f -= d;
      a[ch++] = d+'0';
      if (f < fx[wp])  break;
      if (f+fx[wp] >= 1.0)
	{
	  a[ch-1]++;
	  break;
	}
      f *= 10.0;
      if (!(--dpt))  a[ch++] = '.';
    }
  while (wp--);

  if (dpt > 0)
    {
      if ((dpt > 4) && (exp > 6))
	{
	  d = (a[0] == '-' ? 2 : 1);
	  for (i = ch++; i > d; i--)
	    a[i] = a[i-1];
	  a[d] = '.';
	  efmt = 1;
	}
      else
	{
	  while (--dpt)  a[ch++] = '0';
	  a[ch++] = '.';
	}
    }

  if (a[ch-1]=='.')  a[ch++]='0'; /* trailing zero */
  if (efmt && exp)
    {
      a[ch++] = 'E';
      if (exp < 0)
	{
	  exp = -exp;
	  a[ch++] = '-';
	}
      for (i = 10; i <= exp; i *= 10);
      for (i /= 10; i; i /= 10)
	{
	  a[ch++] = exp/i + '0';
	  exp %= i;
	}
    }

  a[ch] = '\0';
  return a;
}


/*
 * str_reverse
 *	      .
 *	  fixnum2str
 */

static void
str_reverse (char* str)
{
  char *lp = str, *rp = str + strlen (str) - 1;
  while (lp < rp)
    {
      char c = *lp;
      *lp++ = *rp;
      *rp-- = c;
    }
}


/*
 * fixnum2str
 *	    
 *	  ksi_num2str
 */

static char*
fixnum2str (ksi_obj b, unsigned radix)
{
  int size, sign, k=0, i, d;
  char *buffer, *str;

  if (KSI_BIGNUM_P (b))
    {
      sign = ((ksi_bignum) b) -> sign;
      size = ABS (sign);
    }
  else
    {
      sign = KSI_SINT_COD (b);
      size = 1;
    }
  switch (radix)
    {
    case 2:  k = (size * KSI_INT_SIZE);				break;
    case 8:  k = (size * KSI_INT_SIZE) / 3;			break;
    case 10: k = (size * KSI_INT_SIZE * 241L) / 800 + 1;	break;
    case 16: k = (size * KSI_INT_SIZE) / 4;			break;
    default:
      ksi_exn_error ("range", ksi_int2num (radix),
		     "fixnum->string: invalid radix");
    }

  buffer = (char*) ksi_malloc_data (k+2);
  str = buffer;
  if (sign < 0)
    *str++ = '-';

  i = 0;
  if (KSI_BIGNUM_P (b))
    {
      while (i < k)
        {
          int r = radix;
	  ksi_bignum x = (ksi_bignum) b;

          b = divmod_big_i (x -> sign, x -> digs, &r);
          str[i++] = (char) (r < 10 ? r + '0' : (r-10) + 'a');
          if (b == ksi_zero)
            goto finish;
          if (KSI_SINT_P (b))
            break;
        }
    }

  d = KSI_SINT_COD (b);
  d = ABS (d);
  while (i < k)
    {
      int r = d % radix;
      d = d / radix;
      str[i++] = (char) (r < 10 ? r + '0' : (r-10) + 'a');
      if (d == 0)
        break;
    }

finish:
  str[i] = '\0';
  str_reverse (str);
  return buffer;
}


/*
 * ksi_num2str
 *	   .
 *
 * Return
 *	      
 *	 X   RADIX.
 *	   __ .
 *	    .
 *
 * Bugs
 *	     RADIX    10.
 *
 *	       ,
 *	  char*,   const char*.    
 *	  .
 */

char*
ksi_num2str (ksi_obj x, int radix)
{
  if (radix == 0)
    radix = 10;
  if (KSI_SINT_P (x) || KSI_BIGNUM_P (x))
    return fixnum2str (x, radix);

  if (radix != 10)
    ksi_exn_error ("range", ksi_int2num (radix),
		   "number->string: can't use radix `%d' for flonum", radix);

  if (KSI_FLONUM_P (x))
    return double2str (KSI_REPART (x));

  if (KSI_COMPLEX_P (x))
    {
      char* rp = double2str (KSI_REPART (x));
      char* ip = double2str (KSI_IMPART (x));
      return ksi_aprintf ("%s%s%si",
			  rp,
			  (*ip == '-' || *ip == '+') ? "" : "+",
			  ip);
    }

  return "#<invalid number>";
}


/*
 * ksi_double2num
 *	  "   "
 *	   V.
 *
 */

ksi_obj
ksi_double2num (double v)
{
  ksi_flonum num = (ksi_flonum) ksi_malloc_data (sizeof *num);
  num->itag = KSI_TAG_FLONUM;
  num->real = v;
  return (ksi_obj) num;
}

ksi_obj
ksi_rectangular (double r, double i)
{
  ksi_complex num;

  if (i == 0.0 && copysign (1.0, i) > 0.0)
    return ksi_double2num (r);

  num = (ksi_complex) ksi_malloc_data (sizeof *num);
  num->itag = KSI_TAG_COMPLEX;
  num->real = r;
  num->imag = i;
  return (ksi_obj) num;
}

ksi_obj
ksi_polar (double x, double a)
{
  return ksi_rectangular (x * cos (a), x * sin (a));
}

ksi_obj
ksi_int2num (long v)
{
  if (KSI_MIN_INT_TYPE < v && v < KSI_MAX_INT_TYPE)
    return KSI_MK_SINT (v);

#if SIZEOF_INT == SIZEOF_LONG
  {
    ksi_bignum num = alloc_big (1);
    num->sign = (v < 0 ? -1 : 1);
    num->digs[0] = ABS (v);
    return adjust_big (num);
  }
#else
  {
    unsigned *digs;
    int size, sign, i;

    size = (sizeof v) / (sizeof *digs);
    sign = (v < 0 ? -size : size);
    v    = (v < 0 ? -v : v);
    digs = (unsigned*) alloca (size * (sizeof *digs));
    for (i = 0; i < size; i++)
      {
	digs[i] = (unsigned) (v & KSI_INT_MASK);
	v >>= KSI_INT_SIZE;
      }

    return new_big (sign, digs);
  }
#endif
}

ksi_obj
ksi_uint2num (unsigned long v)
{
  if (v < KSI_MAX_INT_TYPE)
    return KSI_MK_SINT (v);

#if SIZEOF_INT == SIZEOF_LONG
  {
    ksi_bignum num = alloc_big (1);
    num->sign = 1;
    num->digs[0] = v;
    return adjust_big (num);
  }
#else
  {
    unsigned *digs;
    int size, i;

    size = (sizeof v) / (sizeof *digs);
    digs = (unsigned*) alloca (size * (sizeof *digs));
    for (i = 0; i < size; i++)
      {
	digs[i] = (unsigned) (v & KSI_INT_MASK);
	v >>= KSI_INT_SIZE;
      }

    return new_big (size, digs);
  }
#endif
}

#ifdef LONG_LONG
ksi_obj
ksi_long2num (LONG_LONG v)
{
  unsigned *digs;
  int size, sign, i;

  size = (sizeof v) / (sizeof *digs);
  sign = (v < 0 ? -size : size);
  v    = (v < 0 ? -v : v);
  digs = (unsigned*) alloca (size * (sizeof *digs));
  for (i = 0; i < size; i++)
    {
      digs[i] = (unsigned) (v & KSI_INT_MASK);
      v >>= KSI_INT_SIZE;
    }

  return new_big (sign, digs);
}

ksi_obj
ksi_ulong2num (unsigned LONG_LONG v)
{
  unsigned *digs;
  int size, i;

  size = (sizeof v) / (sizeof *digs);
  digs = (unsigned*) alloca (size * (sizeof *digs));
  for (i = 0; i < size; i++)
    {
      digs[i] = (unsigned) (v & KSI_INT_MASK);
      v >>= KSI_INT_SIZE;
    }

  return new_big (size, digs);
}
#endif

ksi_obj
ksi_make_bignum (int sign, unsigned* digits)
{
  return new_big (sign, digits);
}

static double
ksi_2double (ksi_obj x, char* fun)
{
  if (KSI_SINT_P (x))
    return (double) KSI_SINT_COD (x);

  if (KSI_BIGNUM_P (x))
    {
      double d = 0.0;
      int sz = ABS (((ksi_bignum) x) -> sign);
      while (--sz >= 0)
        {
          d *= BIG_RADIX;
          d += ((ksi_bignum) x) -> digs [sz];
        }
      return ((ksi_bignum) x) -> sign < 0 ? -d : d;
    }

  if (KSI_FLONUM_P (x))
    return KSI_REPART (x);

  ksi_exn_error ("type", x, "%s: invalid number", fun);
  return 0.0;
}

double
ksi_real_part (ksi_obj x)
{
  if (KSI_COMPLEX_P (x))
    return KSI_REPART (x);

  return ksi_2double (x, "real-part");
}

double
ksi_imag_part (ksi_obj x)
{
  if (KSI_COMPLEX_P (x))
    return KSI_IMPART (x);

  TYPE_CHECK (x, KSI_NUM_P (x), "imag-part: invalid number");
  return 0.0;
}

double
ksi_magnitude (ksi_obj x)
{
  return ksi_num2double (ksi_abs (x));
}

double
ksi_angle (ksi_obj z)
{
  double x=0.0, y = 0.0;
  if (KSI_SINT_P (z)) x = KSI_SINT_COD (z) >= 0 ? 1.0 : -1.0;
  else if (KSI_BIGNUM_P (z)) x = ((ksi_bignum) z) -> sign > 0 ? 1.0 : -1.0;
  else if (KSI_FLONUM_P (z)) x = KSI_REPART (z);
  else if (KSI_COMPLEX_P (z)) x = KSI_REPART (z), y = KSI_IMPART (z);
  else KSI_EXN_ERR ("type", z, "angle: invalid number");
  return atan2 (y, x);
}

double
ksi_num2double (ksi_obj x)
{
  return ksi_2double (x, "ksi_num2double");
}

long
ksi_num2int (ksi_obj x)
{
  if (KSI_SINT_P (x))
    return (long) KSI_SINT_COD (x);

  if (KSI_BIGNUM_P (x))
    {
#if SIZEOF_INT == SIZEOF_LONG
      long i = (long) (((ksi_bignum) x) -> digs [0]);
      return (long) (((ksi_bignum) x) -> sign < 0 ? -i : i);
#else
      long v = 0;
      int sign = ((ksi_bignum) x) -> sign;
      int size = ABS (sign);
      while (--size >= 0)
	{
	  v <<= KSI_INT_SIZE;
	  v  |= ((ksi_bignum) x) -> digs [size];
	}

      return (sign < 0 ? -v : v);
#endif
    }

  if (KSI_FLONUM_P (x))
    return (long) KSI_REPART (x);

  KSI_EXN_ERR ("type", x, "ksi_num2int: invalid number");
  return 0;
}

unsigned long
ksi_num2uint (ksi_obj x)
{
  if (KSI_SINT_P (x))
    return (unsigned long) KSI_SINT_COD (x);

  if (KSI_BIGNUM_P (x))
    {
#if SIZEOF_INT == SIZEOF_LONG
      unsigned long i = (unsigned long) (((ksi_bignum) x) -> digs [0]);
      return (unsigned long) (((ksi_bignum) x) -> sign < 0 ? -i : i);
#else
      unsigned long v = 0;
      int sign = ((ksi_bignum) x) -> sign;
      int size = ABS (sign);
      while (--size >= 0)
	{
	  v <<= KSI_INT_SIZE;
	  v  |= ((ksi_bignum) x) -> digs [size];
	}

      return (sign < 0 ? -v : v);
#endif
    }

  if (KSI_FLONUM_P (x))
    return (unsigned long) KSI_REPART (x);

  KSI_EXN_ERR ("type", x, "ksi_num2uint: invalid number");
  return 0;
}

#ifdef LONG_LONG
LONG_LONG
ksi_num2long (ksi_obj x)
{
  if (KSI_SINT_P (x))
    return KSI_SINT_COD (x);

  if (KSI_BIGNUM_P (x))
    {
      LONG_LONG v = 0;
      int sign = ((ksi_bignum) x) -> sign;
      int size = ABS (sign);
      while (--size >= 0)
	{
	  v <<= KSI_INT_SIZE;
	  v  |= ((ksi_bignum) x) -> digs [size];
	}

      return (sign < 0 ? -v : v);
    }

  if (KSI_FLONUM_P (x))
    return (LONG_LONG) KSI_REPART (x);

  KSI_EXN_ERR ("type", x, "ksi_num2long: invalid number");
  return 0;
}

unsigned LONG_LONG
ksi_num2ulong (ksi_obj x)
{
  if (KSI_SINT_P (x))
    return KSI_SINT_COD (x);

  if (KSI_BIGNUM_P (x))
    {
      unsigned LONG_LONG v = 0;
      int sign = ((ksi_bignum) x) -> sign;
      int size = ABS (sign);
      while (--size >= 0)
	{
	  v <<= KSI_INT_SIZE;
	  v  |= ((ksi_bignum) x) -> digs [size];
	}

      return v;
    }

  if (KSI_FLONUM_P (x))
    return (unsigned LONG_LONG) KSI_REPART (x);

  KSI_EXN_ERR ("type", x, "ksi_num2ulong: invalid number");
  return 0;
}
#endif

ksi_obj
ksi_double2exact (double d)
{
  double u;
  int sign, i;
  unsigned c, *digs;

  d = floor (d + 0.5);
  if (KSI_MIN_INT_TYPE < d && d < KSI_MAX_INT_TYPE)
    return KSI_MK_SINT (d);

  u = d < 0.0 ? -d : d;
  for (i = 0; 0.0 != floor (u); u /= BIG_RADIX, i++);

  TALLOC (digs, unsigned, i);
  sign = d < 0.0 ? -i : i;

  while (--i >= 0)
    {
      u *= BIG_RADIX;
      c = (unsigned) floor (u);
      u -= c;
      digs [i] = c;
    }

  return new_big (sign, digs);
}


ksi_obj
ksi_exact2inexact (ksi_obj x)
{
  if (KSI_COMPLEX_P (x) || KSI_FLONUM_P (x))
    return x;

  return ksi_double2num (ksi_2double (x, "exact->inexact"));
}

ksi_obj
ksi_inexact2exact (ksi_obj x)
{
  if (KSI_EXACT_P (x))
    return x;

  if (KSI_FLONUM_P (x))
    return ksi_double2exact (KSI_REPART (x));

  KSI_EXN_ERR ("type", x, "inexact->exact: invalid number");
  return 0;
}

ksi_obj
ksi_number2string (ksi_obj num, ksi_obj rad)
{
  char *ptr;

  TYPE_CHECK (num, KSI_NUM_P (num), "number->string: invalid number in arg1");
  if (!rad)
    rad = KSI_MK_SINT (10);
  else
    TYPE_CHECK (rad, KSI_SINT_P (rad), "number->string: invalid radix in arg2");

  ptr = ksi_num2str (num, KSI_SINT_COD (rad));
  return ksi_str02string (ptr);
}

ksi_obj
ksi_string2number (ksi_obj str, ksi_obj rad)
{
  TYPE_CHECK (str, KSI_STR_P (str), "string->number: invalid string in arg1");

  if (!rad)
    rad = ksi_zero;
  else
    TYPE_CHECK (rad, KSI_SINT_P (rad), "string->number: invalid radix in arg2");

  return ksi_str02num (KSI_STR_PTR (str), KSI_SINT_COD (rad));
}


ksi_obj
ksi_number_p (ksi_obj x)
{
  return KSI_NUM_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_complex_p (ksi_obj x)
{
  return KSI_NUM_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_real_p (ksi_obj x)
{
  return KSI_REAL_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_rational_p (ksi_obj x)
{
  return KSI_REAL_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_integer_p (ksi_obj x)
{
  if (KSI_SINT_P (x) || KSI_BIGNUM_P (x))
    return ksi_true;

  if (KSI_FLONUM_P (x))
    {
      if (IS_INT (KSI_REPART (x)))
	return ksi_true;
    }

  return ksi_false;
}

ksi_obj
ksi_exact_p (ksi_obj x)
{
  return KSI_EXACT_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_inexact_p (ksi_obj x)
{
  return KSI_INEXACT_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_zero_p (ksi_obj x)
{
  if (x == ksi_zero)
    return ksi_true;

  if (KSI_SINT_P (x) || KSI_BIGNUM_P (x))
    return ksi_false;

  if (KSI_FLONUM_P (x))
    {
      if (KSI_REPART (x) == 0.0)
	return ksi_true;
      return ksi_false;
    }

  if (KSI_COMPLEX_P (x))
    return ksi_false;

  KSI_EXN_ERR ("type", x, "zero?: invalid number");
  return ksi_false;
}

ksi_obj
ksi_positive_p (ksi_obj x)
{
  if (KSI_SINT_P (x))
    {
      if (KSI_SINT_COD (x) > 0)
        return ksi_true;
      return ksi_false;
    }

  if (KSI_BIGNUM_P (x))
    {
      if (((ksi_bignum) x) -> sign > 0)
        return ksi_true;
      return ksi_false;
    }

  if (KSI_FLONUM_P (x))
    {
      if (KSI_REPART (x) > 0.0)
        return ksi_true;
      return ksi_false;
    }

  KSI_EXN_ERR ("type", x, "positive?: invalid number");
  return ksi_false;
}

ksi_obj
ksi_negative_p (ksi_obj x)
{
  if (KSI_SINT_P (x))
    {
      if (KSI_SINT_COD (x) < 0)
        return ksi_true;
      return ksi_false;
    }

  if (KSI_BIGNUM_P (x))
    {
      if (((ksi_bignum) x) -> sign < 0)
        return ksi_true;
      return ksi_false;
    }

  if (KSI_FLONUM_P (x))
    {
      if (KSI_REPART (x) < 0.0)
        return ksi_true;
      return ksi_false;
    }

  KSI_EXN_ERR ("type", x, "negative?: invalid number");
  return ksi_false;
}

ksi_obj
ksi_odd_p (ksi_obj x)
{
  if (KSI_SINT_P (x))
    return (KSI_SINT_COD (x) & 1) ? ksi_true : ksi_false;

  if (KSI_BIGNUM_P (x))
    {
      if ((((ksi_bignum) x) -> digs [0] & 1) != 0)
        return ksi_true;
      return ksi_false;
    }

  if (KSI_FLONUM_P (x))
    {
      double d = KSI_REPART (x);
      if (IS_INT (d))
	{
	  if (fmod (d, 2.0) == 1.0)
	    return ksi_true;
	  return ksi_false;
	}
    }

  KSI_EXN_ERR ("type", x, "odd?: invalid integer");
  return ksi_false;
}

ksi_obj
ksi_even_p (ksi_obj x)
{
  if (KSI_SINT_P (x))
    return (KSI_SINT_COD (x) & 1) ? ksi_false : ksi_true;

  if (KSI_BIGNUM_P (x))
    {
      if ((((ksi_bignum) x) -> digs [0] & 1) == 0)
        return ksi_true;
      return ksi_false;
    }

  if (KSI_FLONUM_P (x))
    {
      double d = KSI_REPART (x);
      if (IS_INT (d))
	{
	  if (fmod (d, 2.0) == 0.0)
	    return ksi_true;
	  return ksi_false;
	}
    }

  KSI_EXN_ERR ("type", x, "even?: invalid integer");
  return ksi_false;
}


/*
 * ksi_num_eqv_p
 *	  .
 *
 * Return
 *	 ksi_true   ,  ksi_false  
 *	.
 *
 * Side effects
 *	     ,  .
 */

ksi_obj
ksi_num_eqv_p (ksi_obj x1, ksi_obj x2)
{
  int i;

  if (x1 == x2)
    return ksi_true;

  if (KSI_SINT_P (x1))
    {
      /* bignum must be adjusted, so x1 != x2 */
      if (KSI_SINT_P (x2) || KSI_BIGNUM_P (x2))
	return ksi_false;

      i = KSI_SINT_COD (x1);
      if (KSI_FLONUM_P (x2))
	return (((double) i) == KSI_REPART (x2) ? ksi_true : ksi_false);

      if (KSI_COMPLEX_P (x2))
	return ksi_false;

      KSI_ERR (x2, "= : invalid number");
    }

  if (KSI_BIGNUM_P (x1))
    {
      /* bignum must be adjusted, so x1 != x2 */
      if (KSI_SINT_P (x2))
	return ksi_false;

      if (KSI_BIGNUM_P (x2))
	{
	  if (((ksi_bignum) x1) -> sign == ((ksi_bignum) x2) -> sign)
	    {
	      int sz = ABS (((ksi_bignum) x1) -> sign);
	      unsigned *p1 = ((ksi_bignum) x1) -> digs;
	      unsigned *p2 = ((ksi_bignum) x2) -> digs;
	      while (--sz >= 0)
		if (p1 [sz] != p2 [sz])
		  return ksi_false;
	      return ksi_true;
	    }
	  return ksi_false;
	}

      if (KSI_FLONUM_P (x2))
	{
	  return (ksi_2double (x1, "= ") == KSI_REPART (x2)
		  ? ksi_true : ksi_false);
	}

      if (KSI_COMPLEX_P (x2))
	return ksi_false;

      KSI_EXN_ERR ("type", x2, "= : invalid number");
    }

  if (KSI_FLONUM_P (x1))
    {
      if (KSI_COMPLEX_P (x2))
	return ksi_false;

      return (KSI_REPART (x1) == ksi_2double (x2, "= ")
	      ? ksi_true : ksi_false);
    }

  if (KSI_COMPLEX_P (x1))
    {
      if (KSI_COMPLEX_P (x2))
	{
	  if (KSI_REPART (x1) == KSI_REPART (x2) &&
	      KSI_IMPART (x1) == KSI_IMPART (x2))
	    return ksi_true;
	  return ksi_false;
	}

      if (KSI_NUM_P (x2))
	return ksi_false;

      KSI_ERR (x2, "= : invalid number");
    }

  KSI_EXN_ERR ("type", x1, "= : invalid number");
  return ksi_false;
}


/*
 * ksi_num_eq_p
 *	  .
 *
 * Scheme procedure
 *	= argN ...
 *
 * Return
 *	 ksi_true   ,  ksi_false  
 *	.
 *
 * Side effects
 *	     ,  .
 * 
 */

ksi_obj
ksi_num_eq_p (int argc, ksi_obj* argv)
{
  int i;
  for (i = 1; i < argc; i++)
    {
      if (ksi_num_eqv_p (argv [0], argv [i]) == ksi_false)
	return ksi_false;
    }

  return ksi_true;
}


/*
 * ksi_less_p
 *	  
 *
 * Return
 *	 1  X1  X2,  0   .
 *
 * Side effects
 *	     ,  .
 *
 */

static int
ksi_less_p (ksi_obj x1, ksi_obj x2, char* fname)
{
  if (KSI_SINT_P (x1))
    {
      if (KSI_SINT_P (x2))
        return (KSI_SINT_COD (x1) < KSI_SINT_COD (x2));

      if (KSI_BIGNUM_P (x2))
	return (((ksi_bignum) x2) -> sign > 0);

      if (KSI_FLONUM_P (x2))
        return (KSI_SINT_COD (x1) < KSI_REPART (x2));

      x1 = x2;
      goto badnum;
    }

  if (KSI_BIGNUM_P (x1))
    {
      /* bignum must be normalized, so compare only signs if possible */

      if (KSI_SINT_P (x2))
	return (((ksi_bignum) x1) -> sign < 0);

      if (KSI_BIGNUM_P (x2))
	{
          int s1 = ((ksi_bignum) x1) -> sign;
          int s2 = ((ksi_bignum) x2) -> sign;
          if (s1 == s2)
            {
              int i = ABS (s1);
              while (--i >= 0)
                {
                  if (((ksi_bignum) x1) -> digs[i]
		      > ((ksi_bignum) x2) -> digs[i])
                    {
                      s1 *= 2;
                      break;
                    }
                  else if (((ksi_bignum) x1) -> digs[i]
			   < ((ksi_bignum) x2) -> digs[i])
                    {
                      s2 *= 2;
                      break;
                    }
                }
            }

          return (s1 < s2);
        }

      if (KSI_FLONUM_P (x2))
        return (ksi_2double (x1, fname) < KSI_REPART (x2));

      x1 = x2;
      goto badnum;
    }

  if (KSI_FLONUM_P (x1))
    {
      if (KSI_SINT_P (x2))
        return (KSI_REPART (x1) < KSI_SINT_COD (x2));

      if (KSI_BIGNUM_P (x2))
        return (KSI_REPART (x1) < ksi_2double (x2, fname));

      if (KSI_FLONUM_P (x2))
        return (KSI_REPART (x1) < KSI_REPART (x2));

      x1 = x2;
      goto badnum;
    }

badnum:
  ksi_exn_error ("type", x1, "%s: invalid number", fname);
  return 0;
}


/*
 * ksi_num_lt_p, ksi_num_le_p, ksi_num_gt_p, ksi_num_ge_p
 *	  .
 *
 * Scheme procedure
 *	< argN ...
 *	<= argN ...
 *	> argN ...
 *	>= argN ...
 */

#define DEF_CMP(fun, cmp)			\
ksi_obj						\
fun (int ac, ksi_obj* av)			\
{						\
  int i;					\
  for (i = 1; i < ac; i++)			\
    {						\
      if (cmp)					\
        return ksi_false;			\
    }						\
  return ksi_true;				\
}

DEF_CMP (ksi_num_lt_p, !ksi_less_p(av[i-1], av[i], "<"))
DEF_CMP (ksi_num_gt_p, !ksi_less_p(av[i], av[i-1], ">"))
DEF_CMP (ksi_num_le_p, ksi_less_p(av[i], av[i-1], "<="))
DEF_CMP (ksi_num_ge_p, ksi_less_p(av[i-1], av[i], ">="))


/*
 * ksi_min, ksi_max
 *	   
 *
 * Scheme procedure
 *	min arg1 argN ...
 *	max arg1 argN ...
 *
 * Bugs
 */

ksi_obj
ksi_min (int argc, ksi_obj* argv)
{
  ksi_obj x;
  int inexact, i;

  x = argv[0];
  inexact = KSI_INEXACT_P (x);
  for (i = 1; i < argc; i++)
    {
      if (KSI_INEXACT_P (argv[i]))
	inexact = 1;

      if (ksi_less_p (argv[i], x, "min"))
        x = argv[i];
    }

  if (inexact && KSI_EXACT_P (x))
    return ksi_exact2inexact (x);
  return x;
}

ksi_obj
ksi_max (int argc, ksi_obj* argv)
{
  ksi_obj x;
  int inexact, i;

  x = argv[0];
  inexact = KSI_INEXACT_P (x);
  for (i = 1; i < argc; i++)
    {
      if (KSI_INEXACT_P (argv[i]))
	inexact = 1;

      if (ksi_less_p (x, argv[i], "max"))
        x = argv[i];
    }

  if (inexact && KSI_EXACT_P (x))
    return ksi_exact2inexact (x);
  return x;
}


ksi_obj
ksi_add (ksi_obj a, ksi_obj b)
{
  if (KSI_SINT_P (a))
    {
      if (KSI_SINT_P (b))
	return ksi_int2num (KSI_SINT_COD (a) + KSI_SINT_COD (b));

      if (KSI_BIGNUM_P (b))
        {
	  int i = KSI_SINT_COD (a);
          if (i > 0)
            return add_big_ui (((ksi_bignum) b) -> digs,
			       ((ksi_bignum) b) -> sign, i);
          if (i < 0)
            return sub_big_ui (((ksi_bignum) b) -> digs,
			       ((ksi_bignum) b) -> sign, -i);
          return b;
        }

      if (KSI_FLONUM_P (b))
        return ksi_double2num (KSI_SINT_COD (a) + KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	return ksi_rectangular (KSI_SINT_COD (a) + KSI_REPART (b),
				KSI_IMPART (b));

      KSI_EXN_ERR ("type", b, "+ : invalid number");
    }

  if (KSI_BIGNUM_P (a))
    {
      if (KSI_SINT_P (b))
        {
	  int i = KSI_SINT_COD (b);
          if (i > 0)
            return add_big_ui (((ksi_bignum) a) -> digs,
			       ((ksi_bignum) a) -> sign, i);
          if (i < 0)
            return sub_big_ui (((ksi_bignum) a) -> digs,
			       ((ksi_bignum) a) -> sign, -i);
          return a;
        }

      if (KSI_BIGNUM_P (b))
        {
          if (((ksi_bignum) b) -> sign < 0)
            return sub_big ((ksi_bignum) a, (ksi_bignum) b);
	  return add_big ((ksi_bignum) a, (ksi_bignum) b);
        }

      if (KSI_FLONUM_P (b))
        return ksi_double2num (ksi_2double (a, "+ ") + KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	return ksi_rectangular (ksi_2double (a, "+ ") + KSI_REPART (b),
				KSI_IMPART (b));

      KSI_EXN_ERR ("type", b, "+ : invalid number");
    }

  if (KSI_FLONUM_P (a))
    {
      if (KSI_SINT_P (b))
	return ksi_double2num (KSI_REPART (a) + KSI_SINT_COD (b));

      if (KSI_BIGNUM_P (b))
	return ksi_double2num (KSI_REPART (a) + ksi_2double (b, "+ "));

      if (KSI_FLONUM_P (b))
	return ksi_double2num (KSI_REPART (a) + KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	return ksi_rectangular (KSI_REPART (a) + KSI_REPART (b),
				KSI_IMPART (b));

      KSI_EXN_ERR ("type", b, "+ : invalid number");
    }

  if (KSI_COMPLEX_P (a))
    {
      if (KSI_SINT_P (b))
	return ksi_rectangular (KSI_REPART (a) + KSI_SINT_COD (b),
				KSI_IMPART (a));

      if (KSI_BIGNUM_P (b))
	return ksi_rectangular (KSI_REPART (a) + ksi_2double (b, "+ "),
				KSI_IMPART (a));

      if (KSI_FLONUM_P (b))
	return ksi_rectangular (KSI_REPART (a) + KSI_REPART (b),
				KSI_IMPART (a));

      if (KSI_COMPLEX_P (b))
	return ksi_rectangular (KSI_REPART (a) + KSI_REPART (b),
				KSI_IMPART (a) + KSI_IMPART (b));

      KSI_EXN_ERR ("type", b, "+ : invalid number");
    }

  KSI_EXN_ERR ("type", a, "+ : invalid number");
  return ksi_zero;
}


ksi_obj
ksi_sub (ksi_obj a, ksi_obj b)
{
  if (KSI_SINT_P (a))
    {
      if (KSI_SINT_P (b))
	return ksi_int2num (KSI_SINT_COD (a) - KSI_SINT_COD (b));

      if (KSI_BIGNUM_P (b))
        {
	  int i = KSI_SINT_COD (a);
          if (i > 0)
            return add_big_ui (((ksi_bignum) b) -> digs, - ((ksi_bignum) b) -> sign, i);
          if (i < 0)
            return sub_big_ui (((ksi_bignum) b) -> digs, - ((ksi_bignum) b) -> sign, -i);
          return new_big (- ((ksi_bignum) b) -> sign, ((ksi_bignum) b) -> digs);
        }

      if (KSI_FLONUM_P (b))
        return ksi_double2num (KSI_SINT_COD (a) - KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	return ksi_rectangular (KSI_SINT_COD (a) - KSI_REPART (b), KSI_IMPART (b));

      KSI_EXN_ERR ("type", b, "- : invalid number");
    }

  if (KSI_BIGNUM_P (a))
    {
      if (KSI_SINT_P (b))
        {
	  int i = KSI_SINT_COD (b);
          if (i > 0)
            return sub_big_ui (((ksi_bignum) a) -> digs, ((ksi_bignum) a) -> sign, i);
          if (i < 0)
            return add_big_ui (((ksi_bignum) a) -> digs, ((ksi_bignum) a) -> sign, -i);
          return a;
        }

      if (KSI_BIGNUM_P (b))
        {
          if (((ksi_bignum) b) -> sign < 0)
            return add_big ((ksi_bignum) a, (ksi_bignum) b);
	  return sub_big ((ksi_bignum) a, (ksi_bignum) b);
        }

      if (KSI_FLONUM_P (b))
        return ksi_double2num (ksi_2double (a, "- ") - KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	return ksi_rectangular (ksi_2double (a, "- ") - KSI_REPART (b), KSI_IMPART (b));

      KSI_EXN_ERR ("type", b, "- : invalid number");
    }

  if (KSI_FLONUM_P (a))
    {
      if (KSI_SINT_P (b))
	return ksi_double2num (KSI_REPART (a) - KSI_SINT_COD (b));

      if (KSI_BIGNUM_P (b))
	return ksi_double2num (KSI_REPART (a) - ksi_2double (b, "- "));

      if (KSI_FLONUM_P (b))
	return ksi_double2num (KSI_REPART (a) - KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	return ksi_rectangular (KSI_REPART (a) - KSI_REPART (b), KSI_IMPART (b));

      KSI_EXN_ERR ("type", b, "- : invalid number");
    }

  if (KSI_COMPLEX_P (a))
    {
      if (KSI_SINT_P (b))
	return ksi_rectangular (KSI_REPART (a) - KSI_SINT_COD (b),
				KSI_IMPART (a));

      if (KSI_BIGNUM_P (b))
	return ksi_rectangular (KSI_REPART (a) - ksi_2double (b, "- "),
				KSI_IMPART (a));

      if (KSI_FLONUM_P (b))
	return ksi_rectangular (KSI_REPART (a) - KSI_REPART (b),
				KSI_IMPART (a));

      if (KSI_COMPLEX_P (b))
	return ksi_rectangular (KSI_REPART (a) - KSI_REPART (b),
				KSI_IMPART (a) - KSI_IMPART (b));

      KSI_EXN_ERR ("type", b, "- : invalid number");
    }

  KSI_EXN_ERR ("type", a, "- : invalid number");
  return ksi_zero;
}

ksi_obj
ksi_mul (ksi_obj a, ksi_obj b)
{
  if (KSI_SINT_P (a))
    {
      int i = KSI_SINT_COD (a);
      if (KSI_SINT_P (b))
        {
          int j = KSI_SINT_COD (b);
	  unsigned d0, d1;
	  ksi_bignum sum;

          MUL_LI (d1, d0, ABS (i), ABS (j));
	  if (d1 == 0 && d0 < KSI_MAX_INT_TYPE)
	    return (PROD_SIGN (i, j) < 0 ? KSI_MK_SINT (-d0) : KSI_MK_SINT (d0));

	  sum = alloc_big (2);
	  sum -> digs[0] = d0;
	  sum -> digs[1] = d1;
	  sum -> sign = PROD_SIGN (i, j) < 0 ? -2 : 2;
          return adjust_big (sum);
        }

      if (KSI_BIGNUM_P (b))
        return mul_big_i (((ksi_bignum) b) -> sign,
			  ((ksi_bignum) b) -> digs, i);

      if (KSI_FLONUM_P (b))
        return ksi_double2num (KSI_REPART (b) * i);

      if (KSI_COMPLEX_P (b))
	return ksi_rectangular (i * KSI_REPART (b), i * KSI_IMPART (b));

      KSI_EXN_ERR ("type", b, "* : invalid number");
    }

  if (KSI_BIGNUM_P (a))
    {
      if (KSI_SINT_P (b))
        return mul_big_i (((ksi_bignum) a) -> sign,
			  ((ksi_bignum) a) -> digs, KSI_SINT_COD (b));

      if (KSI_BIGNUM_P (b))
        return mul_big ((ksi_bignum) a, (ksi_bignum) b);

      if (KSI_FLONUM_P (b))
        return ksi_double2num (ksi_2double (a, "* ") * KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	{
	  double d = ksi_2double (a, "* ");
	  return ksi_rectangular (d * KSI_REPART (b), d * KSI_IMPART (b));
	}

      KSI_EXN_ERR ("type", b, "* : invalid number");
    }

  if (KSI_FLONUM_P (a))
    {
      if (KSI_SINT_P (b))
	return ksi_double2num (KSI_REPART (a) * KSI_SINT_COD (b));

      if (KSI_BIGNUM_P (b))
	return ksi_double2num (KSI_REPART (a) * ksi_2double (b, "* "));

      if (KSI_FLONUM_P (b))
	return ksi_double2num (KSI_REPART (a) * KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	return ksi_rectangular (KSI_REPART (a) * KSI_REPART (b), KSI_REPART (a) * KSI_IMPART (b));

      KSI_EXN_ERR ("type", b, "* : invalid number");
    }

  if (KSI_COMPLEX_P (a))
    {
      if (KSI_SINT_P (b))
	return ksi_rectangular (KSI_REPART (a) * KSI_SINT_COD (b), KSI_IMPART (a) * KSI_SINT_COD (b));

      if (KSI_BIGNUM_P (b))
	{
	  double d = ksi_2double (b, "* ");
	  return ksi_rectangular (KSI_REPART (a) * d, KSI_IMPART (a) * d);
	}

      if (KSI_FLONUM_P (b))
	return ksi_rectangular (KSI_REPART (a) * KSI_REPART (b), KSI_IMPART (a) * KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	return (ksi_rectangular (KSI_REPART (a) * KSI_REPART (b) - KSI_IMPART (a) * KSI_IMPART (b),
                                 KSI_REPART (a) * KSI_IMPART (b) + KSI_IMPART (a) * KSI_REPART (b)));

      KSI_EXN_ERR ("type", b, "* : invalid number");
    }

  KSI_EXN_ERR ("type", a, "* : invalid number");
  return ksi_one;
}


ksi_obj
ksi_div (ksi_obj a, ksi_obj b)
{
  int i;
  double re, im;

  if (KSI_SINT_P (a))
    {
      i = KSI_SINT_COD (a);

      if (KSI_SINT_P (b))
        {
          int j = KSI_SINT_COD (b);
          if (j == 0 || i % j != 0)
            return ksi_double2num (((double) i) / ((double) j));

	  return KSI_MK_SINT (i / j);
        }

      /* bignum must be normalized there, so ABS (b) > ABS (i) */
      if (KSI_BIGNUM_P (b))
	return ksi_double2num (((double) i) / ksi_2double (b, "/ "));

      if (KSI_FLONUM_P (b))
	return ksi_double2num (((double) i) / KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	{
	  re = (double) i; im = 0.0;
	  goto complex_div;
	}

      KSI_EXN_ERR ("type", b, "/ : invalid number");
    }

  if (KSI_BIGNUM_P (a))
    {
      if (KSI_SINT_P (b))
        {
          i = KSI_SINT_COD (b);
          if (i != 0)
	    {
	      ksi_obj res = divmod_big_i (((ksi_bignum) a) -> sign, ((ksi_bignum) a) -> digs, &i);
	      if (i == 0)
		return res;
	      i = KSI_SINT_COD (b);
	    }

          return ksi_double2num (ksi_2double (a, "/ ") / i);
        }

      if (KSI_BIGNUM_P (b))
        {
          int asize = ABS (((ksi_bignum) a) -> sign);
          int bsize = ABS (((ksi_bignum) b) -> sign);
          if (asize >= bsize)
            {
              ksi_obj q, r;
              div_big ((ksi_bignum) a, (ksi_bignum) b, &q, &r);
              if (r == ksi_zero)
                return q;
            }
          return ksi_double2num (ksi_2double (a, "/") / ksi_2double (b, "/"));
        }

      if (KSI_FLONUM_P (b))
	return ksi_double2num (ksi_2double (a, "/ ") / KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	{
	  re = ksi_2double (a, "/ ");
	  im = 0.0;
	  goto complex_div;
	}

      KSI_EXN_ERR ("type", b, "/ : invalid number");
    }

  if (KSI_FLONUM_P (a))
    {
      if (KSI_SINT_P (b))
	return ksi_double2num (KSI_REPART (a) / ((double) KSI_SINT_COD (b)));

      if (KSI_BIGNUM_P (b))
	return ksi_double2num (KSI_REPART (a) / ksi_2double (b, "/ "));

      if (KSI_FLONUM_P (b))
	return ksi_double2num (KSI_REPART (a) / KSI_REPART (b));

      if (KSI_COMPLEX_P (b))
	{
	  re = KSI_REPART (a);
	  im = 0.0;
	  goto complex_div;
	}

      KSI_EXN_ERR ("type", b, "/ : invalid number");
    }

  if (KSI_COMPLEX_P (a))
    {
      re = KSI_REPART (a);
      im = KSI_IMPART (a);

      if (KSI_SINT_P (b))
	{
	  double d = KSI_SINT_COD (b);
	  return ksi_rectangular (re / d, im / d);
	}

      if (KSI_BIGNUM_P (b))
	{
	  double d = ksi_2double (b, "/ ");
	  return ksi_rectangular (re / d, im / d);
	}

      if (KSI_FLONUM_P (b))
	return ksi_rectangular (re / KSI_REPART (b), im / KSI_REPART (b));
	
      if (KSI_COMPLEX_P (b))
	{
	  double d;
	complex_div:
	  d = (KSI_REPART (b) * KSI_REPART (b) + KSI_IMPART (b) * KSI_IMPART (b));
	  return ksi_rectangular ((re * KSI_REPART (b) + im * KSI_IMPART (b)) / d,
				  (im * KSI_REPART (b) - re * KSI_IMPART (b)) / d);
	}

      KSI_EXN_ERR ("type", b, "/ : invalid number");
    }

  KSI_EXN_ERR ("type", a, "/ : invalid number");
  return ksi_zero;
}


ksi_obj
ksi_quot (ksi_obj a, ksi_obj b)
{
  if (KSI_SINT_P (a))
    {
      if (KSI_SINT_P (b))
        {
          int j = KSI_SINT_COD (b);
          if (j == 0)
            ksi_exn_error ("type", ksi_zero, "quotient: zero divide");
          return KSI_MK_SINT (KSI_SINT_COD (a) / j);
        }

      /* bignum must be normalized there, so ABS (b) > ABS (i) */
      if (KSI_BIGNUM_P (b))
        return ksi_zero;

      KSI_EXN_ERR ("type", b, "quotient: invalid integer");
    }

  if (KSI_BIGNUM_P (a))
    {
      if (KSI_SINT_P (b))
        {
          int i = KSI_SINT_COD (b);
          if (i == 0)
            ksi_exn_error ("type", ksi_zero, "quotient: zero divide");
          return divmod_big_i (((ksi_bignum) a) -> sign, ((ksi_bignum) a) -> digs, &i);
        }

      if (KSI_BIGNUM_P (b))
        {
          int asize = ABS (((ksi_bignum) a) -> sign);
          int bsize = ABS (((ksi_bignum) b) -> sign);
          if (asize >= bsize)
            {
              ksi_obj q;
              div_big ((ksi_bignum) a, (ksi_bignum) b, &q, 0);
              return q;
            }
          return ksi_zero;
        }

      KSI_EXN_ERR ("type", b, "quotient: invalid integer");
    }

  KSI_EXN_ERR ("type", a, "quotient: invalid integer");
  return ksi_zero;
}

static ksi_obj
rem (ksi_obj a, ksi_obj b, char* name)
{
  if (KSI_SINT_P (a))
    {
      if (KSI_SINT_P (b))
        {
          int j = KSI_SINT_COD (b);
          if (j == 0)
            ksi_exn_error ("type", ksi_zero, "%s: zero divide", name);
          return KSI_MK_SINT (KSI_SINT_COD (a) % ABS (j));
        }

      /* bignum must be normalized there, so ABS (b) > ABS (i) */
      if (KSI_BIGNUM_P (b))
        return a;

      ksi_exn_error ("type", b, "%s: invalid integer", name);
    }

  if (KSI_BIGNUM_P (a))
    {
      if (KSI_SINT_P (b))
        {
          int i = KSI_SINT_COD (b);
          if (i == 0)
            ksi_exn_error ("type", ksi_zero, "%s: zero divide", name);

          i = ABS (i);
          divmod_big_i (((ksi_bignum) a) -> sign, ((ksi_bignum) a) -> digs, &i);
          if (((ksi_bignum) a) -> sign < 0)
            i = -i;
          return ksi_int2num (i);
        }

      if (KSI_BIGNUM_P (b))
        {
          int asize = ABS (((ksi_bignum) a) -> sign);
          int bsize = ABS (((ksi_bignum) b) -> sign);
          if (bsize == 0)
            ksi_exn_error ("type", ksi_zero, "%s: zero divide", name);
          if (asize >= bsize)
            {
              ksi_obj r;
              div_big ((ksi_bignum) a, (ksi_bignum) b, 0, &r);
              if (((ksi_bignum) a) -> sign < 0)
		{
		  if (KSI_SINT_P (r))
		    r = KSI_MK_SINT (- KSI_SINT_COD (r));
		  else
		    ((ksi_bignum) r) -> sign = - ((ksi_bignum) r) -> sign;
		}
              return r;
            }
	  return a;
        }

      ksi_exn_error ("type", b, "%s: invalid integer", name);
    }

  ksi_exn_error ("type", a, "%s: invalid integer", name);
  return ksi_zero;
}


static ksi_obj
ksi_plus (int argc, ksi_obj* argv)
{
  ksi_obj val;
  if (argc <= 0)
    return ksi_zero;

  val = *argv++;
  while (--argc > 0)
    val = ksi_add (val, *argv++);
  return val;
}

static ksi_obj
ksi_minus (int argc, ksi_obj* argv)
{
  ksi_obj val;

  val = *argv++;
  if (--argc == 0)
    {
      if (KSI_SINT_P (val))
	return KSI_MK_SINT (- KSI_SINT_COD (val));

      if (KSI_BIGNUM_P (val))
	return new_big (- ((ksi_bignum) val) -> sign,
			((ksi_bignum) val) -> digs);

      if (KSI_FLONUM_P (val))
	return ksi_double2num (- KSI_REPART (val));

      if (KSI_COMPLEX_P (val))
	return ksi_rectangular (- KSI_REPART (val), KSI_IMPART (val));

      KSI_EXN_ERR ("type", val, "- : invalid number");
    }

  val = ksi_sub (val, *argv++);
  while (--argc > 0)
    val = ksi_sub (val, *argv++);
  return val;
}

static ksi_obj
ksi_star (int argc, ksi_obj* argv)
{
  ksi_obj val;
  if (argc <= 0)
    return ksi_one;

  val = *argv++;
  while (--argc > 0)
    val = ksi_mul (val, *argv++);
  return val;
}

static ksi_obj
ksi_slash (int argc, ksi_obj* argv)
{
  ksi_obj val;

  val = *argv++;
  if (--argc == 0)
    return ksi_div (ksi_one, val);

  val = ksi_div (val, *argv++);
  while (--argc > 0)
    val = ksi_div (val, *argv++);
  return val;
}

ksi_obj
ksi_rem (ksi_obj a, ksi_obj b)
{
  return rem (a, b, "remainder");
}

ksi_obj
ksi_mod (ksi_obj a, ksi_obj b)
{
  ksi_obj m = rem (a, b, "modulo");
  if (m == ksi_zero)
    return ksi_zero;

  if (ksi_negative_p (a) != ksi_negative_p (b))
    m = ksi_add (m, b);

  return m;
}

static ksi_obj
gcd2 (ksi_obj n1, ksi_obj n2)
{
  ksi_obj t;

again:
  if (ksi_zero_p (n2) == ksi_true)
    return n1;
  t = n1;
  n1 = n2; n2 = ksi_mod (t, n2);
  goto again;
}

ksi_obj
ksi_gcd (int argc, ksi_obj* argv)
{
  ksi_obj tmp, res;
  int inexact=0;

  if (argc <= 0)
    return ksi_zero;

  tmp = *argv++;
  if (KSI_SINT_P (tmp) || KSI_BIGNUM_P (tmp))
    inexact = 0;
  else if (KSI_FLONUM_P (tmp) && IS_INT (KSI_REPART (tmp)))
    {
      tmp = ksi_inexact2exact (tmp);
      inexact = 1;
    }
  else
    KSI_EXN_ERR ("type", tmp, "gcd: invalid integer");

  res = ksi_abs (tmp);
  while (--argc > 0)
    {
      tmp = *argv++;
      if (KSI_FLONUM_P (tmp) && IS_INT (KSI_REPART (tmp)))
	{
	  tmp = ksi_inexact2exact (tmp);
	  inexact = 1;
	}
      else if (!KSI_SINT_P (tmp) && !KSI_BIGNUM_P (tmp))
	KSI_EXN_ERR ("type", tmp, "gcd: invalid integer");

      res = ksi_abs (gcd2 (res, tmp));
    }

  if (inexact)
    res = ksi_exact2inexact (res);
  return res;
}

ksi_obj
ksi_lcm (int argc, ksi_obj* argv)
{
  ksi_obj tmp, res, gcd;
  int inexact=0;

  if (argc <= 0)
    return ksi_one;

  tmp = *argv++;
  if (KSI_SINT_P (tmp) || KSI_BIGNUM_P (tmp))
    inexact = 0;
  else if (KSI_FLONUM_P (tmp) && IS_INT (KSI_REPART (tmp)))
    {
      tmp = ksi_inexact2exact (tmp);
      inexact = 1;
    }
  else
    KSI_EXN_ERR ("type", tmp, "lcm: invalid integer");

  res = ksi_abs (tmp);
  while (--argc > 0)
    {
      tmp = *argv++;
      if (KSI_FLONUM_P (tmp) && IS_INT (KSI_REPART (tmp)))
	{
	  tmp = ksi_inexact2exact (tmp);
	  inexact = 1;
	}
      else if (!KSI_SINT_P (tmp) && !KSI_BIGNUM_P (tmp))
	KSI_EXN_ERR ("type", tmp, "lcm: invalid integer");

      if (ksi_zero_p (tmp) == ksi_true)
        return inexact ? ksi_double2num (0.0) : ksi_zero;

      /* ! order of assigments is important ! */
      gcd = gcd2 (res, tmp);
      res = ksi_mul (res, tmp);
      res = ksi_div (res, gcd);
      res = ksi_abs (res);
    }

  if (inexact)
    res = ksi_exact2inexact (res);
  return res;
}


ksi_obj
ksi_abs (ksi_obj x)
{
  if (KSI_SINT_P (x))
    return KSI_MK_SINT (ABS (KSI_SINT_COD (x)));

  if (KSI_BIGNUM_P (x))
    {
      if (((ksi_bignum) x) -> sign > 0)
	return x;

      return new_big (- ((ksi_bignum) x) -> sign, ((ksi_bignum) x) -> digs);
    }

  if (KSI_FLONUM_P (x))
    return ksi_double2num (fabs (KSI_REPART (x)));

  if (KSI_COMPLEX_P (x))
    return ksi_double2num (sqrt (KSI_REPART (x) * KSI_REPART (x) +
				 KSI_IMPART (x) * KSI_IMPART (x)));

  KSI_EXN_ERR ("type", x, "abs: invalid number");
  return ksi_void;
}

static ksi_obj
make_rectangular (ksi_obj x, ksi_obj y)
{
  return ksi_rectangular (ksi_2double (x, "make-rectangular"),
			  ksi_2double (y, "make-rectangular"));
}

static ksi_obj
make_polar (ksi_obj x, ksi_obj y)
{
  return ksi_polar (ksi_2double (x, "make-polar"),
		    ksi_2double (y, "make-polar"));
}

static ksi_obj
real_part (ksi_obj x)
{
  if (KSI_SINT_P (x) || KSI_BIGNUM_P (x) || KSI_FLONUM_P (x))
    return x;
  if (KSI_COMPLEX_P (x))
    return ksi_double2num (KSI_REPART (x));

  KSI_EXN_ERR ("type", x, "real-part: invalid number");
  return ksi_zero;
}

static ksi_obj
imag_part (ksi_obj x)
{
  if (KSI_SINT_P (x) || KSI_BIGNUM_P (x) || KSI_FLONUM_P (x))
    return x;
  if (KSI_COMPLEX_P (x))
    return ksi_double2num (KSI_IMPART (x));

  KSI_EXN_ERR ("type", x, "imag-part: invalid number");
  return ksi_zero;
}

ksi_obj
angle (ksi_obj z)
{
  return ksi_double2num (ksi_angle (z));
}

ksi_obj
ksi_floor (ksi_obj x)
{
  TYPE_CHECK (x, KSI_REAL_P (x), "floor: invalid number");
  if (KSI_FLONUM_P (x))
    return ksi_double2num (floor (KSI_REPART (x)));
  return x;
}

ksi_obj
ksi_ceiling (ksi_obj x)
{
  TYPE_CHECK (x, KSI_REAL_P (x), "ceiling: invalid number");
  if (KSI_FLONUM_P (x))
    return ksi_double2num (ceil (KSI_REPART (x)));
  return x;
}

ksi_obj
ksi_truncate (ksi_obj x)
{
  TYPE_CHECK (x, KSI_REAL_P (x), "truncate: invalid number");
  if (KSI_FLONUM_P (x))
    {
      double d = KSI_REPART (x);
      return ksi_double2num (floor (d < 0.0 ? ceil (d) : floor (d)));
    }
  return x;
}

ksi_obj
ksi_round (ksi_obj x)
{
  TYPE_CHECK (x, KSI_REAL_P (x), "round: invalid number");
  if (KSI_FLONUM_P (x))
    {
      double d = KSI_REPART (x);
      double v = d + 0.5;
      double r = floor (v);
      if (v == r && v / 2 != floor (v / 2))
        r = r - 1;
      return ksi_double2num (r);
    }
  return x;
}


ksi_obj
ksi_exp (ksi_obj x)
{
  if (KSI_REAL_P (x))
    return ksi_double2num (exp (ksi_2double (x, "exp")));
  if (KSI_COMPLEX_P (x))
    return ksi_polar (exp (KSI_REPART (x)), KSI_IMPART (x));

  KSI_EXN_ERR ("type", x, "exp: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_log (ksi_obj x)
{
  if (KSI_REAL_P (x))
    {
      double d = ksi_2double (x, "log");
      if (d >= 0.0)
	return ksi_double2num (log (d));
      return ksi_rectangular (log (fabs (d)), ksi_angle (x));
    }

  if (KSI_COMPLEX_P (x))
    return ksi_rectangular (log (ksi_magnitude (x)), ksi_angle (x));

  KSI_EXN_ERR ("type", x, "log: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_sin (ksi_obj x)
{
  if (KSI_REAL_P (x))
    return ksi_double2num (sin (ksi_2double (x, "sin")));
  if (KSI_COMPLEX_P (x))
    return ksi_rectangular (sin (KSI_REPART (x)) * cosh (KSI_IMPART (x)),
			    cos (KSI_REPART (x)) * sinh (KSI_IMPART (x)));

  KSI_EXN_ERR ("type", x, "sin: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_cos (ksi_obj x)
{
  if (KSI_REAL_P (x))
    return ksi_double2num (cos (ksi_2double (x, "cos")));
  if (KSI_COMPLEX_P (x))
    return ksi_rectangular (cos (KSI_REPART (x)) * cosh (KSI_IMPART (x)),
			    - sin (KSI_REPART (x)) * sinh (KSI_IMPART (x)));

  KSI_EXN_ERR ("type", x, "cos: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_tan (ksi_obj z)
{
  if (KSI_REAL_P (z))
    return ksi_double2num (tan (ksi_2double (z, "tan")));
  if (KSI_COMPLEX_P (z))
    {
      double x = 2.0 * KSI_REPART (z);
      double y = 2.0 * KSI_IMPART (z);
      double w = cos (x) + cosh (y);
      return ksi_rectangular (sin (x) / w, sinh (y) / w);
    }

  KSI_EXN_ERR ("type", z, "tan: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_sinh (ksi_obj x)
{
  if (KSI_REAL_P (x))
    return ksi_double2num (sinh (ksi_2double (x, "sinh")));
  if (KSI_COMPLEX_P (x))
    return ksi_rectangular (sinh (KSI_REPART (x)) * cos (KSI_IMPART (x)),
			    cosh (KSI_REPART (x)) * sin (KSI_IMPART (x)));

  KSI_EXN_ERR ("type", x, "sinh: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_cosh (ksi_obj x)
{
  if (KSI_REAL_P (x))
    return ksi_double2num (cosh (ksi_2double (x, "cosh")));
  if (KSI_COMPLEX_P (x))
    return ksi_rectangular (cosh (KSI_REPART (x)) * cos (KSI_IMPART (x)),
			    sinh (KSI_REPART (x)) * sin (KSI_IMPART (x)));

  KSI_EXN_ERR ("type", x, "cosh: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_tanh (ksi_obj z)
{
  if (KSI_REAL_P (z))
    return ksi_double2num (tan (ksi_2double (z, "tanh")));
  if (KSI_COMPLEX_P (z))
    {
      double x = 2.0 * KSI_REPART (z);
      double y = 2.0 * KSI_IMPART (z);
      double w = cosh (x) + cos (y);
      return ksi_rectangular (sinh (x) / w, sin (y) / w);
    }

  KSI_EXN_ERR ("type", z, "tan: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_asinh (ksi_obj x)
{
  if (KSI_REAL_P (x))
    {
      double d = ksi_2double (x, "asinh");
      return ksi_double2num (log (d + sqrt (d * d + 1)));
    }
  if (KSI_COMPLEX_P (x))
    return ksi_log (ksi_add (x, ksi_sqrt (ksi_add (ksi_mul (x, x), ksi_one))));

  KSI_EXN_ERR ("type", x, "asinh: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_acosh (ksi_obj x)
{
  if (KSI_REAL_P (x))
    {
      double d = ksi_2double (x, "acosh");
      if (d >= 1.0)
	return ksi_double2num (log (d + sqrt (d * d - 1)));
      goto compl;
    }
  if (KSI_COMPLEX_P (x))
    {
    compl:
      return ksi_log (ksi_add (x, 
			       ksi_sqrt (ksi_sub (ksi_mul (x, x),
						  ksi_one))));
    }

  KSI_EXN_ERR ("type", x, "asinh: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_atanh (ksi_obj x)
{
  if (KSI_REAL_P (x))
    {
      double d = ksi_2double (x, "atanh");
      if (d > -1.0 && d < 1.0)
	return ksi_double2num (0.5 * log ((1 + d) /(1 - d)));
      goto compl;
    }
  if (KSI_COMPLEX_P (x))
    {
    compl:
      return ksi_div (ksi_log (ksi_div (ksi_add (ksi_one, x),
					ksi_sub (ksi_one, x))),
		      ksi_two);
    }

  KSI_EXN_ERR ("type", x, "atanh: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_asin (ksi_obj x)
{
  ksi_obj i, z;

  if (KSI_REAL_P (x))
    {
      double d = ksi_2double (x, "asin");
      if (d >= -1.0 && d <= 1.0)
	return ksi_double2num (asin (d));
      goto compl;
    }
  if (KSI_COMPLEX_P (x))
    {
    compl:
      i = ksi_rectangular (0.0, 1.0);
      z = ksi_asinh (ksi_mul (i, x));
      KSI_IMPART (i) = -1.0;
      return ksi_mul (i, z);
    }

  KSI_EXN_ERR ("type", x, "asin: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_acos (ksi_obj x)
{
  ksi_obj i, z;

  if (KSI_REAL_P (x))
    {
      double d = ksi_2double (x, "acos");
      if (d >= -1.0 && d <= 1.0)
	return ksi_double2num (acos (d));
      goto compl;
    }
  if (KSI_COMPLEX_P (x))
    {
    compl:
      i = ksi_rectangular (0.0, 1.0);
      z = ksi_mul (i, ksi_asinh (ksi_mul (i, x)));
      KSI_IMPART (i) = -1.0;
      return ksi_add (ksi_div (angle (i), ksi_two), z);
    }

  KSI_EXN_ERR ("type", x, "asin: invalid number");
  return ksi_zero;
}

ksi_obj
ksi_atan (ksi_obj z, ksi_obj y)
{
  if (y)
    return ksi_double2num (atan2 (ksi_2double (z, "atan2"),
				  ksi_2double (y, "atan2")));

  if (KSI_REAL_P (z))
    return ksi_double2num (atan (ksi_2double (z, "atan")));

  if (KSI_COMPLEX_P (z))
    {
      ksi_obj i = ksi_rectangular (0.0, 1.0);
      ksi_obj r = ksi_log (ksi_div (ksi_sub (i, z), ksi_add (i, z)));
      KSI_IMPART (i) = 2.0;
      return ksi_div (r, i);
    }

  KSI_EXN_ERR ("type", z, "atan: invalid number");
  return ksi_zero;
}


ksi_obj
ksi_sqrt (ksi_obj x)
{
  double d;

  if (KSI_SINT_P (x))
    {
      int i = KSI_SINT_COD (x);
      if (i < 0)
	return ksi_rectangular (0.0, sqrt (-i));
      d = sqrt (i);
      return (IS_INT (d) ? ksi_int2num ((int) d) : ksi_double2num (d));
    }

  if (KSI_BIGNUM_P (x))
    {
      ksi_obj root;
      int rem;

      if (((ksi_bignum) x) -> sign < 0)
	return ksi_rectangular (0.0, sqrt (-ksi_2double (x, "sqrt")));

      root = sqrt_big ((ksi_bignum) x, &rem);
      if (!rem)
        return root;
      d = ksi_2double (x, "sqrt");
      return ksi_double2num (sqrt (d));
    }

  if (KSI_FLONUM_P (x))
    {
      d = KSI_REPART (x);
      if (d < 0.0)
	return ksi_rectangular (0.0, sqrt (-d));
      return ksi_double2num (sqrt (d));
    }

  if (KSI_COMPLEX_P (x))
    return ksi_polar (sqrt (ksi_magnitude (x)), ksi_angle (x) / 2.0);

  ksi_exn_error ("type", x, "sqrt: invalid number in arg1");
  return ksi_zero;
}

static ksi_obj
exact_expt (ksi_obj z1, ksi_obj z2)
{
  ksi_obj x;

  if (z2 == ksi_zero)
    return ksi_one;
  if (z2 == ksi_one)
    return z1;
  if (ksi_negative_p (z2) == ksi_true)
    {
      x = ksi_abs (z2);
      x = exact_expt (z1, x);
      return ksi_div (ksi_one, x);
    }
  if (ksi_even_p (z2) == ksi_true)
    {
      /* (z1**(z2/2))**2 */
      x = ksi_div (z2, ksi_two);
      x = exact_expt (z1, x);
      return ksi_mul (x, x);
    }
  /* (z1**(z2-1)) * z1 */
  x = ksi_sub (z2, ksi_one);
  x = exact_expt (z1, x);
  return ksi_mul (x, z1);
}

ksi_obj
ksi_expt (ksi_obj z1, ksi_obj z2)
{
  if (KSI_EXACT_P (z1) && KSI_EXACT_P (z2))
    return exact_expt (z1, z2);

  if (KSI_REAL_P (z1) && KSI_REAL_P (z2)
      && ksi_negative_p (z1) == ksi_false)
    {
      double d1 = ksi_2double (z1, "expt");
      double d2 = ksi_2double (z2, "expt");
      return ksi_double2num (pow (d1, d2));
    }

  TYPE_CHECK (z1, KSI_NUM_P (z1), "expt: invalid number in arg1");
  TYPE_CHECK (z2, KSI_NUM_P (z2), "expt: invalid number in arg2");

  return ksi_exp (ksi_mul (z2, ksi_log (z1)));
}


ksi_obj
ksi_lognot (ksi_obj x)
{
  return ksi_sub (KSI_MK_SINT (-1), x);
}

static void
copy_dec_big (unsigned* sum, unsigned* digs, int num)
{
  while (--num >= 0)
    {
      register unsigned m = *digs++;
      *sum++ = m - 1;
      if (m > 0)
	break;
    }
  while (--num >= 0)
    *sum++ = *digs++;
}

static void
inc_big (unsigned* dig, int num)
{
  while (--num >= 0)
    {
      register unsigned a = *dig + 1;
      *dig++ = a;
      if (a > 0)
	break;
    }
}

static ksi_obj
big_logior (unsigned* dx, int nx, unsigned* dy, int ny)
{
  int i;
  ksi_bignum res;
  unsigned *tmp;

  KSI_ASSERT (nx != 0 && ny != 0 && ABS (nx) <= ABS (ny));
  /*
   * ( x |  y) => (x | y)
   * ( x | -y) => -(~(x | ~(y-1)) + 1) => -((~x & (y-1)) + 1)
   * (-x |  y) => -(~(~(x-1) | y) + 1) => -(((x-1) & ~y) + 1)
   * (-x | -y) => -(~(~(x-1) | ~(y-1)) + 1) => -(((x-1) & (y-1)) + 1)
   */

  if (nx > 0)
    {
      if (ny > 0)
	{
	  res = alloc_big (ny);
	  res -> sign = ny;
	  for (i = 0; i < nx; i++)
	    res -> digs [i] = dx [i] | dy [i];
	  for (/**/ ; i < ny; i++)
	    res -> digs [i] = dy [i];
	}
      else
	{
	  res = alloc_big (-ny);
	  res -> sign = ny;
	  copy_dec_big (res->digs, dy, -ny);
	  for (i = 0; i < nx; i++)
 	    res -> digs [i] &= ~dx [i];
	  inc_big (res->digs, -ny);
	}
    }
  else
    {
      nx = -nx;
      res = alloc_big (nx);
      res -> sign = -nx;
      copy_dec_big (res->digs, dx, nx);

      if (ny > 0)
	{
	  for (i = 0; i < nx; i++)
 	    res -> digs [i] &= ~dy [i];
	}
      else
	{
	  TALLOC (tmp, unsigned, nx);
	  copy_dec_big (tmp, dy, nx);
	  for (i = 0; i < nx; i++)
 	    res -> digs [i] &= tmp [i];
	}
      inc_big (res->digs, nx);
    }

  return adjust_big (res);
}

static ksi_obj
big_logand (unsigned* dx, int nx, unsigned* dy, int ny)
{
  int i;
  ksi_bignum res;
  unsigned *tmp;

  KSI_ASSERT (nx != 0 && ny != 0 && ABS (nx) <= ABS (ny));
  /*
   * ( x &  y) => (x & y)
   * ( x & -y) => (x & ~(y-1))
   * (-x &  y) => (~(x-1) & y)
   * (-x & -y) => -(~(~(x-1) & ~(y-1)) + 1) => -(((x-1) | (y-1)) + 1)
   */

  if (nx > 0)
    {
      res = alloc_big (nx);
      res -> sign = nx;
      if (ny > 0)
	{
	  for (i = 0; i < nx; i++)
	    res -> digs [i] = dx [i] & dy [i];
	}
      else
	{
	  TALLOC (tmp, unsigned, nx);
	  copy_dec_big (tmp, dy, nx);
	  for (i = 0; i < nx; i++)
 	    res -> digs [i] = dx [i] & ~tmp [i];
	}
    }
  else
    {
      nx = -nx;
      TALLOC (tmp, unsigned, nx);
      copy_dec_big (tmp, dx, nx);
      if (ny > 0)
	{
	  res = alloc_big (ny);
	  res->sign = ny;
	  for (i = 0; i < nx; i++)
 	    res -> digs [i] = ~tmp [i] & dy [i];
	  for (/**/;  i < ny; i++)
	    res -> digs [i] = dy [i];
	}
      else
	{
	  res = alloc_big (-ny);
	  res->sign = ny;
	  copy_dec_big (res->digs, dy, -ny);
	  for (i = 0; i < nx; i++)
	    res -> digs [i] |= tmp [i];
	  inc_big (res->digs, -ny);
	}
    }

  return adjust_big (res);
}

static ksi_obj
big_logxor (unsigned* dx, int nx, unsigned* dy, int ny)
{
  int i;
  ksi_bignum res;
  unsigned *tmp;

  KSI_ASSERT (nx != 0 && ny != 0 && ABS (nx) <= ABS (ny));
  /*
   * ( x ^  y) => (x ^ y)
   * ( x ^ -y) => -(~(x ^ ~(y-1)) + 1) => -((x ^ (y-1)) + 1)
   * (-x ^  y) => -(~(~(x-1) ^ y) + 1) => -(((x-1) ^ y) + 1)
   * (-x ^ -y) => (~(~(x-1) ^ ~(y-1)) + 1) => ((x-1) ^ (y-1))
   */

  if (nx > 0)
    {
      if (ny > 0)
	{
	  res = alloc_big (ny);
	  res -> sign = ny;
	  for (i = 0; i < nx; i++)
	    res -> digs [i] = dx [i] ^ dy [i];
	  for (/**/ ; i < ny; i++)
	    res -> digs [i] = dy [i];
	}
      else
	{
	  res = alloc_big (-ny);
	  res -> sign = ny;
	  copy_dec_big (res->digs, dy, -ny);
	  for (i = 0; i < nx; i++)
 	    res -> digs [i] ^= dx [i];
	  inc_big (res->digs, -ny);
	}
    }
  else
    {
      nx = -nx;
      TALLOC (tmp, unsigned, nx);
      copy_dec_big (tmp, dx, nx);
      if (ny > 0)
	{
	  res = alloc_big (ny);
	  res -> sign = -ny;
	  for (i = 0; i < nx; i++)
 	    res -> digs [i] = tmp [i] ^ dy [i];
	  for (/**/ ; i < ny; i++)
	    res -> digs [i] = dy [i];
	  inc_big (res -> digs, ny);
	}
      else
	{
	  res = alloc_big (-ny);
	  res -> sign = -ny;
	  copy_dec_big (res->digs, dy, -ny);
	  for (i = 0; i < nx; i++)
	    res -> digs [i] ^= tmp [i];
	}
    }

  return adjust_big (res);
}

static ksi_obj
big_logtest (unsigned* dx, int nx, unsigned* dy, int ny)
{
  unsigned *tmp;
  KSI_ASSERT (nx != 0 && ny != 0 && ABS (nx) <= ABS (ny));

  /* true if bitwise and of any digits is not zero. */
  if (nx > 0)
    {
      if (ny > 0)
	{
	  while (--nx >= 0)
	    if (dx [nx] & dy [nx])
	      return ksi_true;
	}
      else
	{
	  TALLOC (tmp, unsigned, nx);
	  copy_dec_big (tmp, dy, nx);
	  while (--nx >= 0)
	    if (dx [nx] & ~tmp [nx])
	      return ksi_true;
	}
    }
  else
    {
      if (ny < 0 || ny > -nx)
	return ksi_true;
      /* ny > 0 && ny == -nx */

      TALLOC (tmp, unsigned, nx);
      copy_dec_big (tmp, dx, -nx);
      while (--ny >= 0)
	if (dy [ny] & ~tmp [ny])
	  return ksi_true;
    }

  return ksi_false;
}


#define DEF_LOG(nam,big,str,op)						\
ksi_obj									\
nam (ksi_obj x, ksi_obj y)						\
{									\
  if (KSI_SINT_P (x))							\
    {									\
      if (KSI_SINT_P (y))						\
	return KSI_MK_SINT (KSI_SINT_COD (x) op KSI_SINT_COD (y));	\
									\
      if (KSI_BIGNUM_P (y))						\
	{								\
          int i = KSI_SINT_COD (x);					\
	  unsigned n = ABS (i);						\
	  return big (&n, i < 0 ? -1 : 1,				\
		      ((ksi_bignum) y) -> digs,				\
		      ((ksi_bignum) y) -> sign);			\
	}								\
									\
      KSI_EXN_ERR ("type", y, str ": invalid integer");			\
    }									\
									\
  if (KSI_BIGNUM_P (x))							\
    {									\
      if (KSI_SINT_P (y))						\
	{								\
          int i = KSI_SINT_COD (y);					\
	  unsigned n = ABS (i);						\
	  return big (&n, i < 0 ? -1 : 1,				\
		      ((ksi_bignum) x) -> digs,				\
		      ((ksi_bignum) x) -> sign);			\
	}								\
									\
      if (KSI_BIGNUM_P (y))						\
	{								\
	  if (ABS (((ksi_bignum) x) -> sign)				\
	      < ABS (((ksi_bignum) y) -> sign))				\
	    {								\
	      return big (((ksi_bignum) x) -> digs,			\
			  ((ksi_bignum) x) -> sign,			\
			  ((ksi_bignum) y) -> digs,			\
			  ((ksi_bignum) y) -> sign);			\
	    }								\
	  else								\
	    {								\
	      return big (((ksi_bignum) y) -> digs,			\
			  ((ksi_bignum) y) -> sign,			\
			  ((ksi_bignum) x) -> digs,			\
			  ((ksi_bignum) x) -> sign);			\
	    }								\
	}								\
									\
      KSI_EXN_ERR ("type", y, str ": invalid integer");			\
    }									\
									\
  KSI_EXN_ERR ("type", x, str ": invalid integer");			\
  return ksi_false;							\
}

DEF_LOG(ksi_logior, big_logior, "logior", |)
DEF_LOG(ksi_logand, big_logand, "logand", &)
DEF_LOG(ksi_logxor, big_logxor, "logxor", ^)


ksi_obj
ksi_logtest (ksi_obj x, ksi_obj y)
{
  if (KSI_SINT_P (x))
    {
      if (KSI_SINT_P (y))
	return ((KSI_SINT_COD (x) & KSI_SINT_COD (y)) ? ksi_true : ksi_false);

      if (KSI_BIGNUM_P (y))
	{
	  int i = KSI_SINT_COD (x);
	  unsigned n = ABS (i);
	  return big_logtest (&n, i < 0 ? -1 : 1,
			      ((ksi_bignum) y) -> digs,
			      ((ksi_bignum) y) -> sign);
	}

      KSI_EXN_ERR ("type", y, "logtest: invalid integer");
    }

  if (KSI_BIGNUM_P (x))
    {
      if (KSI_SINT_P (y))
	{
	  int i = KSI_SINT_COD (y);
	  unsigned n = ABS (i);
	  return big_logtest (&n, i < 0 ? -1 : 1,
			      ((ksi_bignum) x) -> digs,
			      ((ksi_bignum) x) -> sign);
	}

      if (KSI_BIGNUM_P (y))
	{
	  if (ABS (((ksi_bignum) x) -> sign) < ABS (((ksi_bignum) y) -> sign))
	    {
	      return big_logtest (((ksi_bignum) x) -> digs,
				  ((ksi_bignum) x) -> sign,
				  ((ksi_bignum) y) -> digs,
				  ((ksi_bignum) y) -> sign);
	    }
	  else
	    {
	      return big_logtest (((ksi_bignum) y) -> digs,
				  ((ksi_bignum) y) -> sign,
				  ((ksi_bignum) x) -> digs,
				  ((ksi_bignum) x) -> sign);
	    }
	}

      KSI_EXN_ERR ("type", y, "logtest: invalid integer");
    }

  KSI_EXN_ERR ("type", x, "logtest: invalid integer");
  return ksi_false;
}

ksi_obj
ksi_logbit_p (ksi_obj ind, ksi_obj x)
{
  int i;
  TYPE_CHECK (ind, KSI_SINT_P (ind) && KSI_SINT_COD (ind) >= 0,
	      "logbit?: invalid bit index in arg1");

  i = KSI_SINT_COD (ind);
  if (KSI_SINT_P (x))
    {
      if (i >= (KSI_INT_SIZE - 2))
	return KSI_SINT_COD (x) < 0 ? ksi_true : ksi_false;
      return (KSI_SINT_COD (x) & (1 << i) ? ksi_true : ksi_false);
    }

  if (KSI_BIGNUM_P (x))
    {
      unsigned *tmp;
      int s = ((ksi_bignum) x) -> sign;

      if (s > 0)
	{
	  if (i >= s * BITSPERDIG)
	    return ksi_false;

	  tmp = ((ksi_bignum) x) -> digs;
	  return (tmp [i / BITSPERDIG] & (1 << (i % BITSPERDIG))
		  ? ksi_true : ksi_false);
	}
      else
	{
	  if (i >= -s * BITSPERDIG)
	    return ksi_true;

	  s = i / BITSPERDIG + 1;
	  TALLOC (tmp, unsigned, s);
	  copy_dec_big (tmp, ((ksi_bignum) x) -> digs, s);
	  return (tmp [i / BITSPERDIG] & (1 << (i % BITSPERDIG))
		  ? ksi_false : ksi_true);
	}
    }

  KSI_ERR (x, "logbit?: invalid integer in arg2");
  return ksi_false;
}

ksi_obj
ksi_ash (ksi_obj x, ksi_obj cnt)
{
  int i;
  TYPE_CHECK (x, KSI_EINT_P (x), "ash: invalid integer in agr1");
  TYPE_CHECK (cnt, KSI_SINT_P (cnt), "ash: invalid count in arg2");

  i = KSI_SINT_COD (cnt);
  if (i < 0)
    {
      cnt = exact_expt (ksi_two, KSI_MK_SINT (-i));
      if (ksi_negative_p (x) != ksi_false)
	return ksi_sub (ksi_quot (ksi_add (x, ksi_one), cnt), ksi_one);
      else
	return ksi_quot (x, cnt);
    }
  else
    return ksi_mul (x, exact_expt (ksi_two, cnt));
}


static struct Ksi_Prim_Def defs [] = 
{
  { "number?",		ksi_number_p,		KSI_CALL_ARG1, 1 },
  { "exact?",		ksi_exact_p,		KSI_CALL_ARG1, 1 },
  { "inexact?",		ksi_inexact_p,		KSI_CALL_ARG1, 1 },
  { "integer?",		ksi_integer_p,		KSI_CALL_ARG1, 1 },
  { "rational?",	ksi_rational_p,		KSI_CALL_ARG1, 1 },
  { "real?",		ksi_real_p,		KSI_CALL_ARG1, 1 },
  { "complex?",		ksi_complex_p,		KSI_CALL_ARG1, 1 },

  { "+",		ksi_plus,		KSI_CALL_REST0, 0 },
  { "-",		ksi_minus,		KSI_CALL_REST0, 1 },
  { "*",		ksi_star,		KSI_CALL_REST0, 0 },
  { "/",		ksi_slash,		KSI_CALL_REST0, 1 },

  { "=",		ksi_num_eq_p,		KSI_CALL_REST0, 0 },
  { "<",		ksi_num_lt_p,		KSI_CALL_REST0, 0 },
  { ">",		ksi_num_gt_p,		KSI_CALL_REST0, 0 },
  { "<=",		ksi_num_le_p,		KSI_CALL_REST0, 0 },
  { ">=",		ksi_num_ge_p,		KSI_CALL_REST0, 0 },

  { "zero?",		ksi_zero_p,		KSI_CALL_ARG1, 1 },
  { "positive?",	ksi_positive_p,		KSI_CALL_ARG1, 1 },
  { "negative?",	ksi_negative_p,		KSI_CALL_ARG1, 1 },
  { "odd?",		ksi_odd_p,		KSI_CALL_ARG1, 1 },
  { "even?",		ksi_even_p,		KSI_CALL_ARG1, 1 },

  { "max",		ksi_max,		KSI_CALL_REST0, 1 },
  { "min",		ksi_min,		KSI_CALL_REST0, 1 },
  { "abs",		ksi_abs,		KSI_CALL_ARG1,  1 },

  { "quotient",		ksi_quot,		KSI_CALL_ARG2, 2 },
  { "remainder",	ksi_rem,		KSI_CALL_ARG2, 2 },
  { "modulo",		ksi_mod,		KSI_CALL_ARG2, 2 },

  { "gcd",		ksi_gcd,		KSI_CALL_REST0, 0 },
  { "lcm",		ksi_lcm,		KSI_CALL_REST0, 0 },

  { "make-rectangular",	make_rectangular,	KSI_CALL_ARG2, 2 },
  { "make-polar",	make_polar,		KSI_CALL_ARG2, 2 },
  { "real-part",	real_part,		KSI_CALL_ARG1, 1 },
  { "imag-part",	imag_part,		KSI_CALL_ARG1, 1 },
  { "magnitude",	ksi_abs,		KSI_CALL_ARG1, 1 },
  { "angle",		angle,			KSI_CALL_ARG1, 1 },

  { "floor",		ksi_floor,		KSI_CALL_ARG1, 1 },
  { "ceiling",		ksi_ceiling,		KSI_CALL_ARG1, 1 },
  { "truncate",		ksi_truncate,		KSI_CALL_ARG1, 1 },
  { "round",		ksi_round,		KSI_CALL_ARG1, 1 },

  { "exp",		ksi_exp,		KSI_CALL_ARG1, 1 },
  { "log",		ksi_log,		KSI_CALL_ARG1, 1 },
  { "sin",		ksi_sin,		KSI_CALL_ARG1, 1 },
  { "cos",		ksi_cos,		KSI_CALL_ARG1, 1 },
  { "tan",		ksi_tan,		KSI_CALL_ARG1, 1 },
  { "sinh",		ksi_sinh,		KSI_CALL_ARG1, 1 },
  { "cosh",		ksi_cosh,		KSI_CALL_ARG1, 1 },
  { "tanh",		ksi_tanh,		KSI_CALL_ARG1, 1 },
  { "asin",		ksi_asin,		KSI_CALL_ARG1, 1 },
  { "acos",		ksi_acos,		KSI_CALL_ARG1, 1 },
  { "atan",		ksi_atan,		KSI_CALL_ARG2, 1 },
  { "asinh",		ksi_asinh,		KSI_CALL_ARG1, 1 },
  { "acosh",		ksi_acosh,		KSI_CALL_ARG1, 1 },
  { "atanh",		ksi_atanh,		KSI_CALL_ARG1, 1 },

  { "sqrt",		ksi_sqrt,		KSI_CALL_ARG1, 1 },
  { "expt",		ksi_expt,		KSI_CALL_ARG2, 2 },

  { "exact->inexact",	ksi_exact2inexact,	KSI_CALL_ARG1, 1 },
  { "inexact->exact",	ksi_inexact2exact,	KSI_CALL_ARG1, 1 },
  { "number->string",	ksi_number2string,	KSI_CALL_ARG2, 1 },
  { "string->number",	ksi_string2number,	KSI_CALL_ARG2, 1 },

  { "lognot",		ksi_lognot,		KSI_CALL_ARG1, 1 },
  { "logtest",		ksi_logtest,		KSI_CALL_ARG2, 2 },
  { "logior",		ksi_logior,		KSI_CALL_ARG2, 2 },
  { "logxor",		ksi_logxor,		KSI_CALL_ARG2, 2 },
  { "logand",		ksi_logand,		KSI_CALL_ARG2, 2 },
  { "logbit?",		ksi_logbit_p,		KSI_CALL_ARG2, 2 },
  { "ash",		ksi_ash,		KSI_CALL_ARG2, 2 },

  { 0 }
};

void
ksi_init_numbers ()
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}


 /* End of code */
