/*
 * ksi_pair.c
 * pairs
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Fri Apr 25 02:42:06 1997
 * Last Update:   Sat Nov 10 00:03:23 2001
 */

/* Comments:
 *
 */

#include "ksi_int.h"

static char rcsid[] = "$Id: ksi_pair.c,v 1.6 2001/11/25 14:30:01 demakov Exp $";


ksi_obj
ksi_cons (ksi_obj car, ksi_obj cdr)
{
  ksi_pair x = NEW_TYPED (struct Ksi_Pair, car, cdr);
  x -> itag = KSI_TAG_PAIR;
  x -> car = car;
  x -> cdr = cdr;
  return (ksi_obj) x;
}

ksi_obj
ksi_acons (ksi_obj car, ksi_obj cdr, ksi_obj alist)
{
  ksi_pair x = NEW_TYPED (struct Ksi_Pair, car, cdr);
  ksi_pair y = NEW_TYPED (struct Ksi_Pair, car, cdr);

  x -> itag = KSI_TAG_PAIR;
  x -> car = car;
  x -> cdr = cdr;

  y -> itag = KSI_TAG_PAIR;
  y -> car = (ksi_obj) x;
  y -> cdr = alist;

  return (ksi_obj) y;
}

ksi_obj
ksi_pair_p (ksi_obj x)
{
  return KSI_PAIR_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_null_p (ksi_obj x)
{
  return KSI_NULL_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_list_p (ksi_obj x)
{
  return KSI_LIST_P (x) ? ksi_true : ksi_false;
}

int
ksi_list_len (ksi_obj x)
{
  int len = 0;
  ksi_obj z = x;
  do
    {
      CHECK_EVENTS;

      if (x == ksi_nil)
	return len;
      len++;
      if (!KSI_PAIR_P (x))
	return -len;
      x = KSI_CDR (x);

      if (x == ksi_nil)
	return len;
      len++;
      if (!KSI_PAIR_P (x))
	return -len;
      x = KSI_CDR (x);

      z = KSI_CDR (z);
    }
  while (x != z);
  return -1;
}


ksi_obj
ksi_car (ksi_obj o)
{
  TYPE_CHECK (o, KSI_PAIR_P (o), "car: invalid pair");

  return KSI_CAR (o);
}

ksi_obj
ksi_cdr (ksi_obj o)
{
  TYPE_CHECK (o, KSI_PAIR_P (o), "cdr: invalid pair");

  return KSI_CDR (o);
}

ksi_obj
ksi_set_car_x (ksi_obj o, ksi_obj v)
{
  TYPE_CHECK (o, KSI_M_PAIR_P (o),
	      "set-car!: invalid or constant pair in arg1");

  KSI_CAR (o) = v;
  return ksi_void;
}

ksi_obj
ksi_set_cdr_x (ksi_obj o, ksi_obj v)
{
  TYPE_CHECK (o, KSI_M_PAIR_P (o),
	      "set-cdr!: invalid or constant pair in arg1");

  KSI_CDR (o) = v;
  return ksi_void;
}

ksi_obj
ksi_length (ksi_obj o)
{
  int len;

  if (o == ksi_nil)
    return ksi_zero;

  TYPE_CHECK (o, KSI_PAIR_P (o), "length: invalid list");

  len = ksi_list_len (o);
  TYPE_CHECK (o, len > 0, "length: improper or circular list");

  return KSI_MK_SINT (len);
}


/*
 * ksi_append
 *	 
 *
 * Scheme procedure
 *	append argN ...
 *
 * Side effects
 *	,   ,  .
 *	  ,  cdr  
 *	. 
 */

ksi_obj
ksi_append (int argc, ksi_obj* argv)
{
  ksi_obj list = ksi_nil, *val = &list;

  if (argc <= 0)
    return ksi_nil;

  while (--argc > 0)
    {
      ksi_obj l, x = *argv++;
      for (l = x; l != ksi_nil; l = CDR (l))
        {
	  KSI_CHECK (x, PAIR_P (l), "append: invalid list");
	  CHECK_EVENTS;

          *val = ksi_cons (CAR (l), ksi_nil);
          val = & CDR (*val);
        }
    }

  *val = *argv;
  return list;
}

/*
 * ksi_append_x
 *	 .
 *
 * Scheme procedure
 *	append! argN ...
 *
 * Side effects
 *	 cdr   ,  
 *	 .
 */

ksi_obj
ksi_append_x (int argc, ksi_obj* argv)
{
  ksi_obj list = ksi_nil, *val = &list;

  if (argc <= 0)
    return ksi_nil;

  while (--argc > 0)
    {
      ksi_obj l, x = *argv++;
      *val = x;
      for (l = x; l != ksi_nil; l = CDR (l))
        {
	  TYPE_CHECK (x, PAIR_P (l), "append!: invalid list");
	  CHECK_EVENTS;

          val = & CDR (l);
        }
    }

  *val = *argv;
  return list;
}


/*
 * ksi_reverse
 *	  ,    -.
 *	   .
 *
 * Scheme procedure
 *	reverse arg
 *
 */

ksi_obj
ksi_reverse (ksi_obj arg)
{
  ksi_obj x, list = ksi_nil;
  for (x = arg; x != ksi_nil; x = CDR (x))
    {
      TYPE_CHECK (arg, PAIR_P (x), "reverse: invalid list");
      CHECK_EVENTS;

      list = ksi_cons (CAR (x), list);
    }

  return list;
}


/*
 * ksi_reverse_x
 *	      .
 *
 * Scheme procedure
 *	reverse! arg
 *
 */

ksi_obj
ksi_reverse_x (ksi_obj list)
{
  /*
   * (define (reverse! list new-tail)
   *   (if (not (pair? list))
   *      list
   *      (let ((old-tail (cdr list)))
   *        (set-cdr! list new-tail)
   *          (if (null? old-tail)
   *             list
   *             (reverse! old-tail list)))))
   */

  ksi_obj old_tail, new_tail = ksi_nil;

again:
  CHECK_EVENTS;
  if (!PAIR_P (list))
    return list;

  old_tail = CDR (list);
  CDR (list) = new_tail;
  if (old_tail == ksi_nil)
    return list;

  new_tail = list;
  list = old_tail;
  goto again;
}


/*
 * ksi_list_tail
 *	   LST,  NUM -
 *
 * Scheme procedure
 *	list-tail lst num
 */

ksi_obj
ksi_list_tail (ksi_obj lst, ksi_obj num)
{
  int k;

  TYPE_CHECK (lst, lst == ksi_nil || KSI_PAIR_P (lst),
	      "list-tail: invalid list in arg1");
  TYPE_CHECK (num, KSI_SINT_P (num), "list-tail: invalid index in arg2");

  k = KSI_SINT_COD (num);
  RANGE_CHECK (num, k >= 0, "list-tail: negative index in arg2");

  while (--k >= 0)
    {
      RANGE_CHECK (num, KSI_PAIR_P (lst), "list-tail: too big index in arg2");
      CHECK_EVENTS;
      lst = CDR (lst);
    }

  return lst;
}

/*
 * ksi_list_head
 *	 NUM   LST
 *
 * Scheme procedure
 *	list-head num lst
 */

ksi_obj
ksi_list_head (ksi_obj lst, ksi_obj num)
{
  int k;
  ksi_obj res = ksi_nil, *loc = &res;

  TYPE_CHECK (lst, lst == ksi_nil || KSI_PAIR_P (lst),
	      "list-head: invalid list in arg1");
  TYPE_CHECK (num, KSI_SINT_P (num), "list-head: invalid index in arg2");

  k = KSI_SINT_COD (num);
  RANGE_CHECK (num, k >= 0, "list-head: negative index in arg2");

  while (--k >= 0)
    {
      RANGE_CHECK (num, KSI_PAIR_P (lst), "list-head: too big index in arg2");
      CHECK_EVENTS;
      *loc = CONS (CAR (lst), ksi_nil);
      loc = & CDR (*loc);
      lst = CDR (lst);
    }

  return res;
}


/*
 * ksi_list_ref
 *	 NUM- -   LST
 *
 * Scheme procedure
 *	list-ref arg1 arg2
 */

ksi_obj
ksi_list_ref (ksi_obj lst, ksi_obj num)
{
  int k;

  TYPE_CHECK (lst, KSI_PAIR_P (lst), "list-ref: invalid list in arg1");
  TYPE_CHECK (num, KSI_SINT_P (num), "list-ref: invalid index in arg2");

  k = KSI_SINT_COD (num);
  RANGE_CHECK (num, k >= 0, "list-ref: negative index in arg2");

  while (--k >= 0)
    {
      lst = KSI_CDR (lst);
      RANGE_CHECK (num, KSI_PAIR_P (lst), "list-ref: too big index in arg2");
      CHECK_EVENTS;
    }

  return KSI_CAR (lst);
}

/*
 * ksi_list_set_x
 *	 NUM- -   LST  VAL.
 *
 * Scheme procedure
 *	list-set! num lst val
 */

ksi_obj
ksi_list_set_x (ksi_obj lst, ksi_obj num, ksi_obj val)
{
  int k;

  TYPE_CHECK (lst, KSI_PAIR_P (lst), "list-set!: invalid list in arg1");
  TYPE_CHECK (num, KSI_SINT_P (num), "list-set!: invalid index in arg2");

  k = KSI_SINT_COD (num);
  RANGE_CHECK (num, k >= 0, "list-set!: negative index in arg2");

  while (--k >= 0)
    {
      lst = KSI_CDR (lst);
      RANGE_CHECK (num, KSI_PAIR_P (lst), "list-set!: too big index in arg2");
      CHECK_EVENTS;
    }

  KSI_CAR (lst) = val;
  return ksi_void;
}

ksi_obj
ksi_last_pair (ksi_obj z)
{
  ksi_obj res = z, x = z;

  if (z == ksi_nil)
    return ksi_nil;

  TYPE_CHECK (z, PAIR_P (z), "last-pair: invalid pair in arg1");

  while (1)
    {
      x = CDR (x);
      if (!PAIR_P (x)) return res;
      res = x;
      x = CDR (x);
      if (!PAIR_P (x)) return res;
      res = x;
      z = CDR (z);
      TYPE_CHECK (z, z != x, "last-pair: circular list in arg1");
      CHECK_EVENTS;
    }
  return res;
}

/*
 * ksi_copy_list
 *	    "" (  cdr).
 *	
 */

ksi_obj
ksi_copy_list (ksi_obj x)
{
  if (KSI_VEC_P (x))
    return ksi_copy_vector (x);

  if (KSI_PAIR_P (x))
    {
      ksi_obj lst, *val = &lst;
      do
        {
	  CHECK_EVENTS;
          *val = ksi_cons (KSI_CAR (x), ksi_nil);
          val = & KSI_CDR (*val);
          x = KSI_CDR (x);
        } while (KSI_PAIR_P (x));

      *val = x;
      return lst;
    }

  return x;
}

/*
 * ksi_copy_tree
 *	    "" (  car  cdr);
 */

ksi_obj
ksi_copy_tree (ksi_obj x)
{
  if (KSI_VEC_P (x))
    {
      int i, len = KSI_VEC_LEN (x);
      ksi_obj vec = ksi_alloc_vector (len, KSI_TAG_VECTOR);
      for (i = 0; i < len; ++i)
	{
	  CHECK_EVENTS;
	  KSI_VEC_REF (vec, i) = ksi_copy_tree (KSI_VEC_REF (x, i));
	}
      return vec;
    }

  if (KSI_PAIR_P (x))
    {
      ksi_obj lst, *val = &lst;
      do
        {
	  CHECK_EVENTS;
          *val = ksi_cons (ksi_copy_tree (KSI_CAR (x)), ksi_nil);
          val = & KSI_CDR (*val);
	  x = KSI_CDR (x);
        } while (KSI_PAIR_P (x));

      *val = x;
      return lst;
    }

  return x;
}


ksi_obj
ksi_memq (ksi_obj o, ksi_obj list)
{
  ksi_obj x = list;
  while (list != ksi_nil)
    {
      TYPE_CHECK (list, KSI_PAIR_P (list), "memq: improper list in arg2");
      if (KSI_EQ_P (o, KSI_CAR (list)))
        return list;
      list = KSI_CDR (list);
      if (list == ksi_nil)
	break;

      TYPE_CHECK (list, KSI_PAIR_P (list), "memq: improper list in arg2");
      if (KSI_EQ_P (o, KSI_CAR (list)))
        return list;
      list = KSI_CDR (list);
      x = KSI_CDR (x);
      if (x == list)
	break;
      CHECK_EVENTS;
    }

  return ksi_false;
}

ksi_obj
ksi_memv (ksi_obj o, ksi_obj list)
{
  ksi_obj x = list;
  while (list != ksi_nil)
    {
      TYPE_CHECK (list, KSI_PAIR_P (list), "memv: improper list in arg2");
      if (KSI_EQV_P (o, KSI_CAR (list)))
        return list;
      list = KSI_CDR (list);
      if (list == ksi_nil)
	break;

      TYPE_CHECK (list, KSI_PAIR_P (list), "memv: improper list in arg2");
      if (KSI_EQV_P (o, KSI_CAR (list)))
        return list;
      list = KSI_CDR (list);
      x = KSI_CDR (x);
      if (x == list)
	break;
      CHECK_EVENTS;
    }

  return ksi_false;
}

ksi_obj
ksi_member (ksi_obj o, ksi_obj list, ksi_obj proc)
{
  ksi_obj x = list;

  if (proc)
    TYPE_CHECK (proc, PROC_P (proc), "member: invalid procedure in arg3");

  while (list != ksi_nil)
    {
      TYPE_CHECK (list, KSI_PAIR_P (list), "member: improper list in arg2");
      if (!proc && KSI_EQUAL_P (o, KSI_CAR (list)))
        return list;
      if (proc && KSI_TRUE_P (ksi_apply_2 (proc, o, KSI_CAR (list))))
	return list;

      list = KSI_CDR (list);
      if (list == ksi_nil)
	break;

      TYPE_CHECK (list, KSI_PAIR_P (list), "member: improper list in arg2");
      if (!proc && KSI_EQUAL_P (o, KSI_CAR (list)))
        return list;
      if (proc && KSI_TRUE_P (ksi_apply_2 (proc, o, KSI_CAR (list))))
	return list;

      list = KSI_CDR (list);
      x = KSI_CDR (x);
      if (x == list)
	break;
      CHECK_EVENTS;
    }

  return ksi_false;
}

ksi_obj
ksi_memq_remove (ksi_obj o, ksi_obj list)
{
  ksi_obj res = ksi_nil, *loc = &res;
  while (list != ksi_nil)
    {
      TYPE_CHECK (list, KSI_PAIR_P (list), "memq-remove: improper list in arg2");
      CHECK_EVENTS;
      if (!KSI_EQ_P (o, KSI_CAR (list)))
	{
	  *loc = CONS (KSI_CAR (list), ksi_nil);
	  loc = & KSI_CDR (*loc);
	}
      list = KSI_CDR (list);
    }

  return res;
}

ksi_obj
ksi_memv_remove (ksi_obj o, ksi_obj list)
{
  ksi_obj res = ksi_nil, *loc = &res;
  while (list != ksi_nil)
    {
      TYPE_CHECK (list, KSI_PAIR_P (list), "memv-remove: improper list in arg2");
      CHECK_EVENTS;
      if (!KSI_EQV_P (o, KSI_CAR (list)))
	{
	  *loc = CONS (KSI_CAR (list), ksi_nil);
	  loc = & KSI_CDR (*loc);
	}
      list = KSI_CDR (list);
    }

  return res;
}

ksi_obj
ksi_member_remove (ksi_obj o, ksi_obj list, ksi_obj proc)
{
  ksi_obj res = ksi_nil, *loc = &res;

  if (proc)
    TYPE_CHECK (proc, PROC_P (proc), "member-remove: invalid procedure in arg3");

  while (list != ksi_nil)
    {
      TYPE_CHECK (list, PAIR_P (list), "member-remove: improper list in arg2");
      CHECK_EVENTS;
      if ((!proc && !KSI_EQUAL_P (o, KSI_CAR (list)))
	  || (proc && KSI_FALSE_P (ksi_apply_2 (proc, o, KSI_CAR (list)))))
	{
	  *loc = CONS (KSI_CAR (list), ksi_nil);
	  loc = & KSI_CDR (*loc);
	}
      list = KSI_CDR (list);
    }

  return res;
}

ksi_obj
ksi_memq_remove_x (ksi_obj o, ksi_obj list)
{
  ksi_obj cur = list, *loc = &list;
  while (cur != ksi_nil)
    {
      TYPE_CHECK (list, PAIR_P (cur), "memq-remove!: invalid list in arg2");
      CHECK_EVENTS;
      if (KSI_EQ_P (o, KSI_CAR (cur)))
	*loc = KSI_CDR (cur);
      else
	loc = & KSI_CDR (cur);
      cur = KSI_CDR (cur);
    }

  return list;
}

ksi_obj
ksi_memv_remove_x (ksi_obj o, ksi_obj list)
{
  ksi_obj cur = list, *loc = &list;
  while (cur != ksi_nil)
    {
      TYPE_CHECK (list, PAIR_P (cur), "memv-remove!: invalid list in arg2");
      CHECK_EVENTS;
      if (KSI_EQV_P (o, KSI_CAR (cur)))
	*loc = KSI_CDR (cur);
      else
	loc = & KSI_CDR (cur);
      cur = KSI_CDR (cur);
    }

  return list;
}

ksi_obj
ksi_member_remove_x (ksi_obj o, ksi_obj list, ksi_obj proc)
{
  ksi_obj cur = list, *loc = &list;

  if (proc)
    TYPE_CHECK (proc, PROC_P (proc), "member-remove!: invalid procedure in arg3");
  while (cur != ksi_nil)
    {
      TYPE_CHECK (list, PAIR_P (cur), "member-remove!: improper list in arg2");
      CHECK_EVENTS;
      if ((!proc && KSI_EQUAL_P (o, KSI_CAR (cur)))
	  || (proc && KSI_TRUE_P (ksi_apply_2 (proc, o, KSI_CAR (cur)))))
	*loc = KSI_CDR (cur);
      else
	loc = & KSI_CDR (cur);
      cur = KSI_CDR (cur);
    }

  return list;
}

ksi_obj
ksi_assq (ksi_obj o, ksi_obj list)
{
  while (list != ksi_nil)
    {
      ksi_obj x;
      TYPE_CHECK (list, KSI_PAIR_P (list), "assq: improper list in arg2");
      CHECK_EVENTS;
      x = KSI_CAR (list);
      TYPE_CHECK (x, KSI_PAIR_P (x), "assq: invalid pair");
      if (KSI_EQ_P (o, KSI_CAR (x)))
        return x;
      list = KSI_CDR (list);
    }
  return ksi_false;
}

ksi_obj
ksi_assv (ksi_obj o, ksi_obj list)
{
  while (list != ksi_nil)
    {
      ksi_obj x;
      TYPE_CHECK (list, KSI_PAIR_P (list), "assv: improper list in arg2");
      CHECK_EVENTS;
      x = KSI_CAR (list);
      TYPE_CHECK (x, KSI_PAIR_P (x), "assv: invalid pair");
      if (KSI_EQV_P (o, KSI_CAR (x)))
        return x;
      list = KSI_CDR (list);
    }
  return ksi_false;
}

ksi_obj
ksi_assoc (ksi_obj o, ksi_obj list, ksi_obj proc)
{
  if (proc)
    TYPE_CHECK (proc, PROC_P (proc), "assoc: invalid procedure in arg3");

  while (list != ksi_nil)
    {
      ksi_obj x;
      TYPE_CHECK (list, KSI_PAIR_P (list), "assoc: improper list in arg2");
      CHECK_EVENTS;
      x = KSI_CAR (list);
      TYPE_CHECK (x, KSI_PAIR_P (x), "assoc: invalid pair");
      if (!proc && KSI_EQUAL_P (o, KSI_CAR (x)))
        return x;
      if (proc && KSI_TRUE_P (ksi_apply_2 (proc, o, KSI_CAR (x))))
	return x;
      list = KSI_CDR (list);
    }
  return ksi_false;
}


ksi_obj
ksi_assq_ref (ksi_obj list, ksi_obj key)
{
  while (KSI_PAIR_P (list))
    {
      ksi_obj x = KSI_CAR (list);
      if (KSI_PAIR_P (x) && KSI_EQ_P (key, KSI_CAR (x)))
        return KSI_CDR (x);
      CHECK_EVENTS;
      list = KSI_CDR (list);
    }

  return ksi_false;
}

ksi_obj
ksi_assv_ref (ksi_obj list, ksi_obj key)
{
  while (KSI_PAIR_P (list))
    {
      ksi_obj x = KSI_CAR (list);
      if (KSI_PAIR_P (x) && KSI_EQV_P (key, KSI_CAR (x)))
        return KSI_CDR (x);
      CHECK_EVENTS;
      list = KSI_CDR (list);
    }

  return ksi_false;
}

ksi_obj
ksi_assoc_ref (ksi_obj list, ksi_obj key, ksi_obj proc)
{
  while (KSI_PAIR_P (list))
    {
      ksi_obj x = KSI_CAR (list);
      if (KSI_PAIR_P (x)
	  && ((!proc && KSI_EQUAL_P (key, KSI_CAR (x)))
	      || (proc && KSI_TRUE_P (ksi_apply_2 (proc, key, KSI_CAR (x))))))
        return KSI_CDR (x);
      CHECK_EVENTS;
      list = KSI_CDR (list);
    }

  return ksi_false;
}

ksi_obj
ksi_assq_set_x (ksi_obj alist, ksi_obj key, ksi_obj val)
{
  ksi_obj list;
  for (list = alist; PAIR_P (list); list = CDR (list))
    {
      ksi_obj x = CAR (list);
      if (PAIR_P (x) && KSI_EQ_P (key, CAR (x)))
	{
	  CDR (x) = val;
	  return alist;
	}
      CHECK_EVENTS;
    }

  return ksi_acons (key, val, alist);
}

ksi_obj
ksi_assv_set_x (ksi_obj alist, ksi_obj key, ksi_obj val)
{
  ksi_obj list;
  for (list = alist; PAIR_P (list); list = CDR (list))
    {
      ksi_obj x = CAR (list);
      if (PAIR_P (x) && KSI_EQV_P (key, CAR (x)))
	{
	  CDR (x) = val;
	  return alist;
	}
      CHECK_EVENTS;
    }

  return ksi_acons (key, val, alist);
}

ksi_obj
ksi_assoc_set_x (ksi_obj alist, ksi_obj key, ksi_obj val, ksi_obj proc)
{
  ksi_obj list;
  for (list = alist; PAIR_P (list); list = CDR (list))
    {
      ksi_obj x = CAR (list);
      if (PAIR_P (x)
	  && ((!proc && KSI_EQUAL_P (key, CAR (x)))
	      || (proc && KSI_TRUE_P (ksi_apply_2 (proc, key, CAR (x))))))
	{
	  CDR (x) = val;
	  return alist;
	}
      CHECK_EVENTS;
    }

  return ksi_acons (key, val, alist);
}

ksi_obj
ksi_assq_remove_x (ksi_obj alist, ksi_obj key)
{
  ksi_obj prev = 0, list;
  for (list = alist; PAIR_P (list); list = CDR (list))
    {
      ksi_obj x = CAR (list);
      if (PAIR_P (x) && KSI_EQ_P (key, CAR (x)))
	{
	  if (list == alist)
	    return CDR (list);
	  CDR (prev) = CDR (list);
	  return alist;
	}
      prev = list;
      CHECK_EVENTS;
    }

  return alist;
}

ksi_obj
ksi_assv_remove_x (ksi_obj alist, ksi_obj key)
{
  ksi_obj prev = 0, list;
  for (list = alist; PAIR_P (list); list = CDR (list))
    {
      ksi_obj x = CAR (list);
      if (PAIR_P (x) && KSI_EQV_P (key, CAR (x)))
	{
	  if (list == alist)
	    return CDR (list);
	  CDR (prev) = CDR (list);
	  return alist;
	}
      prev = list;
      CHECK_EVENTS;
    }

  return alist;
}

ksi_obj
ksi_assoc_remove_x (ksi_obj alist, ksi_obj key, ksi_obj proc)
{
  ksi_obj prev = 0, list;
  for (list = alist; PAIR_P (list); list = CDR (list))
    {
      ksi_obj x = CAR (list);
      if (PAIR_P (x)
	  && ((!proc && KSI_EQUAL_P (key, CAR (x)))
	      || (proc && KSI_TRUE_P (ksi_apply_2 (proc, key, CAR (x))))))
	{
	  if (list == alist)
	    return CDR (list);
	  CDR (prev) = CDR (list);
	  return alist;
	}
      prev = list;
      CHECK_EVENTS;
    }

  return alist;
}


ksi_obj
ksi_new_list (int argc, ksi_obj* argv)
{
  ksi_obj args = ksi_nil;
  while (--argc >= 0)
    args = CONS (argv[argc], args);
  return args;
}

ksi_obj
ksi_make_list (ksi_obj len, ksi_obj init)
{
  int l;
  ksi_obj res = ksi_nil;

  TYPE_CHECK (len, KSI_SINT_P (len) && KSI_SINT_COD (len) >= 0,
	      "make-list: invalid integer in arg1");

  if (init == ksi_null)
    init = ksi_void;

  l = KSI_SINT_COD (len);
  while (--l >= 0)
    res = CONS (init, res);

  return res;
}


static ksi_obj
eval_cxr (char* data, ksi_obj val)
{
  char* ptr;

  ptr = data + strlen (data) - 1;
  for (--ptr; *ptr != 'c'; --ptr)
    {
#if CAUTIOUS >= 1
      if (!KSI_PAIR_P (val))
	ksi_exn_error ("type", val, "%s: invalid pair", data);
#endif

      val = (*ptr == 'd') ? KSI_CDR (val) : KSI_CAR (val);
    }

  return val;
}

static ksi_obj
ksi_memq_p (ksi_obj x, ksi_obj list)
{
  return KSI_TRUE_P (ksi_memq (x, list)) ? ksi_true : ksi_false;
}

static ksi_obj
ksi_memv_p (ksi_obj x, ksi_obj list)
{
  return KSI_TRUE_P (ksi_memv (x, list)) ? ksi_true : ksi_false;
}

static ksi_obj
ksi_member_p (ksi_obj x, ksi_obj list, ksi_obj p)
{
  return KSI_TRUE_P (ksi_member (x, list, p)) ? ksi_true : ksi_false;
}

static ksi_obj
ksi_xcons (ksi_obj cdr, ksi_obj car)
{
  ksi_pair x = NEW_TYPED (struct Ksi_Pair, car, cdr);
  x -> itag = KSI_TAG_PAIR;
  x -> car = car;
  x -> cdr = cdr;
  return (ksi_obj) x;
}

static ksi_obj
ksi_cons_a (int argc, ksi_obj* argv)
{
  ksi_obj args = argv[--argc];

  while (--argc >= 0)
    args = CONS (argv[argc], args);
  return args;
}


static struct Ksi_Prim_Def defs [] = 
{
  { "xcons",		ksi_xcons,		KSI_CALL_ARG2, 2 },
  { "acons",		ksi_acons,		KSI_CALL_ARG3, 3 },
  { "cons*",		ksi_cons_a,		KSI_CALL_REST0, 1 },

  { "set-car!",		ksi_set_car_x,		KSI_CALL_ARG2, 2 },
  { "set-cdr!",		ksi_set_cdr_x,		KSI_CALL_ARG2, 2 },
  { "length",		ksi_length,		KSI_CALL_ARG1, 1 },
  { "make-list",	ksi_make_list,		KSI_CALL_ARG2, 1 },
  { "append",		ksi_append,		KSI_CALL_REST0, 0 },
  { "append!",		ksi_append_x,		KSI_CALL_REST0, 0 },
  { "reverse",		ksi_reverse,		KSI_CALL_ARG1, 1 },
  { "reverse!",		ksi_reverse_x,		KSI_CALL_ARG1, 1 },
  { "list-tail",	ksi_list_tail,		KSI_CALL_ARG2, 2 },
  { "list-head",	ksi_list_head,		KSI_CALL_ARG2, 2 },
  { "list-ref",		ksi_list_ref,		KSI_CALL_ARG2, 2 },
  { "last-pair",	ksi_last_pair,		KSI_CALL_ARG1, 1 },
  { "list-set!",	ksi_list_set_x,		KSI_CALL_ARG3, 3 },

  { "memq?",		ksi_memq_p,		KSI_CALL_ARG2, 2 },
  { "memv?",		ksi_memv_p,		KSI_CALL_ARG2, 2 },
  { "member?",		ksi_member_p,		KSI_CALL_ARG3, 2 },

  { "member",		ksi_member,		KSI_CALL_ARG3, 2 },
  { "memq-remove",	ksi_memq_remove,	KSI_CALL_ARG2, 2 },
  { "memv-remove",	ksi_memv_remove,	KSI_CALL_ARG2, 2 },
  { "member-remove",	ksi_member_remove,	KSI_CALL_ARG3, 2 },
  { "memq-remove!",	ksi_memq_remove_x,	KSI_CALL_ARG2, 2 },
  { "memv-remove!",	ksi_memv_remove_x,	KSI_CALL_ARG2, 2 },
  { "member-remove!",	ksi_member_remove_x,	KSI_CALL_ARG3, 2 },

  { "assq",		ksi_assq,		KSI_CALL_ARG2, 2 },
  { "assv",		ksi_assv,		KSI_CALL_ARG2, 2 },
  { "assoc",		ksi_assoc,		KSI_CALL_ARG3, 2 },
  { "assq-ref",		ksi_assq_ref,		KSI_CALL_ARG2, 2 },
  { "assv-ref",		ksi_assv_ref,		KSI_CALL_ARG2, 2 },
  { "assoc-ref",	ksi_assoc_ref,		KSI_CALL_ARG3, 2 },
  { "assq-set!",	ksi_assq_set_x,		KSI_CALL_ARG3, 3 },
  { "assv-set!",	ksi_assv_set_x,		KSI_CALL_ARG3, 3 },
  { "assoc-set!",	ksi_assoc_set_x,	KSI_CALL_ARG4, 3 },
  { "assq-remove!",	ksi_assq_remove_x,	KSI_CALL_ARG2, 2 },
  { "assv-remove!",	ksi_assv_remove_x,	KSI_CALL_ARG2, 2 },
  { "assoc-remove!",	ksi_assoc_remove_x,	KSI_CALL_ARG3, 2 },

  { "copy-list",	ksi_copy_list,		KSI_CALL_ARG1, 1 },
  { "copy-tree",	ksi_copy_tree,		KSI_CALL_ARG1, 1 },

  { 0 }
};

void
ksi_init_pairs ()
{
  int l, x;

  ksi_ignore (rcsid);
  ksi_reg_unit (defs);

  ksi_defsym ("null?",	ksi_new_imm (KSI_TAG_NULLP), 0);
  ksi_defsym ("pair?",	ksi_new_imm (KSI_TAG_PAIRP), 0);
  ksi_defsym ("list?",	ksi_new_imm (KSI_TAG_LISTP), 0);
  ksi_defsym ("cons",	ksi_new_imm (KSI_TAG_CONS), 0);
  ksi_defsym ("car",	ksi_new_imm (KSI_TAG_CAR), 0);
  ksi_defsym ("cdr",	ksi_new_imm (KSI_TAG_CDR), 0);
  ksi_defsym ("memq",	ksi_new_imm (KSI_TAG_MEMQ), 0);
  ksi_defsym ("memv",	ksi_new_imm (KSI_TAG_MEMV), 0);

  KSI->list_proc = ksi_new_imm (KSI_TAG_LIST);
  ksi_defsym ("list", KSI->list_proc, 0);

  /* c(a|d)*r defined separately */
  for (l = 2; l <= 6; ++l)
    for (x = 0; x < (1 << l); ++x)
      {
	ksi_obj prim;
        char* name = (char*) ksi_malloc_data (l+3);
        int i = 0;

        for (name [0] = 'c'; i++ < l; )
          name[i] = x & (1 << (l-i)) ? 'd' : 'a';
        name[i++] = 'r';
        name[i] = 0;

	prim = ksi_defun (name, eval_cxr, KSI_CALL_ARG2, 2);
        ksi_defsym (name, ksi_close_prim (prim, 1, (ksi_obj*) &name), 0);
      }
}


 /* End of code */
