/*
 * ksi_prim.c
 * primitives
 *
 * Copyright (C) 1997-2000,2001, Ivan Demakov.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Wed Dec  3 08:16:27 1997
 * Last Update:   Sun Mar 25 04:25:07 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_prim.h"


static char rcsid[] = "$Id: ksi_prim.c,v 1.5.6.1 2009/03/01 10:39:13 ksion Exp $";


ksi_obj
ksi_apply_prim (ksi_obj prim, int ac, ksi_obj *av)
{
  int nary;

  KSI_TYPE_CHECK2 (prim, KSI_PRIM_P (prim),
		   "ksi_apply_prim: invalid primitive");

  WNA_CHECK (ac >= KSI_PRIM_REQV (prim), prim, KSI_PRIM_NAME (prim));

#if 1

  nary = ac * 2 * KSI_CALL_NUM + KSI_PRIM_CALL (prim);
  switch (nary)
    {
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_ARG0:
      return ((ksi_call_arg0) KSI_PRIM_PROC (prim)) ();
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_ARG1:
      return ((ksi_call_arg1) KSI_PRIM_PROC (prim)) (0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_ARG2:
      return ((ksi_call_arg2) KSI_PRIM_PROC (prim)) (0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_ARG3:
      return ((ksi_call_arg3) KSI_PRIM_PROC (prim)) (0, 0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_ARG4:
      return ((ksi_call_arg4) KSI_PRIM_PROC (prim)) (0, 0, 0,
						     0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_ARG5:
      return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (0, 0, 0,
						     0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_ARG6:
      return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (0, 0, 0,
						     0, 0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_ARG7:
      return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (0, 0, 0,
						     0, 0, 0,
						     0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_ARG8:
      return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (0, 0, 0,
						     0, 0, 0,
						     0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_ARG9:
      return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (0, 0, 0,
						     0, 0, 0,
						     0, 0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_REST0:
      return ((ksi_call_rest0) KSI_PRIM_PROC (prim)) (0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_REST1:
      return ((ksi_call_rest1) KSI_PRIM_PROC (prim)) (0,
						      0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_REST2:
      return ((ksi_call_rest2) KSI_PRIM_PROC (prim)) (0, 0,
						      0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_REST3:
      return ((ksi_call_rest3) KSI_PRIM_PROC (prim)) (0, 0, 0,
						      0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_REST4:
      return ((ksi_call_rest4) KSI_PRIM_PROC (prim)) (0, 0, 0,
						      0,
						      0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_REST5:
      return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (0, 0, 0,
						      0, 0,
						      0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_REST6:
      return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (0, 0, 0,
						      0, 0, 0,
						      0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_REST7:
      return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (0, 0, 0,
						      0, 0, 0,
						      0,
						      0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_REST8:
      return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (0, 0, 0,
						      0, 0, 0,
						      0, 0,
						      0, 0);
    case 0 * 2 * KSI_CALL_NUM + KSI_CALL_REST9:
      return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (0, 0, 0,
						      0, 0, 0,
						      0, 0, 0,
						      0, 0);

    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_ARG1:
      return ((ksi_call_arg1) KSI_PRIM_PROC (prim)) (av[0]);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_ARG2:
      return ((ksi_call_arg2) KSI_PRIM_PROC (prim)) (av[0], 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_ARG3:
      return ((ksi_call_arg3) KSI_PRIM_PROC (prim)) (av[0], 0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_ARG4:
      return ((ksi_call_arg4) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						     0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_ARG5:
      return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						     0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_ARG6:
      return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						     0, 0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_ARG7:
      return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						     0, 0, 0,
						     0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_ARG8:
      return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						     0, 0, 0,
						     0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_ARG9:
      return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						     0, 0, 0,
						     0, 0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_REST1:
      return ((ksi_call_rest1) KSI_PRIM_PROC (prim)) (av[0],
						      0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_REST2:
      return ((ksi_call_rest2) KSI_PRIM_PROC (prim)) (av[0], 0,
						      0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_REST3:
      return ((ksi_call_rest3) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						      0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_REST4:
      return ((ksi_call_rest4) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						      0,
						      0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_REST5:
      return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						      0, 0,
						      0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_REST6:
      return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						      0, 0, 0,
						      0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_REST7:
      return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						      0, 0, 0,
						      0,
						      0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_REST8:
      return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						      0, 0, 0,
						      0, 0,
						      0, 0);
    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_REST9:
      return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
						      0, 0, 0,
						      0, 0, 0,
						      0, 0);

    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_ARG2:
      return ((ksi_call_arg2) KSI_PRIM_PROC (prim)) (av[0], av[1]);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_ARG3:
      return ((ksi_call_arg3) KSI_PRIM_PROC (prim)) (av[0], av[1], 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_ARG4:
      return ((ksi_call_arg4) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						     0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_ARG5:
      return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						     0, 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_ARG6:
      return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						     0, 0, 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_ARG7:
      return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						     0, 0, 0,
						     0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_ARG8:
      return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						     0, 0, 0,
						     0, 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_ARG9:
      return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						     0, 0, 0,
						     0, 0, 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_REST2:
      return ((ksi_call_rest2) KSI_PRIM_PROC (prim)) (av[0], av[1],
						      0, 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_REST3:
      return ((ksi_call_rest3) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						      0, 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_REST4:
      return ((ksi_call_rest4) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						      0,
						      0, 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_REST5:
      return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						      0, 0,
						      0, 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_REST6:
      return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						      0, 0, 0,
						      0, 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_REST7:
      return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						      0, 0, 0,
						      0,
						      0, 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_REST8:
      return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						      0, 0, 0,
						      0, 0,
						      0, 0);
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_REST9:
      return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
						      0, 0, 0,
						      0, 0, 0,
						      0, 0);

    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_ARG3:
      return ((ksi_call_arg3) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2]);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_ARG4:
      return ((ksi_call_arg4) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_ARG5:
      return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     0, 0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_ARG6:
      return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     0, 0, 0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_ARG7:
      return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     0, 0, 0,
						     0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_ARG8:
      return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     0, 0, 0,
						     0, 0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_ARG9:
      return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     0, 0, 0,
						     0, 0, 0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_REST3:
      return ((ksi_call_rest3) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      0, 0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_REST4:
      return ((ksi_call_rest4) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      0,
						      0, 0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_REST5:
      return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      0, 0,
						      0, 0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_REST6:
      return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      0, 0, 0,
						      0, 0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_REST7:
      return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      0, 0, 0,
						      0,
						      0, 0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_REST8:
      return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      0, 0, 0,
						      0, 0,
						      0, 0);
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_REST9:
      return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      0, 0, 0,
						      0, 0, 0,
						      0, 0);

    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_ARG4:
      return ((ksi_call_arg4) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3]);
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_ARG5:
      return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], 0);
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_ARG6:
      return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], 0, 0);
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_ARG7:
      return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], 0, 0,
						     0);
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_ARG8:
      return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], 0, 0,
						     0, 0);
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_ARG9:
      return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], 0, 0,
						     0, 0, 0);
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_REST4:
      return ((ksi_call_rest4) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3],
						      0, 0);
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_REST5:
      return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], 0,
						      0, 0);
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_REST6:
      return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], 0, 0,
						      0, 0);
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_REST7:
      return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], 0, 0,
						      0,
						      0, 0);
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_REST8:
      return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], 0, 0,
						      0, 0,
						      0, 0);
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_REST9:
      return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], 0, 0,
						      0, 0, 0,
						      0, 0);

    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_ARG5:
      return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4]);
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_ARG6:
      return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], 0);
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_ARG7:
      return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], 0,
						     0);
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_ARG8:
      return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], 0,
						     0, 0);
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_ARG9:
      return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], 0,
						     0, 0, 0);
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_REST5:
      return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4],
						      0, 0);
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_REST6:
      return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], 0,
						      0, 0);
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_REST7:
      return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], 0,
						      0,
						      0, 0);
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_REST8:
      return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], 0,
						     0, 0);
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_REST9:
      return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], 0,
						      0, 0, 0,
						      0, 0);

    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_ARG6:
      return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], av[5]);
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_ARG7:
      return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], av[5],
						     0);
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_ARG8:
      return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], av[5],
						     0, 0);
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_ARG9:
      return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], av[5],
						     0, 0, 0);
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_REST6:
      return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      0, 0);
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_REST7:
      return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      0,
						      0, 0);
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_REST8:
      return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      0, 0,
						      0, 0);
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_REST9:
      return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      0, 0, 0,
						      0, 0);

    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_ARG7:
      return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], av[5],
						     av[6]);
    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_ARG8:
      return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], av[5],
						     av[6], 0);
    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_ARG9:
      return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], av[5],
						     av[6], 0, 0);
    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_REST7:
      return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      av[6],
						      0, 0);
    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_REST8:
      return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      av[6], 0,
						      0, 0);
    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_REST9:
      return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      av[6], 0, 0,
						      0, 0);

    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_ARG8:
      return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], av[5],
						     av[6], av[7]);
    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_ARG9:
      return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], av[5],
						     av[6], av[7], 0);
    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_REST8:
      return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      av[6], av[7],
						      0, 0);
    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_REST9:
      return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      av[6], av[7], 0,
						      0, 0);

    case 9 * 2 * KSI_CALL_NUM + KSI_CALL_ARG9:
      return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						     av[3], av[4], av[5],
						     av[6], av[7], av[8]);
    case 9 * 2 * KSI_CALL_NUM + KSI_CALL_REST9:
      return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      av[6], av[7], av[8],
						      0, 0);

    case 1 * 2 * KSI_CALL_NUM + KSI_CALL_REST0:
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_REST0:
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_REST0:
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_REST0:
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_REST0:
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_REST0:
    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_REST0:
    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_REST0:
    case 9 * 2 * KSI_CALL_NUM + KSI_CALL_REST0:
      goto rest0;
      
    case 2 * 2 * KSI_CALL_NUM + KSI_CALL_REST1:
    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_REST1:
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_REST1:
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_REST1:
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_REST1:
    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_REST1:
    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_REST1:
    case 9 * 2 * KSI_CALL_NUM + KSI_CALL_REST1:
      goto rest1;

    case 3 * 2 * KSI_CALL_NUM + KSI_CALL_REST2:
    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_REST2:
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_REST2:
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_REST2:
    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_REST2:
    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_REST2:
    case 9 * 2 * KSI_CALL_NUM + KSI_CALL_REST2:
      goto rest2;

    case 4 * 2 * KSI_CALL_NUM + KSI_CALL_REST3:
    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_REST3:
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_REST3:
    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_REST3:
    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_REST3:
    case 9 * 2 * KSI_CALL_NUM + KSI_CALL_REST3:
      goto rest3;

    case 5 * 2 * KSI_CALL_NUM + KSI_CALL_REST4:
    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_REST4:
    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_REST4:
    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_REST4:
    case 9 * 2 * KSI_CALL_NUM + KSI_CALL_REST4:
      goto rest4;

    case 6 * 2 * KSI_CALL_NUM + KSI_CALL_REST5:
    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_REST5:
    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_REST5:
    case 9 * 2 * KSI_CALL_NUM + KSI_CALL_REST5:
      goto rest5;

    case 7 * 2 * KSI_CALL_NUM + KSI_CALL_REST6:
    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_REST6:
    case 9 * 2 * KSI_CALL_NUM + KSI_CALL_REST6:
      goto rest6;

    case 8 * 2 * KSI_CALL_NUM + KSI_CALL_REST7:
    case 9 * 2 * KSI_CALL_NUM + KSI_CALL_REST7:
      goto rest7;

    case 9 * 2 * KSI_CALL_NUM + KSI_CALL_REST8:
      goto rest8;
    }

  switch (KSI_PRIM_CALL (prim))
    {
    case KSI_CALL_REST0: rest0:
      return ((ksi_call_rest0) KSI_PRIM_PROC (prim)) (ac, av);
    case KSI_CALL_REST1: rest1:
      return ((ksi_call_rest1) KSI_PRIM_PROC (prim)) (av[0],
						      ac-1, av+1);
    case KSI_CALL_REST2: rest2:
      return ((ksi_call_rest2) KSI_PRIM_PROC (prim)) (av[0], av[1],
						      ac-2, av+2);
    case KSI_CALL_REST3: rest3:
      return ((ksi_call_rest3) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      ac-3, av+3);
    case KSI_CALL_REST4: rest4:
      return ((ksi_call_rest4) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3],
						      ac-4, av+4);

    case KSI_CALL_REST5: rest5:
      return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4],
						      ac-5, av+5);
    case KSI_CALL_REST6: rest6:
      return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      ac-6, av+6);
    case KSI_CALL_REST7: rest7:
      return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      av[6],
						      ac-7, av+7);
    case KSI_CALL_REST8: rest8:
      return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      av[6], av[7],
						      ac-8, av+8);
    case KSI_CALL_REST9:
      return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
						      av[3], av[4], av[5],
						      av[6], av[7], av[8],
						      ac-9, av+9);

    case KSI_CALL_ARG0:
    case KSI_CALL_ARG1:
    case KSI_CALL_ARG2:
    case KSI_CALL_ARG3:
    case KSI_CALL_ARG4:
    case KSI_CALL_ARG5:
    case KSI_CALL_ARG6:
    case KSI_CALL_ARG7:
    case KSI_CALL_ARG8:
    case KSI_CALL_ARG9:
      ksi_exn_error ("arity", prim, ksi_wna_s, KSI_PRIM_NAME (prim));

    default:
      break;
    }

#else

  switch (KSI_PRIM_CALL (prim))
    {
    case KSI_CALL_ARG0:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_arg0) KSI_PRIM_PROC (prim)) ();
	default:
	wna_err:
	  ksi_exn_error ("arity", prim, ksi_wna_s, KSI_PRIM_NAME (prim));
	}
    case KSI_CALL_ARG1:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_arg1) KSI_PRIM_PROC (prim)) (0);
	case 1:
	  return ((ksi_call_arg1) KSI_PRIM_PROC (prim)) (av[0]);
	default:
	  goto wna_err;
	}
    case KSI_CALL_ARG2:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_arg2) KSI_PRIM_PROC (prim)) (0, 0);
	case 1:
	  return ((ksi_call_arg2) KSI_PRIM_PROC (prim)) (av[0], 0);
	case 2:
	  return ((ksi_call_arg2) KSI_PRIM_PROC (prim)) (av[0], av[1]);
	default:
	  goto wna_err;
	}
    case KSI_CALL_ARG3:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_arg3) KSI_PRIM_PROC (prim)) (0, 0, 0);
	case 1:
	  return ((ksi_call_arg3) KSI_PRIM_PROC (prim)) (av[0], 0, 0);
	case 2:
	  return ((ksi_call_arg3) KSI_PRIM_PROC (prim)) (av[0], av[1], 0);
	case 3:
	  return ((ksi_call_arg3) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2]);
	default:
	  goto wna_err;
	}
    case KSI_CALL_ARG4:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_arg4) KSI_PRIM_PROC (prim)) (0, 0, 0,
							 0);
	case 1:
	  return ((ksi_call_arg4) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							 0);
	case 2:
	  return ((ksi_call_arg4) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							 0);
	case 3:
	  return ((ksi_call_arg4) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 0);
	case 4:
	  return ((ksi_call_arg4) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3]);
	default:
	  goto wna_err;
	}
    case KSI_CALL_ARG5:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (0, 0, 0,
							 0, 0);
	case 1:
	  return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							 0, 0);
	case 2:
	  return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							 0, 0);
	case 3:
	  return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 0, 0);
	case 4:
	  return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], 0);
	case 5:
	  return ((ksi_call_arg5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4]);
	default:
	  goto wna_err;
	}
    case KSI_CALL_ARG6:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (0, 0, 0,
							 0, 0, 0);
	case 1:
	  return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							 0, 0, 0);
	case 2:
	  return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							 0, 0, 0);
	case 3:
	  return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 0, 0, 0);
	case 4:
	  return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], 0, 0);
	case 5:
	  return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], 0);
	case 6:
	  return ((ksi_call_arg6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], av[5]);
	default:
	  goto wna_err;
	}
    case KSI_CALL_ARG7:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (0, 0, 0,
							 0, 0, 0,
							 0);
	case 1:
	  return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							 0, 0, 0,
							 0);
	case 2:
	  return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							 0, 0, 0,
							 0);
	case 3:
	  return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 0, 0, 0,
							 0);
	case 4:
	  return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], 0, 0,
							 0);
	case 5:
	  return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], 0,
							 0);
	case 6:
	  return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], av[5],
							 0);
	case 7:
	  return ((ksi_call_arg7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], av[5],
							 av[6]);
	default:
	  goto wna_err;
	}
    case KSI_CALL_ARG8:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (0, 0, 0,
							 0, 0, 0,
							 0, 0);
	case 1:
	  return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							 0, 0, 0,
							 0, 0);
	case 2:
	  return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							 0, 0, 0,
							 0, 0);
	case 3:
	  return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 0, 0, 0,
							 0, 0);
	case 4:
	  return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], 0, 0,
							 0, 0);
	case 5:
	  return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], 0,
							 0, 0);
	case 6:
	  return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], av[5],
							 0, 0);
	case 7:
	  return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], av[5],
							 av[6], 0);
	case 8:
	  return ((ksi_call_arg8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], av[5],
							 av[6], av[7]);
	default:
	  goto wna_err;
	}
    case KSI_CALL_ARG9:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (0, 0, 0,
							 0, 0, 0,
							 0, 0, 0);
	case 1:
	  return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							 0, 0, 0,
							 0, 0, 0);
	case 2:
	  return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							 0, 0, 0,
							 0, 0, 0);
	case 3:
	  return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 0, 0, 0,
							 0, 0, 0);
	case 4:
	  return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], 0, 0,
							 0, 0, 0);
	case 5:
	  return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], 0,
							 0, 0, 0);
	case 6:
	  return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], av[5],
							 0, 0, 0);
	case 7:
	  return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], av[5],
							 av[6], 0, 0);
	case 8:
	  return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], av[5],
							 av[6], av[7], 0);
	case 9:
	  return ((ksi_call_arg9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							 av[3], av[4], av[5],
							 av[6], av[7], av[8]);
	default:
	  goto wna_err;
	}

    case KSI_CALL_REST0:
      return ((ksi_call_rest0) KSI_PRIM_PROC (prim)) (ac, av);
    case KSI_CALL_REST1:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_rest1) KSI_PRIM_PROC (prim)) (0,
							  0, 0);
	default:
	  return ((ksi_call_rest1) KSI_PRIM_PROC (prim)) (av[0],
							  ac-1, av+1);
	}
    case KSI_CALL_REST2:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_rest2) KSI_PRIM_PROC (prim)) (0, 0,
							  0, 0);
	case 1:
	  return ((ksi_call_rest2) KSI_PRIM_PROC (prim)) (av[0], 0,
							  0, 0);
	default:
	  return ((ksi_call_rest2) KSI_PRIM_PROC (prim)) (av[0], av[1],
							  ac-2, av+2);
	}
    case KSI_CALL_REST3:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_rest3) KSI_PRIM_PROC (prim)) (0, 0, 0,
							  0, 0);
	case 1:
	  return ((ksi_call_rest3) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							  0, 0);
	case 2:
	  return ((ksi_call_rest3) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							  0, 0);
	default:
	  return ((ksi_call_rest3) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  ac-3, av+3);
	}
    case KSI_CALL_REST4:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_rest4) KSI_PRIM_PROC (prim)) (0, 0, 0,
							  0,
							  0, 0);
	case 1:
	  return ((ksi_call_rest4) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							  0,
							  0, 0);
	case 2:
	  return ((ksi_call_rest4) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							  0,
							  0, 0);
	case 3:
	  return ((ksi_call_rest4) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  0,
							  0, 0);
	default:
	  return ((ksi_call_rest4) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3],
							  ac-4, av+4);
	}
    case KSI_CALL_REST5:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (0, 0, 0,
							  0, 0,
							  0, 0);
	case 1:
	  return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							  0, 0,
							  0, 0);
	case 2:
	  return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							  0, 0,
							  0, 0);
	case 3:
	  return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  0, 0,
							  0, 0);
	case 4:
	  return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], 0,
							  0, 0);
	default:
	  return ((ksi_call_rest5) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4],
							  ac-5, av+5);
	}
    case KSI_CALL_REST6:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (0, 0, 0,
							  0, 0, 0,
							  0, 0);
	case 1:
	  return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							  0, 0, 0,
							  0, 0);
	case 2:
	  return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							  0, 0, 0,
							  0, 0);
	case 3:
	  return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  0, 0, 0,
							  0, 0);
	case 4:
	  return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], 0, 0,
							  0, 0);
	case 5:
	  return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], 0,
							  0, 0);
	default:
	  return ((ksi_call_rest6) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], av[5],
							  ac-6, av+6);
	}
    case KSI_CALL_REST7:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (0, 0, 0,
							  0, 0, 0,
							  0,
							  0, 0);
	case 1:
	  return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							  0, 0, 0,
							  0,
							  0, 0);
	case 2:
	  return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							  0, 0, 0,
							  0,
							  0, 0);
	case 3:
	  return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  0, 0, 0,
							  0,
							  0, 0);
	case 4:
	  return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], 0, 0,
							  0,
							  0, 0);
	case 5:
	  return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], 0,
							  0,
							  0, 0);
	case 6:
	  return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], av[5],
							  0,
							  0, 0);
	default:
	  return ((ksi_call_rest7) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], av[5],
							  av[6],
							  ac-7, av+7);
	}
    case KSI_CALL_REST8:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (0, 0, 0,
							  0, 0, 0,
							  0, 0,
							  0, 0);
	case 1:
	  return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							  0, 0, 0,
							  0, 0,
							  0, 0);
	case 2:
	  return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							  0, 0, 0,
							  0, 0,
							  0, 0);
	case 3:
	  return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  0, 0, 0,
							  0, 0,
							  0, 0);
	case 4:
	  return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], 0, 0,
							  0, 0,
							  0, 0);
	case 5:
	  return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], 0,
							  0, 0,
							  0, 0);
	case 6:
	  return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], av[5],
							  0, 0,
							  0, 0);
	case 7:
	  return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], av[5],
							  av[6], 0,
							  0, 0);
	default:
	  return ((ksi_call_rest8) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], av[5],
							  av[6], av[7],
							  ac-8, av+8);
	}
    case KSI_CALL_REST9:
      switch (ac)
	{
	case 0:
	  return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (0, 0, 0,
							  0, 0, 0,
							  0, 0, 0,
							  0, 0);
	case 1:
	  return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], 0, 0,
							  0, 0, 0,
							  0, 0, 0,
							  0, 0);
	case 2:
	  return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], 0,
							  0, 0, 0,
							  0, 0, 0,
							  0, 0);
	case 3:
	  return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  0, 0, 0,
							  0, 0, 0,
							  0, 0);
	case 4:
	  return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], 0, 0,
							  0, 0, 0,
							  0, 0);
	case 5:
	  return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], 0,
							  0, 0, 0,
							  0, 0);
	case 6:
	  return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], av[5],
							  0, 0, 0,
							  0, 0);
	case 7:
	  return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], av[5],
							  av[6], 0, 0,
							  0, 0);
	case 8:
	  return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], av[5],
							  av[6], av[7], 0,
							  0, 0);
	default:
	  return ((ksi_call_rest9) KSI_PRIM_PROC (prim)) (av[0], av[1], av[2],
							  av[3], av[4], av[5],
							  av[6], av[7], av[8],
							  ac-9, av+9);
	}
    default:
      break;
    }
#endif

  return (ksi_obj) KSI_PRIM_PROC (prim);
}


ksi_obj
ksi_defun (const char *name, ksi_proc_t proc, ksi_call_t call, int reqv)
{
  ksi_obj  sym = (name ? ksi_str02sym (name) : 0);
  ksi_prim val = NEW_TYPED (struct Ksi_Prim, name, name);

  if (call == KSI_CALL_ARG0 && reqv == 0)
    val->itag = KSI_TAG_PRIM_0;
  else if (call == KSI_CALL_ARG1 && reqv == 1)
    val->itag = KSI_TAG_PRIM_1;
  else if (call == KSI_CALL_ARG2 && reqv == 2)
    val->itag = KSI_TAG_PRIM_2;
  else if (call == KSI_CALL_REST0)
    val->itag = KSI_TAG_PRIM_r;
  else
    val->itag = KSI_TAG_PRIM;

  val->proc = proc;
  val->call = call;
  val->reqv = reqv;
  val->name = (sym ? KSI_SYM_PTR (sym) : ksi_aprintf ("#<prim %p>", val));

  if (sym)
    ksi_define (sym, (ksi_obj) val, 0);

  return (ksi_obj) val;
}

ksi_obj
ksi_reg_prim (struct Ksi_Prim_Def *def)
{
  return ksi_defun (def->name, def->proc, def->call, def->reqv);
}

void
ksi_reg_unit (struct Ksi_Prim_Def *defs)
{
  ksi_ignore (rcsid);
  while (defs->name)
    {
      ksi_defun (defs->name, defs->proc, defs->call, defs->reqv);
      defs++;
    }
}

ksi_obj
ksi_close_prim (ksi_obj prim, int argc, ksi_obj *argv)
{
  ksi_prim_closure clos;
  int i;

  KSI_TYPE_CHECK2 (prim, KSI_PRIM_P (prim),
		   "ksi_close_prim: invalid primitive");

  if (argc <= 0)
    return prim;

  clos = NEW_TYPED_ARR (struct Ksi_Prim_Closure, argc, prim, argv);
  clos->itag = KSI_TAG_PRIM_CLOSURE;
  clos->prim = prim;
  clos->argc = argc;
  for (i = 0; i < argc; i++)
    clos->argv[i] = argv[i];

  return (ksi_obj) clos;
}

ksi_obj
ksi_apply_prim_closure (ksi_obj clos, int argc, ksi_obj *argv)
{
  int n, i;
  ksi_obj *v;

  n = argc + ((ksi_prim_closure) clos) -> argc;
  v = (ksi_obj*) alloca (sizeof (ksi_obj) * n);

  for (n = 0; n < ((ksi_prim_closure) clos) -> argc; n++)
    v[n] = ((ksi_prim_closure) clos) -> argv[n];
  for (i = 0; i < argc; n++, i++)
    v[n] = argv[i];

  return ksi_apply_prim (((ksi_prim_closure) clos) -> prim, n, v);
}


ksi_obj
ksi_make_prim (ksi_proc_t proc, void *data, const char *name)
{
  ksi_prim val = NEW_TYPED (struct Ksi_Prim, name, name);
  val->itag = KSI_TAG_PRIM;
  val->proc = proc;
  val->call = KSI_CALL_REST1;
  val->reqv = 1;
  val->name = (name ? name : ksi_aprintf ("#<prim %p>", val));

  return ksi_close_prim ((ksi_obj) val, 1, (ksi_obj*) &data);
}


 /* End of code */
