/*
 * ksi_proc.c
 * Misc procedures
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Wed Aug  6 22:26:25 1997
 * Last Update:   Fri Jul 27 22:36:51 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_klos.h"

static char rcsid[] = "$Id: ksi_proc.c,v 1.7 2001/07/27 18:48:05 demakov Exp $";

ksi_obj
ksi_void_p (ksi_obj x)
{
  return x == ksi_void ? ksi_true : ksi_false;
}

ksi_obj
ksi_procedure_p (ksi_obj x)
{
  if (!KSI_OBJ_P (x))
    return ksi_false;

  switch (x -> itag)
    {
    case KSI_TAG_PRIM:
    case KSI_TAG_PRIM_0:
    case KSI_TAG_PRIM_1:
    case KSI_TAG_PRIM_2:
    case KSI_TAG_PRIM_r:
    case KSI_TAG_CLOSURE:
    case KSI_TAG_PRIM_CLOSURE:
    case KSI_TAG_CONS:
    case KSI_TAG_CAR:
    case KSI_TAG_CDR:
    case KSI_TAG_NOT:
    case KSI_TAG_EQP:
    case KSI_TAG_EQVP:
    case KSI_TAG_EQUALP:
    case KSI_TAG_NULLP:
    case KSI_TAG_PAIRP:
    case KSI_TAG_LISTP:
    case KSI_TAG_VECTORP:
    case KSI_TAG_LIST:
    case KSI_TAG_APPLY:
    case KSI_TAG_CALL_CC:
    case KSI_TAG_CALL_WITH_VALUES:
    case KSI_TAG_MEMQ:
    case KSI_TAG_MEMV:
    case KSI_TAG_MEMBER:
      return ksi_true;
    }

  if (GENERIC_P (x))
    return ksi_true;

  if (NEXT_METHOD_P (x))
    return ksi_true;

  return ksi_false;
}

ksi_obj
ksi_procedure_arity (ksi_obj proc)
{
  TYPE_CHECK (proc, PROC_P (proc), "procedure-arity: invalid procedure in arg1");
  return ksi_true;
}

ksi_obj
ksi_procedure_has_arity_p (ksi_obj proc, ksi_obj min_args, ksi_obj max_args)
{
  if (!max_args)
    max_args = min_args;

  TYPE_CHECK (proc, PROC_P (proc), "procedure-has-arity?: invalid procedure in arg1");
  TYPE_CHECK (min_args, KSI_SINT_P (min_args), "procedure-has-arity?: invalid integer in arg2");
  TYPE_CHECK (max_args, max_args == ksi_true || KSI_SINT_P (max_args), "procedure-has-arity?: invalid integer in arg3");

  return ksi_true;
}

ksi_obj
ksi_map (ksi_obj proc, int argc, ksi_obj* argv)
{
  ksi_obj res = ksi_nil, *loc = &res;

  TYPE_CHECK (proc, PROC_P (proc), "map: invalid procedure in arg1");

  if (argc == 1)
    {
      ksi_obj x, xs = argv[0];
      for (;;)
	{
	  CHECK_EVENTS;
	  if (xs == ksi_nil)
	    return res;
	  if (PAIR_P (xs))
	    {
	      x    = CAR (xs);
	      xs   = CDR (xs);
	      *loc = CONS (ksi_apply_proc (proc, 1, &x), ksi_nil);
	      loc  = & CDR (*loc);
	    }
	  else
	    {
	      *loc = ksi_apply_proc (proc, 1, &xs);
	      return res; 
	    }
	}
    }
  else
    {
      int i;
      ksi_obj *av, *xs;
      xs = (ksi_obj*) alloca (argc * sizeof *argv);
      av = (ksi_obj*) alloca (argc * sizeof *argv);
      memcpy (xs, argv, argc * sizeof *argv);

      for (;;)
	{
	  CHECK_EVENTS;
	  if (!PAIR_P (xs[0]))
	    {
	      if (xs[0] == ksi_nil)
		return res;
#if CAUTIOUS >= 1
	      for (i = 1; i < argc; i++)
		{
		  if (xs[i] == ksi_nil || PAIR_P (xs[i]))
		    ksi_exn_error ("type", argv[0],
				   "map: improper list in arg2");
		}
#endif
	      *loc = ksi_apply_proc (proc, argc, xs);
	      return res;
	    }

	  for (i = 0; i < argc; i++)
	    {
	      if (xs[i] == ksi_nil)
		return res;
#if CAUTIOUS >= 1
	      if (!PAIR_P (xs[i]))
		ksi_exn_error ("type", argv[i],
			       "map: improper list in arg%d", i+2);
#endif
	      av[i] = CAR (xs[i]);
	      xs[i] = CDR (xs[i]);
	    }

	  *loc = CONS (ksi_apply_proc (proc, argc, av), ksi_nil);
	  loc = & CDR (*loc);
	}
    }
}

ksi_obj
ksi_for_each (ksi_obj proc, int argc, ksi_obj* argv)
{
  TYPE_CHECK (proc, PROC_P (proc), "for-each: invalid procedure in arg1");

  if (argc == 1)
    {
      ksi_obj x, xs = argv[0];
      for (;;)
	{
	  CHECK_EVENTS;
	  if (!PAIR_P (xs))
	    return ksi_void;
	  x  = CAR (xs);
	  xs = CDR (xs);
	  ksi_apply_proc (proc, 1, &x);
	}
    }
  else
    {
      int i;
      ksi_obj *av, *xs;
      xs = (ksi_obj*) alloca (argc * sizeof *argv);
      av = (ksi_obj*) alloca (argc * sizeof *argv);
      memcpy (xs, argv, argc * sizeof *argv);

      for (;;)
	{
	  CHECK_EVENTS;
	  for (i = 0; i < argc; i++)
	    {
	      if (!PAIR_P (xs[i]))
		return ksi_void;
	      av[i] = CAR (xs[i]);
	      xs[i] = CDR (xs[i]);
	    }

	  ksi_apply_proc (proc, argc, av);
	}
    }
}

ksi_obj
ksi_fold_left (ksi_obj kons, ksi_obj knil, int argc, ksi_obj* argv)
{
  int i;
  ksi_obj *xs, *av;

  TYPE_CHECK (kons, PROC_P (kons), "foldl: invalid procedure in arg1");

  xs = (ksi_obj*) alloca (argc * sizeof *argv);
  av = (ksi_obj*) alloca ((argc + 1) * sizeof *av);
  memcpy (xs, argv, argc * sizeof *argv);

  av[argc] = knil;
  for (;;)
    {
      CHECK_EVENTS;
      for (i = 0; i < argc; i++)
	{
	  if (!PAIR_P (xs[i]))
	    return av[argc];

	  av[i] = CAR (xs[i]);
	  xs[i] = CDR (xs[i]);
	}

      av[argc] = ksi_apply_proc (kons, argc+1, av);
    }

  /* never reached */
}

ksi_obj
ksi_fold_right (ksi_obj kons, ksi_obj knil, int argc, ksi_obj* argv)
{
  int i;
  ksi_obj *av;

  TYPE_CHECK (kons, PROC_P (kons), "foldl: invalid procedure in arg1");

  av = (ksi_obj*) alloca ((argc + 1) * sizeof *argv);
  for (i = 0; i < argc; i++)
    {
      if (!PAIR_P (argv[i]))
	return knil;
      av[i] = CDR (argv[i]);
    }

  av[argc] = ksi_fold_right (kons, knil, argc, av);
  for (i = 0; i < argc; i++)
    av[i] = CAR (argv[i]);

  CHECK_EVENTS;
  return ksi_apply_proc (kons, argc+1, av);
}

static ksi_obj
ksi_void_proc (int ac, ksi_obj *av)
{
  return ksi_void;
}

static ksi_obj
ksi_true_proc (int ac, ksi_obj *av)
{
  return ksi_true;
}

static ksi_obj
ksi_false_proc (int ac, ksi_obj *av)
{
  return ksi_false;
}


static struct Ksi_Prim_Def defs [] = 
{
  {"procedure?",		ksi_procedure_p,		KSI_CALL_ARG1, 1},
  {"procedure-arity",		ksi_procedure_arity,		KSI_CALL_ARG1, 1},
  {"procedure-has-arity?",	ksi_procedure_has_arity_p,	KSI_CALL_ARG2, 3},

  {"void?",			ksi_void_p,			KSI_CALL_ARG1, 1},
  {"void",			ksi_void_proc,			KSI_CALL_REST0, 0},
  {"true",			ksi_true_proc,			KSI_CALL_REST0, 0},
  {"false",			ksi_false_proc,			KSI_CALL_REST0, 0},

  {"map",			ksi_map,			KSI_CALL_REST1, 2},
  {"for-each",			ksi_for_each,			KSI_CALL_REST1, 2},
  {"foldl",			ksi_fold_left,			KSI_CALL_REST2, 3},
  {"foldr",			ksi_fold_right,			KSI_CALL_REST2, 3},

  {0}
};

void
ksi_init_procs (void)
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);

  KSI->void_proc  = ksi_spec_val ("void", 0, 0);
  KSI->true_proc  = ksi_spec_val ("true", 0, 0);
  KSI->false_proc = ksi_spec_val ("false", 0, 0);
}


 /* End of code */
