/*
 * ksi_repl.c
 * read-eval-print loop
 *
 * Copyright (C) 1997-2000,2001, Ivan Demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Sun Aug 10 22:52:47 1997
 * Last Update:   Sat Mar 17 16:51:59 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"

static char rcsid[] = "$Id: ksi_repl.c,v 1.7 2001/03/23 21:01:38 demakov Exp $";


int
ksi_call_with_dynamic_context (int (*proc) (void *data),
			       int (*handler) (void *data),
			       void *data)
{
  volatile KSI_STACK_ELEM top;
  volatile ksi_stack stack;
  volatile ksi_wind the_exit;
  volatile ksi_obj in, out, err;
  volatile int retval, old_interactive;
  ksi_obj val, tag;

  ksi_ignore (rcsid);

  old_interactive = KSI->interactive;
  stack = KSI -> stack;
  the_exit = KSI -> exit_catch;
  in  = ksi_current_input_port ();
  out = ksi_current_output_port ();
  err = ksi_current_error_port ();

again:
  KSI -> stack = (ksi_stack) &top;
  KSI -> exit_catch = ksi_add_catch (ksi_true, 0, 0);
  ksi_set_current_input_port (in);
  ksi_set_current_output_port (out);
  ksi_set_current_error_port (err);

  FLUSH_REGISTER_WINDOWS;
  if (setjmp (KSI -> exit_catch -> the_catch -> jmp . j_buf) == 0)
    {
      retval = proc (data);
      ksi_del_catch (KSI -> exit_catch);
    }
  else
    {
      tag = KSI -> exit_catch -> the_catch -> thrown;
      val = KSI -> exit_catch -> the_catch -> value;

      if (KSI_EXIT_P (tag))
	retval = KSI_EXIT_COD (tag);
      else
	{
	  ksi_handle_error (tag, val);

	  if (!handler)
	    retval = KSI_ERROR_EXIT_CODE;
	  else
	    {
	      retval = handler (data);
	      if (retval == KSI_NORMAL_EXIT_CODE)
		goto again;
	    }
	}
    }

  KSI->interactive = old_interactive;
  KSI -> stack = stack;
  KSI -> exit_catch = the_exit;
  ksi_set_current_input_port (in);
  ksi_set_current_output_port (out);
  ksi_set_current_error_port (err);

  return retval;
}


 /* End of code */
