/*
 * ksi_str.c
 * strings
 *
 * Copyright (C) 1997-2000,2001,2002, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Sat Jan 25 00:25:23 1997
 * Last Update:   Thu Feb  7 21:50:09 2002
 */

/* Comments:
 *
 */

#include "ksi_int.h"

static char rcsid[] = "$Id: ksi_str.c,v 1.7 2002/02/07 15:58:45 ksion Exp $";


static char* empty_str = "";

ksi_obj
ksi_string_p (ksi_obj x)
{
  return KSI_STR_P (x) ? ksi_true : ksi_false;
}

const char*
ksi_safe_string_ptr (ksi_obj x)
{
  char* ptr;

  if (KSI_C_STR_P (x))
    return KSI_STR_PTR (x);
  if (KSI_SYM_P (x))
    return KSI_SYM_PTR (x);

  TYPE_CHECK (x, KSI_M_STR_P (x), "ksi_safe_string_ptr: invalid string");

  ptr = (char*) ksi_malloc_data (KSI_STR_LEN (x) + 1);
  memcpy (ptr, KSI_STR_PTR (x), KSI_STR_LEN (x));
  ptr [KSI_STR_LEN (x)] = '\0';
  return ptr;
}

const char*
ksi_string2str (ksi_obj x)
{
  int len, i, k, num;
  const char* str;
  char* ptr;
  int c8 = KSI_TRUE_P (ksi_spec_val ("*print-8bit-chars*", ksi_false, 0));

  TYPE_CHECK (x, KSI_STR_P (x), "ksi_string2str: invalid string");

  len = KSI_STR_LEN (x);
  str = KSI_STR_PTR (x);

  if (len <= 0)
    return "\"\"";

  for (i = 0, k = 0, num = 0; i < len; ++i)
    {
      int c = str[i] & 0xff;
      switch (c)
	{
	case '\0':
	  num+=3;
	  break;
	case '\n': case '\r': case '\t': case '\f': case '\b': case '\a':
	case 0x1b:
	  ++k; ++num;
	  break;
	case '"': case '\\':
	  ++num;
	  break;
	default:
	  if (c < ' ' || (c >= 127 && !isalpha (c) && !c8))
	    {
	      ++k;
	      num += 3;
	    }
	}
    }

  ptr = (char*) ksi_malloc_data (len + num + 4);
  ptr [k=0] = '"';
  for (i = 0; i < len; ++i)
    {
      int c = str[i] & 0xff;
      switch (c)
	{
	case '\0':
	  c = '0';
	  ptr[++k] = '\\';
	  ptr[++k] = '0';
	  ptr[++k] = '0';
	  break;
	case '\n':
	  c = 'n';
	  ptr[++k] = '\\';
	  break;
	case '\r':
	  c = 'r';
	  ptr[++k] = '\\';
	  break;
	case '\t':
	  c = 't';
	  ptr[++k] = '\\';
	  break;
	case '\f':
	  c = 'f';
	  ptr[++k] = '\\';
	  break;
	case '\b':
	  c = 'b';
	  ptr[++k] = '\\';
	  break;
	case '\a':
	  c = 'a';
	  ptr[++k] = '\\';
	  break;
	case 0x1b:
	  c = 'e';
	  ptr[++k] = '\\';
	  break;
	case '"': case '\\':
	  ptr[++k] = '\\';
	  break;
	default:
	  if (c < ' ' || (c >= 127 && !isalpha (c) && !c8))
	    {
	      ptr[++k] = '\\';
	      ptr[++k] = (char) ('0' + (c / (8 * 8)) % 8);
	      ptr[++k] = (char) ('0' + (c / 8) % 8);
	      ptr[++k] = (char) ('0' + (c % 8));
	      continue;
	    }
	}
      ptr[++k] = c;
    }

  ptr[++k] = '"';
  ptr[++k] = '\0';

  return ptr;
}

ksi_obj
ksi_str2string (const char* s, int l)
{
  ksi_string str;

  if (s == 0 || l <= 0)
    {
      str = NEW_TYPED (struct Ksi_String, ptr, ptr);
      str -> itag = KSI_TAG_STRING;
      str -> ptr = empty_str;
      str -> len = 0;
    }
  else
    {
      str = NEW_TYPED_EXT (struct Ksi_String, l+1, ptr, ptr);
      str -> itag = KSI_TAG_STRING;
      str -> ptr = (char*) str + sizeof *str;
      memcpy (str -> ptr, s, l);
      str -> ptr [l] = '\0';
      str -> len = l;
    }

  return (ksi_obj) str;
}

ksi_obj
ksi_str02string (const char* s)
{
  return ksi_str2string (s, s ? strlen (s) : 0);
}

ksi_obj
ksi_ptr02string (const char* s)
{
  ksi_string str = NEW_TYPED (struct Ksi_String, ptr, ptr);
  str -> itag = KSI_TAG_CONST_STRING;
  if (s == 0 || *s == 0)
    {
      str -> ptr = empty_str;
      str -> len = 0;
    }
  else
    {
      str -> ptr = (char*) s;
      str -> len = strlen (s);
    }

  return (ksi_obj) str;
}

ksi_obj
ksi_ptr2string (const char* s, int len)
{
  ksi_string str = NEW_TYPED (struct Ksi_String, ptr, ptr);
  str -> itag = KSI_TAG_CONST_STRING;
  if (s == 0 || *s == 0 || len <= 0)
    {
      str -> ptr = empty_str;
      str -> len = 0;
    }
  else
    {
      str -> ptr = (char*) s;
      str -> len = len;
    }

  return (ksi_obj) str;
}

ksi_obj
ksi_make_string (int len, char c)
{
  ksi_string str = NEW_TYPED_EXT (struct Ksi_String, len+1, ptr, ptr);
  str -> itag = KSI_TAG_STRING;
  str -> ptr = (char*) str + sizeof *str;
  if (len)
    memset (str -> ptr, c, len);
  str -> ptr [len] = '\0';
  str -> len = len;

  return (ksi_obj) str;
}


ksi_obj
ksi_string_length (ksi_obj str)
{
  TYPE_CHECK (str, KSI_STR_P (str), "string-length: invalid string");

  return ksi_int2num (KSI_STR_LEN (str));
}

ksi_obj
ksi_string_ref (ksi_obj str, ksi_obj k)
{
  int ind;

  TYPE_CHECK (str, KSI_STR_P (str), "string-ref: invalid string in arg1");
  TYPE_CHECK (k, KSI_SINT_P (k), "string-ref: invalid index in arg2");

  ind = KSI_SINT_COD (k);

  RANGE_CHECK (k, 0 <= ind && ind < KSI_STR_LEN (str),
	       "string-ref: invalid index in arg2");

  return KSI_MK_CHAR ((unsigned char) KSI_STR_PTR (str) [ind]);
}

ksi_obj
ksi_string_set_x (ksi_obj str, ksi_obj k, ksi_obj ch)
{
  int ind;

  TYPE_CHECK (str, KSI_STR_P (str), "string-set!: invalid string in arg1");
  TYPE_CHECK (str, !KSI_C_STR_P (str), "string-set!: constant string in arg1");
  TYPE_CHECK (k, KSI_SINT_P (k), "string-set!: invalid index in arg2");
  TYPE_CHECK (ch, KSI_CHAR_P (ch), "string-set!: invalid char in arg3");

  ind = KSI_SINT_COD (k);

  RANGE_CHECK (k, 0 <= ind && ind < KSI_STR_LEN (str),
	       "string-set!: invalid index in arg2");

  KSI_STR_PTR (str) [ind] = KSI_CHAR_COD (ch);
  return ksi_void;
}

ksi_obj
ksi_substring (ksi_obj str, ksi_obj start, ksi_obj end)
{
  int s, e;

  TYPE_CHECK (str, KSI_STR_P (str), "substring: invalid string in arg1");
  TYPE_CHECK (start, KSI_SINT_P (start), "substring: invalid index in arg2");
  TYPE_CHECK (end, !end || KSI_SINT_P (end), "substring: invalid index in arg3");

  s = KSI_SINT_COD (start);
  e = (end ? KSI_SINT_COD (end) : KSI_STR_LEN (str));

  RANGE_CHECK (start, 0 <= s && s <= e, "substring: invalid index in arg2");
  RANGE_CHECK (end, e <= KSI_STR_LEN (str), "substring: invalid index in arg3");

  return ksi_str2string (KSI_STR_PTR (str) + s, e - s);
}

ksi_obj
ksi_string_append (int argc, ksi_obj* args)
{
  ksi_obj str;
  int i, len = 0;
  char* ptr;

  for (i = 0; i < argc; i++)
    {
      if (KSI_CHAR_P (args[i]))
	len += 1;
      else
	{
	  TYPE_CHECK (args[i], KSI_STR_P (args[i]),
		      "string-append: invalid string");
	  len += KSI_STR_LEN (args[i]);
	}
    }

  str = ksi_make_string (len, '\0');
  ptr = KSI_STR_PTR (str);
  for (i = 0; i < argc; i++)
    {
      if (KSI_CHAR_P (args[i]))
	*ptr++ = KSI_CHAR_COD (args[i]);
      else if (KSI_STR_LEN (args[i]) > 0)
	{
	  memcpy (ptr, KSI_STR_PTR (args[i]), KSI_STR_LEN (args[i]));
	  ptr += KSI_STR_LEN (args[i]);
	}
    }

  return str;
}

ksi_obj
ksi_string2list (ksi_obj str)
{
  int len;
  char* ptr;
  ksi_obj list;

  TYPE_CHECK (str, KSI_STR_P (str), "string->list: invalid string");

  len = KSI_STR_LEN (str);
  ptr = KSI_STR_PTR (str);
  list = ksi_nil;
  while (len--)
    list = ksi_cons (KSI_MK_CHAR ((unsigned char) ptr [len]), list);

  return list;
}

ksi_obj
ksi_list2string (ksi_obj list)
{
  int i, len = ksi_list_len (list);
  ksi_obj str;
  char* ptr;

  TYPE_CHECK (list, len >= 0, "list->string: invalid list in arg1");

  str = ksi_make_string (len, '\0');
  ptr = KSI_STR_PTR (str);
  for (i = 0; i < len; ++i, list = KSI_CDR (list))
    {
      ksi_obj x = KSI_CAR (list);
      TYPE_CHECK (x, KSI_CHAR_P (x), "list->string: invalid char");

      ptr [i] = KSI_CHAR_COD (x);
    }

  return str;
}

ksi_obj
ksi_new_string (int argc, ksi_obj* argv)
{
  int i;
  ksi_obj str = ksi_make_string (argc, '\0');
  char* ptr = KSI_STR_PTR (str);

  for (i = 0; i < argc; ++i)
    {
      TYPE_CHECK (argv[i], KSI_CHAR_P (argv[i]), "string: invalid char");
      ptr [i] = KSI_CHAR_COD (argv[i]);
    }

  return str;
}

ksi_obj
ksi_string_copy (ksi_obj str)
{
  TYPE_CHECK (str, KSI_STR_P (str), "string-copy: invalid string");
  return ksi_str2string (KSI_STR_PTR (str), KSI_STR_LEN (str));
}

ksi_obj
ksi_string_fill_x (ksi_obj str, ksi_obj c)
{
  int i;

  TYPE_CHECK (str, KSI_STR_P (str), "string-fill!: invalid string in arg1");
  TYPE_CHECK (str, !KSI_C_STR_P (str), "string-fill!: constant string in arg1");
  TYPE_CHECK (c, KSI_CHAR_P (c), "string-fill!: invalid char in arg2");

  for (i = 0; i < KSI_STR_LEN (str); ++i)
    KSI_STR_PTR (str) [i] = KSI_CHAR_COD (c);

  return ksi_void;
}

ksi_obj
ksi_string_eqv_p (ksi_obj s1, ksi_obj s2)
{
  int l1, l2;
  char *p1, *p2;

  TYPE_CHECK (s1, KSI_STR_P (s1), "string=?: invalid string in arg1");
  TYPE_CHECK (s2, KSI_STR_P (s2), "string=?: invalid string in arg2");

  l1 = KSI_STR_LEN (s1);
  l2 = KSI_STR_LEN (s2);
  p1 = KSI_STR_PTR (s1);
  p2 = KSI_STR_PTR (s2);

  if (l1 == l2 && p1 == p2)
    return ksi_true;

  return ksi_false;
}

ksi_obj
ksi_string_equal_p (ksi_obj s1, ksi_obj s2)
{
  int l1, l2;
  char *p1, *p2;

  TYPE_CHECK (s1, KSI_STR_P (s1), "string=?: invalid string");
  TYPE_CHECK (s2, KSI_STR_P (s2), "string=?: invalid string");

  l1 = KSI_STR_LEN (s1);
  l2 = KSI_STR_LEN (s2);
  p1 = KSI_STR_PTR (s1);
  p2 = KSI_STR_PTR (s2);

  if (l1 == l2)
    {
      if (p1 != p2)
	while (l1--)
	  if (*p1++ != *p2++)
	    return ksi_false;

      return ksi_true;
    }

  return ksi_false;
}

ksi_obj
ksi_string_ci_equal_p (ksi_obj s1, ksi_obj s2)
{
  int l1, l2;
  unsigned char *p1, *p2;

  TYPE_CHECK (s1, KSI_STR_P (s1), "string-ci=?: invalid string");
  TYPE_CHECK (s2, KSI_STR_P (s2), "string-ci=?: invalid string");

  l1 = KSI_STR_LEN (s1);
  l2 = KSI_STR_LEN (s2);
  p1 = (unsigned char*) KSI_STR_PTR (s1);
  p2 = (unsigned char*) KSI_STR_PTR (s2);

  if (l1 == l2)
    {
      if (p1 != p2)
	while (l1--)
	  if (tolower (*p1++) != tolower (*p2++))
	    return ksi_false;

      return ksi_true;
    }

  return ksi_false;
}

ksi_obj
ksi_string_eq_p (int argc, ksi_obj* args)
{
  int i;
  for (i = 1; i < argc; i++)
    {
      if (ksi_string_equal_p (args[0], args[i]) == ksi_false)
	return ksi_false;
    }

  return ksi_true;
}

ksi_obj
ksi_string_ci_eq_p (int argc, ksi_obj* args)
{
  int i;
  for (i = 1; i < argc; i++)
    {
      if (ksi_string_ci_equal_p (args[0], args[i]) == ksi_false)
	return ksi_false;
    }

  return ksi_true;
}


#ifdef HAVE_STRCOLL
static int
string_less_p (ksi_obj s1, ksi_obj s2, char* fname)
{
  int l1, l2;
  unsigned char *p1, *p2;
  char a1[2], a2[2];
  a1[1] = '\0';
  a2[1] = '\0';

#if CAUTIOUS >= 1
  if (!KSI_STR_P (s1))
    ksi_exn_error ("type", s1, "%s: invalid string", fname);
  if (!KSI_STR_P (s2))
    ksi_exn_error ("type", s2, "%s: invalid string", fname);
#endif

  l1 = KSI_STR_LEN (s1);
  l2 = KSI_STR_LEN (s2);
  p1 = (unsigned char*) KSI_STR_PTR (s1);
  p2 = (unsigned char*) KSI_STR_PTR (s2);

  for (; l1 && l2; --l1, --l2)
    {
      int c;
      a1[0] = *p1++;
      a2[0] = *p2++;
      c = strcoll (a1, a2);
      if (c > 0) return 0;
      if (c < 0) return 1;
    }

  return (l1 < l2);
}

static int
string_ci_less_p (ksi_obj s1, ksi_obj s2, char* fname)
{
  int l1, l2;
  unsigned char *p1, *p2;
  char a1[2], a2[2];
  a1[1] = '\0';
  a2[1] = '\0';

#if CAUTIOUS >= 1
  if (!KSI_STR_P (s1))
    ksi_exn_error ("type", s1, "%s: invalid string", fname);
  if (!KSI_STR_P (s2))
    ksi_exn_error ("type", s2, "%s: invalid string", fname);
#endif

  l1 = KSI_STR_LEN (s1);
  l2 = KSI_STR_LEN (s2);
  p1 = (unsigned char*) KSI_STR_PTR (s1);
  p2 = (unsigned char*) KSI_STR_PTR (s2);

  for (; l1 && l2; --l1, --l2)
    {
      int c;
      a1[0] = tolower (*p1); p1++;
      a2[0] = tolower (*p2); p2++;
      c = strcoll (a1, a2);
      if (c > 0) return 0;
      if (c < 0) return 1;
    }

  return (l1 < l2);
}

#else
static int
string_less_p (ksi_obj s1, ksi_obj s2, char* fname)
{
  int l1, l2;
  unsigned char *p1, *p2;
#if CAUTIOUS >= 1
  if (!KSI_STR_P (s1))
    ksi_exn_error ("type", s1, "%s: invalid string", fname);
  if (!KSI_STR_P (s2))
    ksi_exn_error ("type", s2, "%s: invalid string", fname);
#endif
  l1 = KSI_STR_LEN (s1);
  l2 = KSI_STR_LEN (s2);
  p1 = (unsigned char*) KSI_STR_PTR (s1);
  p2 = (unsigned char*) KSI_STR_PTR (s2);
  for (; l1 && l2; --l1, --l2)
    {
      int c = *p1++ - *p2++;
      if (c > 0) return 0;
      if (c < 0) return 1;
    }
  return (l1 < l2);
}

static int
string_ci_less_p (ksi_obj s1, ksi_obj s2, char* fname)
{
  int l1, l2;
  unsigned char *p1, *p2;
#if CAUTIOUS >= 1
  if (!KSI_STR_P (s1))
    ksi_exn_error ("type", s1, "%s: invalid string", fname);
  if (!KSI_STR_P (s2))
    ksi_exn_error ("type", s2, "%s: invalid string", fname);
#endif
  l1 = KSI_STR_LEN (s1);
  l2 = KSI_STR_LEN (s2);
  p1 = (unsigned char*) KSI_STR_PTR (s1);
  p2 = (unsigned char*) KSI_STR_PTR (s2);
  for (; l1 && l2; --l1, --l2)
    {
      int c = tolower (*p1) - tolower (*p2);
      p1++; p2++;
      if (c > 0) return 0;
      if (c < 0) return 1;
    }
  return (l1 < l2);
}
#endif

#define DEF_CMP(fun, cmp)			\
ksi_obj						\
fun (int ac, ksi_obj* av)			\
{						\
  int i;					\
  for (i = 1; i < ac; i++)			\
    {						\
      if (cmp)					\
        return ksi_false;			\
    }						\
  return ksi_true;				\
}

DEF_CMP (ksi_string_ls_p, !string_less_p (av[i-1], av[i], "string<?"))
DEF_CMP (ksi_string_gt_p, !string_less_p (av[i], av[i-1], "string>?"))
DEF_CMP (ksi_string_le_p, string_less_p (av[i], av[i-1], "string<=?"))
DEF_CMP (ksi_string_ge_p, string_less_p (av[i-1], av[i], "string>=?"))

DEF_CMP (ksi_string_ci_ls_p, !string_ci_less_p (av[i-1], av[i], "string<?"))
DEF_CMP (ksi_string_ci_gt_p, !string_ci_less_p (av[i], av[i-1], "string>?"))
DEF_CMP (ksi_string_ci_le_p, string_ci_less_p (av[i], av[i-1], "string<=?"))
DEF_CMP (ksi_string_ci_ge_p, string_ci_less_p (av[i-1], av[i], "string>=?"))


static ksi_obj
make_string (ksi_obj k, ksi_obj c)
{
  if (c == ksi_null)
    c = KSI_MK_CHAR ('\0');

  TYPE_CHECK (k, KSI_SINT_P (k), "make-string: invalid integer in arg1");
  TYPE_CHECK (c, KSI_CHAR_P (c), "make-string: invalid char in arg2");

  return ksi_make_string (KSI_SINT_COD (k), KSI_CHAR_COD (c));
}


static ksi_obj
ksi_string_index (ksi_obj str, ksi_obj chr, ksi_obj beg)
{
  const char* ptr;
  int l, b;

  if (beg == ksi_null)
    beg = ksi_zero;

  TYPE_CHECK (str, KSI_STR_P (str), "string-index: invalid string in arg1");
  TYPE_CHECK (chr, KSI_CHAR_P (chr), "string-index: invalid char in arg2");
  TYPE_CHECK (beg, KSI_SINT_P (beg) && KSI_SINT_COD (beg) >= 0,
	      "string-index: invalid integer in arg3");

  ptr = KSI_STR_PTR (str);
  l = KSI_STR_LEN (str);
  b = KSI_SINT_COD (beg);
  if (b < l)
    {
      char* where = (char*) memchr (ptr+b, KSI_CHAR_COD (chr), l-b);
      if (where)
	return KSI_MK_SINT (where - ptr);
    }

  return ksi_false;
}

static ksi_obj
ksi_string_rindex (ksi_obj str, ksi_obj chr, ksi_obj beg)
{
  int len, pos;

  TYPE_CHECK (str, KSI_STR_P (str), "string-rindex: invalid string in arg1");
  TYPE_CHECK (chr, KSI_CHAR_P (chr), "string-rindex: invalid char in arg2");

  len = KSI_STR_LEN (str);
  if (beg == ksi_null)
    pos = len;
  else
    {
      TYPE_CHECK (beg, KSI_SINT_P (beg) && KSI_SINT_COD (beg) >= 0,
		  "string-rindex: invalid integer in arg3");
      pos = MIN (len, KSI_SINT_COD (beg));
    }

  while (--pos >= 0)
    if (KSI_STR_PTR (str) [pos] == KSI_CHAR_COD (chr))
      return KSI_MK_SINT (pos);

  return ksi_false;
}

ksi_obj
ksi_string_upcase_x (ksi_obj s)
{
  int i;

  TYPE_CHECK (s, KSI_STR_P (s), "string-upcase!: invalid string in arg1");
  TYPE_CHECK (s, !KSI_C_STR_P (s), "string-upcase!: constant string in arg1");

#ifdef MSWIN32
  if (!ksi_win_console ())
    CharUpper (KSI_STR_PTR (s));
  else
    {
      OemToCharBuff (KSI_STR_PTR (s), KSI_STR_PTR (s), KSI_STR_LEN (s));
      CharUpper (KSI_STR_PTR (s));
      CharToOemBuff (KSI_STR_PTR (s), KSI_STR_PTR (s), KSI_STR_LEN (s));
    }
#else
  for (i = 0; i < KSI_STR_LEN (s); i++)
    KSI_STR_PTR (s) [i] = toupper ((unsigned char) KSI_STR_PTR (s) [i]);
#endif
  return s;
}

ksi_obj
ksi_string_downcase_x (ksi_obj s)
{
  int i;

  TYPE_CHECK (s, KSI_STR_P (s), "string-downcase!: invalid string in arg1");
  TYPE_CHECK (s, !KSI_C_STR_P (s), "string-downcase!: constant string in arg1");

#ifdef MSWIN32
  if (!ksi_win_console ())
    CharLower (KSI_STR_PTR (s));
  else
    {
      OemToCharBuff (KSI_STR_PTR (s), KSI_STR_PTR (s), KSI_STR_LEN (s));
      CharLower (KSI_STR_PTR (s));
      CharToOemBuff (KSI_STR_PTR (s), KSI_STR_PTR (s), KSI_STR_LEN (s));
    }
#else
  for (i = 0; i < KSI_STR_LEN (s); i++)
    KSI_STR_PTR (s) [i] = tolower ((unsigned char) KSI_STR_PTR (s) [i]);
#endif
  return s;
}

ksi_obj
ksi_string_capitalize_x (ksi_obj str)
{
  char* ptr;
  int i, len, fl;

  TYPE_CHECK (str, KSI_STR_P (str), "string-capitalize!: invalid string in arg1");
  TYPE_CHECK (str, !KSI_C_STR_P (str), "string-capitalize!: constant string in arg1");

  ptr = KSI_STR_PTR (str);
  len = KSI_STR_LEN (str);
  fl = 0;
  for (i = 0; i < len; i++)
    {
      if (isalpha ((unsigned char) ptr [i]))
	{
	  if (fl)
	    ptr [i] = tolower ((unsigned char) ptr [i]);
	  else
	    {
	      ptr [i] = toupper ((unsigned char) ptr [i]);
	      fl = 1;
	    }
	}
      else
	fl = 0;
    }

  return str;
}


static struct Ksi_Prim_Def defs[] =
{
  { "string?",		ksi_string_p,		KSI_CALL_ARG1, 1 },

  { "make-string",	make_string,		KSI_CALL_ARG2, 1 },
  { "string-length",	ksi_string_length,	KSI_CALL_ARG1, 1 },
  { "string-ref",	ksi_string_ref,		KSI_CALL_ARG2, 2 },
  { "string-set!",	ksi_string_set_x,	KSI_CALL_ARG3, 3 },
  { "substring",	ksi_substring,		KSI_CALL_ARG3, 2 },
  { "string->list",	ksi_string2list,	KSI_CALL_ARG1, 1 },
  { "list->string",	ksi_list2string,	KSI_CALL_ARG1, 1 },
  { "string-copy",	ksi_string_copy,	KSI_CALL_ARG1, 1 },
  { "string-fill!",	ksi_string_fill_x,	KSI_CALL_ARG2, 2 },
  { "string",		ksi_new_string,		KSI_CALL_REST0, 0 },
  { "string-append",	ksi_string_append,	KSI_CALL_REST0, 0 },

  { "string=?",		ksi_string_eq_p,	KSI_CALL_REST0, 0 },
  { "string-ci=?",	ksi_string_ci_eq_p,	KSI_CALL_REST0, 0 },
  { "string<?",		ksi_string_ls_p,	KSI_CALL_REST0, 0 },
  { "string>?",		ksi_string_gt_p,	KSI_CALL_REST0, 0 },
  { "string<=?",	ksi_string_le_p,	KSI_CALL_REST0, 0 },
  { "string>=?",	ksi_string_ge_p,	KSI_CALL_REST0, 0 },
  { "string-ci<?",	ksi_string_ci_ls_p,	KSI_CALL_REST0, 0 },
  { "string-ci>?",	ksi_string_ci_gt_p,	KSI_CALL_REST0, 0 },
  { "string-ci<=?",	ksi_string_ci_le_p,	KSI_CALL_REST0, 0 },
  { "string-ci>=?",	ksi_string_ci_ge_p,	KSI_CALL_REST0, 0 },

  // non R5RS string utils
  { "string-index",		ksi_string_index,	KSI_CALL_ARG3, 2 },
  { "string-rindex",		ksi_string_rindex,	KSI_CALL_ARG3, 2 },
  { "string-upcase!",		ksi_string_upcase_x,	KSI_CALL_ARG1, 1 },
  { "string-downcase!",		ksi_string_downcase_x,	KSI_CALL_ARG1, 1 },
  { "string-capitalize!",	ksi_string_capitalize_x,KSI_CALL_ARG1, 1 },

  { 0 }
};

void
ksi_init_strings ()
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}


 /* End of code */
