/*
 * ksi_symb.c
 * symbols
 *
 * Copyright (C) 1997-2000, Ivan Demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Fri Jan 24 19:08:53 1997
 * Last Update:   Sat Jul 29 18:24:58 2000
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_vtab.h"

static char rcsid[] = "$Id: ksi_symb.c,v 1.2 2000/08/14 18:07:58 demakov Exp $";


/*
 * ksi_intern
 *	    NAME   LEN   
 *	.    ,  
 *	.
 *
 * Return
 *	   
 *
 * Side effects
 *	      ,   ,
 *	       .
 */

ksi_obj
ksi_intern (const char* name, int len)
{
  ksi_symbol sym;

  BASE_CHECK_INIT;
  SYMTAB_CHECK_INIT;

  sym = ksi_lookup_sym (KSI->obarray, name, len);
  if (sym == 0)
    {
      sym = (ksi_symbol) ksi_malloc_data (len + 1 + sizeof *sym);
      sym -> itag = KSI_TAG_SYMBOL;
      sym -> len = len;
      sym -> ptr = (char*) sym + sizeof *sym;
      if (len)
	memcpy ((char*) sym -> ptr, name, len);
      ((char*) sym -> ptr) [len] = 0;

      ksi_append_sym (KSI->obarray, sym);
    }

  return (ksi_obj) sym;
}


/*
 * ksi_ptr02sym
 *	    NAME   
 *	.    ,  
 *	.
 *
 * Return
 *	   
 *
 * Side effects
 *	      ,   ,
 *	     .
 */

ksi_obj
ksi_ptr02sym (const char* name)
{
  int len;
  ksi_symbol sym;

  BASE_CHECK_INIT;
  SYMTAB_CHECK_INIT;

  len = strlen (name);
  sym = ksi_lookup_sym (KSI->obarray, name, len);
  if (sym == 0)
    {
      sym = NEW_TYPED (struct Ksi_Symbol, ptr, ptr);
      sym -> itag = KSI_TAG_SYMBOL;
      sym -> len = len;
      sym -> ptr = name;

      ksi_append_sym (KSI->obarray, sym);
    }

  return (ksi_obj) sym;
}


/*
 * ksi_find_symbol
 *	    NAME   LEN   
 *	.
 *
 * Return
 *	 ,  ksi_null    .
 *
 * Side effects
 *	None
 */

ksi_obj
ksi_find_symbol (const char* name, int len)
{
  BASE_CHECK_INIT;
  SYMTAB_CHECK_INIT;

  return (ksi_obj) ksi_lookup_sym (KSI->obarray, name, len);
}


ksi_obj
ksi_symbol_p (ksi_obj x)
{
  return KSI_SYM_P (x) ? ksi_true : ksi_false;
}

const char*
ksi_symbol2str (ksi_obj sym)
{
  int i, k, num, len;
  const char* nam;
  char* ptr;

  TYPE_CHECK (sym, KSI_SYM_P (sym), "ksi_symbol2str: invalid symbol");

  len = KSI_SYM_LEN (sym);
  if (len <= 0)
    return "#||";

  nam = KSI_SYM_PTR (sym);
  for (i = 0, k = 0, num = 0; i < len; ++i)
    {
      switch (nam[i])
	{
	case '\0':
	  ++k; num+=3;
	  break;
	case '\n': case '\r': case '\t': case '\f': case '\b': case '\a':
	  ++k; ++num;
	  break;

	case ' ': case '(': case ')': case '[': case ']': case '{': case '}':
	case ';': case '`': case ',': case '\'': case '"':
	  ++k;
	  break;

	case '#':
	case '1': case '2': case '3': case '4': case '5':
	case '6': case '7': case '8': case '9': case '0':
	  if (i == 0)
	    ++k;
	  break;

	case '|': case '\\':
	  ++num;
	  break;

	default:
	  if (isupper (nam[i]))
	    ++k;
	}
    }
  if (k == 0 && nam [len] == '\0')
    return nam;

  ptr = (char*) ksi_malloc_data (len + num + 4);
  ptr[k=0] = '#';
  ptr[++k] = '|';
  for (i = 0; i < len; ++i)
    {
      char c = nam[i];
      switch (c)
	{
	case '\0':
	  c = '0';
	  ptr[++k] = '\\';
	  if ('0' <= nam[i+1] && nam[i+1] <= '7')
	    {
	      ptr[++k] = '0';
	      ptr[++k] = '0';
	    }
	  break;
	case '\n':
	  c = 'n';
	  ptr[++k] = '\\';
	  break;
	case '\r':
	  c = 'r';
	  ptr[++k] = '\\';
	  break;
	case '\t':
	  c = 't';
	  ptr[++k] = '\\';
	  break;
	case '\f':
	  c = 'f';
	  ptr[++k] = '\\';
	  break;
	case '\b':
	  c = 'b';
	  ptr[++k] = '\\';
	  break;
	case '\a':
	  c = 'a';
	  ptr[++k] = '\\';
	  break;
	case '|': case '\\':
	  ptr[++k] = '\\';
	  break;
	}
      ptr[++k] = c;
    }
  ptr[++k] = '|';
  ptr[++k] = '\0';

  return ptr;
}


ksi_obj
ksi_symbol2string (ksi_obj sym)
{
  TYPE_CHECK (sym, KSI_SYM_P (sym), "symbol->string: invalid symbol");

  return ksi_ptr02string (KSI_SYM_PTR (sym));
}

ksi_obj
ksi_string2symbol (ksi_obj str)
{
  TYPE_CHECK (str, KSI_STR_P (str), "string->symbol: invalid string");

  if (KSI_C_STR_P (str))
    return ksi_ptr02sym (KSI_STR_PTR (str));
  return ksi_intern (KSI_STR_PTR (str), KSI_STR_LEN (str));
}


ksi_obj
ksi_gensym (ksi_obj name, ksi_obj unused)
{
  const char *str = 0;
  char *buf = 0;
  int len = 0, sz = 0;

  if (KSI->gensym_num == 0)
    KSI->gensym_num = ksi_one;

  if (name && ksi_spec_val ("*gensym-with-prefix*", ksi_false, 0) != ksi_false)
    {
      if (KSI_STR_P (name))
	{
	  str = KSI_STR_PTR (name);
	  len = KSI_STR_LEN (name);
	}
      else if (KSI_SYM_P (name))
	{
	  str = KSI_SYM_PTR (name);
	  len = KSI_SYM_LEN (name);
	}
      else
	KSI_EXN_ERR ("type", name, "gensym: invalid object in arg1");
    }

  for (;;)
    {
      char* ptr = ksi_num2str (KSI->gensym_num, 10);
      int l = strlen (ptr) + 1;
      if (buf == 0 || sz <= len+l)
	buf = (char*) ksi_malloc_data ((sz = len+l) + 1);

      if (len)
	memcpy (buf, str, len);
      buf [len] = '_';
      memcpy (buf+len+1, ptr, l);
      KSI->gensym_num = ksi_add (KSI->gensym_num, ksi_one);

      if (!ksi_find_symbol (buf, len+l))
	return ksi_ptr02sym (buf);
    }
}


static struct Ksi_Prim_Def defs[] =
{
  { "symbol?",		ksi_symbol_p,		KSI_CALL_ARG1, 1 },
  { "symbol->string",	ksi_symbol2string,	KSI_CALL_ARG1, 1 },
  { "string->symbol",	ksi_string2symbol,	KSI_CALL_ARG1, 1 },
  { "gensym",		ksi_gensym,		KSI_CALL_ARG2, 0 },

  { 0 }
};

void
ksi_init_symbols ()
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}


 /* End of code */
