/*
 * ksi_vect.c
 * vectors
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Fri Apr 25 03:59:34 1997
 * Last Update:   Sat Nov 10 00:09:54 2001
 */

/* Comments:
 *
 */

#include "ksi_int.h"

static char rcsid[] = "$Id: ksi_vect.c,v 1.4 2001/11/25 14:30:01 demakov Exp $";


ksi_obj
ksi_alloc_vector (int dim, int tag)
{
  ksi_vector vec;
  vec = NEW_TYPED_ARR (struct Ksi_Vector, dim, arr, arr);
  vec -> itag = tag;
  vec -> dim = dim;
  return (ksi_obj) vec;
}

ksi_obj
ksi_copy_vector (ksi_obj x)
{
  ksi_obj vec;
  int i;

  TYPE_CHECK (x, KSI_VEC_P (x), "copy-vector: invalid vector");

  vec = ksi_alloc_vector (KSI_VEC_LEN (x), KSI_TAG_VECTOR);
  for (i = 0; i < KSI_VEC_LEN (x); i++)
    KSI_VEC_REF (vec, i) = KSI_VEC_REF (x, i);

  return vec;
}

ksi_obj
ksi_vector_p (ksi_obj x)
{
  return KSI_VEC_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_new_vector (int argc, ksi_obj* argv)
{
  int i;
  ksi_obj vec = ksi_alloc_vector (argc, KSI_TAG_VECTOR);

  for (i = 0; i < argc; ++i)
    KSI_VEC_REF (vec, i) = argv [i];

  return vec;
}

ksi_obj
ksi_list2vector (ksi_obj x)
{
  ksi_obj vec;
  int i, len = KSI_LIST_LEN (x);
  TYPE_CHECK (x, len >= 0, "list->vector: invalid list");

  vec = ksi_alloc_vector (len, KSI_TAG_VECTOR);
  for (i = 0; i < len; ++i, x = CDR (x))
    KSI_VEC_REF (vec, i) = CAR (x);

  return vec;
}

ksi_obj
ksi_vector2list (ksi_obj v)
{
  ksi_obj list = ksi_nil;
  int len;

  TYPE_CHECK (v, KSI_VEC_P (v), "vector->list: invalid vector");

  len = KSI_VEC_LEN (v);
  while (len--)
    list = ksi_cons (KSI_VEC_REF (v, len), list);

  return list;
}

ksi_obj
ksi_make_vector (ksi_obj k, ksi_obj fill)
{
  ksi_obj vec;
  int i;

  TYPE_CHECK (k, KSI_SINT_P (k) && KSI_SINT_COD (k) >= 0,
	      "make-vector: invalid integer in arg1");

  if (fill == ksi_null)
    fill = ksi_void;

  vec = ksi_alloc_vector (KSI_SINT_COD (k), KSI_TAG_VECTOR);
  for (i = 0; i < KSI_SINT_COD (k); ++i)
    KSI_VEC_REF (vec, i) = fill;

  return vec;
}

ksi_obj
ksi_vector_length (ksi_obj v)
{
  TYPE_CHECK (v, KSI_VEC_P (v), "vector-length: invalid vector");
  return KSI_MK_SINT (KSI_VEC_LEN (v));
}

ksi_obj
ksi_vector_ref (ksi_obj vec, ksi_obj k)
{
  int idx = 0;

  if (KSI_SINT_P (k))
    idx = KSI_SINT_COD (k);
  else if (KSI_CHAR_P (k))
    idx = KSI_CHAR_COD (k);
  else
    KSI_EXN_ERR ("type", k, "vector-ref: invalid index in arg2");

  TYPE_CHECK (vec, KSI_VEC_P (vec), "vector-ref: invalid vector in arg1");
  RANGE_CHECK (k, 0 <= idx && idx < KSI_VEC_LEN (vec),
	       "vector-ref: index out of range in arg2");

  return KSI_VEC_REF (vec, idx);
}

ksi_obj
ksi_vector_set_x (ksi_obj vec, ksi_obj k, ksi_obj val)
{
  int idx = 0;

  if (KSI_SINT_P (k))
    idx = KSI_SINT_COD (k);
  else if (KSI_CHAR_P (k))
    idx = KSI_CHAR_COD (k);
  else
    KSI_EXN_ERR ("type", k, "vector-set!: invalid index in arg2");

  TYPE_CHECK (vec, KSI_M_VEC_P (vec),
	      "vector-set!: invalid or constant vector in arg1");
  RANGE_CHECK (k, 0 <= idx && idx < KSI_VEC_LEN (vec),
	       "vector-set!: index out of range in arg2");

  KSI_VEC_REF (vec, idx) = val;
  return ksi_void;
}

ksi_obj
ksi_vector_fill_x (ksi_obj vec, ksi_obj fill)
{
  int len;

  TYPE_CHECK (vec, KSI_M_VEC_P (vec),
	      "vector-fill!: invalid or constant vector in arg1");

  len = KSI_VEC_LEN (vec);
  while (--len >= 0)
    KSI_VEC_REF (vec, len) = fill;

  return ksi_void;
}

ksi_obj
ksi_vector_map (ksi_obj proc, ksi_obj vec)
{
  int len;
  ksi_obj res;

  TYPE_CHECK (proc, PROC_P (proc), "vector-map: invalid procedure in arg1");
  TYPE_CHECK (vec, KSI_VEC_P (vec), "vector-map: invalid vector in arg2");

  len = KSI_VEC_LEN (vec);
  res = ksi_alloc_vector (len, KSI_TAG_VECTOR);
  while (--len >= 0)
    KSI_VEC_REF (res, len) = ksi_apply_1 (proc, KSI_VEC_REF (vec, len));

  return res;
}

ksi_obj
ksi_vector_for_each (ksi_obj proc, ksi_obj vec)
{
  int i;

  TYPE_CHECK (proc, PROC_P (proc), "vector-for-each: invalid procedure in arg1");
  TYPE_CHECK (vec, KSI_VEC_P (vec), "vector-for-each: invalid vector in arg2");

  for (i = 0; i < KSI_VEC_LEN (vec); i++)
    ksi_apply_1 (proc, KSI_VEC_REF (vec, i));

  return ksi_void;
}


static struct Ksi_Prim_Def defs[] =
{
  { "vector",		ksi_new_vector,		KSI_CALL_REST0, 0 },
  { "list->vector",	ksi_list2vector,	KSI_CALL_ARG1, 1 },
  { "vector->list",	ksi_vector2list,	KSI_CALL_ARG1, 1 },
  { "make-vector",	ksi_make_vector,	KSI_CALL_ARG2, 1 },
  { "vector-length",	ksi_vector_length,	KSI_CALL_ARG1, 1 },
  { "vector-ref",	ksi_vector_ref,		KSI_CALL_ARG2, 2 },
  { "vector-set!",	ksi_vector_set_x,	KSI_CALL_ARG3, 3 },
  { "vector-fill!",	ksi_vector_fill_x,	KSI_CALL_ARG2, 2 },
  { "vector-map",	ksi_vector_map,		KSI_CALL_ARG2, 2 },
  { "vector-for-each",	ksi_vector_for_each,	KSI_CALL_ARG2, 2 },

  { 0 }
};

void
ksi_init_vectors ()
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);

  ksi_defsym ("vector?", ksi_new_imm (KSI_TAG_VECTORP), 0);
}


 /* End of code */
