/*
 * ksi_writ.c
 * write
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Fri Aug  1 04:08:03 1997
 * Last Update:   Sun Nov 18 01:16:06 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_vtab.h"
#include "ksi_comp.h"
#include "ksi_mod.h"
#include "ksi_klos.h"


static char rcsid[] = "$Id: ksi_writ.c,v 1.6 2001/11/25 14:30:01 demakov Exp $";



static const char*
ksi_imm2str (ksi_obj o)
{
  int cod = KSI_TC8_TAG (o);
  switch (cod)
    {
    case KSI_TC8_CHAR:	return ksi_char2str (o);
    case KSI_TC8_BOOL:	return KSI_BOOL_COD (o) ? "#t" : "#f";
    case KSI_TC8_ERR:	return "#error";
    case KSI_TC8_VOID:	return "#void";
    case KSI_TC8_EOF:	return "#eof";
    case KSI_TC8_NIL:	return "()";

    case KSI_TC8_EXIT:
      cod = KSI_EXIT_COD (o);
      if (cod == KSI_NORMAL_EXIT_CODE)
	return "#exit-success";
      if (cod == KSI_ERROR_EXIT_CODE)
	return "#exit-fail";
      return ksi_aprintf ("#exit-%d", cod);
    }

  return ksi_aprintf ("#imm-%d", cod);
}

static char*
expand_str (char* ptr, int* size, int len)
{
  if (*size <= len)
    {
      *size = (len + 32 + 31) & ~31;
      ptr = (char*) ksi_realloc (ptr, *size);
    }
  return ptr;
}

static char*
one_pair (char* ptr, int* size, int* len, ksi_obj* x,
	  const char* (*print) (ksi_obj))

{
  const char* p = print (KSI_CAR (*x));
  int l = strlen (p);
  ptr = expand_str (ptr, size, *len+l);
  memcpy (ptr + *len, p, l);
  *len += l;

  *x = KSI_CDR (*x);
  if (*x == ksi_nil)
    return ptr;

  if (!KSI_PAIR_P (*x))
    {
      p = print (*x);
      l = strlen (p);
      ptr = expand_str (ptr, size, *len + l + 3);
      ptr[*len] = ' '; *len += 1;
      ptr[*len] = '.'; *len += 1;
      ptr[*len] = ' '; *len += 1;
      memcpy (ptr + *len, p, l);
      *len += l;
      return ptr;
    }

  ptr = expand_str (ptr, size, *len+1);
  ptr[*len] = ' '; *len += 1;
  return ptr;
}

static const char*
ksi_print_pair (char* prefix, ksi_obj x, char* suffix,
		const char* (*print) (ksi_obj))
{
  int size = strlen (prefix) + 32, len = 0;
  char* ptr = (char*) ksi_malloc_data (size);
  ksi_obj z = x;

  if (PAIR_P (CDR (x)) && CDR (CDR (x)) == ksi_nil)
    {
      if (CAR (x) == KSI->sym_quote)
	  {
	    ptr [len++] = '\'';
	    x = CDR (x);
	    ptr = one_pair (ptr, &size, &len, &x, print);
	    ptr = expand_str (ptr, &size, len+1);
	    ptr [len] = '\0';
	    return ptr;
	  }
      if (CAR (x) == KSI->sym_quasiquote)
	  {
	    ptr [len++] = '`';
	    x = CDR (x);
	    ptr = one_pair (ptr, &size, &len, &x, print);
	    ptr = expand_str (ptr, &size, len+1);
	    ptr [len] = '\0';
	    return ptr;
	  }
      if (CAR (x) == KSI->sym_unquote)
	  {
	    ptr [len++] = ',';
	    x = CDR (x);
	    ptr = one_pair (ptr, &size, &len, &x, print);
	    ptr = expand_str (ptr, &size, len+1);
	    ptr [len] = '\0';
	    return ptr;
	  }
      if (CAR (x) == KSI->sym_unquote_splicing)
	  {
	    ptr [len++] = ',';
	    ptr [len++] = '@';
	    x = CDR (x);
	    ptr = one_pair (ptr, &size, &len, &x, print);
	    ptr = expand_str (ptr, &size, len+1);
	    ptr [len] = '\0';
	    return ptr;
	  }
    }

  strcpy (ptr, prefix);
  len = strlen (ptr);

  for (;;)
    {
      ptr = one_pair (ptr, &size, &len, &x, print);
      if (!KSI_PAIR_P (x))
	break;
      ptr = one_pair (ptr, &size, &len, &x, print);
      if (!KSI_PAIR_P (x))
	break;

      z = KSI_CDR (z);
      if (z == x)
	{
	  ptr = expand_str (ptr, &size, len+5);
	  ptr[len++] = '.';
	  ptr[len++] = ' ';
	  ptr[len++] = '.';
	  ptr[len++] = ' ';
	  ptr[len++] = '.';
	  break;
	}
    }

  ptr = expand_str (ptr, &size, len + strlen (suffix) + 1);
  strcpy (ptr + len, suffix);

  return ptr;
}

static const char*
ksi_print_vec (char* beg, ksi_obj x, char* end, const char* (*print) (ksi_obj))
{
  int n;
  int bl = strlen (beg);
  int el = strlen (end);
  int size = bl + el + 32, len = bl;
  char* ptr = (char*) ksi_malloc_data (size);

  memcpy (ptr, beg, bl);
  if (KSI_VEC_LEN (x) > 0)
    for (n = 0;;)
      {
	const char* p = print (KSI_VEC_REF (x, n));
	int l = strlen (p);
	ptr = expand_str (ptr, &size, len+l);
	memcpy (ptr + len, p, l);
	len += l;

	if (++n >= KSI_VEC_LEN (x))
	  break;

	ptr = expand_str (ptr, &size, len+1);
	ptr[len++] = ' ';
      }

  ptr = expand_str (ptr, &size, len+el+1);
  strcpy (ptr+len, end);

  return ptr;
}

static const char*
ksi_print_values (ksi_obj x)
{
  if (VALUES_VALS (x) == ksi_nil)
    return "#<values ()>";
  else
    return ksi_aprintf ("#<values %w>", VALUES_VALS (x));
}

static const char*
ksi_print_port (ksi_obj o)
{
  return ksi_aprintf ("#<%s%s%s%s>",
		      ((ksi_port) o) -> closed ? "closed-" : "",
		      ((ksi_port) o) -> input  ? "input-"  : "",
		      ((ksi_port) o) -> output ? "output-" : "",
		      ((ksi_port) o) -> ops -> name ((ksi_port) o));
}

static const char*
ksi_print_prim (ksi_obj x)
{
  if (KSI_PRIM_NAME (x))
    return ksi_aprintf ("#<primitive %s >", KSI_PRIM_NAME (x));

  return ksi_aprintf ("#<primitive %p>", x);
}

static const char*
ksi_print_promise (ksi_obj x)
{
  ksi_obj z = PROMISE_VAL (x) ? PROMISE_VAL (x) : PROMISE_BODY (x);
  return ksi_aprintf ("#<promise %w>", z);
}

static const char*
ksi_print_inst (ksi_obj x)
{
  ksi_obj str = ksi_make_string (0, 0);
  ksi_port p = ksi_new_str_port (str);
  p->output = 1;
  ksi_write_inst (x, p, 1);
  return KSI_STR_PTR (str);
}

static const char*
ksi_print_ptr (ksi_obj x)
{
#define BEG "#<pointer"
#define END ">"

  if (ksi_spec_val ("*print-pointer-body*", 0, 0) != ksi_false)
    {
      int i, size = KSI_CPTR_SZ (x) * 3 + sizeof BEG + sizeof END;
      char *ptr = (char*) ksi_malloc_data (size), *loc = ptr + sizeof BEG - 1;

      memcpy (ptr, BEG, sizeof BEG - 1);
      for (i = 0; i < KSI_CPTR_SZ (x); i++)
	{
	  static char digs[] = { '0', '1', '2', '3', '4', '5', '6', '7',
				 '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };

	  unsigned d = ((unsigned char*) KSI_CPTR_VAL (x)) [i];
	  *loc++ = ' ';
	  *loc++ = digs[(d >> 4) & 0xf];
	  *loc++ = digs[(d & 0xf)];
	}
      memcpy (loc, END, sizeof END);

      return ptr;
    }

  return ksi_aprintf ("%s %p[%d]>", BEG, KSI_CPTR_VAL (x), KSI_CPTR_SZ (x));

#undef BEG
#undef END
}

static const char*
ksi_print_env (ksi_obj x)
{
  return ksi_aprintf ("#<environ %p>", x);
}


static const char*
ksi_print_hashtab (ksi_obj x)
{
  return ksi_aprintf ("#<hash-table %p>", x);
}


static const char*
ksi_print_code (char* beg, ksi_obj x, char* end)
{
  int bl = strlen (beg);
  int el = strlen (end);
  int size = bl + el + 32, len = bl;
  char* ptr = (char*) ksi_malloc_data (size);
  int n;

  memcpy (ptr, beg, bl);
  for (n = 0;;)
    {
      const char* p = ksi_obj2str (CODE_VAL (x, n));
      int l = strlen (p);
      ptr = expand_str (ptr, &size, len+l);
      memcpy (ptr + len, p, l);
      len += l;

      if (++n >= CODE_NUM (x) + 1)
	break;

      ptr = expand_str (ptr, &size, len+1);
      ptr[len++] = ' ';
    }

  ptr = expand_str (ptr, &size, len+el+1);
  strcpy (ptr+len, end);

  return ptr;
}

static const char*
ksi_print_closure (ksi_obj x)
{
  if (ksi_spec_val ("*print-closure-body*", 0, 0) != ksi_false)
    return ksi_aprintf ("#<closure %w>", CLOS_BODY (x));
  if (SYM_P (CLOS_DOC (x)))
    return ksi_aprintf ("#<closure %w>", CLOS_DOC (x));
  if (PAIR_P (CLOS_DOC (x)))
    return ksi_aprintf ("#<closure %w>", CAR (CLOS_DOC (x)));

  return ksi_aprintf ("#<closure %p>", x);
}

static const char*
ksi_print_module (ksi_obj x)
{
  return ksi_aprintf ("#<module %w>", MOD_NAME (x));
}

static const char*
ksi_print_obj (char* prefix, ksi_obj x, char* suffix)
{
  return ksi_aprintf ("%s%w%s", prefix, x, suffix);
}

static const char*
ksi_print_varbox (const char* p, ksi_obj x)
{
  return ksi_aprintf ("#{%s %d %d}", p, VAR_LEV (x), VAR_NUM (x));
}

static const char*
ksi_print_event (ksi_obj x)
{
  const char* p = ((ksi_event) x) -> ops -> name ((ksi_event) x);
  return ksi_aprintf ("#<event %s>", p);
}

const char*
ksi_obj2str (ksi_obj o)
{
  if (o == ksi_null)
    return "#<null>";

  switch (KSI_COD (o) & 3)
    {
    case KSI_TC2_INT:
      return ksi_num2str (o, 10);

    case KSI_TC2_IMM:
      return ksi_imm2str (o);

    case KSI_TC2_PTR:
      switch (o->itag)
	{
	case KSI_TAG_PRIM:
	case KSI_TAG_PRIM_0:
	case KSI_TAG_PRIM_1:
	case KSI_TAG_PRIM_2:
	case KSI_TAG_PRIM_r:
	  return ksi_print_prim (o);

	case KSI_TAG_CONS:
	  return "#cons";

	case KSI_TAG_CAR:
	  return "#car";

	case KSI_TAG_CDR:
	  return "#cdr";

	case KSI_TAG_NOT:
	  return "#not";

	case KSI_TAG_EQP:
	  return "#eq?";

	case KSI_TAG_EQVP:
	  return "#eqv?";

	case KSI_TAG_EQUALP:
	  return "#equal?";

	case KSI_TAG_MEMQ:
	  return "#memq";

	case KSI_TAG_MEMV:
	  return "#memv";

	case KSI_TAG_MEMBER:
	  return "#member";

	case KSI_TAG_NULLP:
	  return "#null?";

	case KSI_TAG_PAIRP:
	  return "#pair?";

	case KSI_TAG_LISTP:
	  return "#list?";

	case KSI_TAG_LIST:
	  return "#list";

	case KSI_TAG_APPLY:
	  return "#apply";

	case KSI_TAG_CALL_CC:
	  return "#call/cc";

	case KSI_TAG_CALL_WITH_VALUES:
	  return "#call/vs";

	case KSI_TAG_VECTORP:
	  return "#vector?";

	case KSI_TAG_CLOSURE:
	  return ksi_print_closure (o);

	case KSI_TAG_PRIM_CLOSURE:
	  return "#<prim-closure>";

	case KSI_TAG_QUOTE:
	  return ksi_print_code ("#{quote ", o, "}");

	case KSI_TAG_AND:
	  return ksi_print_code ("#{and ", o, "}");

	case KSI_TAG_OR:
	  return ksi_print_code ("#{or ", o, "}");

	case KSI_TAG_BEGIN:
	  return ksi_print_code ("#{begin ", o, "}");

	case KSI_TAG_IF2:
	case KSI_TAG_IF3:
	  return ksi_print_code ("#{if ", o, "}");

	case KSI_TAG_DEFINE:
	  return ksi_print_code ("#{define ", o, "}");

	case KSI_TAG_SET:
	  return ksi_print_code ("#{set! ", o, "}");

	case KSI_TAG_DELAY:
	  return ksi_print_code ("#{delay ", o, "}");

	case KSI_TAG_CALL:
	  return ksi_print_code ("#{call ", o, "}");

	case KSI_TAG_FRAME:
	  return ksi_print_code ("#{frame ", o, "}");

	case KSI_TAG_LAMBDA:
	  return ksi_print_code ("#{lambda ", o, "}");

	case KSI_TAG_INSTANCE:
	  return ksi_print_inst (o);

	case KSI_TAG_NEXT_METHOD:
	  return "#next-method";

	case KSI_TAG_VAR0:
	case KSI_TAG_VAR1:
	case KSI_TAG_VAR2:
	case KSI_TAG_VARN:
	  return ksi_print_varbox ("var", o);

	case KSI_TAG_LOCAL:
	  return ksi_print_obj ("#{local ", VAR_SYM (o), "}");

	case KSI_TAG_IMPORTED:
	  return ksi_print_obj ("#{imported ", VAR_SYM (o), "}");

	case KSI_TAG_GLOBAL:
	  return ksi_print_obj ("#{global ", VAR_SYM (o), "}");

	case KSI_TAG_FREEVAR:
	  return ksi_print_obj ("#{freevar ", VAR_SYM (o), "}");

	case KSI_TAG_SYMBOL:
	  return ksi_symbol2str (o);

	case KSI_TAG_STRING:
	case KSI_TAG_CONST_STRING:
	  return ksi_string2str (o);

	case KSI_TAG_BIGNUM:
	case KSI_TAG_FLONUM:
	case KSI_TAG_RATIONAL:
	case KSI_TAG_COMPLEX:
	  return ksi_num2str (o, 10);

	case KSI_TAG_KEYWORD:
	  return ksi_key2str (o);

	case KSI_TAG_POINTER:
	case KSI_TAG_CONST_POINTER:
	  return ksi_print_ptr (o);

	case KSI_TAG_PAIR:
	case KSI_TAG_CONST_PAIR:
	  return ksi_print_pair ("(", o, ")", ksi_obj2str);

	case KSI_TAG_VECTOR:
	case KSI_TAG_CONST_VECTOR:
	  return ksi_print_vec ("#(", o, ")", ksi_obj2str);

	case KSI_TAG_ANNOTATION:
	  return ksi_print_vec ("#<annotation ", o, ">", ksi_obj2str);

	case KSI_TAG_PORT:
	  return ksi_print_port (o);

	case KSI_TAG_PROMISE:
	  return ksi_print_promise (o);

	case KSI_TAG_VALUES:
	  return ksi_print_values (o);

	case KSI_TAG_ENVIRON:
	  return ksi_print_env (o);

	case KSI_TAG_HASHTAB:
	  return ksi_print_hashtab (o);

	case KSI_TAG_MODULE:
	  return ksi_print_module (o);

	case KSI_TAG_EXN:
	  return ksi_print_vec ("#<exn ", o, ">", ksi_obj2str);

	case KSI_TAG_EVENT:
	  return ksi_print_event (o);

	case KSI_TAG_CORE:
	  return ksi_print_vec ("#<core-", o, ">", ksi_obj2str);

	case KSI_TAG_EXTENDED:
	case KSI_TAG_BROKEN:
	  return KSI_EXT_TAG (o) -> print (KSI_EXT_TAG (o), o, 1);
	}

      ksi_error ("object->string: invalid object tag: %d(0x%x)",
		 o->itag, o->itag);
      break;
    }

  return "#impossible";
}

const char*
ksi_obj2name (ksi_obj x)
{
  if (x == ksi_null)
    return "#<null>";

  if (KSI_OBJ_P (x))
    {
      switch (x->itag)
	{
	case KSI_TAG_SYMBOL:
	  return KSI_SYM_PTR (x);

	case KSI_TAG_STRING:
	case KSI_TAG_CONST_STRING:
	  return ksi_safe_string_ptr (x);

	case KSI_TAG_PAIR:
	case KSI_TAG_CONST_PAIR:
	  return ksi_print_pair ("", x, "", ksi_obj2name);

	case KSI_TAG_VECTOR:
	case KSI_TAG_CONST_VECTOR:
	  return ksi_print_vec ("", x, "", ksi_obj2name);
	}
    }

  return ksi_obj2str (x);
}


void
ksi_init_write (void)
{
  ksi_ignore (rcsid);
}

 /* End of code */
